import { Component, Input, OnInit } from '@angular/core';
import { AbstractControl, Validators } from '@angular/forms';
import { CurrencyService } from '@core/services/currency.service';
import { SpecialHandlingService } from '@core/services/special-handling.service';
import { SpinnerService } from '@core/services/spinner.service';
import { TimeZoneService } from '@core/services/time-zone.service';
import { ApplicationForUi } from '@core/typings/application.typing';
import { Budget, BudgetFundingSourceCombo, FundingSourceTypes, RemainingAmountBudgetMap } from '@core/typings/budget.typing';
import { OrganizationForInfoPanel } from '@core/typings/organization.typing';
import { ProcessingTypes } from '@core/typings/payment.typing';
import { ProgramDetailFromApi } from '@core/typings/program.typing';
import { ApplicationViewService } from '@features/application-view/application-view.service';
import { AwardControlGroup } from '@features/awards/award-amount-control/award-amount-control.component';
import { AwardService } from '@features/awards/award.service';
import { AwardModalResolver } from '@features/awards/resolvers/award-modal.resolver';
import { ApproveAwardPayModalPayload, AwardForApproveAwardPay } from '@features/awards/typings/award.typing';
import { ApplicationBudgetInfo } from '@features/budget-assignments/budget-assignments.typing';
import { BudgetService } from '@features/budgets/budget.service';
import { ClientSettingsService } from '@features/client-settings/client-settings.service';
import { InKindService } from '@features/in-kind/in-kind.service';
import { InKindItemToAwardOrPay } from '@features/in-kind/in-kind.typing';
import { EmailService } from '@features/system-emails/email.service';
import { EmailNotificationType } from '@features/system-emails/email.typing';
import { EmailValidator, GenericFile, OrganizationEligibleForGivingStatus, Tab, TextFriendlySpecialCharCleaner, TypeaheadSelectOption, TypeSafeFormBuilder, TypeSafeFormGroup, TypeToken } from '@yourcause/common';
import { I18nService } from '@yourcause/common/i18n';
import { ExchangeRate } from '@yourcause/common/masking';
import { YCModalComponent } from '@yourcause/common/modals';
import moment from 'moment';
interface CashFormGroup {
  cashAmount: string;
  cashScheduledDate: string;
  cashBudgetIdFundingSource: BudgetFundingSourceCombo;
}
interface InKindFormGroup {
  inKindAmount: string;
  inKindScheduledDate: string;
  inKindBudgetIdFundingSource: BudgetFundingSourceCombo;
}
interface DetailsFormGroup {
  automaticallyAwardAndPay: boolean;
  notifyApplicant: boolean;
  clientEmailTemplateId: number;
  cc: string[];
  bcc: string[];
  attachments: GenericFile[];
  customMessage: string;
}

@Component({
  selector: 'gc-approve-award-pay-modal',
  templateUrl: './approve-award-pay-modal.component.html',
  styleUrls: ['./approve-award-pay-modal.component.scss']
})
export class ApproveAwardPayModalComponent extends YCModalComponent<ApproveAwardPayModalPayload> implements OnInit {
  @Input() applicationId: number;
  @Input() programId: number;

  formGroup: TypeSafeFormGroup<DetailsFormGroup>;
  cashFormGroup: TypeSafeFormGroup<CashFormGroup>;
  inKindFormGroup: TypeSafeFormGroup<InKindFormGroup>;
  emailModalOpen = false;
  application: ApplicationForUi;
  cashAmountRequested: number;
  cashAmountRequestedEquiv: number;
  inKindAmountRequested: number;
  defaultCurrency = this.clientSettingsService.defaultCurrency;
  showConversions = false;
  exchangeRate: ExchangeRate;
  potentialBudgets: Budget[] = [];
  budgetFsCashOptions: TypeaheadSelectOption<BudgetFundingSourceCombo>[] = [];
  budgetFsInKindOptions: TypeaheadSelectOption<BudgetFundingSourceCombo>[] = [];
  awardType: FundingSourceTypes; // if null, allows both awardTypes
  itemsToAwardAndPay: InKindItemToAwardOrPay[] = [];
  tabs: Tab[] = [];
  isCashView = true;
  budgetMapCash: RemainingAmountBudgetMap = {};
  budgetMapInKind: RemainingAmountBudgetMap = {};
  program: ProgramDetailFromApi;
  cashBudgetFsCombo: string;
  inKindBudgetFsCombo: string;
  currentEmailActive: boolean;
  editingItems = false;
  awardAmountBaseHelpText = this.i18n.translate(
    'AWARDS:textTheAmountYouIntendToAwardAndPay',
    {},
    'The amount you intend to award and pay'
  );
  canMoveFundsCashText = '';
  canMoveFundsInKindText = '';
  inactiveRequestedItems = '';
  appReservedInfo: ApplicationBudgetInfo;
  createPaymentAlertHelper: string;
  hasSpecialHandling = false;
  recommendedFundingAmount: number;
  recommendedFundingAmountEquiv: number;
  isMasked = false;
  organization: OrganizationForInfoPanel;
  $awardGroup = new TypeToken<TypeSafeFormGroup<AwardControlGroup>>();

  constructor (
    private formBuilder: TypeSafeFormBuilder,
    private spinnerService: SpinnerService,
    private i18n: I18nService,
    private clientSettingsService: ClientSettingsService,
    private currencyService: CurrencyService,
    private awardService: AwardService,
    private inKindService: InKindService,
    private budgetService: BudgetService,
    private emailService: EmailService,
    private timeZoneService: TimeZoneService,
    private awardModalResolver: AwardModalResolver,
    private specialHandlingService: SpecialHandlingService,
    private applicationViewService: ApplicationViewService
  ) {
    super();
  }

  get hasUnits () {
    return !this.awardType ||  this.awardType === FundingSourceTypes.UNITS;
  }

  get hasCash () {
    return !this.awardType || this.awardType === FundingSourceTypes.DOLLARS;
  }

  get unallocatedSourceMap () {
    return this.budgetService.unallocatedSourceMap;
  }

  get customMessage () {
    return this.formGroup ? this.formGroup.value.customMessage : '';
  }

  get allItemsMap () {
    return this.inKindService.allItemsMap;
  }

  get currentEmailType () {
    if (this.formGroup) {
      if (this.automaticallyAwardAndPay) {
        return this.getCurrentEmailType();
      }
    }

    return EmailNotificationType.ApplicationApproved;
  }

  get automaticallyAwardAndPay () {
    if (this.formGroup) {
      return this.formGroup.value.automaticallyAwardAndPay;
    }

    return true;
  }

  get cashBudgetIdFundingSource () {
    return this.cashFormGroup?.get('cashBudgetIdFundingSource')?.value;
  }

  get inKindBudgetIdFundingSource () {
    return this.inKindFormGroup?.get('inKindBudgetIdFundingSource')?.value;
  }

  get selectedCashBudgetClosed () {
    return this.cashBudgetIdFundingSource?.isClosed;
  }

  get selectedInKindBudgetClosed () {
    return this.inKindBudgetIdFundingSource?.isClosed;
  }

  get isYcProcessed () {
    return this.automaticallyAwardAndPay &&
      this.cashBudgetIdFundingSource?.fundingSource?.processingTypeId === ProcessingTypes.YourCause;
  }


  get modalInvalid () {
    if (this.automaticallyAwardAndPay) {
      const cashInvalid = this.cashFormGroup ? this.cashFormGroup.invalid : false;
      const inKindInvalid = this.inKindFormGroup ? this.inKindFormGroup.invalid : false;

      return cashInvalid || inKindInvalid || this.formGroup.invalid;
    }

    return this.formGroup.invalid;
  }

  async ngOnInit () {
    this.spinnerService.startSpinner();
    const [
      application
    ] = await Promise.all([
      this.applicationViewService.getApplicationDetail(
        this.applicationId
      ),
      await this.awardModalResolver.resolve()
    ]);
    this.application = application;
    this.isMasked = this.application.isMasked &&
      !this.application.canViewMaskedApplicantInfo;
    this.hasSpecialHandling = this.specialHandlingService.hasSpecialHandling(
      this.application.specialHandling
    );
    await Promise.all([
      this.setOrganization(),
      this.setAmountInfo(),
      this.getBudgetsAndOptions()
    ]);
    this.spinnerService.stopSpinner();
  }

  async setOrganization () {
    if (this.application.organizationId) {
      const specialHandling = await this.specialHandlingService.getDefaultSpecialHandling(
        this.application.organizationId,
        this.application.nonprofitGuid,
        this.application.organizationName,
        this.application.applicantId,
        this.application.applicantFirstName,
        this.application.applicantLastName,
        this.application.applicantEmail,
        this.application.applicantPhoneNumber
      );
      this.organization = this.applicationViewService.getOrgForInfoPanel(
        this.application,
        specialHandling
      );
    }
  }

  async setAmountInfo () {
    if (this.application.reservedFunds) {
      this.appReservedInfo = {
        budgetId: this.application.budgetId,
        fundingSourceId: this.application.fundingSourceId,
        reservedFunds: this.application.reservedFunds,
        amountReserved: this.application.amountRequested
      };
    }
    const amountRequested = this.application.amountRequested;
    const equiv = this.application.currencyRequestedAmountEquivalent;
    this.showConversions = !!equiv && (amountRequested !== equiv);
    if (this.showConversions) {
      this.exchangeRate = await this.currencyService.getExchangeRate(
        this.application.currencyRequested,
        this.defaultCurrency
      );
    }
    this.cashAmountRequested = !this.showConversions ? amountRequested : equiv;
    this.cashAmountRequestedEquiv = !this.showConversions ? null : amountRequested;
    this.inKindAmountRequested = this.application.inKindAmountRequested;
    const recommendedAmount = this.application.recommendedFundingAmount;
    if (recommendedAmount) {
      this.recommendedFundingAmount = !this.showConversions ?
        recommendedAmount :
        (recommendedAmount / this.exchangeRate.rate);
      this.recommendedFundingAmountEquiv = !this.showConversions ?
        null :
        recommendedAmount;
    }
  }

  async getBudgetsAndOptions () {
    const {
      allowCash,
      allowUnits,
      filteredBudgets
    } = await this.budgetService.getBudgetsFilteredByProgramAndProcessor(
      this.application.programId,
      this.application.organizationEligibleForGivingStatus === OrganizationEligibleForGivingStatus.ELIGIBLE,
      this.application.grantProgramCycle.id
    );
    this.potentialBudgets = filteredBudgets;
    await this.setAwardType(allowCash, allowUnits);
    this.setFinalBudgetListsAndOptions();
    if (this.hasUnits) {
      this.setInKindItemsFields();
    }
    await this.setFormGroupsAndTabs();
  }

  async setAwardType (allowCash: boolean, allowUnits: boolean) {
    const response = this.awardService.getAwardType(
      allowCash,
      allowUnits,
      null,
      [],
      false
    );
    this.awardType = response.awardType;
  }

  setInKindItemsFields () {
    this.itemsToAwardAndPay = this.application.inKindItems.map((item) => {
      const identification = item.itemIdentification;
      const map = this.allItemsMap[identification];

      return {
        identification,
        name: map.name,
        value: map.value,
        unitsRemaining: map.unitsRemaining,
        categoryId: map.categoryId,
        unitsEntered: item.count,
        inUse: map.inUse
      };
    }).filter((item) => item.inUse);
    this.inactiveRequestedItems = this.inKindService.getInactiveRequestedItemsString(
      this.application.inKindItems
    );
  }

  setCreatePaymentAlertHelper () {
    const budgetFs = this.cashFormGroup?.value?.cashBudgetIdFundingSource;
    this.createPaymentAlertHelper = this.awardService.getCreatePaymentAlertHelper(
      budgetFs?.fundingSource?.processingTypeId === ProcessingTypes.YourCause,
      this.hasSpecialHandling,
      true
    );
  }

  setFinalBudgetListsAndOptions () {
    const potentialCash = this.potentialBudgets.filter((budget) => {
      return budget.fundingSourceType === FundingSourceTypes.DOLLARS;
    });
    const potentialInKind = this.potentialBudgets.filter((budget) => {
      return budget.fundingSourceType === FundingSourceTypes.UNITS;
    });
    this.budgetFsCashOptions = this.hasCash &&
      (this.cashAmountRequested || this.application.recommendedFundingAmount) ?
        this.budgetService.getBudgetFundingSourceComboOptions(
          potentialCash,
          [this.application.organizationEligibleForGivingStatus]
        ) :
        [];
    this.budgetFsInKindOptions = this.hasUnits ?
      this.budgetService.getBudgetFundingSourceComboOptions(
        potentialInKind,
        [this.application.organizationEligibleForGivingStatus]
      ) :
      [];
  }

  async setFormGroupsAndTabs () {
    if (this.budgetFsCashOptions.length > 0) {
      const {
        defaultBudgetId,
        defaultFsId
      } = await this.getDefaultBudgetAndFs(true);
      const budgetFs = this.getBudgetFormControlValue(true, defaultBudgetId, defaultFsId);
      let unformattedAmount = this.showConversions ?
        this.application.currencyRequestedAmountEquivalent :
        this.application.amountRequested;
      if (this.recommendedFundingAmount) {
        unformattedAmount = this.recommendedFundingAmount;
      }
      const amount = this.awardService.getAwardAmountForFormGroup(
        unformattedAmount,
        this.application.currencyRequested,
        budgetFs.isClosed
      );
      this.cashBudgetFsCombo = `${defaultBudgetId}-${defaultFsId}`;
      this.cashFormGroup = this.formBuilder.group<CashFormGroup>({
        cashAmount: [amount, this.awardMinimumValidator()],
        cashBudgetIdFundingSource: [
          budgetFs,
          Validators.required
        ],
        cashScheduledDate: [
          moment().toString(),
          Validators.required
        ]
      }, {
        validator: [
          this.sourceAvailableCash()
        ]
      });
      this.setCreatePaymentAlertHelper();
    }
    if (this.budgetFsInKindOptions.length > 0) {
      const amount = this.awardService.getAwardAmountForFormGroup(
        this.application.inKindAmountRequested,
        this.defaultCurrency
      );
      const {
        defaultBudgetId,
        defaultFsId
      } = await this.getDefaultBudgetAndFs(false);
      this.inKindBudgetFsCombo = `${defaultBudgetId}-${defaultFsId}`;
      this.inKindFormGroup = this.formBuilder.group<InKindFormGroup>({
        inKindAmount: [amount, this.awardMinimumValidator()],
        inKindBudgetIdFundingSource: [
          this.getBudgetFormControlValue(false, defaultBudgetId, defaultFsId),
          Validators.required
        ],
        inKindScheduledDate: [
          moment().toString(),
          Validators.required
        ]
      }, {
        validator: [
          this.sourceAvailableInKind()
        ]
      });
    }
    if (this.inKindFormGroup && this.cashFormGroup) {
      this.tabs = [{
        label: this.i18n.translate('GLOBAL:textCash'),
        active: true
      }, {
        label: this.i18n.translate(
          'GLOBAL:hdrInKind',
          {},
          'In Kind'
        ),
        active: false
      }];
    } else if (this.inKindFormGroup) {
      this.isCashView = false;
    }
    this.setRemainingAmountBudgetMap();

    this.formGroup = this.formBuilder.group<DetailsFormGroup>({
      automaticallyAwardAndPay: true,
      notifyApplicant: this.application.applicantCanReceiveEmails,
      clientEmailTemplateId: 0,
      cc: [[], EmailValidator()],
      bcc: [[], EmailValidator()],
      attachments: [],
      customMessage: ''
    });
    this.currentEmailActive = await this.emailService.isProgramEmailActive(
      this.currentEmailType,
      this.programId
    );
  }

  getCurrentEmailType () {
    if (this.cashFormGroup && this.inKindFormGroup) {
      return EmailNotificationType.ApplicationApprovedAndAwardedInKindAndCash;
    } else if (this.cashFormGroup) {
      return EmailNotificationType.ApplicationApprovedAndAwarded;
    } else if (this.inKindFormGroup) {
      return EmailNotificationType.ApplicationApprovedAndAwardedInKind;
    }

    return EmailNotificationType.ApplicationApproved;
  }

  async getDefaultBudgetAndFs (isCash = true) {
    const array = isCash ? this.budgetFsCashOptions : this.budgetFsInKindOptions;
    const budgetFundingSource = await this.budgetService.getDefaultBudgetForNewPayment(
      array,
      !isCash,
      this.programId,
      this.application.grantProgramCycle.id,
      this.application.budgetId,
      this.application.fundingSourceId
    );

    return {
      defaultBudgetId: budgetFundingSource.budget.id,
      defaultFsId: budgetFundingSource.fundingSource.fundingSourceId
    };
  }

  getBudgetFormControlValue (
    isCash: boolean,
    defaultBudgetId: number,
    defaultFsId: number
  ) {
    const array = isCash ? this.budgetFsCashOptions : this.budgetFsInKindOptions;

    return array.find(option => {
      return option.value.budget.id === defaultBudgetId &&
        option.value.fundingSource.fundingSourceId === defaultFsId;
    }).value;
  }

  setRemainingAmountBudgetMap () {
    if (this.cashFormGroup) {
      const currentCashAmount = this.awardService.unmaskAndConvertFromRequestedToDefault(
        this.cashFormGroup.value.cashAmount,
        this.getCurrency(true),
        this.getRate(true)
      );
      this.budgetMapCash = this.budgetService.getRemainingAmountBudgetMap(
        this.budgetFsCashOptions,
        this.cashFormGroup.value.cashBudgetIdFundingSource,
        currentCashAmount,
        null,
        false,
        false,
        this.automaticallyAwardAndPay && !this.canMoveFundsInKindText,
        this.appReservedInfo ? [this.appReservedInfo] : [],
        null,
        null
      );
    }
    if (this.inKindFormGroup) {
      const currentUnitAmount = this.awardService.unmaskAndConvertFromRequestedToDefault(
        this.inKindFormGroup.value.inKindAmount,
        this.getCurrency(false),
        this.getRate(false)
      );
      this.budgetMapInKind = this.budgetService.getRemainingAmountBudgetMap(
        this.budgetFsInKindOptions,
        this.inKindFormGroup.value.inKindBudgetIdFundingSource,
        currentUnitAmount,
        null,
        false,
        true,
        this.automaticallyAwardAndPay && !this.canMoveFundsCashText,
        [],
        null,
        null
      );
    }
  }

  getRate (isCash = true) {
    return isCash ?
      (this.exchangeRate ? this.exchangeRate.rate : 1) :
      1;
  }

  getExchangeRateId (isCash = true) {
    return isCash ?
      (this.exchangeRate ? this.exchangeRate.id : null) :
      null;
  }

  getCurrency (isCash = true) {
    return isCash ?
      this.application.currencyRequested || this.defaultCurrency :
      this.defaultCurrency;
  }

  awardPaymentAmountChange (isCash = true) {
    this.setRemainingAmountBudgetMap();
    if (isCash && this.inKindFormGroup) {
      this.inKindFormGroup.updateValueAndValidity();
    } else if (this.cashFormGroup) {
      this.cashFormGroup.updateValueAndValidity();
    }
  }

  onTotalItemsChange (total: number) {
    const value = this.awardService.getAwardAmountForFormGroup(
      total,
      this.defaultCurrency
    );
    this.inKindFormGroup.get('inKindAmount').setValue(
      value
    );
  }

  onItemsChange (items: InKindItemToAwardOrPay[]) {
    this.itemsToAwardAndPay = [ ...items ];
  }

  sourceChanged (isCash = true) {
    const budgetFs = this.getBudgetFs(isCash);
    const attr = isCash ? 'cashBudgetFsCombo' : 'inKindBudgetFsCombo';
    this[attr] = `${budgetFs.budget.id}-${budgetFs.fundingSource.fundingSourceId}`;
    this.setRemainingAmountBudgetMap();
    if (isCash) {
      this.cashFormGroup.get('cashAmount').updateValueAndValidity();
    } else {
      this.inKindFormGroup.get('inKindAmount').updateValueAndValidity();
    }
    this.setCreatePaymentAlertHelper();
  }

  onSubmit (moveFunds = false) {
    const formValue = this.formGroup.value;
    const awards = this.automaticallyAwardAndPay ?
      this.generateAwardsPayload(moveFunds) :
      undefined;
    if (moveFunds) {
      this.budgetService.resetFundingSources();
    }
    this.closeModal.emit({
      applicationId: this.applicationId,
      sendEmail: formValue.notifyApplicant,
      clientEmailTemplateId: formValue.clientEmailTemplateId,
      customMessage: this.customMessage,
      emailOptionsModel: {
        bccEmails: formValue.bcc,
        ccEmails: formValue.cc,
        attachments: formValue.attachments
      },
      awards,
      usedOverage: moveFunds
    });
  }

  generateAwardsPayload (moveFunds = false) {
    const awards: AwardForApproveAwardPay[] = [];
    if (this.cashFormGroup) {
      const cashAward = this.generateAward(true, moveFunds);
      awards.push(cashAward);
    }
    if (this.inKindFormGroup) {
      const inKindAward = this.generateAward(false, moveFunds);
      awards.push(inKindAward);
    }

    return awards;
  }

  getBudgetFs (isCash = true) {
    return isCash ?
      this.cashFormGroup.value.cashBudgetIdFundingSource :
      this.inKindFormGroup.value.inKindBudgetIdFundingSource;
  }

  getAmount (isCash = true) {
    return isCash ?
      this.cashFormGroup.value.cashAmount :
      this.inKindFormGroup.value.inKindAmount;
  }

  getScheduledDate (isCash = true) {
    return isCash ?
      this.cashFormGroup.value.cashScheduledDate :
      this.inKindFormGroup.value.inKindScheduledDate;
  }

  generateAward (isCash = true, moveFunds = false) {
    const awardType = isCash ? FundingSourceTypes.DOLLARS : FundingSourceTypes.UNITS;
    const budgetFs = this.getBudgetFs(isCash);
    const rate = this.getRate(isCash);
    const currency = this.getCurrency(isCash);
    const formValAmount = this.getAmount(isCash);
    let amount = this.awardService.unmaskAndConvertFromRequestedToDefault(
      formValAmount,
      currency,
      rate
    );
    let amountEquivalent = this.awardService.getAmountInRequested(
      formValAmount,
      currency
    );
    if (moveFunds) {
      const availToMove = this.unallocatedSourceMap[
        budgetFs.fundingSource.fundingSourceId
      ];
      const remaining = this.getRemainingAmount(budgetFs);
      const canCoverAll = (remaining + availToMove) >= 0;
      if (!canCoverAll) {
        amount  = this.awardService.getNewPaymentAmountFromMovingFunds(
          budgetFs,
          remaining,
          formValAmount,
          currency,
          rate
        );
        const equiv = this.currencyService.fixNumberOfDecimals(
          amount / rate,
          currency
        );
        amountEquivalent = +equiv;
      }
    }
    const inKindItems = !isCash ?
      this.itemsToAwardAndPay.filter((item) => {
        return +item.unitsEntered > 0;
      }).map((item) => {
        return {
          itemIdentification: item.identification,
          count: +item.unitsEntered,
          value: item.value
        };
      }) :
      [];
    const award: AwardForApproveAwardPay =  {
      awardType,
      awardAmount: amount,
      currencyRequested: this.getCurrency(isCash),
      amountEquivalent,
      currencyExchangeRate: this.getExchangeRateId(isCash),
      inKindAwardItemsRequested: inKindItems,
      description: '',
      awardDate: this.timeZoneService.returnMidnightUTCDate(),
      customMessage: this.customMessage,
      payment: {
        applicationId: this.applicationId,
        awardId: null,
        scheduledDate: this.timeZoneService.returnMidnightUTCDate(
          this.getScheduledDate(isCash)
        ),
        budgetId: budgetFs.budget.id,
        fundingSourceId: budgetFs.fundingSource.fundingSourceId,
        fundingSourceType: awardType,
        totalAmount: amount,
        notes: '',
        paymentDesignation: this.application.designation ? TextFriendlySpecialCharCleaner(this.application.designation) : '',
        careOf: this.application?.careOf ?? '',
        currencyRequested: this.getCurrency(isCash),
        currencyExchangeRate: this.getExchangeRateId(isCash),
        totalAmountEquivalent: amountEquivalent,
        differentThanConversion: false,
        inKindItems
      }
    };

    return award;
  }

  getRemainingAmount (
    budgetFundingSource: BudgetFundingSourceCombo,
    isCash = true
  ) {
    const amount = this.getAmount(isCash);
    const currentPaymentAmount = this.awardService.unmaskAndConvertFromRequestedToDefault(
      amount,
      this.getCurrency(isCash),
      this.getRate(isCash)
    );
    const remaining = this.budgetService.getRemainingAmountForBudgetFs(
      budgetFundingSource,
      budgetFundingSource.budget.id,
      budgetFundingSource.fundingSource.fundingSourceId,
      currentPaymentAmount,
      null,
      isCash && this.appReservedInfo ? [this.appReservedInfo] : [],
      null,
      null
    );

    return remaining;
  }

  awardMinimumValidator () {
    return (control: AbstractControl) => {
      const amount = this.awardService.getUnmaskedAmount(
        control.value,
        this.application.currencyRequested || this.defaultCurrency
      );
      if (amount < 0) {
        return {
          awardMinOfZero: {
            i18nKey: 'common:textPleaseEnterANumberEqualToOrGreaterThanZero',
            defaultValue: 'Amount must be equal to or greater than zero.'
          }
        };
      }

      return null;
    };
  }

  sourceAvailableCash () {
    return (group: AbstractControl) => {
      let returnVal = null;
      const control = group.get('cashBudgetIdFundingSource');
      if (control && this.cashFormGroup) {
        returnVal = this.handleSourceValidation(control, true);
      }

      return returnVal;
    };
  }

  sourceAvailableInKind () {
    return (group: AbstractControl) => {
      let returnVal = null;
      const control = group.get('inKindBudgetIdFundingSource');
      if (control && this.inKindFormGroup) {
        returnVal = this.handleSourceValidation(control, false);
      }

      return returnVal;
    };
  }

  handleSourceValidation (control: AbstractControl, isCash = true) {
    const value: BudgetFundingSourceCombo = control.value;
    const attr = isCash ? 'canMoveFundsCashText' : 'canMoveFundsInKindText';
    if (value && value.fundingSource) {
      const remaining = this.getRemainingAmount(value, isCash);
      if (remaining < 0) {
        const amount = this.getAmount(isCash);
        const allowOverage = this.automaticallyAwardAndPay &&
          (isCash ? !this.canMoveFundsInKindText : !this.canMoveFundsCashText);
        const response = this.awardService.handlePaymentSourceValidation(
          value,
          amount,
          remaining,
          this.getCurrency(isCash),
          this.getRate(isCash),
          !isCash,
          allowOverage
        );
        this[attr] = response.canMoveFundsText;

        return response.error;
      }
    }
    this[attr] = '';

    return null;
  }

}
