import { Component, Input, OnInit } from '@angular/core';
import { AbstractControl, Validators } from '@angular/forms';
import { CurrencyService } from '@core/services/currency.service';
import { SpinnerService } from '@core/services/spinner.service';
import { TimeZoneService } from '@core/services/time-zone.service';
import { ApplicationFromPaginated, BulkApplicationDetail } from '@core/typings/application.typing';
import { Budget, BudgetFundingSourceCombo, FundingSourceTypes, RemainingAmountBudgetMap } from '@core/typings/budget.typing';
import { ProcessingTypes } from '@core/typings/payment.typing';
import { ProgramDetailFromApi } from '@core/typings/program.typing';
import { ApplicationManagerService } from '@features/application-manager/services/application-manager/application-manager.service';
import { AwardService } from '@features/awards/award.service';
import { AwardModalResolver } from '@features/awards/resolvers/award-modal.resolver';
import { ApplicationAward, BulkApproveAwardPayPayload, BulkApproveModalResponse, BulkAward } from '@features/awards/typings/award.typing';
import { ApplicationBudgetInfo } from '@features/budget-assignments/budget-assignments.typing';
import { BudgetService } from '@features/budgets/budget.service';
import { ClientSettingsService } from '@features/client-settings/client-settings.service';
import { InKindService } from '@features/in-kind/in-kind.service';
import { ApplicantInKindItem, InKindItemToAwardOrPay } from '@features/in-kind/in-kind.typing';
import { EmailService } from '@features/system-emails/email.service';
import { EmailNotificationType } from '@features/system-emails/email.typing';
import { TagsGroup } from '@features/system-tags/system-tags-input/system-tags-input.component';
import { SystemTags } from '@features/system-tags/typings/system-tags.typing';
import { EmailValidator, GenericFile, OrganizationEligibleForGivingStatus, SimpleStringMap, Tab, TextFriendlySpecialCharCleaner, TypeaheadSelectOption, TypeSafeFormBuilder, TypeSafeFormGroup, TypeToken } from '@yourcause/common';
import { I18nService } from '@yourcause/common/i18n';
import { ExchangeRate } from '@yourcause/common/masking';
import { YCModalComponent } from '@yourcause/common/modals';
import moment from 'moment';
interface CashFormGroup {
  cashBudgetIdFundingSource: BudgetFundingSourceCombo;
  cashScheduledDate: string;
}
interface InKindFormGroup {
  inKindBudgetIdFundingSource: BudgetFundingSourceCombo;
  inKindScheduledDate: string;
}
interface DetailsGroup extends TagsGroup {
  automaticallyAwardAndPay: boolean;
  awardDate: moment.Moment;
  description: string;
  notifyApplicantCash: boolean;
  notifyApplicantInKind: boolean;
  customMessage: string;
  comment: string;
  cashAttachments: GenericFile[];
  inKindAttachments: GenericFile[];
  cashCC: string[];
  cashBCC: string[];
  inKindCC: string[];
  inKindBCC: string[];
  notes: string;
  cashClientEmailTemplateId: number;
  inKindClientEmailTemplateId: number;
}

@Component({
  selector: 'gc-bulk-approve-award-pay-modal',
  templateUrl: './bulk-approve-award-pay-modal.component.html',
  styleUrls: ['./bulk-approve-award-pay-modal.component.scss']
})
export class BulkApproveAwardPayModalComponent extends YCModalComponent<
  BulkApproveAwardPayPayload|BulkApproveModalResponse
> implements OnInit {
  @Input() applications: ApplicationFromPaginated[];
  @Input() programId: number;
  @Input() cycleId: number;
  @Input() currencyRequested: string;
  @Input() awardOnly = false;

  modalHeader = '';
  formGroup: TypeSafeFormGroup<DetailsGroup>;
  cashFormGroup: TypeSafeFormGroup<CashFormGroup>;
  inKindFormGroup: TypeSafeFormGroup<InKindFormGroup>;
  emailModalOpen = false;
  cashAmountRequested: number;
  inKindAmountRequested: number;
  cashAmountRequestedText: string;
  inKindAmountRequestedText: string;
  defaultCurrency = this.clientSettingsService.defaultCurrency;
  showConversions = false;
  exchangeRate: ExchangeRate;
  potentialBudgets: Budget[] = [];
  budgetFsCashOptions: TypeaheadSelectOption<BudgetFundingSourceCombo>[] = [];
  budgetFsInKindOptions: TypeaheadSelectOption<BudgetFundingSourceCombo>[] = [];
  awardType: FundingSourceTypes; // if null, allows both awardTypes
  itemsToAwardAndPay: InKindItemToAwardOrPay[] = [];
  tabs: Tab[] = [];
  budgetMapCash: RemainingAmountBudgetMap = {};
  budgetMapInKind: RemainingAmountBudgetMap = {};
  program: ProgramDetailFromApi;
  cashBudgetFsCombo: string;
  inKindBudgetFsCombo: string;
  awardAmountBaseHelpText = this.i18n.translate(
    'AWARDS:textTheAmountYouIntendToAwardAndPay',
    {},
    'The amount you intend to award and pay'
  );
  showNotesAndTags = false;
  availableTags: TypeaheadSelectOption[] = [];
  tagType = SystemTags.Buckets.Payment;
  activeTabIndex = 0;
  cashIndex: number;
  inKindIndex: number;
  optionsIndex: number;
  cashEmailType: EmailNotificationType;
  inKindEmailType: EmailNotificationType;
  inactiveRequestedItems: string;
  orgEligibleStatusArray: OrganizationEligibleForGivingStatus[];
  submitting = false;
  applicationIds: number[] = [];
  appReservedInfo: ApplicationBudgetInfo[];
  recommendedFundingAmount: number;
  recommendedFundingAmountText: string;
  addingNewTag = false;
  appDetailsById: Record<string, BulkApplicationDetail> = {};
  $tagsGroup = new TypeToken<TypeSafeFormGroup<TagsGroup>>();

  constructor (
    private formBuilder: TypeSafeFormBuilder,
    private spinnerService: SpinnerService,
    private i18n: I18nService,
    private clientSettingsService: ClientSettingsService,
    private currencyService: CurrencyService,
    private awardService: AwardService,
    private inKindService: InKindService,
    private budgetService: BudgetService,
    private emailService: EmailService,
    private timeZoneService: TimeZoneService,
    private awardModalResolver: AwardModalResolver,
    private applicationManagerService: ApplicationManagerService
  ) {
    super();
  }

  get hasUnits () {
    return !this.awardType ||  this.awardType === FundingSourceTypes.UNITS;
  }

  get hasCash () {
    return !this.awardType || this.awardType === FundingSourceTypes.DOLLARS;
  }

  get customMessage () {
    return this.formGroup ? this.formGroup.value.customMessage : '';
  }

  get allItemsMap () {
    return this.inKindService.allItemsMap;
  }

  get automaticallyAwardAndPay () {
    if (this.formGroup) {
      return this.formGroup.value.automaticallyAwardAndPay;
    }

    return true;
  }

  get cashBudgetIdFundingSource () {
    return this.cashFormGroup?.value?.cashBudgetIdFundingSource;
  }

  get inKindBudgetIdFundingSource () {
    return this.inKindFormGroup?.value?.inKindBudgetIdFundingSource;
  }

  get selectedCashBudgetClosed () {
    return this.cashBudgetIdFundingSource?.isClosed;
  }

  get selectedInKindBudgetClosed () {
    return this.inKindBudgetIdFundingSource?.isClosed;
  }

  get modalInvalid () {
    if (this.automaticallyAwardAndPay) {
      const cashInvalid = this.cashFormGroup ? this.cashFormGroup.invalid : false;
      const inKindInvalid = this.inKindFormGroup ? this.inKindFormGroup.invalid : false;

      return cashInvalid || inKindInvalid || this.addingNewTag;
    }

    return false;
  }

  get isYcProcessed () {
    return this.automaticallyAwardAndPay &&
      this.cashBudgetIdFundingSource?.fundingSource?.processingTypeId === ProcessingTypes.YourCause;
  }

  async ngOnInit () {
    this.spinnerService.startSpinner();
    const [
      awardResponse,
      appDetailsById
    ] = await Promise.all([
      this.awardModalResolver.resolve(),
      this.applicationManagerService.getBulkApplicationDetailsMap(
        this.applications.map((app) => app.applicationId)
      )
    ]);
    this.appDetailsById = appDetailsById;
    this.availableTags = awardResponse.availableTags;
    this.applicationIds = this.applications.map((app) => app.applicationId);
    this.orgEligibleStatusArray = this.applications.map((app) => app.organizationEligibleForGivingStatus);
    this.appReservedInfo = this.applications.map((app) => {
      return {
        budgetId: app.budgetId,
        fundingSourceId: app.fundingSourceId,
        reservedFunds: app.reservedFunds,
        amountReserved: app.amountRequested
      };
    }).filter((item) => item.reservedFunds);
    await this.setTopLevelStats();
    await this.getBudgetsAndOptions();
    this.setModalHeader();
    this.spinnerService.stopSpinner();
  }

  async setTopLevelStats () {
    this.cashAmountRequested = this.applications.reduce((acc, row) => {
      const appDetail = this.appDetailsById[row.applicationId];
      if (!appDetail.hasCashAward) {
        // There could be a mixed bag.
        // Calculate the total that will come just from amount requested
        const amount = row.recommendedFundingAmount ? 0 : row.amountRequested;

        return acc + (amount || 0);
      }

      return acc;
    }, 0);
    const equiv = this.applications.reduce((acc, row) => {
      const appDetail = this.appDetailsById[row.applicationId];
      if (!appDetail.hasCashAward) {
        const amount = row.recommendedFundingAmount ?
          0 :
          row.currencyRequestedAmountEquivalent;

        return acc + (amount || 0);
      }

      return acc;
    }, 0);
    this.showConversions = !!equiv && (this.cashAmountRequested !== equiv);
    if (this.showConversions) {
      this.exchangeRate = await this.currencyService.getExchangeRate(
        this.currencyRequested,
        this.defaultCurrency
      );
    }
    const cashAmountRequested = !this.showConversions ? this.cashAmountRequested : equiv;
    const cashAmountRequestedEquivText = !this.showConversions || !cashAmountRequested ?
      '' :
      this.currencyService.formatMoney(
        this.cashAmountRequested,
        this.defaultCurrency
      );
    this.cashAmountRequestedText = cashAmountRequested ?
      this.currencyService.formatMoney(
        cashAmountRequested,
        this.currencyRequested,
        this.showConversions
      ) + (cashAmountRequestedEquivText ? ` (${cashAmountRequestedEquivText})` : '') :
      '';
    this.recommendedFundingAmount = this.applications.reduce((acc, row) => {
      const appDetail = this.appDetailsById[row.applicationId];
      if (!appDetail.hasCashAward) {
        return acc + (row.recommendedFundingAmount || 0);
      }

      return acc;
    }, 0);
    const recommendedFundingAmount = !this.showConversions ?
      this.recommendedFundingAmount :
      this.recommendedFundingAmount / this.exchangeRate.rate;
    const recommendedFundingAmountEquivText = !this.showConversions ||
      !recommendedFundingAmount ?
      '' :
      this.currencyService.formatMoney(
        this.recommendedFundingAmount,
        this.defaultCurrency
      );
    this.recommendedFundingAmountText = recommendedFundingAmount ?
      this.currencyService.formatMoney(
        recommendedFundingAmount,
        this.currencyRequested,
        this.showConversions
      ) + (recommendedFundingAmountEquivText ? ` (${recommendedFundingAmountEquivText})` : '') :
      '';
    this.inKindAmountRequested = this.applications.reduce((acc, row) => {
      const appDetail = this.appDetailsById[row.applicationId];
      if (!appDetail.hasInKindAward) {
        return acc + (row.inKindAmountRequested || 0);
      }

      return acc;
    }, 0);
    this.inKindAmountRequestedText = this.inKindAmountRequested ?
      this.currencyService.formatMoney(
        this.inKindAmountRequested
      ) : '';
  }

  setModalHeader () {
    if (this.awardOnly) {
      this.modalHeader = this.i18n.translate(
        'AWARDS:hdrAwardAndPayApplications',
        {},
        'Award and Pay Applications'
      );
    } else {
      if (this.automaticallyAwardAndPay) {
        this.modalHeader = this.i18n.translate(
          'AWARDS:hdrApproveAwardAndPay',
          {},
          'Approve, Award, and Pay'
        );
      } else {
        this.modalHeader = this.i18n.translate(
          'AWARDS:hdrApproveApplication'
        );
      }
    }
  }

  async getBudgetsAndOptions () {
    const {
      allowCash,
      allowUnits,
      filteredBudgets
    } = await this.budgetService.getBudgetsFilteredByProgramAndProcessor(
      this.programId,
      this.orgEligibleStatusArray.every((status) => status === OrganizationEligibleForGivingStatus.ELIGIBLE),
      this.cycleId
    );
    if (!this.awardOnly) {
      this.potentialBudgets = filteredBudgets;
      await this.setAwardType(allowCash, allowUnits);
    } else {
      // in this scenario, there may already be existing awards we need to account for
      let allowCashOverride = allowCash;
      let allowInKindOverride = allowUnits;
      if (allowCash && !this.cashAmountRequested && !this.recommendedFundingAmount) {
        allowCashOverride = false;
      }
      if (allowUnits && !(this.inKindAmountRequested || this.inKindAmountRequested === 0)) {
        allowInKindOverride = false;
      }
      this.potentialBudgets = filteredBudgets.filter((budget) => {
        return budget.fundingSourceType === FundingSourceTypes.DOLLARS ?
          allowCashOverride :
          allowInKindOverride;
      });
      this.setAwardType(allowCashOverride, allowInKindOverride);
    }
    this.setFinalBudgetListsAndOptions();
    if (this.hasUnits) {
      this.setInKindItemsFields();
    }
    await this.setFormGroupsAndTabs();
  }

  async setAwardType (allowCash: boolean, allowUnits: boolean) {
    const response = this.awardService.getAwardType(
      allowCash,
      allowUnits,
      null,
      [],
      false
    );
    this.awardType = response.awardType;
  }

  setInKindItemsFields () {
    let itemCountMap: SimpleStringMap<number> = {};

    Object.keys(this.appDetailsById).forEach((key) => {
      const app = this.appDetailsById[+key];
      app.inKindItems.forEach((item) => {
        const count = itemCountMap[item.itemIdentification];
        if (count) {
          itemCountMap = {
            ...itemCountMap,
            [item.itemIdentification]: count + item.count
          };
        } else {
          itemCountMap = {
            ...itemCountMap,
            [item.itemIdentification]: item.count
          };
        }
      });
    });
    const itemsToAwardAndPay = Object.keys(itemCountMap).map((identification) => {
      const map = this.allItemsMap[identification];

      return {
        identification,
        name: map.name,
        value: map.value,
        unitsRemaining: map.unitsRemaining,
        categoryId: map.categoryId,
        unitsEntered: itemCountMap[identification],
        inUse: map.inUse
      };
    });
    this.inactiveRequestedItems = this.inKindService.getInactiveRequestedItemsString(
      itemsToAwardAndPay.map((item) => {
        return {
          count: item.unitsEntered,
          itemIdentification: item.identification
        };
      })
    );
    this.itemsToAwardAndPay = itemsToAwardAndPay.filter((item) => item.inUse);
  }

  setFinalBudgetListsAndOptions () {
    const potentialCash = this.potentialBudgets.filter((budget) => {
      return budget.fundingSourceType === FundingSourceTypes.DOLLARS;
    });
    const potentialInKind = this.potentialBudgets.filter((budget) => {
      return budget.fundingSourceType === FundingSourceTypes.UNITS;
    });
    this.budgetFsCashOptions = this.hasCash &&
      (this.cashAmountRequested || this.recommendedFundingAmount) ?
      this.budgetService.getBudgetFundingSourceComboOptions(
        potentialCash,
        this.orgEligibleStatusArray
      ) :
      [];
    this.budgetFsInKindOptions = this.hasUnits && (this.inKindAmountRequested || this.inKindAmountRequested === 0) ?
      this.budgetService.getBudgetFundingSourceComboOptions(
        potentialInKind,
        this.orgEligibleStatusArray
      ) :
      [];
  }

  async setFormGroupsAndTabs () {
    if (this.budgetFsCashOptions.length > 0) {
      const {
        defaultBudgetId,
        defaultFsId
      } = await this.getDefaultBudgetAndFs(true);
      const budgetFs = this.getBudgetFormControlValue(true, defaultBudgetId, defaultFsId);
      this.cashBudgetFsCombo = `${defaultBudgetId}-${defaultFsId}`;
      this.cashFormGroup = this.formBuilder.group<CashFormGroup>({
        cashBudgetIdFundingSource: [
          budgetFs, Validators.required
        ],
        cashScheduledDate: [
          moment().toString(),
          Validators.required
        ]
      }, {
        validator: [
          this.sourceAvailableCash()
        ]
      });
    }
    if (this.budgetFsInKindOptions.length > 0) {
      const {
        defaultBudgetId,
        defaultFsId
      } = await this.getDefaultBudgetAndFs(false);
      this.inKindBudgetFsCombo = `${defaultBudgetId}-${defaultFsId}`;
      this.inKindFormGroup = this.formBuilder.group<InKindFormGroup>({
        inKindBudgetIdFundingSource: [
          this.getBudgetFormControlValue(false, defaultBudgetId, defaultFsId),
          Validators.required
        ],
        inKindScheduledDate: [
          moment().toString(),
          Validators.required
        ]
      }, {
        validator: [
          this.sourceAvailableInKind()
        ]
      });
    }
    const additionalOptions = {
      label: this.i18n.translate(
        'GLOBAL:textAdditionalOptions',
        {},
        'Additional options'
      ),
      active: false
    };
    const financialInformation = this.i18n.translate(
      'GLOBAL:textFinancialInformation',
      {},
      'Financial information'
    );
    const cash = {
      label: !this.inKindFormGroup ?
         financialInformation :
         this.i18n.translate('GLOBAL:textCash'),
      active: true
    };
    const inKind = {
      label: !this.cashFormGroup ?
        financialInformation :
        this.i18n.translate(
          'GLOBAL:hdrInKind',
          {},
          'In Kind'
        ),
      active: false
    };
    this.setTabs(cash, inKind, additionalOptions);
    this.setRemainingAmountBudgetMap();
    await this.setEmailTypes();
    this.formGroup = this.formBuilder.group<DetailsGroup>({
      automaticallyAwardAndPay: true,
      awardDate: moment(),
      description: '',
      notifyApplicantCash: false,
      notifyApplicantInKind: false,
      customMessage: '',
      comment: '',
      tags: [[]],
      newTagName: '',
      cashAttachments: [[]],
      inKindAttachments: [[]],
      cashCC: [[], EmailValidator()],
      cashBCC: [[], EmailValidator()],
      inKindCC: [[], EmailValidator()],
      inKindBCC: [[], EmailValidator()],
      notes: '',
      cashClientEmailTemplateId: 0,
      inKindClientEmailTemplateId: 0
    });
  }

  setTabs (cash: Tab, inKind: Tab, additionalOptions: Tab) {
    if (this.inKindFormGroup && this.cashFormGroup) {
      this.tabs = [ cash, inKind, additionalOptions ];
      this.cashIndex = 0;
      this.inKindIndex = 1;
      this.optionsIndex = 2;
    } else if (this.inKindFormGroup) {
      inKind.active = true;
      this.tabs = [ inKind, additionalOptions ];
      this.inKindIndex = 0;
      this.optionsIndex = 1;
    } else {
      this.tabs = [ cash, additionalOptions ];
      this.cashIndex = 0;
      this.optionsIndex = 1;
    }
  }

  async setEmailTypes () {
    if (this.formGroup && !this.awardOnly) {
      if (this.formGroup.value.automaticallyAwardAndPay) {
        this.setCashAndInKindEmailTypes();
      } else {
        this.cashEmailType = EmailNotificationType.ApplicationApproved;
        this.inKindEmailType = null;
      }
    } else {
      this.setCashAndInKindEmailTypes();
    }
    if (this.cashEmailType) {
      const cashEmailActive = await this.emailService.isProgramEmailActive(
        this.cashEmailType,
        this.programId
      );
      if (!cashEmailActive) {
        this.cashEmailType = null;
      }
    }
    if (this.inKindEmailType) {
      const inKindEmailActive = await this.emailService.isProgramEmailActive(
        this.inKindEmailType,
        this.programId
      );
      if (!inKindEmailActive) {
        this.inKindEmailType = null;
      }
    }
  }

  setCashAndInKindEmailTypes () {
    if (this.cashFormGroup) {
      this.cashEmailType = this.awardOnly ?
        EmailNotificationType.ApplicationAwarded :
        EmailNotificationType.ApplicationApprovedAndAwarded;
    }
    if (this.inKindFormGroup) {
      this.inKindEmailType = this.awardOnly ?
        EmailNotificationType.ApplicationAwardedInKind :
        EmailNotificationType.ApplicationApprovedAndAwardedInKind;
    }
  }

  async getDefaultBudgetAndFs (isCash = true) {
    const array = isCash ? this.budgetFsCashOptions : this.budgetFsInKindOptions;
    const budgetFundingSource = await this.budgetService.getDefaultBudgetForNewPayment(
      array,
      !isCash,
      this.programId,
      this.cycleId
    );

    return {
      defaultBudgetId: budgetFundingSource.budget.id,
      defaultFsId: budgetFundingSource.fundingSource.fundingSourceId
    };
  }

  getBudgetFormControlValue (
    isCash: boolean,
    defaultBudgetId: number,
    defaultFsId: number
  ) {
    const array = isCash ? this.budgetFsCashOptions : this.budgetFsInKindOptions;

    return array.find(option => {
      return option.value.budget.id === defaultBudgetId &&
        option.value.fundingSource.fundingSourceId === defaultFsId;
    }).value;
  }

  setRemainingAmountBudgetMap () {
    if (this.cashFormGroup) {
      this.budgetMapCash = this.budgetService.getRemainingAmountBudgetMap(
        this.budgetFsCashOptions,
        this.cashFormGroup.value.cashBudgetIdFundingSource,
        this.cashAmountRequested + this.recommendedFundingAmount,
        null,
        false,
        false,
        false,
        this.appReservedInfo,
        null,
        null
      );
    }
    if (this.inKindFormGroup) {
      this.budgetMapInKind = this.budgetService.getRemainingAmountBudgetMap(
        this.budgetFsInKindOptions,
        this.inKindFormGroup.value.inKindBudgetIdFundingSource,
        this.inKindAmountRequested,
        null,
        false,
        true,
        false,
        [],
        null,
        null
      );
    }
  }

  getRate (isCash = true) {
    return isCash ?
      (this.exchangeRate ? this.exchangeRate.rate : 1) :
      1;
  }

  getExchangeRateId (isCash = true) {
    return isCash ?
      (this.exchangeRate ? this.exchangeRate.id : null) :
      null;
  }

  getCurrency (isCash = true) {
    return isCash ?
      this.currencyRequested || this.defaultCurrency :
      this.defaultCurrency;
  }

  automaticallyAwardAndPayChanged () {
    if (!this.awardOnly) {
      if (this.formGroup.value.automaticallyAwardAndPay) {
        this.setEmailTypes();
      } else {
        this.cashEmailType = EmailNotificationType.ApplicationApproved;
        this.inKindEmailType = null;
      }
    }
  }

  awardPaymentAmountChange (isCash = true) {
    this.setRemainingAmountBudgetMap();
    if (isCash && this.inKindFormGroup) {
      this.inKindFormGroup.updateValueAndValidity();
    } else if (this.cashFormGroup) {
      this.cashFormGroup.updateValueAndValidity();
    }
  }

  getBudgetFs (isCash = true) {
    return isCash ?
      this.cashFormGroup.value.cashBudgetIdFundingSource :
      this.inKindFormGroup.value.inKindBudgetIdFundingSource;
  }

  sourceChanged (isCash = true) {
    const budgetFs = this.getBudgetFs(isCash);
    const attr = isCash ? 'cashBudgetFsCombo' : 'inKindBudgetFsCombo';
    this[attr] = `${budgetFs.budget.id}-${budgetFs.fundingSource.fundingSourceId}`;
    this.setRemainingAmountBudgetMap();
  }

  async onSubmit () {
    this.submitting = true;
    if (this.automaticallyAwardAndPay) {
      this.spinnerService.startSpinner();
      const payload = await this.generateBulkAwardPayload();
      this.spinnerService.stopSpinner();
      this.closeModal.emit(payload);
    } else {
      this.closeModal.emit({
        sendEmail: this.formGroup.value.notifyApplicantCash,
        clientEmailTemplateId: this.formGroup.value.cashClientEmailTemplateId,
        customMessage: this.formGroup.value.customMessage,
        comment: this.formGroup.value.comment,
        emailOptionsModel: {
          bccEmails: this.formGroup.value.cashBCC,
          ccEmails: this.formGroup.value.cashCC,
          attachments: this.formGroup.value.cashAttachments
        }
      });
    }
    this.submitting = false;
  }

  async generateBulkAwardPayload (): Promise<BulkApproveAwardPayPayload> {
    const applications: ApplicationAward[] = [];
    await Promise.all(this.applications.map(async (app) => {
      const awards: BulkAward[] = [];
      const appDetail = this.appDetailsById[app.applicationId];
      if (
        !appDetail.hasCashAward &&
        app.amountRequested &&
        this.cashFormGroup
      ) {
        const hasRecommendedFunding = !!app.recommendedFundingAmount;
        const cashAward = await this.generateAward(
          app.applicationId,
          true,
          hasRecommendedFunding ?
            app.recommendedFundingAmount :
            app.amountRequested,
          hasRecommendedFunding ?
            app.recommendedFundingAmount :
            app.currencyRequestedAmountEquivalent,
          this.cashFormGroup.value.cashBudgetIdFundingSource,
          app.designation,
          app.careOf
        );
        awards.push(cashAward);
      }
      if (
        !appDetail.hasInKindAward &&
        (app.inKindAmountRequested || app.inKindAmountRequested === 0) &&
        this.inKindFormGroup
      ) {
        const inKindAward = await this.generateAward(
          app.applicationId,
          false,
          app.inKindAmountRequested,
          app.inKindAmountRequested,
          this.inKindFormGroup.value.inKindBudgetIdFundingSource,
          app.designation,
          app.careOf,
          this.appDetailsById[app.applicationId].inKindItems.filter((item) => {
            return item.count > 0;
          }).map((item) => {
            return {
              itemIdentification: item.itemIdentification,
              count: item.count,
              value: this.allItemsMap[item.itemIdentification].value
            };
          })
        );
        awards.push(inKindAward);
      }
      applications.push({
        applicationId: app.applicationId,
        awards
      });
    }));
    const cashBudgetFs: BudgetFundingSourceCombo = this.cashFormGroup ?
      this.cashFormGroup.value.cashBudgetIdFundingSource :
      null;
    const inKindBudgetFs: BudgetFundingSourceCombo = this.inKindFormGroup ?
      this.inKindFormGroup.value.inKindBudgetIdFundingSource :
      null;

    return {
      applications,
      cashBudgetId: cashBudgetFs ? cashBudgetFs.budget.id : null,
      cashFundingSourceId: cashBudgetFs ?
        cashBudgetFs.fundingSource.fundingSourceId :
        null,
      inKindBudgetId: inKindBudgetFs ? inKindBudgetFs.budget.id : null,
      inKindFundingSourceId: inKindBudgetFs ?
        inKindBudgetFs.fundingSource.fundingSourceId :
        null,
      customMessage: this.formGroup.value.customMessage,
      paymentTagIds: this.formGroup.value.tags
    };
  }

  async generateAward (
    applicationId: number,
    isCash: boolean,
    amount: number,
    amountEquivalent: number,
    budgetFs: BudgetFundingSourceCombo,
    designation: string,
    careOf: string,
    inKindItems?: ApplicantInKindItem[]
  ): Promise<BulkAward> {
    const scheduledDate = isCash ?
      this.cashFormGroup.value.cashScheduledDate :
      this.inKindFormGroup.value.inKindScheduledDate;
    const awardType = isCash ? FundingSourceTypes.DOLLARS : FundingSourceTypes.UNITS;
    const formValue =  this.formGroup.value;
    const sendEmail = isCash ?
      formValue.notifyApplicantCash :
      formValue.notifyApplicantInKind;
    const award: BulkAward = {
      sendEmail,
      clientEmailTemplateId: sendEmail ?
        (isCash ?
         formValue.cashClientEmailTemplateId :
         formValue.inKindClientEmailTemplateId
        ) :
        null,
      emailOptions: {
        ccEmails: sendEmail ?
          (isCash ? formValue.cashCC : formValue.inKindCC) :
          [],
        bccEmails: sendEmail ?
          (isCash ? formValue.cashBCC : formValue.inKindBCC) :
          [],
        attachments: sendEmail ?
          await this.emailService.returnIdsFromMixedAttachments(
            isCash ? formValue.cashAttachments : formValue.inKindAttachments
          ) :
          []
      },
      awardType,
      awardAmount: amount,
      currencyRequested: this.getCurrency(isCash),
      amountEquivalent,
      currencyExchangeRate: this.getExchangeRateId(isCash),
      inKindAwardItemsRequested: inKindItems,
      description: '',
      awardDate: this.timeZoneService.returnMidnightUTCDate(formValue.awardDate),
      customMessage: formValue.customMessage,
      payment: {
        applicationId,
        awardId: null,
        scheduledDate: this.timeZoneService.returnMidnightUTCDate(scheduledDate),
        budgetId: budgetFs.budget.id,
        fundingSourceId: budgetFs.fundingSource.fundingSourceId,
        fundingSourceType: awardType,
        totalAmount: amount,
        notes: formValue.notes || '',
        paymentDesignation: TextFriendlySpecialCharCleaner(designation) || '',
        currencyRequested: this.getCurrency(isCash),
        currencyExchangeRate: this.getExchangeRateId(isCash),
        totalAmountEquivalent: amountEquivalent,
        differentThanConversion: false,
        inKindItems,
        careOf,
        accountNumber: budgetFs?.budget?.accountNumber ?? ''
      }
    };

    return award;
  }


  getRemainingAmount (
    budgetFundingSource: BudgetFundingSourceCombo,
    isCash = true
  ) {
    const remaining = this.budgetService.getRemainingAmountForBudgetFs(
      budgetFundingSource,
      budgetFundingSource.budget.id,
      budgetFundingSource.fundingSource.fundingSourceId,
      isCash ?
        (this.cashAmountRequested + this.recommendedFundingAmount) :
        this.inKindAmountRequested,
      null,
      isCash ? this.appReservedInfo : [],
      null,
      null
    );

    return remaining;
  }

  sourceAvailableCash () {
    return (group: AbstractControl) => {
      let returnVal = null;
      const control = group.get('cashBudgetIdFundingSource');
      if (control && this.cashFormGroup) {
        returnVal = this.handleSourceValidation(control, true);
      }

      return returnVal;
    };
  }

  sourceAvailableInKind () {
    return (group: AbstractControl) => {
      let returnVal = null;
      const control = group.get('inKindBudgetIdFundingSource');
      if (control && this.inKindFormGroup) {
        returnVal = this.handleSourceValidation(control, false);
      }

      return returnVal;
    };
  }

  handleSourceValidation (control: AbstractControl, isCash = true) {
    const value: BudgetFundingSourceCombo = control.value;
    if (value && value.fundingSource) {
      const remaining = this.getRemainingAmount(value, isCash);
      if (remaining < 0) {
        return {
          budgetIdFundingSource: { // control name
            amountExceeded: { // type of error
              i18nKey: 'AWARDS:textCannotExceedAmountAvailableDesc',
              defaultValue: 'Cannot exceed amount available'
            }
          }
        };
      }
    }

    return null;
  }

}
