import { Component, OnDestroy, OnInit } from '@angular/core';
import { Validators } from '@angular/forms';
import { ActivatedRoute, Router } from '@angular/router';
import { SpinnerService } from '@core/services/spinner.service';
import { FooterState } from '@core/states/footer.state';
import { AutomationAPI } from '@core/typings/api/automation.typing';
import { GrantProgramCycleBudgetFundingSource } from '@core/typings/program.typing';
import { Automation } from '@core/typings/ui/automation.typing';
import { BudgetService } from '@features/budgets/budget.service';
import { ProgramService } from '@features/programs/program.service';
import { RuleAutomationService } from '@features/rule-automation/rule-automation.service';
import { ArrayHelpersService, FooterStates, TypeaheadSelectOption, TypeSafeFormBuilder, TypeSafeFormGroup } from '@yourcause/common';
import { I18nService } from '@yourcause/common/i18n';
import { uniq } from 'lodash';
import { Subscription } from 'rxjs';
import { BudgetAssignmentsService } from '../budget-assignments.service';
import { BudgetAssignmentRuleSetDetail, SaveBudgetAssignmentRuleSet } from '../budget-assignments.typing';

@Component({
  selector: 'gc-manage-budget-assignment-page',
  templateUrl: './manage-budget-assignment-page.component.html',
  styleUrls: ['./manage-budget-assignment-page.component.scss']
})
export class ManageBudgetAssignmentPageComponent implements OnInit, OnDestroy {
  sub: Subscription;
  id = this.activatedRoute.snapshot.params.id;
  objects: Automation.ObjectConfig[] = [];
  isNew = !this.id || (this.id === 'new');
  detail: BudgetAssignmentRuleSetDetail;
  title = this.isNew ?
    this.i18n.translate('CONFIG:textCreateBudgetAssignment', {}, 'Create Budget Assignment') :
    this.i18n.translate('CONFIG:textEditBudgetAssignment', {}, 'Edit Budget Assignment');
  formGroup: TypeSafeFormGroup<{
    program: number;
    budget: number;
    fundingSource: number;
    name: string;
    description: string;
  }>;
  programOptions: TypeaheadSelectOption<number>[] = this.programService.allPublishedActivePrograms
    .map(program => ({
      label: program.grantProgramName,
      value: +program.grantProgramId
    }));
  budgetOptions: TypeaheadSelectOption<number>[];
  fundingSourceOptions: TypeaheadSelectOption<number>[];
  rulesValid: boolean;
  uiRules: Automation.CriteriaFormState[];
  applyRulesWithOr = false;

  constructor (
    private router: Router,
    private automationService: RuleAutomationService,
    private footerState: FooterState,
    private spinnerService: SpinnerService,
    private budgetAssignmentService: BudgetAssignmentsService,
    private programService: ProgramService,
    private budgetService: BudgetService,
    private formBuilder: TypeSafeFormBuilder,
    private activatedRoute: ActivatedRoute,
    private i18n: I18nService,
    private arrayHelper: ArrayHelpersService
  ) {
    this.footerState.set('footerState', FooterStates.ACTION);
    this.footerState.set('footerActionLabel', i18n.translate('common:btnSave'));
    this.footerState.set('footerAction', this.saveRuleSet);
    this.footerState.setCancelLabel(this.i18n.translate('common:btnCancel'));
    this.footerState.setCancelAction(() => this.navigateBack());
  }

  async ngOnInit () {
    this.setDetailAndFormGroup();
    if (!this.isNew) {
      await this.programChanged(this.detail.grantProgramId);
      await this.budgetChanged(this.detail.budgetId);
    }
  }

  setDetailAndFormGroup () {
    this.detail = this.budgetAssignmentService.get('detailMap')[this.id];
    this.applyRulesWithOr = this.detail?.applyRulesWithOr ?? false;
    this.formGroup = this.formBuilder.group({
      program: [this.detail?.grantProgramId ?? 0, Validators.required],
      budget: [this.detail?.budgetId, Validators.required],
      fundingSource: [this.detail?.fundingSourceId, Validators.required],
      name: [this.detail?.name ?? '', Validators.required],
      description: this.detail?.description ?? ''
    });
    this.sub = this.formGroup.statusChanges.subscribe(() => {
      this.ruleValidationChanged(this.rulesValid);
    });
  }

  saveRuleSet = async () => {
    this.spinnerService.startSpinner();
    this.footerState.set('footerActionDisabled', true);
    this.footerState.set('footerSecondaryDisabled', true);
    const formValue = this.formGroup.value;
    const uiRules: Automation.CriteriaFormState[] = this.uiRules;
    let rules: AutomationAPI.SaveAutomationRuleSetExpressionModel[] = [];
    if (uiRules) {
      rules = uiRules
        .reduce((acc, rule) => [
          ...acc,
          ...this.automationService.mapCriteriaToAPIRule(
            rule
          )
        ], [] as AutomationAPI.SaveAutomationRuleSetExpressionModel[]);
    }
    const saveInfo: SaveBudgetAssignmentRuleSet = {
      budgetAssignmentAutomationRuleSetId: !this.isNew ? +this.id : null,
      description: formValue.description,
      grantProgramId: formValue.program,
      budgetId: formValue.budget,
      fundingSourceId: formValue.fundingSource,
      name: formValue.name,
      applyRulesWithOr: this.applyRulesWithOr,
      rules,
      formId: null
    };
    const passed = await this.budgetAssignmentService.saveRuleSet(
      saveInfo
    );
    if (passed) {
      this.navigateBack();
    }
    this.spinnerService.stopSpinner();
    this.footerState.set('footerActionDisabled', false);
    this.footerState.set('footerSecondaryDisabled', false);
  };

  private navigateBack () {
    this.router.navigate(
      [`../`],
      {
        relativeTo: this.activatedRoute
      }
    );
  }

  ruleValidationChanged (valid: boolean) {
    this.rulesValid = valid;
    const invalid = !valid || this.formGroup.invalid;
    this.footerState.set(
      'footerActionDisabled',
      invalid
    );
  }

  async programChanged (programId: number) {
    this.spinnerService.startSpinner();
    this.budgetOptions = null;
    const result = await this.programService.getProgramCycleBudgets(programId);
    this.budgetOptions = this.arrayHelper.sort(uniq(this.flattenProgramBudgetFS(result).map(row => row.budgetId))
      .map(budgetOption => {
        return this.budgetService.get('cashBudgetOptions')
          .find(option => option.value === budgetOption);
      })
      .filter(budgetOption => {
        return budgetOption && this.getFsOptions(result, budgetOption.value).length > 0;
      }), 'label');

    if (!this.budgetOptions.some(option => option.value === this.formGroup.value.budget)) {
      this.formGroup.get('budget').setValue(null);
    }
    this.objects = await this.budgetAssignmentService.getRuleObjectsByProgramId(programId);
    this.spinnerService.stopSpinner();
  }


  async budgetChanged (budgetId: number) {
    this.spinnerService.startSpinner();
    this.fundingSourceOptions = null;

    const result = await this.programService.getProgramCycleBudgets(this.formGroup.value.program);
    const options = this.getFsOptions(result, budgetId);
    this.fundingSourceOptions = this.arrayHelper.sort(options.filter((option, index) => {
      return options.findIndex(findOption => {
        return option.value === findOption.value;
      }) === index;
    }), 'label');
    if (!this.fundingSourceOptions.some(option => option.value === this.formGroup.value.fundingSource)) {
      this.formGroup.get('fundingSource').setValue(null);
    }
    this.spinnerService.stopSpinner();
  }

  private getFsOptions (
    result: GrantProgramCycleBudgetFundingSource[],
    budgetId: number
  ) {
    return this.flattenProgramBudgetFS(result)
      .filter(row => row.budgetId === budgetId)
      .map(fsOption => {
        return this.budgetService.get('cashSourceOptions')
          .find(option => option.value === fsOption.fundingSourceId);
      })
      .filter(option => !!option);
  }

  flattenProgramBudgetFS (result: GrantProgramCycleBudgetFundingSource[]) {
    return result.reduce<{ budgetId: number; fundingSourceId: number }[]>((acc, item) => {
      return [
        ...acc,
        ...item.grantProgramCycleBudgetFundingSources
      ];
    }, []);
  }

  ngOnDestroy () {
    this.programService.resetProgramCycleBudgetMap();
    this.automationService.clearFormRuleObjectMap();
    this.footerState.clearAll();
    this.sub.unsubscribe();
  }
}
