
import { Component, OnDestroy, OnInit } from '@angular/core';
import { ActivatedRoute, Router } from '@angular/router';
import { BudgetDetail, BudgetFundingSource, BudgetFundingSourceModalResponse, FundingSourceTypes } from '@core/typings/budget.typing';
import { ProcessingTypes } from '@core/typings/payment.typing';
import { ClientSettingsService } from '@features/client-settings/client-settings.service';
import { I18nService } from '@yourcause/common/i18n';
import { ConfirmationModalComponent, ModalFactory } from '@yourcause/common/modals';
import { Subscription } from 'rxjs';
import { BudgetFundingSourceModalComponent } from '../budget-funding-source-modal/budget-funding-source-modal.component';
import { BudgetService } from '../budget.service';

@Component({
  selector: 'gc-budget-funding-sources-tab',
  templateUrl: './budget-funding-sources-tab.component.html',
  styleUrls: ['./budget-funding-sources-tab.component.scss']
})
export class BudgetFundingSourcesTabComponent implements OnInit, OnDestroy {
  SOURCE_TYPES = FundingSourceTypes;
  ProcessingTypes = ProcessingTypes;
  id = this.activatedRoute.snapshot.parent.params.id;
  isNew = this.id === 'new';
  budget: BudgetDetail;
  originalBudgetFundingSources: BudgetFundingSource[];
  clientProcessText = this.i18n.translate(
    'BUDGET:textClientProcessError',
    {},
    'This budget is attached to a program requiring at least one client processed funding source. Please add a new client funded source before removing.'
  );
  yourCauseProcessText = this.i18n.translate(
    'BUDGET:textYourCauseProcessError',
    {},
    'This budget is attached to a program requiring at least one YourCause processed funding source. Please add a new YourCause funding source before removing.'
  );
  sub = new Subscription();

  constructor (
    private budgetService: BudgetService,
    private i18n: I18nService,
    private modalFactory: ModalFactory,
    private activatedRoute: ActivatedRoute,
    private clientSettingsService: ClientSettingsService,
    private router: Router
  ) {
    this.sub.add(
      this.budgetService.changesTo$('budgetEditMap').subscribe(() => {
        this.budget = this.budgetService.get('budgetEditMap')[this.id];
      })
    );
  }

  get canReserveFunds () {
    return this.clientSettingsService.get('clientSettings').reserveFunds;
  }

  ngOnInit () {
    this.originalBudgetFundingSources = [
      ...this.budget.budgetFundingSources
    ];
  }

  async addEditFundingSource (source: BudgetFundingSource, index: number) {
    const response = await this.modalFactory.open(
      BudgetFundingSourceModalComponent, {
        existingType: this.budget.budgetFundingSources.length ?
          this.budget.budgetFundingSources[0].fundingSourceType :
          null,
        existingSources: this.budget.budgetFundingSources,
        existingSource: source,
        originalExistingSources: this.originalBudgetFundingSources || []
      }
    );
    if (response) {
      if (response instanceof String) {
        this.router.navigate([response]);
      } else {
        this.budget.budgetFundingSources = this.budgetService.handleBudgetFundingSourceModalResponse(
          response as BudgetFundingSourceModalResponse,
          this.budget,
          index
        );
        this.budgetService.attachEditBudget({
          ...this.budget
        });
      }
    }
  }

  async removeFundingSource (source: BudgetFundingSource) {
    if (!this.isNew) {
      const proceed = await this.modalFactory.open(
        ConfirmationModalComponent, {
        confirmButtonText: this.i18n.translate(
          'common:btnDelete',
          {},
          'Delete'
        ),
        modalHeader: this.i18n.translate(
          'BUDGET:hdrDeleteFundingSource',
          {},
          'Delete Funding Source'
        ),
        modalSubHeader: source.fundingSourceName,
        confirmText: this.i18n.translate(
          'BUDGET:textAreYouSureDeleteFundingSource',
          {},
          'Are you sure you want to delete this funding source? This action cannot be undone.'
        )
      });
      if (proceed) {
        this.handleRemoveSource(source);
      }
    } else {
      this.handleRemoveSource(source);
    }
  }

  handleRemoveSource (source: BudgetFundingSource) {
    this.budget.budgetFundingSources = this.budget.budgetFundingSources
      .filter((fundingSource) => {
        return fundingSource.fundingSourceId !== source.fundingSourceId;
      });
    this.budgetService.attachEditBudget({
      ...this.budget
    });
  }

  ngOnDestroy () {
    this.sub.unsubscribe();
  }
}
