import { Component, OnInit } from '@angular/core';
import { ClientSettingsPages } from '@core/typings/brand.typing';
import { ColorPaletteType } from '@core/typings/branding.typing';
import { ChartService, PanelTypes, TypeaheadSelectOption, TypeSafeFormBuilder, TypeSafeFormGroup } from '@yourcause/common';
import { I18nService } from '@yourcause/common/i18n';
import { ChartOptions } from 'chart.js';
import { ClientSettingsService } from '../client-settings.service';
interface ColorFormGroup {
  primary: string;
  secondary: string;
  utility: string;
  chartPrimary: string;
  chartSecondary: string;
  chartUtility: string;
  colorPalette: ColorPaletteType;
}

@Component({
  selector: 'gc-branding-colors-page',
  templateUrl: './branding-colors-page.component.html',
  styleUrls: ['./branding-colors-page.component.scss']
})
export class BrandingColorsPageComponent implements OnInit {
  type = ClientSettingsPages.COLORS;
  PanelTypes = PanelTypes;
  sampleChartData = this.chartService.getSampleChartData(20, 1);
  sampleChartLabels = this.chartService.getSampleChartData(20, '1');
  chartOptions: ChartOptions = {
    responsive: true,
    plugins: {
      legend: {
        display: false
      },
      tooltip: {
        enabled: false
      }
    }
  };
  colorPalettteOptions: TypeaheadSelectOption[] = [{
    label: this.i18n.translate(
      'common:textCreateShades',
      {},
      'Create shades'
    ),
    value: ColorPaletteType.SHADES
  }, {
    label: this.i18n.translate(
      'common:textStandardPalette',
      {},
      'Standard palette'
    ),
    value: ColorPaletteType.STANDARD
  }, {
    label: this.i18n.translate(
      'common:textAlternatePalette',
      {},
      'Alternate palette'
    ),
    value: ColorPaletteType.ALT
  }];
  colorFormGroup: TypeSafeFormGroup<ColorFormGroup> = this.formBuilder.group({
    primary: this.clientBranding.brandPrimary,
    secondary: this.clientBranding.brandSecondary,
    utility: this.clientBranding.brandUtility,
    chartPrimary: this.clientBranding.chartPrimary || this.clientBranding.brandPrimary,
    chartSecondary: this.clientBranding.chartSecondary || this.clientBranding.brandSecondary,
    chartUtility: this.clientBranding.chartUtility || this.clientBranding.brandUtility,
    colorPalette: this.clientBranding.colorPalette || ColorPaletteType.SHADES
  });
  showAccessibilityAlert = false;
  alternativeColorText = '';
  chartBackgroundColors: string[] = [];

  constructor (
    private clientSettingsService: ClientSettingsService,
    private formBuilder: TypeSafeFormBuilder,
    private chartService: ChartService,
    private i18n: I18nService
  ) { }

  get clientBranding () {
    return this.clientSettingsService.get('clientBranding');
  }

  ngOnInit () {
    this.checkPrimaryAccessibility();
    this.setChartBackgroundColors();
  }

  checkPrimaryAccessibility () {
    const primary = this.colorFormGroup.value.primary;
    const contrast = this.chartService.getContrastByHex(primary);
    this.showAccessibilityAlert = !this.chartService.doesContrastPass(contrast);
    if (this.showAccessibilityAlert) {
      const altHex = this.chartService.getAlternativeColorHex(
        primary
      );
      this.alternativeColorText = this.i18n.translate(
        'common:textAccessibilityFriendly',
        {},
        'Accessibility friendly'
      ) + `: ${altHex}`;
    } else {
      this.alternativeColorText = '';
    }
  }

  setChartBackgroundColors () {
    const formVal = this.colorFormGroup.value;
    const primary = formVal.chartPrimary || formVal.primary;
    const secondary = formVal.chartSecondary || formVal.secondary;
    const utility = formVal.chartUtility || formVal.utility;
    const colors = [primary, secondary, utility].filter((color) => color);
    const createShades = formVal.colorPalette === ColorPaletteType.SHADES;
    const mode = formVal.colorPalette === ColorPaletteType.STANDARD ?
      'rgb' :
      'hsl';
    this.chartBackgroundColors = this.chartService.getFixedAmountOfColors(
      20,
      colors,
      createShades,
      mode
    );
  }

  colorChanged () {
    this.checkPrimaryAccessibility();
    this.setChartBackgroundColors();
    const formVal = this.colorFormGroup.value;
    this.clientSettingsService.setConfigureMap({
      ...this.clientSettingsService.configureSettingsMap,
      colors: {
        primary: formVal.primary,
        secondary: formVal.secondary,
        utility: formVal.utility,
        chartPrimary: formVal.chartPrimary,
        chartSecondary: formVal.chartSecondary,
        chartUtility: formVal.chartUtility,
        colorPalette: formVal.colorPalette
      }
    });
  }
}
