import { Component, OnInit } from '@angular/core';
import { EmailService } from '@features/system-emails/email.service';
import { EmailCommGroup, EmailFrequency, EmailNotificationType, EmailSetupPrefs } from '@features/system-emails/email.typing';
import { UserService } from '@features/users/user.service';
import { PanelTypes } from '@yourcause/common';
import { I18nService } from '@yourcause/common/i18n';

@Component({
  selector: 'gc-communication-preferences',
  templateUrl: './communication-preferences.component.html',
  styleUrls: ['./communication-preferences.component.scss']
})

export class CommunicationPreferencesComponent implements OnInit {
  PanelTypes = PanelTypes;
  preferenceEmails: EmailSetupPrefs[] = [{
    type: EmailNotificationType.ExportNotification,
    canUnsubscribe: true,
    canChangeFrequency: false,
    name: this.i18n.translate(
      'ACCOUNT:textExportStatus',
      {},
      'Export Status'
    ),
    description: this.i18n.translate(
      'ACCOUNT:textExportStatusDesc',
      {},
      'Received when you export a report from Data Hub'
    ),
    frequencyOptions: [],
    frequency: null,
    receiveEmail: true
  }, {
    type: EmailNotificationType.WorkflowLevelSummaryForClientUser,
    canUnsubscribe: true,
    canChangeFrequency: true,
    name: this.i18n.translate(
      'ACCOUNT:textWorkflowLevelReminder',
      {},
      'Workflow Level Reminder'
    ),
    description: this.i18n.translate(
      'ACCOUNT:textWorkflowLevelReminderDesc3',
      {},
      'Email is sent at the frequency set here and will contain a list of submissions in your workflow levels that may require action.'
    ),
    frequency: EmailFrequency.WEEKLY,
    frequencyOptions: [{
      label: this.i18n.translate(
        'GLOBAL:textDaily'
      ),
      value: EmailFrequency.DAILY
    }, {
      label: this.i18n.translate(
        'GLOBAL:textWeekly'
      ),
      value: EmailFrequency.WEEKLY
    }, {
      label: this.i18n.translate(
        'GLOBAL:textMonthly'
      ),
      value: EmailFrequency.MONTHLY
    }],
    receiveEmail: true
  }, {
    type: EmailNotificationType.ApplicationHasEnteredYourWorkFlowLevel,
    canUnsubscribe: true,
    canChangeFrequency: false,
    name: this.i18n.translate(
      'ACCOUNT:textImmediateWorkflowLevelApplication',
      {},
      'Immediate Workflow Level for Applications'
    ),
    description: this.i18n.translate(
      'ACCOUNT:textImmediateWorkflowLevelApplicationDesc',
      {},
      'Sent immediately when an application enters your workflow level.'
    ),
    frequency: EmailFrequency.IMMEDIATE,
    frequencyOptions: [],
    receiveEmail: false
  }, this.userService.user.clientHasNominations ? {
    type: EmailNotificationType.NominationHasEnteredYourWorkFlowLevel,
    canUnsubscribe: true,
    canChangeFrequency: false,
    name: this.i18n.translate(
      'ACCOUNT:textImmediateWorkflowLevelNomination',
      {},
      'Immediate Workflow Level for Nominations'
    ),
    description: this.i18n.translate(
      'ACCOUNT:textImmediateWorkflowLevelNominationDesc',
      {},
      'Sent immediately when a nomination enters your workflow level.'
    ),
    frequency: EmailFrequency.IMMEDIATE,
    frequencyOptions: [],
    receiveEmail: false
  } : null].filter(item => !!item);
  emailGroup: EmailCommGroup = {
    name: '',
    group: []
  };

  constructor (
    private userService: UserService,
    private emailService: EmailService,
    private i18n: I18nService
  ) { }

  get userEmail () {
    return this.userService.userEmail;
  }

  get emails () {
    return this.emailService.emails;
  }

  get communicationPrefs () {
    return this.emailService.communicationPrefs || [];
  }

  ngOnInit () {
    this.communicationPrefs.forEach((pref) => {
      const index = this.preferenceEmails.findIndex((email) => {
        return email.type === pref.emailNotificationType;
      });
      if (index > -1) {
        this.preferenceEmails[index].receiveEmail = !pref.optedOut;
        this.preferenceEmails[index].frequency = pref.frequency;
      }
    });
    this.preferenceEmails.forEach((email) => {
      this.handlePush(email);
    });
  }

  handlePush (email: EmailSetupPrefs) {
    const found = this.emails.find((item) => {
      return item.emailNotificationType === email.type;
    });
    if (found && !found.isDisabled) {
      this.emailGroup.group.push(email);
    }
  }

  async onPreferenceChange (response: {
    receiveEmail: boolean;
    frequency: EmailFrequency;
    index: number;
  }) {
    const email = this.emailGroup.group[response.index];
    const updateFrequency = email.canChangeFrequency &&
      (email.frequency !== response.frequency);
    const updateReceive = email.canUnsubscribe &&
      (email.receiveEmail !== response.receiveEmail);
    if (updateFrequency || updateReceive) {
      const optedOut = email.canUnsubscribe ?
        !response.receiveEmail :
        false;
      const frequency = email.canChangeFrequency && !optedOut ?
        response.frequency || EmailFrequency.WEEKLY :
        null;
      email.frequency = frequency;
      email.receiveEmail = !optedOut;
      const newPref = {
        emailNotificationType: email.type,
        frequency,
        optedOut
      };
      await this.emailService.updateCommunicationsPrefs(newPref, email);
    }
  }

}
