
import { Component, Input, OnDestroy, OnInit, ViewEncapsulation } from '@angular/core';
import { ActivatedRoute, Router } from '@angular/router';
import { UnsavedChangesModalComponent } from '@core/components/unsaved-changes-modal/unsaved-changes-modal.component';
import { SpinnerService } from '@core/services/spinner.service';
import { I18nService } from '@yourcause/common/i18n';
import { ModalFactory } from '@yourcause/common/modals';
import { GridsterConfig, GridsterItem } from 'angular-gridster2';
import { filter, map, Subscription } from 'rxjs';
import { DashboardsService } from '../dashboards.service';
import { Dashboards, GCDashboards } from '../dashboards.typing';
import { ManageWidgetModalComponent } from '../manage-widget-modal/manage-widget-modal.component';
import { RemoveWidgetModalComponent } from '../remove-widget-modal/remove-widget-modal.component';
import { WidgetDrilldownModalComponent } from '../widget-drilldown-modal/widget-drilldown-modal.component';
@Component({
  selector: 'gc-home-dashboard-page',
  templateUrl: './home-dashboard-page.component.html',
  styleUrls: ['./home-dashboard-page.component.scss'],
  encapsulation: ViewEncapsulation.None
})
export class HomeDashboardPageComponent implements OnInit, OnDestroy {
  @Input() isPreview = false;
  editing = false;
  gridConfig: GridsterConfig = {
    ...this.dashboardsService.defaultGridsterConfig,
    resizable: {
      enabled: this.editing
    },
    draggable: {
      enabled: this.editing
    },
    itemChangeCallback: (item) => this.widgetPositionChanged(item),
    itemResizeCallback: (item) => this.widgetPositionChanged(item)
  };
  id = +this.activatedRoute.snapshot.params.id;
  sub = new Subscription();
  widgets$ = this.dashboardsService.changesTo$('dashboardDetails').pipe(
    filter(dashboard => !!dashboard[this.id]),
    map(dashboard => {
      return dashboard[this.id].widgets;
    }));
  zeroStateTitle = this.i18n.translate(
    'common:textAddWidgetToDashboard',
    {},
    'Add widgets to your dashboard'
  );
  zeroStateDescription = this.i18n.translate(
    'common:textAddWidgetToDashboardDetail',
    {},
    'Add widgets to your dashboard by clicking "Add widget" from the dropdown actions above'
  );

  constructor (
    private dashboardsService: DashboardsService,
    private modalFactory: ModalFactory,
    private spinnerService: SpinnerService,
    private activatedRoute: ActivatedRoute,
    private router: Router,
    private i18n: I18nService
  ) {
    this.sub.add(
      this.dashboardsService.changesTo$('editing')
        .subscribe((editing) => {
          this.editing = editing;
          this.updateGridOptions();
        })
    );
  }

  get dashboard () {
    return this.dashboardsService.get('dashboardDetails')[this.id];
  }

  ngOnInit () {
    if (!this.id) {
      this.router.navigate([
        this.dashboardsService.homeRoute
      ]);

      return;
    }
  }

  updateGridOptions () {
    this.gridConfig = {
      ...this.dashboardsService.defaultGridsterConfig,
      resizable: {
        enabled: this.editing
      },
      draggable: {
        enabled: this.editing
      },
      itemChangeCallback: (item) => this.widgetPositionChanged(item),
      itemResizeCallback: (item) => this.widgetPositionChanged(item)
    };
  }

  async widgetDrilldown (e: {
    clickEvent: Dashboards.ChartClickEvent;
    widget: GCDashboards.Widget;
  }) {
    const {
      clickEvent,
      widget
    } = e;
    let tableDataFactory = widget.tableDataFactory;
    if (!tableDataFactory) {
      tableDataFactory = this.dashboardsService.getTableDataFactory(
        clickEvent.chart,
        clickEvent
      );
    }
    // clickEvent is null for stat widgets since we don't care where they clicked
    const subHeader = clickEvent ?
      this.dashboardsService.getHeaderFromClickEvent(clickEvent) :
      null;

    this.modalFactory.open(WidgetDrilldownModalComponent, {
      modalHeader: widget.name,
      drilldownValue: subHeader,
      tableDataFactory,
      drilldownColumns: widget.drilldownColumns,
      object: widget.object,
      rowsPerPage: widget.rowsPerPage
    });
  }

  widgetPositionChanged (item: GridsterItem) {
    this.dashboardsService.handleWidgetPositionChange(
      item,
      this.dashboard.oldWidgets
    );
  }

  async removeWidget (widget: GCDashboards.WidgetConfigFromApi) {
    const response = await this.modalFactory.open(
      RemoveWidgetModalComponent,
      {
        widget
      }
    );
    if (response) {
      this.spinnerService.startSpinner();
      await this.dashboardsService.removeWidget(
        widget.id,
        this.id,
        response.keepTemplate
      );
      this.spinnerService.stopSpinner();
    }
  }

  async widgetModal (existingWidget?: GCDashboards.Widget) {
    let widgetDetail: GCDashboards.WidgetConfig;
    if (existingWidget) {
      this.spinnerService.startSpinner();
      widgetDetail = await this.dashboardsService.getWidgetDetail(existingWidget.id);
      const editMap = this.dashboardsService.widgetEditMap;
      if (editMap && editMap[existingWidget.id]) {
        const updated = editMap[existingWidget.id];
        const updatedConfig: GCDashboards.WidgetConfig = JSON.parse(updated.chartConfig);
        widgetDetail.x = updatedConfig.x;
        widgetDetail.y = updatedConfig.y;
        widgetDetail.height = updatedConfig.height;
        widgetDetail.width = updatedConfig.width;
      }
      this.spinnerService.stopSpinner();
    }
    const response = await this.modalFactory.open(
      ManageWidgetModalComponent,
      {
        existingWidget: widgetDetail,
        dashboardId: this.id,
        isUpdate: true
      },
      { class: 'modal-full-size'}
    );
    if (response) {
      this.dashboardsService.setWidgetDisplay(existingWidget.id, null);
    }
  }

  async canDeactivate () {
    if (this.editing) {
      if (this.dashboardsService.shouldPromptForUnsavedChanges()) {
        this.dashboardsService.setPromptingForUnsavedChanges(true);
        const response = await this.modalFactory.open(
          UnsavedChangesModalComponent,
          {}
        );
        this.dashboardsService.setPromptingForUnsavedChanges(false);
        const canDeactivate = response !== GCDashboards.UnsavedChangesResponse.STAY;
        if (canDeactivate) {
          if (response === GCDashboards.UnsavedChangesResponse.SAVE_AND_LEAVE) {
            this.spinnerService.startSpinner();
            await this.dashboardsService.saveDashboardEdits();
            this.dashboardsService.cancelEditMode();
            this.spinnerService.stopSpinner();
          } else {
            this.dashboardsService.cancelEditMode();
          }
        }

        return canDeactivate;
      } else {
        this.dashboardsService.cancelEditMode();

        return true;
      }
    } else {
      return true;
    }
  }

  ngOnDestroy () {
    this.sub.unsubscribe();
  }
}
