import { Injectable } from '@angular/core';
import { AdHocReportingUI } from '@core/typings/ui/ad-hoc-reporting.typing';
import { Automation } from '@core/typings/ui/automation.typing';
import { ArrayHelpersService, Column, TypeaheadSelectOption } from '@yourcause/common';
import { I18nService } from '@yourcause/common/i18n';
import { LogService } from '@yourcause/common/logging';
import { NotifierService } from '@yourcause/common/notifier';
import { AttachYCState, BaseYCService } from '@yourcause/common/state';
import { EmployeeSSOFieldsResources } from './employee-sso-fields.resources';
import { EmployeeSSOFieldsState } from './employee-sso-fields.state';
import { EmployeeSSOField, EmployeeSSOFields, EmployeeSSOFieldsData } from './employee-sso-fields.typing';


@AttachYCState(EmployeeSSOFieldsState)
@Injectable({ providedIn: 'root' })
export class EmployeeSSOFieldsService extends BaseYCService<EmployeeSSOFieldsState> {

   constructor (
    private logger: LogService,
    private employeeSSOFieldsResources: EmployeeSSOFieldsResources,
    private i18n: I18nService,
    private notifier: NotifierService,
    private arrayHelper: ArrayHelpersService
  ) {
    super();
  }

  get employeeSSOFields () {
    return this.get('employeeSSOFields');
  }

  get employeeSSOFieldData () {
    return this.get('employeeSSOFieldData');
  }

  get dependentSsoFilteringOptions () {
    return this.get('dependentSsoFilteringOptions');
  }

  async getEmployeeSSOFieldsForApp (appId: number) {
    const employeeInfo = await this.employeeSSOFieldsResources.getEmployeeSSOFieldsForApp(appId);
    Object.keys(employeeInfo)
      .map(key => key as keyof EmployeeSSOFieldsData)
      .forEach((key: keyof EmployeeSSOFieldsData) => {
        if (!employeeInfo[key]) {
          employeeInfo[key] = '';
        }
      });

    return employeeInfo;
  }

  async resetEmployeeSSOFields () {
    this.set('employeeSSOFields', undefined);
    await this.setEmployeeSSOFields();
  }

  async setEmployeeSSOFields () {
    if (!this.employeeSSOFields) {
      const columns = await this.employeeSSOFieldsResources.getEmployeeSSOFields();
      columns.forEach((column) => {
        this.setColumnNameAndAttr(column);
      });
      this.set(
        'employeeSSOFields',
        this.arrayHelper.sort(columns, 'column')
      );
      this.setDependentSsoFilteringOptions();
    }
  }

  async getEmployeeSsoFieldsBulk (
    applicationIds: number[]
  ) {
    const data = await this.employeeSSOFieldsResources.getEmployeeSSOFieldsBulk(applicationIds);
    data.forEach((employeeInfo) => {
      Object.keys(employeeInfo)
        .map(key => key as keyof EmployeeSSOFieldsData)
        .forEach((key: keyof EmployeeSSOFieldsData) => {
          if (!employeeInfo[key]) {
            employeeInfo[key] = '';
          }
        });
    });

    return data;
  }

  setDependentSsoFilteringOptions () {
    const options = this.employeeSSOFields.map<TypeaheadSelectOption<Column>>((field) => {
      const label = field.name || field.columnName;

      return {
        label,
        value: {
          label,
          visible: true,
          type: 'text',
          labelOnly: true,
          prop: field.attr,
          options: [],
          dependentColumnOptions: [],
          columnName: label,
          filterOnly: true
        }
      };
    });

    this.set(
      'dependentSsoFilteringOptions',
      this.arrayHelper.sort(options, 'label')
    );
  }

  getEmployeeSSOFieldsForLookup (): Record<EmployeeSSOFields, EmployeeSSOField> {
    const fieldMap: Record<EmployeeSSOFields, EmployeeSSOField> = this.get('employeeSSOFields')
      .reduce((acc, hrField) => {
          return {
            ...acc,
            [hrField.column]: hrField
          };
      }, {} as Record<EmployeeSSOFields, EmployeeSSOField>);

      return fieldMap;
  }

  setColumnNameAndAttr (column: EmployeeSSOField) {
    column.type = 'text';
    switch (column.column) {
      case EmployeeSSOFields.DepartmentName:
        column.columnName =  this.i18n.translate(
          'GLOBAL:textDepartmentName',
          {},
          'Department name'
        );
        column.attr = 'departmentName';
        break;
      case EmployeeSSOFields.Division:
        column.columnName = this.i18n.translate(
          'GLOBAL:textDivision',
          {},
          'Division'
        );
        column.attr = 'division';
        break;
      case EmployeeSSOFields.EmployeeTypeCode:
        column.columnName = this.i18n.translate(
          'GLOBAL:textEmployeeTypeCode',
          {},
          'Employee type code'
        );
        column.attr = 'employeeTypeCode';
        break;
      case EmployeeSSOFields.Region:
        column.columnName =  this.i18n.translate(
          'GLOBAL:textRegion',
          {},
          'Region'
        );
        column.attr = 'region';
        break;
      case EmployeeSSOFields.ReportGroupName:
        column.columnName = this.i18n.translate(
          'GLOBAL:textReportGroupName',
          {},
          'Report group name'
        );
        column.attr = 'reportGroupName';
        break;
      case EmployeeSSOFields.CustomChar1:
        column.columnName = this.i18n.translate(
          'GLOBAL:textCustomColumn1',
          {},
          'Custom column 1'
        );
        column.attr = 'customChar1';
        break;
      case EmployeeSSOFields.CustomChar2:
        column.columnName = this.i18n.translate(
          'GLOBAL:textCustomColumn2',
          {},
          'Custom column 2'
        );
        column.attr = 'customChar2';
        break;
      case EmployeeSSOFields.CustomChar3:
        column.columnName = this.i18n.translate(
          'GLOBAL:textCustomColumn3',
          {},
          'Custom column 3'
        );
        column.attr = 'customChar3';
        break;
      case EmployeeSSOFields.CustomChar4:
        column.columnName = this.i18n.translate(
          'GLOBAL:textCustomColumn4',
          {},
          'Custom column 4'
        );
        column.attr = 'customChar4';
        break;
      case EmployeeSSOFields.CustomChar5:
        column.columnName = this.i18n.translate(
          'GLOBAL:textCustomColumn5',
          {},
          'Custom column 5'
        );
        column.attr = 'customChar5';
        break;
      case EmployeeSSOFields.CustomChar6:
        column.columnName = this.i18n.translate(
          'GLOBAL:textCustomColumn6',
          {},
          'Custom column 6'
        );
        column.attr = 'customChar6';
        break;
      case EmployeeSSOFields.CustomChar7:
        column.columnName = this.i18n.translate(
          'GLOBAL:textCustomColumn7',
          {},
          'Custom column 7'
        );
        column.attr = 'customChar7';
        break;
      case EmployeeSSOFields.CustomChar8:
        column.columnName = this.i18n.translate(
          'GLOBAL:textCustomColumn8',
          {},
          'Custom column 8'
        );
        column.attr = 'customChar8';
        break;
      case EmployeeSSOFields.CustomChar9:
        column.columnName = this.i18n.translate(
          'GLOBAL:textCustomColumn9',
          {},
          'Custom column 9'
        );
        column.attr = 'customChar9';
        break;
      case EmployeeSSOFields.CustomChar10:
        column.columnName = this.i18n.translate(
          'GLOBAL:textCustomColumn10',
          {},
          'Custom column 10'
        );
        column.attr = 'customChar10';
        break;
      case EmployeeSSOFields.ExternalEmployeeId2:
        column.columnName = this.i18n.translate(
          'GLOBAL:textExternalEmployeeId2',
          {},
          'External employee ID 2'
        );
        column.attr = 'externalEmployeeId2';
        break;
      case EmployeeSSOFields.FullTimeCodeId:
        column.columnName = this.i18n.translate(
          'GLOBAL:textFullTimeCodeId',
          {},
          'Full time code ID'
        );
        column.attr = 'fullTimeCodeId';
        break;
      case EmployeeSSOFields.HireDate:
         column.columnName = this.i18n.translate(
          'GLOBAL:textHireDate',
          {},
          'Hire date'
        );
        column.attr = 'hireDate';
        column.type = 'date';
        break;
      case EmployeeSSOFields.JobCodeId:
        column.columnName = this.i18n.translate(
          'GLOBAL:textJobCodeId',
          {},
          'Job code ID'
        );
        column.attr = 'jobCodeId';
        break;
      case EmployeeSSOFields.SiteCodeId:
        column.columnName = this.i18n.translate(
          'GLOBAL:textSiteCodeID',
          {},
          'Site code ID'
        );
        column.attr = 'siteCodeId';
        break;
      case EmployeeSSOFields.ExternalEmployeeId:
        column.columnName = this.i18n.translate(
          'GLOBAL:textExternalEmployeeId',
          {},
          'External employee ID'
        );
        column.attr = 'externalEmployeeId';
        break;
      default:
        column.columnName = '';
        column.attr = '';
        break;
    }
  }

  async setEmployeeSSOFieldData () {
    if (!this.employeeSSOFieldData) {
      const data = await this.employeeSSOFieldsResources.getEmployeeSSOFieldData();

      this.set('employeeSSOFieldData', data);
    }
  }

  async updateEmployeeSSOFields (payload: EmployeeSSOField[]) {
    try {
      await this.employeeSSOFieldsResources.updateEmployeeSSOFields(
        payload
      );
      this.notifier.success(this.i18n.translate(
        'CONFIG:textSuccessfullyUpdatedEmployeeSSOFields',
        {},
        'Successfully updated employee SSO fields'
      ));

      return true;
    } catch (e) {
      this.logger.error(e);
      this.notifier.error(this.i18n.translate(
        'CONFIG:textErrorUpdatingEmployeeSSOFields',
        {},
        'There was an error updating employee SSO fields'
      ));

      return false;
    }
  }

  setEmployeeColumnsForAutomation () {
    const columns: Automation.BaseObjectColumnType[] = this.employeeSSOFields.filter((column) => {
      return column.inUse;
    }).map((column) => {
      return {
        label: column.name || column.columnName,
        key: column.attr,
        comparison: Automation.Comparisons.TextIs
      };
    });
    this.set('employeeColumnsForAutomation', columns);
  }

  async setApplicantColumnsForReporting () {
    if (!this.employeeSSOFields) {
      await this.setEmployeeSSOFields();
    }
    const columns = this.employeeSSOFields.filter((column) => {
      return column.inUse;
    }).map((column) => {
      const returnObj: AdHocReportingUI.ColumnDefinition = {
        display: column.name || column.columnName,
        i18nKey: '',
        type: 'text',
        column: column.attr
      };
      if (column.type === 'date') {
        (returnObj as any).type = 'date';
        (returnObj as any).format = 'date';
      }

      return returnObj;
    });
    this.set('applicantColumnsForReporting', columns);
  }


}


