import { CdkDropList } from '@angular/cdk/drag-drop';
import { AfterViewInit, ChangeDetectorRef, Component, EventEmitter, Input, OnChanges, OnDestroy, OnInit, Output, SimpleChanges } from '@angular/core';
import { ComponentLimitMetModalComponent } from '@core/components/component-limit-met-modal/component-limit-met-modal.component';
import { SpinnerService } from '@core/services/spinner.service';
import { FooterState } from '@core/states/footer.state';
import { ReferenceFieldAPI } from '@core/typings/api/reference-fields.typing';
import { BaseApplication } from '@core/typings/application.typing';
import { ClientSettingsService } from '@features/client-settings/client-settings.service';
import { Form, FormDefinitionComponent, FormDefinitionForUi, FormFieldPasteLocation, MAX_COMPONENTS_PER_PAGE } from '@features/configure-forms/form.typing';
import { ReportFieldLogicResolver } from '@features/configure-forms/resolvers/report-field-logic.resolver';
import { FormsService } from '@features/configure-forms/services/forms/forms.service';
import { ComponentHelperService } from '@features/formio/services/component-helper/component-helper.service';
import { CreateEditReferenceFieldModalComponent } from '@features/reference-fields/create-edit-reference-field-modal/create-edit-reference-field-modal.component';
import { ReferenceFieldsService } from '@features/reference-fields/services/reference-fields.service';
import { CallMaker, CallMakerFactory, NoResultsOptions, PanelSection, PanelStyles, SimpleStringMap } from '@yourcause/common';
import { ToolboxPosition } from '@yourcause/common/expandable-toolbox';
import { I18nService } from '@yourcause/common/i18n';
import { ConfirmationModalComponent, ModalFactory } from '@yourcause/common/modals';
import { cloneDeep } from 'lodash';
import { Subscription } from 'rxjs';
import { BucketComp, FormBuilderActionEvent, FormBuilderActions } from '../form-builder.typing';
import { FormBuilderService } from '../services/form-builder/form-builder.service';

@Component({
  selector: 'gc-form-builder',
  templateUrl: './form-builder.component.html',
  styleUrls: ['./form-builder.component.scss']
})
export class FormBuilderComponent implements AfterViewInit, OnInit, OnChanges, OnDestroy {
  @Input() form: Form;
  @Input() tabIndex = 0;
  @Output() formChange = new EventEmitter<Form>();
  @Output() numberOfCompsTrackerChanged = new EventEmitter<Record<number, number>>();
  @Output() allowMoreCompsThanMaxChanged = new EventEmitter<boolean>();
  @Output() onCreateNewPage = new EventEmitter();
  @Output() formDetailClicked = new EventEmitter();
  @Output() saveClicked = new EventEmitter();
  @Output() cancelClicked = new EventEmitter();
  @Output() onTabIndexChanged = new EventEmitter<number>();

  sampleExternalFields: Partial<BaseApplication> = {
    amountRequested: null,
    currencyRequestedAmountEquivalent: null,
    referenceFields: {}
  };
  toolboxPanelSections: PanelSection<never, BucketComp[]>[] = [];
  ToolboxPosition = ToolboxPosition;
  compHiddenMap: SimpleStringMap<string> = {};
  clientSettings = this.clientSettingsService.get('clientSettings');
  hasSSO = this.clientSettings.hasSSO;
  afterInit = false;
  numberOfCompsTracker: Record<number, number> = {};
  allowMoreCompsThanMax = false;
  dragBucketId = this.formBuilderService.bucketId;
  typeAudienceMap = this.formService.typeAudienceMap;
  toolboxOpen = true;
  lastValidFormState: FormDefinitionForUi[];
  allDropLists: (CdkDropList|string)[] = [];
  PanelStyles = PanelStyles;
  searchTerm = '';
  height: string;
  noResultsOptions: NoResultsOptions = {
    leftMessage: this.i18n.translate('common:lblNoResultsFound', {}, 'No results found'),
    rightHeader: this.i18n.translate('common:textNoFieldsFound2', {}, 'No fields found'),
    rightDescription: this.i18n.translate(
      'common:textCreateFormFieldByClickingBelow',
      {},
      'Create a form field by clicking below'
    ),
    rightButtonText: this.i18n.translate('common:textCreateNewFormField', {}, 'Create new form field'),
    rigthButtonAction: () => {
      this.addFormField();
    }
  };
  sub = new Subscription();
  toolboxCallMaker: CallMaker<string, boolean> = CallMakerFactory.create(
    (_) => this.doToolboxFiltering(),
    250,
    false,
    true
  );

  constructor (
    private footerState: FooterState,
    private formService: FormsService,
    private clientSettingsService: ClientSettingsService,
    private referenceFieldService: ReferenceFieldsService,
    private spinnerService: SpinnerService,
    private modalFactory: ModalFactory,
    private componentHelper: ComponentHelperService,
    private i18n: I18nService,
    private formBuilderService: FormBuilderService,
    private cdr: ChangeDetectorRef,
    private reportFieldLogicResolver: ReportFieldLogicResolver
  ) {
    this.sub.add(this.formBuilderService.changesTo$('componentToViewOrEdit').subscribe(() => {
      const component = this.formBuilderService.componentToViewOrEdit;
      if (!!component) {
        this.viewOrEditComponent(component.compKey, component.isEdit);
      }
    }));
  }

  get currentFormDefinition () {
    return this.form.formDefinition[this.tabIndex];
  }

  async ngAfterViewInit () {
    this.reportFieldLogicResolver.resolve();
    this.setNumberOfCompsTracker();
    this.setAllowMoreCompsThanMax();
    this.assignFooterState();
    this.afterInit = true;
  }

  ngOnInit () {
    this.setCurrentFormRefFieldsInfo();
    this.setToolboxPanelSections();
    this.lastValidFormState = cloneDeep(this.form.formDefinition);
  }

  ngOnChanges (changes: SimpleChanges) {
    if (changes.form) {
      this.setCurrentFormRefFieldsInfo();
      this.setToolboxPanelSections();
    }
  }

  handleBucketReady (dropList: CdkDropList) {
    this.formBuilderService.patchDragDrop(dropList._dropListRef);
  }

  viewOrEditComponent (compKey: string, isEdit: boolean) {
    // Find the tab the component lives on
    let foundIndex: number;
    this.form.formDefinition.forEach((tab, index) => {
      this.componentHelper.eachComponent(tab.components, (comp) => {
        if (comp.key === compKey) {
          foundIndex = index;
        }
      });
    });
    if (foundIndex || foundIndex === 0) {
      this.onTabIndexChanged.emit(foundIndex);
      setTimeout(() => {
        const compClass = `apiKey_${compKey}`;
        const componentElement = document.getElementsByClassName(compClass)[0];
        if (componentElement) {
          componentElement.scrollIntoView();
        }
        if (isEdit) {
          this.handleComponentAction({
            component: this.componentHelper.findComponentByKey(this.form.formDefinition, compKey),
            action: FormBuilderActions.Edit
          });
        }
      });
    }
    this.formBuilderService.setComponentToViewOrEdit(undefined);
  }

  async addFormField () {
    const result = await this.modalFactory.open(
      CreateEditReferenceFieldModalComponent,
      {
        addingFieldToForm: true,
        forcedAudienceSelection: this.formBuilderService.currentFormBuilderFormAudience,
        saveText: this.i18n.translate(
          'common:textSaveAndConfigure',
          {},
          'Save and configure'
        ),
        secondarySaveText: this.i18n.translate(
          'common:textSaveAndViewForm',
          {},
          'Save and view form'
        )
      }
    );
    if (result) {
      const comp = await this.formBuilderService.addNewFieldToForm(result);
      if (comp) {
        this.handleFormChange(this.form.formDefinition);

        // New Fields get added at the bottom of the page, so scroll to it
        setTimeout(() => {
          window.scrollTo(0, document.body.scrollHeight);
        });
      }
    }
  }

  assignFooterState () {
    this.footerState.setPrimaryDisabled(false);
    this.footerState.setFooterState(this.footerState.FOOTER_STATES.ACTION);
    this.footerState.setPrimaryAction(async () => this.saveClicked.emit());
    this.footerState.setActionLabel(this.i18n.translate('common:btnSave', {}, 'Save'));
    this.footerState.setSecondaryLabel(this.i18n.translate('common:textFormDetail', {}, 'Form detail'));
    this.footerState.setSecondaryAction(() => this.formDetailClicked.emit());
    this.footerState.setTertiaryLabel(this.i18n.translate('common:textAddNewFormFieldToPage', {}, 'Add new form field to page'));
    this.footerState.setTertiaryAction(() => this.addFormField());
    this.footerState.setCancelLabel(this.i18n.translate('common:btnCancel', {}, 'Cancel'));
    this.footerState.setCancelAction(() => this.cancelClicked.emit());
  }

  async doToolboxFiltering () {
    this.setToolboxPanelSections();

    return true;
  }

  setToolboxPanelSections () {
    this.toolboxPanelSections = this.formBuilderService.getToolboxPanelSections(
      this.form.formType,
      this.formBuilderService.currentFormBuilderFormAudience,
      this.searchTerm
    );
  }

  onSearchTermChange (searchTerm: string) {
    this.searchTerm = searchTerm;
    this.toolboxCallMaker.invoke(() => searchTerm);
    if (!this.toolboxOpen) {
      this.toolboxOpen = true;
    }
  }

  setCurrentFormRefFieldsInfo () {
    const components = this.componentHelper.getAllComponents(this.form.formDefinition);
    this.referenceFieldService.setCurrentFormRefFieldsInfo(components);
  }

  handleDropListsChanged (dropLists: Iterable<CdkDropList|string>) {
    this.allDropLists = [...dropLists];
    this.cdr.markForCheck();
  }

  async handleComponentAction (
    event: FormBuilderActionEvent
  ) {
    let emitChange = false;
    switch (event.action) {
      case FormBuilderActions.Copy:
        this.handleCopy(event.component);
        break;
      case FormBuilderActions.Cut:
        this.handleCut(event.component);
        break;
      case FormBuilderActions.Paste_Below:
        emitChange = await this.handlePaste(
          event.component,
          FormFieldPasteLocation.BELOW
        );
        break;
      case FormBuilderActions.Paste_Above:
        emitChange = await this.handlePaste(
          event.component,
          FormFieldPasteLocation.ABOVE
        );
        break;
      case FormBuilderActions.Paste_Into_Container:
        emitChange = await this.handlePaste(
          event.component,
          FormFieldPasteLocation.INSIDE_CONTAINER
        );
        break;
      case FormBuilderActions.Edit:
        emitChange = await this.handleEdit(event.component);
        break;
      case FormBuilderActions.Remove:
        emitChange = await this.handleRemove(event.component);
        break;
    }

    if (emitChange) {
      this.handleFormChange(this.form.formDefinition);
    }
  }

  async handleFormChange (
    formDefinition: FormDefinitionForUi[]
  ) {
    let preventChange = false;
    if (await this.handleDependentPicklistValidity()) {
      preventChange = true;
    } else if (await this.hasTooManyComponentsOnPage()) {
      preventChange = true;
    }
    if (preventChange) {
      // this sets the builder state to match the original formDefinition before the change was emitted
      this.form.formDefinition = this.lastValidFormState;
    } else {
      // Allowed to proceed with updating form definition
      this.lastValidFormState = cloneDeep(formDefinition);
      this.form.formDefinition = formDefinition;
    }
    this.form.formDefinition = [
      ...this.form.formDefinition
    ];

    this.setCurrentFormRefFieldsInfo();
    this.setNumberOfCompsTracker();
    this.setToolboxPanelSections();
    this.formChange.emit(this.form);
  }

  setNumberOfCompsTracker () {
    const tracker: Record<number, number> = {};
    this.form.formDefinition.forEach((tab, index) => {
      let numberOfComponents = 0;
      this.componentHelper.eachComponent(tab.components, (component) => {
        if (component.type !== 'button') {
          numberOfComponents = numberOfComponents + 1;
        }
      });
      tracker[index] = numberOfComponents;
    });
    this.numberOfCompsTracker = tracker;
    this.numberOfCompsTrackerChanged.emit(this.numberOfCompsTracker);
  }

  /* If the form already had more than 100 components to begin with */
  setAllowMoreCompsThanMax () {
    let allowedMoreCompsThanMax = false;
    Object.keys(this.numberOfCompsTracker).forEach((key) => {
      if (this.numberOfCompsTracker[+key] > MAX_COMPONENTS_PER_PAGE) {
        allowedMoreCompsThanMax = true;
      }
    });
    this.allowMoreCompsThanMax = allowedMoreCompsThanMax;
    this.allowMoreCompsThanMaxChanged.emit(this.allowMoreCompsThanMax);
  }

  async hasTooManyComponentsOnPage (): Promise<boolean> {
    let preventChange = false;
    if (!this.allowMoreCompsThanMax) {
      let numberOfComps = 0;
      this.componentHelper.eachComponent(this.currentFormDefinition.components, (comp) => {
        if (comp.type !== 'button') {
          numberOfComps = numberOfComps + 1;
        }
      });
      if (numberOfComps > MAX_COMPONENTS_PER_PAGE) {
        const newPage = await this.modalFactory.open(
          ComponentLimitMetModalComponent,
          {}
        );
        if (newPage) {
          this.onCreateNewPage.emit();
        }
        preventChange = true;
      }
    }

    return preventChange;
  }

  async handleEdit (
    component: FormDefinitionComponent
  ): Promise<boolean> {
    const updatedComponent = await this.formBuilderService.editComponentModal(
      component,
      this.form.id,
      this.form.formDefinition
    );
    if (updatedComponent) {
      this.componentHelper.removeOrReplaceFormComponent(
        component,
        this.currentFormDefinition,
        updatedComponent
      );

      return true;
    }

    return false;
  }

  handleCopy (
    component: FormDefinitionComponent
  ) {
    this.formBuilderService.handleCopyComponent(component);
  }

  handleCut (
    component: FormDefinitionComponent
  ) {
    this.formBuilderService.handleCutComponent(component, this.currentFormDefinition);

    this.form.formDefinition = [...this.form.formDefinition];
  }

/**
 *
 * @param relativeComponent component above, below, or containing the pasted component
 * @param pasteLocation where the pasted component should be insterted
 * @returns boolean for whether the component was successfully copied (saved in database if needed) and pasted
 */
  async handlePaste (
    relativeComponent: FormDefinitionComponent,
    pasteLocation: FormFieldPasteLocation
  ): Promise<boolean> {
    const componentToPaste = this.componentHelper.prepComponentForPaste(
      this.formBuilderService.copiedComponent,
      this.form.formDefinition
    );

    const refComponents = this.componentHelper.getRefFieldComponentsToCopy(
      componentToPaste,
      this.form.formDefinition
    );
    let proceed = false;
    if (refComponents.length > 0) {
      proceed = await this.confirmReferenceFieldPaste(
        refComponents,
        componentToPaste,
        relativeComponent,
        pasteLocation
      );
    } else {
      this.formBuilderService.handlePasteComponent(
        relativeComponent,
        this.currentFormDefinition,
        pasteLocation
      );
      proceed = true;
    }

    return proceed;
  }

  async handleRemove (
    component: FormDefinitionComponent
  ): Promise<boolean> {
    const result = await this.modalFactory.open(
      ConfirmationModalComponent,
      {
        modalHeader: this.i18n.translate(
          'CONFIG:hdrRemoveComponent',
          {},
          'Remove Component'
        ),
        modalSubHeader: component.label,
        confirmText: this.i18n.translate(
          'CONFIG:textAreYouSureRemoveComponent',
          {},
          'Are you sure you want to remove this component'
        ),
        confirmButtonText: this.i18n.translate(
          'common:textRemove',
          {},
          'Remove'
        )
      }
    );

    if (result) {
      this.componentHelper.removeOrReplaceFormComponent(
        component,
        this.currentFormDefinition
      );

      return true;
    }

    return false;
  }

  async confirmReferenceFieldPaste (
    refComponents: FormDefinitionComponent[],
    pastedComponent: FormDefinitionComponent,
    relativeComponent: FormDefinitionComponent,
    pasteLocation: FormFieldPasteLocation
  ): Promise<boolean> {
    let singleField: ReferenceFieldAPI.ReferenceFieldDisplayModel;
    if (refComponents.length === 1) {
      const key = this.componentHelper.getRefFieldKeyFromCompType(
        refComponents[0].type
      );
      singleField = this.referenceFieldService.getReferenceFieldByKey(
        key
      );
    }
    const result = await this.modalFactory.open(
      ConfirmationModalComponent,
      {
        modalHeader: this.i18n.translate(
          singleField ?
            'FORMS:hdrPasteFormField' :
            'FORMS:hdrPasteFormFields',
          {},
          singleField ?
            'Paste Form Field' :
            'Paste Form Fields'
        ),
        modalSubHeader: singleField?.name || '',
        confirmText: singleField ?
          this.i18n.translate(
            'GLOBAL:textAreYouSurePasteRefField',
            {},
            'Are you sure you want to paste the form field? Pasting a copy of the form field will create a new form field with the same configuration.'
          ) :
          this.i18n.translate(
            'GLOBAL:textAreYouSurePasteFormFields',
            {},
            'Are you sure you want to paste the form fields? This will create new form fields with the same configurations.'
          ),
        confirmButtonText: this.i18n.translate(
          'common:textPaste',
          {},
          'Paste'
        )
      }
    );

    if (result) {
      this.spinnerService.startSpinner();
      await this.formBuilderService.proceedWithPasteRefFields(
        refComponents,
        pastedComponent,
        relativeComponent,
        this.currentFormDefinition,
        pasteLocation
      );
      this.spinnerService.stopSpinner();

      return true;
    }

    return false;
  }

  async handleDependentPicklistValidity () {
    // check that all child comps have parent on form
    // if child without parent, return both and open modal
    let preventChange = false;
    const invalidObj = this.referenceFieldService.getDependencyValidationFromFormDef(
      this.currentFormDefinition
    );
    if (invalidObj.parentPicklist) {
      // if invalid, we pop modal saying child comps exist
      await this.modalFactory.open(
        ConfirmationModalComponent,
        {
          modalHeader: this.i18n.translate(
            'FORMS:hdrCannotRemovePicklist',
            {},
            'Cannot Remove Picklist'
          ),
          confirmButtonText: this.i18n.translate(
            'common:textClose',
            {},
            'Close'
          ),
          modalSubHeader: invalidObj.parentPicklist.name,
          confirmText: this.i18n.translate(
            'FORMS:textUnableToRemoveParentPicklist',
            {
              childPicklistName: invalidObj.childPicklist.name
            },
            'The picklist you are trying to remove is a parent picklist of the child __childPicklistName__. Because of this dependency you must remove the child picklist before removing the parent.'
          )
        }
      );
      preventChange = true;
    }

    return preventChange;
  }

  ngOnDestroy () {
     // clear entire picklist map
    this.referenceFieldService.resetParentPicklistValueMap();
    this.footerState.clearAll();
  }
}
