/* eslint-disable @typescript-eslint/no-non-null-assertion */
import { coerceElement } from '@angular/cdk/coercion';
import { CdkDragDrop, DragRef, DropListRef, moveItemInArray, transferArrayItem } from '@angular/cdk/drag-drop';
import { Injectable } from '@angular/core';
import { SpinnerService } from '@core/services/spinner.service';
import { ReferenceFieldAPI } from '@core/typings/api/reference-fields.typing';
import { AdHocReportingUI } from '@core/typings/ui/ad-hoc-reporting.typing';
import { ReferenceFieldsUI } from '@core/typings/ui/reference-fields.typing';
import { ClientSettingsService } from '@features/client-settings/client-settings.service';
import { BuilderConfig, BuilderConfigGroupType, ComponentTabIndexMap, FormAudience, FormDefinitionComponent, FormDefinitionForUi, FormFieldPasteLocation, FormTypes } from '@features/configure-forms/form.typing';
import { EmployeeSSOFieldsService } from '@features/employee-sso-fields/employee-sso-fields.service';
import { ComponentConfigurationModalComponent } from '@features/formio/component-configuration/component-configuration-modal/component-configuration-modal.component';
import { ComponentHelperService, EMPLOYEE_SSO_TYPE_PREFIX } from '@features/formio/services/component-helper/component-helper.service';
import { FormHelperService } from '@features/formio/services/form-helper/form-helper.service';
import { ReferenceFieldsService, REF_COMPONENT_TYPE_PREFIX } from '@features/reference-fields/services/reference-fields.service';
import { ArrayHelpersService, DragAndDropItem, Panel, PanelSection, ValueComparisonService } from '@yourcause/common';
import { I18nService } from '@yourcause/common/i18n';
import { LogService } from '@yourcause/common/logging';
import { ModalFactory } from '@yourcause/common/modals';
import { NotifierService } from '@yourcause/common/notifier';
import { AttachYCState, BaseYCService } from '@yourcause/common/state';
import { GuidService } from '@yourcause/common/utils';
import { FormBuilderState } from '../../form-builder.state';
import { BucketComp, ComponentToViewOrEdit, FormBuilderDropEvent } from '../../form-builder.typing';

@AttachYCState(FormBuilderState)
@Injectable({ providedIn: 'root' })
export class FormBuilderService  extends BaseYCService<FormBuilderState> {
  readonly bucketId = this.formHelperService.componentBucketIdPrefix +
    '-' +
    this.guidService.short();

  constructor (
    private clientSettingsService: ClientSettingsService,
    private employeeSSOFieldsService: EmployeeSSOFieldsService,
    private arrayHelperService: ArrayHelpersService,
    private modalFactory: ModalFactory,
    private formHelperService: FormHelperService,
    private guidService: GuidService,
    private componentHelper: ComponentHelperService,
    private referenceFieldService: ReferenceFieldsService,
    private i18n: I18nService,
    private valueComparisonService: ValueComparisonService,
    private logger: LogService,
    private notifier: NotifierService,
    private spinnerService: SpinnerService
  ) {
    super();
  }

  get copiedComponent () {
    return this.get('copiedComponent');
  }

  get currentFormBuilderFormAudience () {
    return this.get('currentFormBuilderFormAudience');
  }

  get currentFormBuilderFormId () {
    return this.get('currentFormBuilderFormId');
  }

  get currentFormBuilderDefinition () {
    return this.get('currentFormBuilderDefinition');
  }

  get currentFormBuilderIndex () {
    return this.get('currentFormBuilderIndex');
  }

  get openQuickAddModal () {
    return this.get('openQuickAddModal');
  }

  get inFormBuilder () {
    return this.get('inFormBuilder');
  }

  get componentToViewOrEdit () {
    return this.get('componentToViewOrEdit');
  }

  /**
   * Sets the "componentToViewOrEdit" attribute on the state
   *
   * @param value: Value to set
   */
  setComponentToViewOrEdit (value: ComponentToViewOrEdit) {
    this.set('componentToViewOrEdit', value);
  }
  /**
   * Set open quick add modal
   */
  setOpenQuickAddModal (openQuickAddModal: boolean) {
    this.set('openQuickAddModal', openQuickAddModal);
  }

  /**
   * Sets the current form builder definition
   *
   * @param definition: form definition
   */
  setCurrentFormBuilderDefinition (definition: FormDefinitionForUi[]) {
    this.set('currentFormBuilderDefinition', definition);
  }

  /**
   * Sets the current form builder audience
   *
   * @param audience: form audience
   */
  setCurrentFormBuilderFormAudience (
    audience: FormAudience
  ) {
    this.set('currentFormBuilderFormAudience', audience);
  }

  /**
   * Sets the current form builder index
   *
   * @param index: form builder tab index
   */
  setCurrentFormBuilderIndex (index: number) {
    this.set('currentFormBuilderIndex', index);
  }

  /**
   * Sets the current form builder form ID
   *
   * @param formId: form id
   */
  setCurrentFormBuilderFormId (formId: number) {
    this.set('currentFormBuilderFormId', formId);
  }

  /**
   * Sets the inFormBuilder attr
   *
   * @param inFormBuilder: are we in the form builder?
   */
  setInFormBuilder (inFormBuilder: boolean) {
    this.set('inFormBuilder', inFormBuilder);
  }

  /**
   * Sets the copiedComponent attr
   *
   * @param component: Copied component to set
   */
  setCopiedComponent (component: FormDefinitionComponent) {
    this.set('copiedComponent', component);
  }

  resetCopiedComponent () {
    this.setCopiedComponent(null);
  }

  /**
   * Returns the component tab index map
   *
   * @param tabs form definition tabs
   * @returns a map with component key on the left, tabIndex & refKey on the right
   */
  getComponentTabIndexMap (tabs: FormDefinitionForUi[]) {
    const map: ComponentTabIndexMap = {};
    tabs.forEach((tab, index) => {
      this.componentHelper.eachComponent(tab.components, (component) => {
        map[component.key] = {
          tabIndex: index,
          refKey: this.componentHelper.getRefFieldKeyFromCompType(
            component.type
          )
        };
      }, true);
    });

    return map;
  }

  /**
   * Given the builder config group type, this returns the panels for that type
   *
   * @param type: builder config group type
   * @param formType: Form Type
   * @param formAudience: Form Audience
   * @param searchTerm: the search term to filter by
   * @returns the panels for the given group type
   */
  getPanelsByType (
    type: BuilderConfigGroupType,
    formType: FormTypes,
    formAudience: FormAudience,
    searchTerm: string
  ): Panel<never, BucketComp[]>[] {
    switch (type) {
      case BuilderConfigGroupType.Capture:
        return [
          this.getReferenceFieldsPanel(true, formType, formAudience, searchTerm),
          this.getStandardPanel(formType, formAudience, searchTerm)
        ].filter((item) => !!item);
      case BuilderConfigGroupType.Display:
        return [
          this.getSsoPanel(formAudience, searchTerm),
          this.getReferenceFieldsPanel(false, formType, formAudience, searchTerm),
          this.getReportPanel(formType, formAudience, searchTerm),
          this.getCustomContentPanel(formAudience, searchTerm)
        ].filter((item) => !!item);
      case BuilderConfigGroupType.Layout:
        return [
          this.getLayoutPanel(formAudience, searchTerm)
        ].filter((item) => !!item);
      case BuilderConfigGroupType.OnForm:
        return [
          this.getOnFormPanel(searchTerm)
        ].filter((item) => !!item);
    }
  }

  /**
   * Gets the toolbox panel sections for the form builder
   *
   * @param formType: Form Type
   * @param formAudience: Form Audience
   * @param searchTerm: the search term to filter by
   * @returns the toolbox panel sections
   */
  getToolboxPanelSections (
    formType: FormTypes,
    formAudience: FormAudience,
    searchTerm: string
  ): PanelSection<never, BucketComp[]>[] {
    const groups = this.getBuilderGroups();

    return Object.keys(groups).map<PanelSection<never, BucketComp[]>>((key) => {
      const type = +key as BuilderConfigGroupType;
      const panels = this.getPanelsByType(type, formType, formAudience, searchTerm);
      const numberOfComponents = panels.reduce((acc, panel) => {
        const numberOfComps = panel.context.length;

        return acc + numberOfComps;
      }, 0);

      return {
        name: groups[type] + this.getFilterNumberToAppend(searchTerm, numberOfComponents),
        panels,
        open: type === BuilderConfigGroupType.Capture
      };
    }).filter((item) => {
      return item.panels.length > 0;
    });
  }

  /**
   * Returns a map of the builder group config and it's label
   *
   * @returns the Builder Group Config
   */
  getBuilderGroups (): BuilderConfig {
    return {
      [BuilderConfigGroupType.Layout]: this.i18n.translate('common:hdrLayout', {}, 'Layout'),
      [BuilderConfigGroupType.Capture]: this.i18n.translate('common:hdrCapture', {}, 'Capture'),
      [BuilderConfigGroupType.Display]: this.i18n.translate('FORMS:textDisplay', {}, 'Display'),
      [BuilderConfigGroupType.OnForm]: this.i18n.translate('common:hdrCurrentlyOnForm', {}, 'Currently on form')
    };
  }

  /**
   * Gets the filter number string we append to the groups
   *
   * @param searchTerm: Search Term we are filtering by
   * @param numberOfResults: Number of results in the group
   * @returns the string to append to the group
   */
  getFilterNumberToAppend (
    searchTerm: string,
    numberOfResults: number
  ) {
    return !!searchTerm ? ` (${numberOfResults})` : '';
  }

  /**
   * Checks if the array has any visible components
   *
   * @param components: Components to check
   * @returns if there are visible components to show
   */
  hasVisibleComponents (components: BucketComp[]) {
    return components.some((comp) => {
      return !comp.hidden;
    });
  }

  /**
   * Get Panel Obj for Hidden Panels
   *
   * For On Form Panel & Layout Panel
   * The panel itself is hidden because there's only 1 in the group,
   * So it's visible by just showing the section header
   * In this case, we pass a unique ID so it can be differentiated from other panels like this
   *
   * @param type
   * @param components
   * @returns
   */
  getHiddenPanelObj (
    type: BuilderConfigGroupType,
    components: BucketComp[]
  ) {
    return {
      name: '',
      description: '',
      icon: '',
      iconClass: '',
      context: components,
      uniqueId: `${type}`
    };
  }

  /**
   * Returns the layout panel
   *
   * @param formAudience: Form Audience
   * @param searchTerm: the search term to filter by
   * @returns the layout panel
   */
  getLayoutPanel (
    formAudience: FormAudience,
    searchTerm: string
  ): Panel<never, BucketComp[]> {
    const components = this.getLayoutComponents(formAudience, searchTerm);
    if (this.hasVisibleComponents(components)) {
      return this.getHiddenPanelObj(BuilderConfigGroupType.Layout, components);
    }

    return null;
  }

  /**
   * Returns the layout components
   *
   * @param formAudience: Form audience
   * @param searchTerm: the search term to filter by
   * @returns the layout components
   */
  getLayoutComponents (
    formAudience: FormAudience,
    searchTerm: string
  ): BucketComp[] {
    const layoutComponents: BucketComp[] = [{
      key: 'columns',
      label: this.i18n.translate(
        'common:textColumns',
        {},
        'Columns'
      ),
      name: 'Columns',
      type: 'columns',
      fieldAudience: formAudience,
      icon: this.getFieldIcon('columns'),
      tooltip: this.getFieldTooltip('columns'),
      isReferenceField: false
    }, {
      key: 'fieldset',
      label: this.i18n.translate(
        'common:textFieldSet',
        {},
        'Field set'
      ),
      name: 'Field Set',
      type: 'fieldset',
      icon: this.getFieldIcon('fieldset'),
      tooltip: this.getFieldTooltip('fieldset'),
      fieldAudience: formAudience,
      isReferenceField: false
    }, {
      key: 'panel',
      label: this.i18n.translate(
        'common:textPanel',
        {},
        'Panel'
      ),
      name: 'Panel',
      type: 'panel',
      icon: this.getFieldIcon('panel'),
      tooltip: this.getFieldTooltip('panel'),
      fieldAudience: formAudience,
      isReferenceField: false
    }, {
      key: 'table',
      label:  this.i18n.translate(
        'FORMS:textTable',
        {},
        'Table'
      ),
      name: 'Table',
      type: 'table',
      icon: this.getFieldIcon('table'),
      tooltip: this.getFieldTooltip('table'),
      fieldAudience: formAudience,
      isReferenceField: false
    }, {
      key: 'well',
      label: this.i18n.translate(
        'common:textWell',
        {},
        'Well'
      ),
      name: 'Well',
      type: 'well',
      icon: this.getFieldIcon('well'),
      tooltip: this.getFieldTooltip('well'),
      fieldAudience: formAudience,
      isReferenceField: false
    }].filter((comp) => {
      return this.valueComparisonService.compareTerms(searchTerm, [comp.label]);
    });

    return this.arrayHelperService.sort(layoutComponents, 'label');
  }

  /**
   * Gets the Refrence Fields Panel
   *
   * @param isCapture: is for capture?
   * @param formType: Form Type
   * @param formAudience: Form Audience
   * @param searchTerm: the search term to filter by
   * @returns the reference field panel
   */
  getReferenceFieldsPanel (
    isCapture: boolean,
    formType: FormTypes,
    formAudience: FormAudience,
    searchTerm: string
  ): Panel<never, BucketComp[]> {
    let name: string;
    const appResponsesText = this.i18n.translate(
      'common:textApplicantResponses',
      {},
      'Applicant responses'
    );
    const gmResponsesText = this.i18n.translate(
      'common:textGrantManagerResponses',
      {},
      'Grant manager responses'
    );
    if (isCapture) {
      const showApplicantForCapture = formAudience === FormAudience.APPLICANT;
      name = showApplicantForCapture ? appResponsesText : gmResponsesText;

    } else {
      const showManagerForDisplay = formAudience === FormAudience.APPLICANT;
      name = showManagerForDisplay ? gmResponsesText : appResponsesText;
    }
    const components = this.getReferenceFieldComponents(isCapture, formType, formAudience, searchTerm);
    if (this.hasVisibleComponents(components)) {
      return {
        name: name + this.getFilterNumberToAppend(searchTerm, components.length),
        description: '',
        icon: '',
        iconClass: '',
        context: components
      };
    }

    return null;
  }

  /**
   * Returns the reference field components
   *
   * @param isCapture: is for capture?
   * @param formType: Form Type
   * @param searchTerm: the search term to filter by
   * @param formAudience: Form Audience
   * @returns the reference field components
   */
  getReferenceFieldComponents (
    isCapture: boolean,
    formType: FormTypes,
    formAudience: FormAudience,
    searchTerm: string
  ): BucketComp[] {
    const fields = this.getFilteredFieldsForBuilder(isCapture, formType, formAudience, searchTerm);

    const components = fields.map<BucketComp>((field) => {
      return this.getReferenceFieldBucketComp(field);
    });

    return this.arrayHelperService.sort(components, 'name');
  }

  /**
   * Adapts the field to a bucket component
   *
   * @param field: Reference field
   * @returns the bucket component for the field
   */
  getReferenceFieldBucketComp (
    field: ReferenceFieldAPI.ReferenceFieldDisplayModel
  ) {
    const type = `${REF_COMPONENT_TYPE_PREFIX}${field.key}`;

    return {
      key: field.key,
      label: field.name,
      name: field.name,
      type,
      icon: this.getFieldIcon(type),
      tooltip: this.getFieldTooltip(type),
      fieldAudience: field.formAudience,
      isReferenceField: true,
      hidden: !this.componentIsAvailableToAdd(type, [])
    };
  }

  /**
   * Returns the filtered fields for the builder
   *
   * @param isCapture: is for capture?
   * @param formType: Form Type
   * @param formAudience: Form Audience
   * @param searchTerm: the search term to filter by
   * @returns the filtered fields
   */
  getFilteredFieldsForBuilder (
    isCapture: boolean,
    formType: FormTypes,
    formAudience: FormAudience,
    searchTerm: string
  ): ReferenceFieldAPI.ReferenceFieldDisplayModel[] {
    const isRoutingOrEligibility = [
      FormTypes.ROUTING,
      FormTypes.ELIGIBILITY
    ].includes(formType);

    return this.referenceFieldService.allReferenceFields.filter((field) => {
      let audiencePassed = false;
      if (isCapture) {
        // We capture responses based on the same audience
        audiencePassed = field.formAudience === formAudience;
      } else {
        // We display fields of the opposite audience that are single response
        audiencePassed = field.formAudience !== formAudience &&
          field.isSingleResponse;
      }

      const nonRoutingOrEligibilityFields = [
        ReferenceFieldsUI.ReferenceFieldTypes.FileUpload,
        ReferenceFieldsUI.ReferenceFieldTypes.ExternalAPI,
        ReferenceFieldsUI.ReferenceFieldTypes.Aggregate,
        ReferenceFieldsUI.ReferenceFieldTypes.Table,
        ReferenceFieldsUI.ReferenceFieldTypes.Subset
      ];
      const fieldPassed = isRoutingOrEligibility ?
        !nonRoutingOrEligibilityFields.includes(field.type) :
        true;
      const isValidType = ![
        ReferenceFieldsUI.ReferenceFieldTypes.DataPoint
      ].includes(field.type);
      const searchPassed = this.valueComparisonService.compareTerms(searchTerm, [field.name, field.key]);

      return audiencePassed && fieldPassed && isValidType && searchPassed;
    });
  }

  /**
   * Returns the standard components panel
   *
   * @param formType: Form type
   * @param formAudience: Form audience
   * @param searchTerm: the search term to filter by
   * @returns the standard panel
   */
  getStandardPanel (
    formType: FormTypes,
    formAudience: FormAudience,
    searchTerm: string
  ): Panel<never, BucketComp[]> {
    const components = this.getStandardComponents(formType, formAudience, [], searchTerm);
    if (this.hasVisibleComponents(components)) {
      return {
        name: this.i18n.translate(
          'common:hdrStandardComponents',
          {},
          'Standard components'
        ) + this.getFilterNumberToAppend(searchTerm, components.length),
        description: '',
        icon: '',
        iconClass: '',
        context: components
      };
    }

    return null;
  }

  /**
   * Returns the standard components
   *
   * @param formType: Form Type
   * @param formAudience: Form Audience
   * @param searchTerm: the search term to filter by
   * @returns the standard components
   */
  getStandardComponents (
    formType: FormTypes,
    formAudience: FormAudience,
    componentsTypesToAdd: string[],
    searchTerm: string
  ) {
    const isEligibility = formType === FormTypes.ELIGIBILITY;
    const isRouting = formType === FormTypes.ROUTING;
    const isNomination = formType === FormTypes.NOMINATION;
    const notRoutingEligibilityOrNom = !isEligibility && !isRouting && !isNomination;
    const clientSettings = this.clientSettingsService.clientSettings;
    const standardComponents: BucketComp[] = [{
      key: 'cashAmountRequested',
      label: this.i18n.translate(
        'GLOBAL:lblCashAmountRequested',
        {},
        'Cash amount requested'
      ),
      name: 'Cash amount requested',
      type: 'amountRequested',
      icon: this.getFieldIcon('amountRequested'),
      tooltip: this.getFieldTooltip('amountRequested'),
      fieldAudience: formAudience,
      isReferenceField: false,
      hidden: !this.componentIsAvailableToAdd('amountRequested', componentsTypesToAdd)
    }, !isEligibility && !isRouting ? {
      key: 'inKindAmountRequested',
      label: this.i18n.translate(
        'common:hdrInKindAmountRequested',
        {},
        'In kind amount requested'
      ),
      name: 'In-kind amount requested',
      type: 'inKindItems',
      icon: this.getFieldIcon('inKindItems'),
      tooltip: this.getFieldTooltip('inKindItems'),
      fieldAudience: formAudience,
      isReferenceField: false,
      hidden: !this.componentIsAvailableToAdd('inKindItems', componentsTypesToAdd)
    } : undefined, notRoutingEligibilityOrNom ? {
      key: 'attention',
      label: this.i18n.translate(
        'FORMS:textAttention',
        {},
        'Attention'
      ),
      name: 'Attention',
      type: 'careOf',
      fieldAudience: formAudience,
      icon: this.getFieldIcon('careOf'),
      tooltip: this.getFieldTooltip('careOf'),
      isReferenceField: false,
      hidden: !this.componentIsAvailableToAdd('careOf', componentsTypesToAdd)
    } : undefined, notRoutingEligibilityOrNom && clientSettings.paymentDesignations ? {
      key: 'designation',
      label: this.i18n.translate(
        'GLOBAL:textDesignation',
        {},
        'Designation'
      ),
      name: 'Designation',
      type: 'designation',
      icon: this.getFieldIcon('designation'),
      tooltip: this.getFieldTooltip('designation'),
      fieldAudience: formAudience,
      isReferenceField: false,
      hidden: !this.componentIsAvailableToAdd('designation', componentsTypesToAdd)
    } : undefined, notRoutingEligibilityOrNom && clientSettings.specialHandling ? {
      key: 'specialHandling',
      label: this.i18n.translate(
        'ADMIN:hdrAlternateAddress',
        {},
        'Alternate Address'
      ),
      name: 'Special handling',
      type: 'specialHandling',
      icon: this.getFieldIcon('specialHandling'),
      tooltip: this.getFieldTooltip('specialHandling'),
      fieldAudience: formAudience,
      isReferenceField: false,
      hidden: !this.componentIsAvailableToAdd('specialHandling', componentsTypesToAdd)
    } : undefined, formAudience === FormAudience.MANAGER ? {
      key: 'decision2',
      label: this.i18n.translate(
        'GLOBAL:lblDecision',
        {},
        'Decision'
      ),
      name: 'Decision',
      type: 'decision',
      fieldAudience: formAudience,
      isReferenceField: false,
      tooltip: this.getFieldTooltip('decision'),
      icon: this.getFieldIcon('decision'),
      hidden: !this.componentIsAvailableToAdd('decision', componentsTypesToAdd)
    } : undefined, formAudience === FormAudience.MANAGER ? {
      key: 'reviewerFundingRecommendation',
      label: this.i18n.translate(
        'GLOBAL:textReviewerRecommendedFundingAmount',
        {},
        'Reviewer recommended funding amount'
      ),
      name: 'Reviewer funding recommendation',
      type: 'reviewerRecommendedFundingAmount',
      fieldAudience: formAudience,
      icon: this.getFieldIcon('reviewerRecommendedFundingAmount'),
      tooltip: this.getFieldTooltip('reviewerRecommendedFundingAmount'),
      isReferenceField: false,
      hidden: !this.componentIsAvailableToAdd('reviewerRecommendedFundingAmount', componentsTypesToAdd)
    } : undefined].filter((item) => {
      return !!item &&
        this.valueComparisonService.compareTerms(searchTerm, [item.name]);
    });

    return this.arrayHelperService.sort(standardComponents, 'label');
  }

  /**
   * Checks if the given component type can be added to the form
   *
   * @param compType: Component type
   * @param componentTypesToAdd: Component types to add
   * @returns if the component is available to add to the form
   */
  componentIsAvailableToAdd (
    compType: string,
    componentTypesToAdd: string[]
  ) {
    const currentFormBuilderComps = this.componentHelper.getAllComponents(this.currentFormBuilderDefinition);
    const currentCompsPass = currentFormBuilderComps.every((comp) => {
      return comp.type !== compType;
    });
    const newCompsPass = !componentTypesToAdd ||
      componentTypesToAdd?.length === 0 ||
      componentTypesToAdd.every((type) => {
        return compType !== type;
      });

    return currentCompsPass && newCompsPass;
  }

  /**
   * Returns the SSO panel
   *
   * @param formAudience: Form Audience
   * @param searchTerm: the search term to filter by
   * @returns the sso panel
   */
  getSsoPanel (
    formAudience: FormAudience,
    searchTerm: string
  ): Panel<never, BucketComp[]> {
    if (this.clientSettingsService.clientSettings.hasSSO) {
      const components = this.getSsoComponents(formAudience, searchTerm);
      if (this.hasVisibleComponents(components)) {
        return {
          name: this.i18n.translate(
            'common:textEmployeeHRData',
            {},
            'Employee HR data'
          ) + this.getFilterNumberToAppend(searchTerm, components.length),
          description: '',
          icon: '',
          iconClass: '',
          context: components
        };
      }
    }

    return null;
  }

  /**
   * Returns the SSO components
   *
   * @param formAudience: Form Audience
   * @param searchTerm: the search term to filter by
   * @returns the SSO components
   */
  getSsoComponents (
    formAudience: FormAudience,
    searchTerm: string
  ) {
    // get SSO fields - these can be renamed/relabeled by the client
    const ssoFields = this.employeeSSOFieldsService.employeeSSOFields;
    // adapt to our model for the toolbox
    const components: BucketComp[] = ssoFields.map((field) => {
      const type = `${EMPLOYEE_SSO_TYPE_PREFIX}-${field.attr}`;

      return {
        key: field.attr,
        label: field.columnName,
        name: field.columnName,
        type,
        fieldAudience: formAudience,
        icon: this.getFieldIcon(type),
        tooltip: this.getFieldTooltip(type),
        isReferenceField: false,
        hidden: !this.componentIsAvailableToAdd(type, [])
      };
    }).filter((field) => {
      return this.valueComparisonService.compareTerms(searchTerm, [field.key, field.label]);
    });

    return this.arrayHelperService.sort(components, 'label');
  }

  /**
   * Returns the report panel
   *
   * @param formType: Form Type
   * @param formAudience: Form Audience
   * @param searchTerm: the search term to filter by
   * @returns the report panel
   */
  getReportPanel (
    formType: FormTypes,
    formAudience: FormAudience,
    searchTerm: string
  ): Panel<never, BucketComp[]> {
    const isEligibility = formType === FormTypes.ELIGIBILITY;
    const isRouting = formType === FormTypes.ROUTING;
    if (!isEligibility && !isRouting) {
      const reportFieldLabel = this.i18n.translate(
        'common:textReportField',
        {},
        'Report field'
      );
      const type = 'reportField';
      const components = [{
        key: 'report',
        label: reportFieldLabel,
        name: reportFieldLabel,
        type,
        fieldAudience: formAudience,
        icon: this.getFieldIcon(type),
        tooltip: this.getFieldTooltip(type),
        isReferenceField: false
      }].filter((item) => {
        return this.valueComparisonService.compareTerms(searchTerm, [item.label, item.key]);
      });

      if (this.hasVisibleComponents(components)) {
        return {
          name: this.i18n.translate(
            'common:textReportComponents',
            {},
            'Report components'
          ) + this.getFilterNumberToAppend(searchTerm, components.length),
          description: '',
          icon: '',
          iconClass: '',
          context: components
        };
      }
    }

    return null;
  }

  /**
   * Returns the custom content panel
   *
   * @param formAudience: Form Audience
   * @param searchTerm: the search term to filter by
   * @returns the Custom content panel
   */
  getCustomContentPanel (
    formAudience: FormAudience,
    searchTerm: string
  ): Panel<never, BucketComp[]> {
    const type = 'content';
    const components = [{
      key: 'content',
      label: this.i18n.translate(
        'common:labelContent',
        {},
        'Content'
      ),
      name: 'Content',
      type,
      icon: this.getFieldIcon(type),
      fieldAudience: formAudience,
      tooltip: this.getFieldTooltip(type),
      isReferenceField: false
    }].filter((field) => {
      return this.valueComparisonService.compareTerms(searchTerm, [field.key, field.label]);
    });
    if (this.hasVisibleComponents(components)) {
      return {
        name: this.i18n.translate(
          'common:textCustomText',
          {},
          'Custom text'
        ) + this.getFilterNumberToAppend(searchTerm, components.length),
        description: '',
        icon: '',
        iconClass: '',
        context: components
      };
    }

    return null;
  }

  /**
   * Returns the components on form panel
   *
   * @param searchTerm: the search term to filter by
   * @returns the on Form Panel
   */
  getOnFormPanel (searchTerm: string): Panel<never, BucketComp[]> {
    const components = this.getOnFormComponents(searchTerm);
    if (this.hasVisibleComponents(components)) {
      return this.getHiddenPanelObj(BuilderConfigGroupType.OnForm, components);
    }

    return null;
  }

  /**
   * Returns components already on the form
   *
   * @param searchTerm: the search term to filter by
   * @returns list of components on the form
   */
  getOnFormComponents (searchTerm: string): BucketComp[] {
    const currentFormBuilderComps = this.componentHelper.getAllComponents(this.currentFormBuilderDefinition);
    const comps = currentFormBuilderComps.filter((comp) => {
      const field = this.referenceFieldService.getReferenceFieldFromCompType(comp.type);

      return comp.type !== 'button' &&
        comp.type !== 'htmlelement' &&
        this.valueComparisonService.compareTerms(searchTerm, [comp.label, field?.key, field?.name]);
    }).map<BucketComp>((comp) => {
      return {
        key: comp.key,
        label: comp.label,
        name: comp.label,
        type: comp.type,
        icon: this.getFieldIcon(comp.type),
        tooltip: this.getFieldTooltip(comp.type),
        fieldAudience: comp.formAudience,
        notDraggable: true,
        isReferenceField: this.componentHelper.isReferenceFieldComp(comp.type),
        actions: [{
          icon: 'cog',
          tooltip: this.i18n.translate(
            'CONFIG:btnEditComponent',
            {},
            'Edit component'
          ),
          onClick: (item: DragAndDropItem) => {
            this.setComponentToViewOrEdit({
              isEdit: true,
              compType: item.type,
              compKey: item.key
            });
          }
        }, {
          icon: 'external-link',
          tooltip: this.i18n.translate(
            'common:textGoToComponent',
            {},
            'Go to component'
          ),
          onClick: (item: DragAndDropItem) => {
            this.setComponentToViewOrEdit({
              isEdit: false,
              compType: item.type,
              compKey: item.key
            });
          }
        }]
      };
    });

    return this.arrayHelperService.sort(comps, 'label');
  }

  /**
   * Gets the tooltip for the field/component in the toolbox
   *
   * @param type: Component type
   * @returns the components tooltip for the form toolbox
   */
  getFieldTooltip (type: string) {
    let returnVal = '';
    if (this.componentHelper.isReferenceFieldComp(type)) {
      const nameText = this.i18n.translate('common:lblName', {}, 'Name');
      const keyText = this.i18n.translate('common:hdrKey', {}, 'Key');
      const typeText = this.i18n.translate('common:hdrType', {}, 'Type');
      const categoryText = this.i18n.translate('GLOBAL:textCategory', {}, 'Category');
      const descriptionText = this.i18n.translate('common:textDescription', {}, 'Description');
      const field = this.referenceFieldService.getReferenceFieldFromCompType(type);
      const typeTranslated = this.referenceFieldService.getFieldTypeTranslatedWithIcon(field.type).label;
      const category = this.referenceFieldService.categoryNameMap[field.categoryId];

      returnVal = `
        <div><strong>${nameText}</strong>: ${field.name}</div>
        <div><strong>${keyText}</strong>: ${field.key}</div>
        <div><strong>${typeText}</strong>: ${typeTranslated}</div>
        <div><strong>${categoryText}</strong>: ${category}</div>
      `;

      if (field.description) {
        returnVal = returnVal +
          `<div><strong>${descriptionText}</strong>: ${field.description}</div>`;
      }
    }

    return returnVal;
  }

  /**
   * Gets the icon to display with the component in the toolbox
   *
   * @param type: Component type
   * @returns the component's icon to display on the toolbox
   */
  getFieldIcon (type: string): string {
    if (type.startsWith(EMPLOYEE_SSO_TYPE_PREFIX)) {
      return 'user';
    }
    if (this.componentHelper.isReferenceFieldComp((type))) {
      const field = this.referenceFieldService.getReferenceFieldFromCompType(type);

      return this.referenceFieldService.getFieldTypeTranslatedWithIcon(field.type).icon;
    }
    switch (type) {
      case 'content':
        return 'align-center';
      case 'reportField':
        return 'analytics';
      case 'reviewerRecommendedFundingAmount':
        return 'comments-dollar';
      case 'decision':
        return 'thumbs-up';
      case 'specialHandling':
        return 'address-card';
      case 'designation':
        return 'star';
      case 'careOf':
        return 'font';
      case 'inKindItems':
        return 'gift';
      case 'amountRequested':
        return 'money-bill-wave-alt';
      case 'well':
        return 'draw-square';
      case 'table':
        return 'table';
      case 'panel':
        return 'heading';
      case 'fieldset':
        return 'bars';
      case 'columns':
        return 'columns';
    }

    return '';
  }

  /**
   * Stores a component on the state
   *
   * @param component The component being copied
   */
  handleCopyComponent (
    component: FormDefinitionComponent
  ) {
    this.setCopiedComponent(component);
  }

  /**
   * Copies the component and removes it from the definition
   *
   * @param component Component to be cut
   * @param formDef Form definition the component is being cut from
   */
  handleCutComponent (
    component: FormDefinitionComponent,
    formDef: FormDefinitionForUi
  ) {
    this.handleCopyComponent(component);
    this.componentHelper.removeOrReplaceFormComponent(component, formDef);
  }

  /**
   * Pastes the copied component relative to `relativeComponent`
   *
   * @param relativeComponent Relative component for paste
   * @param formDef Form definition accepting pasted component
   * @param pasteLocation Where to insert the form component
   */
  handlePasteComponent (
    relativeComponent: FormDefinitionComponent,
    formDef: FormDefinitionForUi,
    pasteLocation: FormFieldPasteLocation
  ) {
    if (this.copiedComponent) {
      this.componentHelper.insertFormComponent(
        this.copiedComponent,
        formDef,
        relativeComponent,
        pasteLocation
      );
      this.setCopiedComponent(null);
    }
  }

  async proceedWithPasteRefFields (
    refComponents: FormDefinitionComponent[],
    pastedComponent: FormDefinitionComponent,
    relativeComponent: FormDefinitionComponent,
    currentFormDefinition: FormDefinitionForUi,
    pasteLocation: FormFieldPasteLocation
  ) {
    // If confirmed, create the new reference field(s) and register them
    const createdFields = await this.referenceFieldService.handleCopyComponents(
      refComponents
    );

    await this.finishPasteRefFields(
      refComponents,
      pastedComponent,
      relativeComponent,
      createdFields,
      currentFormDefinition,
      pasteLocation
    );
  }

  async finishPasteRefFields (
    refComponents: FormDefinitionComponent[],
    pastedComponent: FormDefinitionComponent,
    relativeComponent: FormDefinitionComponent,
    createdFields: ReferenceFieldsUI.FormFieldComponentCopy[],
    formDefinition: FormDefinitionForUi,
    pasteLocation: FormFieldPasteLocation
  ) {
    if (createdFields) { // create fields passed
      this.componentHelper.insertFormComponent(
        pastedComponent,
        formDefinition,
        relativeComponent,
        pasteLocation
      );
      this.componentHelper.eachComponent(formDefinition.components, (formComp) => {
        // Find the original component by key
        const found = refComponents.find((compToCopy) => {
          return formComp === compToCopy;
        });
        if (found) {
          // Find newly created key and update the type
          const newKey = createdFields.find((copy) => {
            return formComp.type === `referenceFields-${copy.oldKey}`;
          }).newKey;
          formComp.type = `referenceFields-${newKey}`;
        }
      });
      this.setCopiedComponent(null);
    }
  }

  /**
   * Adds the newly created reference field to the form
   *
   * @param modalResponse: Modal response
   */
  async addNewFieldToForm (modalResponse: ReferenceFieldsUI.ModalReturn) {
    try {
      // Create the new field
      this.spinnerService.startSpinner();
      const newField = await this.referenceFieldService.handleCreateOrUpdateField(
        null,
        modalResponse.field,
        modalResponse.tableFields,
        true
      );
      const bucketComp = this.getReferenceFieldBucketComp(newField);
      const comp = this.generateComponent(this.currentFormBuilderDefinition, bucketComp);
      this.spinnerService.stopSpinner();
      const updatedComp = await this.addNewComponent(
        comp,
        this.currentFormBuilderFormId,
        this.currentFormBuilderDefinition,
        modalResponse.isSecondarySave
      );
      if (updatedComp) {
        // Insert at the bottom of the current page
        this.componentHelper.insertFormComponent(
          updatedComp,
          this.currentFormBuilderDefinition[this.currentFormBuilderIndex]
        );
      }

      return updatedComp;
    } catch (e) {
      this.spinnerService.stopSpinner();
      this.logger.error(e);
      this.notifier.error(this.i18n.translate(
        'common:textThereWasAnErrorSaving',
        {},
        'There was an error saving'
      ));

      return null;
    }
  }

  /**
   * Adds the newly created reference field component to the form
   *
   * @param newComp: New component to add
   * @param formId: Form ID
   * @param formDef: Form Definition
   * @param skipEditModal: Skip opening the edit modal
   * @returns the adapted component
   */
  async addNewComponent (
    newComp: FormDefinitionComponent,
    formId: number,
    formDef: FormDefinitionForUi[],
    skipEditModal = false
  ) {
    const adaptedComp: FormDefinitionComponent = {
      ...newComp,
      value: this.formHelperService.getValueFromComponent(
        newComp,
        { referenceFields: {} },
        false
      )
    };
    if (!skipEditModal) {
      return this.editComponentModal(adaptedComp, formId, formDef);
    }

    return adaptedComp;
  }

  /**
   * Opens the modal to edit a component
   *
   * @param component: Component to edit
   * @param formId: Form ID
   * @param formDefinition: Form Definition
   * @returns the updated component
   */
  async editComponentModal (
    component: FormDefinitionComponent,
    formId: number,
    formDefinition: FormDefinitionForUi[]
  ): Promise<FormDefinitionComponent> {
    const updatedComponent = await this.modalFactory.open(
      ComponentConfigurationModalComponent,
      {
        component,
        formId,
        formDefinition
      }
    );

    return updatedComponent;
  }

  /**
   * Generates a form component
   *
   * @param formDef: Form Definition
   * @param bucket: Component from Bucket
   * @returns the adapted form component
   */
  generateComponent (
    formDef: FormDefinitionForUi[],
    bucket: BucketComp
  ): FormDefinitionComponent {
    const allComps = this.componentHelper.getAllComponents(formDef, true);

    const comp: FormDefinitionComponent = {
      key: bucket.key,
      label: bucket.label,
      type: bucket.type,
      placeholder: '',
      validate: {
        required: false
      },
      tooltip: '',
      title: '',
      legend: '',
      description: '',
      prefix: '',
      suffix: '',
      displayType: AdHocReportingUI.DisplayTypes.TextField
    };

    switch (bucket.type) {
      case 'panel':
      case 'fieldset':
      case 'well':
        comp.components = [];
        comp.input = false;
        break;
      case 'columns':
        comp.columns = [{
          components: [],
          width: 6,
          offset: 0,
          push: 0,
          pull: 0
        } as FormDefinitionComponent, {
          components: [],
          width: 6,
          offset: 0,
          push: 0,
          pull: 0
        } as FormDefinitionComponent],
        comp.input = false;
        break;
      case 'table':
        comp.rows = this.getEmptyGridRows(3, 3);
        comp.input = false;
        break;
      case 'content':
        comp.input = false;
        break;
    }

    comp.key = this.componentHelper.guessKey(comp, allComps);

    return comp;
  }

  /**
   * Gets a set of empty grid rows
   *
   * @param numRows: Number of rows
   * @param numCols: Number of columns
   * @returns the set of empty grid rows
   */
  getEmptyGridRows (
    numRows: number,
    numCols: number
  ): FormDefinitionComponent[][] {
    const rows: FormDefinitionComponent[][] = [];
    for (let i = 0; i < numRows; i++) {
      const cols = [];
      for (let j = 0; j < numCols; j++) {
        const component = {
          components: []
        } as FormDefinitionComponent;
        cols.push(component);
      }
      rows.push(cols);
    }

    return rows;
  }

  /**
   * Handles the drop event from @angular/cdk/drag-drop
   *
   * @param drop Drop event from CDK
   * @param formDef Full form definition
   * @returns Whether or not the def was changed
   */
  async handleComponentDrop (
    drop: CdkDragDrop<FormDefinitionComponent[]>,
    formId: number,
    formDef: FormDefinitionForUi[]
  ): Promise<FormBuilderDropEvent> {
      const isMovingFromBucket = drop.previousContainer.id === this.bucketId;
      const isMovingWithinContainer = drop.previousContainer.id === drop.container.id &&
        drop.previousIndex !== drop.currentIndex;
      const isMovingBetweenContainers = !isMovingWithinContainer && !isMovingFromBucket;
      // if moving from bucket to a container - copy from bucket to container
      if (isMovingFromBucket) {
        const bucketComp = drop.item.data as BucketComp;
        const newComp = this.generateComponent(formDef, bucketComp);
        const comp = await this.addNewComponent(newComp, formId, formDef);
        if (comp) {
          const { data } = drop.container;
          data.splice(drop.currentIndex, 0, comp);

          return FormBuilderDropEvent.NewComponent;
        }
      // if dropped within the same container
      } else if (isMovingWithinContainer) {
        // and the index has changed - move within container
        moveItemInArray(
          drop.container.data,
          drop.previousIndex,
          drop.currentIndex
        );
        // otherwise element is moved from one container (not bucket) to another - transfer item over

        return FormBuilderDropEvent.ReorderComponent;
      } else if (isMovingBetweenContainers) {
        transferArrayItem(
          drop.previousContainer.data,
          drop.container.data,
          drop.previousIndex,
          drop.currentIndex
        );

        return FormBuilderDropEvent.ReorderComponent;
      }

    // otherwise, they did not drop the component in a valid spot
    return FormBuilderDropEvent.InvalidDrop;
  }


  // Not possible to improt isInsideClientRect from @angular/cdk/drag-drop/client-rect
  isInsideClientRect (clientRect: ClientRect, x: number, y: number) {
    const { top, bottom, left, right } = clientRect;

    return y >= top && y <= bottom && x >= left && x <= right;
  }

  //#region nested patch
  // copied from https://github.com/angular/components/pull/21526/files
  patchDragDrop (dropListRef: DropListRef) {
    // A few lines of code used for debugging (saved to avoid having to re-write them)

    const self = this;

    dropListRef._getSiblingContainerFromPosition = function (
      item: DragRef,
      x: number,
      y: number
    ): DropListRef | undefined {
      const targets = [this, ...this['_siblings']];

      // Only consider targets where the drag postition is within the client rect
      // (this avoids calling enterPredicate on each possible target)
      const matchingTargets = targets.filter(ref => {
        return ref._clientRect && self.isInsideClientRect(ref._clientRect, x, y);
      });

      // Stop if no targets match the coordinates
      if (matchingTargets.length === 0) {
        return undefined;
      }

      // Order candidates by DOM hierarchy and z-index
      const orderedMatchingTargets = self.orderByHierarchy(matchingTargets);

      // The drop target is the last matching target in the list
      const matchingTarget =
        orderedMatchingTargets[orderedMatchingTargets.length - 1];

      // Only return matching target if it is a sibling
      if (matchingTarget === this) {
        return undefined;
      }

      // Can the matching target receive the item?
      if (!matchingTarget._canReceive(item, x, y)) {
        return undefined;
      }

      // Return matching target
      return matchingTarget;
    };
  }

// Order a list of DropListRef so that for nested pairs, the outer DropListRef
// is preceding the inner DropListRef. Should probably be ammended to also
// sort by Z-level.
  orderByHierarchy (refs: DropListRef[]) {
    // Build a map from HTMLElement to DropListRef
    const refsByElement: Map<HTMLElement, DropListRef> = new Map();
    refs.forEach(ref => {
      refsByElement.set(coerceElement(ref.element), ref);
    });

    // Function to identify the closest ancestor among th DropListRefs
    const findAncestor = (ref: DropListRef) => {
      let ancestor = coerceElement(ref.element).parentElement;

      while (ancestor) {
        if (refsByElement.has(ancestor)) {
          return refsByElement.get(ancestor);
        }
        ancestor = ancestor.parentElement;
      }

      return undefined;
    };

    // Node type for tree structure
    interface NodeType { ref: DropListRef; parent?: NodeType; children: NodeType[] }

    // Add all refs as nodes to the tree
    const tree: Map<DropListRef, NodeType> = new Map();
    refs.forEach(ref => {
      tree.set(ref, { ref, children: [] });
    });

    // Build parent-child links in tree
    refs.forEach(ref => {
      const parent = findAncestor(ref);

      if (parent) {
        const node = tree.get(ref);
        const parentNode = tree.get(parent);

        node!.parent = parentNode;
        parentNode!.children.push(node!);
      }
    });

    // Find tree roots
    const roots = Array.from(tree.values()).filter(node => !node.parent);

    // Function to recursively build ordered list from roots and down
    const buildOrderedList = (nodes: NodeType[], list: DropListRef[]) => {
      list.push(...nodes.map(node => node.ref));
      nodes.forEach(node => {
        buildOrderedList(node.children, list);
      });
    };

    // Build and return the ordered list
    const ordered: DropListRef[] = [];
    buildOrderedList(roots, ordered);

    return ordered;
  }

  //#regionend nested patch
}
