import { Component, Input, OnChanges, OnDestroy, OnInit, Optional } from '@angular/core';
import { AbstractControl, FormControl, FormGroup, FormGroupDirective, ValidatorFn } from '@angular/forms';
import { AddressFormGroup } from '@core/components/address-block/address-block.component';
import { ApplicationFileService } from '@core/services/application-file.service';
import { SpinnerService } from '@core/services/spinner.service';
import { BaseApplication } from '@core/typings/application.typing';
import { ApplicationAttachmentService } from '@features/application-view/application-attachments/application-attachments.service';
import { ClientSettingsService } from '@features/client-settings/client-settings.service';
import { FormBuilderService } from '@features/formio/form-builder/services/form-builder/form-builder.service';
import { GcFormRendererComponent } from '@features/formio/form-renderer/gc-form-renderer/gc-form-renderer.component';
import { ComponentHelperService } from '@features/formio/services/component-helper/component-helper.service';
import { TypeSafeFormBuilder, YcFile } from '@yourcause/common';
import { I18nService } from '@yourcause/common/i18n';
import { LogService } from '@yourcause/common/logging';
import { NotifierService } from '@yourcause/common/notifier';
import { isEqual } from 'lodash';
import { Subscription } from 'rxjs';
import { BaseFormioComponent } from '../../base/base.component';
import { FILE_LABEL, NAME_LABEL, NOTES_LABEL, REASON_LABEL, SPECIAL_HANDLING_DESC, SPECIAL_HANDLING_FIELDS_MAP, SPECIAL_HANDLING_REQUIRED_DESC } from './special-handling.constants';
interface SpecialHandlingFormGroup extends AddressFormGroup {
  handlingName?: string;
  handlingOn?: boolean;
  notes?: string;
  reason?: string;
  file?: any;
}

export interface SpecialHandling {
  name?: string;
  address1?: string;
  address2?: string;
  city?: string;
  state?: string;
  country?: string;
  postalCode?: string;
  notes?: string;
  fileUrl: string;
  reason: string;
}

export interface SpecialHandlingFromApi {
  specialHandlingName: string;
  specialHandlingAddress1: string;
  specialHandlingAddress2: string;
  specialHandlingCity: string;
  specialHandlingCountry: string;
  specialHandlingPostalCode: string;
  specialHandlingStateProvinceRegion: string;
  specialHandlingNotes: string;
  specialHandlingFileUrl: string;
  specialHandlingReason: string;
}

export const BlankSpecialHandling: SpecialHandling = {
  name: '',
  address1: '',
  address2: '',
  city: '',
  state: '',
  postalCode: '',
  country: '',
  notes: '',
  fileUrl: '',
  reason: ''
};

@Component({
  selector: 'gc-special-handling',
  templateUrl: './gc-special-handling.component.html',
  styleUrls: ['./gc-special-handling.component.scss']
})
export class GcSpecialHandlingComponent extends BaseFormioComponent<SpecialHandling> implements OnInit, OnChanges, OnDestroy {
  @Input() disabled: boolean;
  @Input() requireFileUpload: boolean;
  @Input() requireReason: boolean;
  @Input() specialHandlingInstructions: string;

  SPECIAL_HANDLING_DESC = SPECIAL_HANDLING_DESC;
  SPECIAL_HANDLING_REQUIRED_DESC = SPECIAL_HANDLING_REQUIRED_DESC;
  sub = new Subscription();
  invalidSpecialHandlingError = this.i18n.translate(
    'APPLY:textPleaseEnsureAllRequiredFieldsAreCompleted',
    {},
    'Please ensure all required fields are completed'
  );
  showSpecialHandling = this.clientSettingsService.clientSettings?.specialHandling;
  NAME_LABEL = NAME_LABEL;
  FILE_LABEL = FILE_LABEL;
  NOTES_LABEL = NOTES_LABEL;
  REASON_LABEL = REASON_LABEL;
  invalidText: string;

  constructor (
    @Optional() formGroupDir: FormGroupDirective,
    private clientSettingsService: ClientSettingsService,
    private i18n: I18nService,
    private applicationFileService: ApplicationFileService,
    private applicationAttachmentService: ApplicationAttachmentService,
    private spinnerService: SpinnerService,
    public formBuilder: TypeSafeFormBuilder,
    public formBuilderService: FormBuilderService,
    public componentHelper: ComponentHelperService,
    private logger: LogService,
    private notifier: NotifierService,
    @Optional() formRenderer: GcFormRendererComponent<BaseApplication>
  ) {
    super(formRenderer, formGroupDir, formBuilder, formBuilderService, componentHelper);
  }

  get specialHandlingGroup () {
    return this.control as FormGroup;
  }

  get handlingFromParent () {
    let handling = {} as SpecialHandling;
    if (this.parentFields) {
      handling = this.parentFields.specialHandling || {} as SpecialHandling;
    }

    return handling;
  }

  ngOnInit () {
    super.ngOnInit();
    this.setFormGroup(this.data, []);
    let previousChanges = this.specialHandlingGroup.value;
    this.sub.add(this.specialHandlingGroup.valueChanges.subscribe((changes) => {
      if (!isEqual(changes, previousChanges)) {
        if (
          !changes.handlingOn &&
          changes.handlingOn !== previousChanges.handlingOn
        ) {
          previousChanges = changes;
          this.clearFormGroupValues();
          this.kickOffUpdate();
        } else {
          previousChanges = changes;
          this.kickOffUpdate();
        }
      }
    }));

    this.prepareForTranslate();
  }

  generateFormGroup () {
    let validator: ValidatorFn[] = [];
    if (!this.inFormBuilder) {
      validator = [
        this.requiredSpecialHandlingValidator()
      ];

      if (this.comp.customValidation) {
        validator.push(this.formGroupCustomValidator());
      }
    }
    this.formGroup = this.formBuilder.group({
      [this.compKey]: this.formBuilder.group({
        address1: [this.handlingFromParent.address1 || ''],
        address2: this.handlingFromParent.address2 || '',
        city: [this.handlingFromParent.city || ''],
        stateProvRegCode: [this.handlingFromParent.state || ''],
        countryCode: [this.handlingFromParent.country || 'US'],
        postalCode: [this.handlingFromParent.postalCode || '']
      })
    }, {
      validator
    });
    const isHandlingOn = this.parentHasHandling();

    this.specialHandlingGroup.addControl('handlingOn', new FormControl(
      isHandlingOn
    ));
    this.specialHandlingGroup.removeControl('name');
    this.specialHandlingGroup.addControl('handlingName', new FormControl(
      this.handlingFromParent.name || ''
    ));
    this.specialHandlingGroup.addControl('notes', new FormControl(
      this.handlingFromParent.notes || ''
    ));
    this.specialHandlingGroup.addControl('reason', new FormControl(
      this.handlingFromParent.reason || '',
      []
    ));
    let formControlFile: YcFile = null;
    const file = this.applicationFileService.breakDownloadUrlDownToObject(
      this.handlingFromParent.fileUrl
    );
    if (file) {
      formControlFile  = new YcFile(
        file.fileName,
        null,
        this.handlingFromParent.fileUrl,
        +file.fileId
      );
    }
    this.specialHandlingGroup.addControl('file', new FormControl(
      formControlFile,
      []
    ));
  }

  emitValidity () {
    this.onValidChange.emit({
      isValid: this.specialHandlingGroup.valid,
      key: this.compKey
    });
  }

  clearFormGroupValues () {
    this.specialHandlingGroup.setValue({
      handlingOn: false,
      handlingName: '',
      address1: '',
      address2: '',
      city: '',
      stateProvRegCode: '',
      countryCode: '',
      postalCode: '',
      notes: '',
      reason: '',
      file: null
    });
  }

  async getHandlingFromFormGroup () {
    if (this.specialHandlingGroup) {
      const formVal = this.specialHandlingGroup.value;
      if (formVal.handlingOn) {
        let fileUrl: string = null;
        if (formVal.file instanceof YcFile) {
          fileUrl = await this.getFileUrl(formVal.file);
        }
        const handling: SpecialHandling = {
          name: formVal.handlingName || '',
          address1: formVal.address1 || '',
          address2: formVal.address2 || '',
          city: formVal.city || '',
          state: formVal.stateProvRegCode || '',
          country: formVal.countryCode || '',
          postalCode: formVal.postalCode || '',
          notes: formVal.notes || '',
          reason: formVal.reason || '',
          fileUrl
        };
        if (this.isEmpty(handling)) {
          return this.setDefaultHandling();
        } else {
          return handling;
        }
      }
    }

    return BlankSpecialHandling;
  }

  patchValue () {
    if (this.specialHandlingGroup && this.parentFields) {
      const handlingOn = !this.isEmpty(this.data);
      const handling = this.data || {} as SpecialHandling;
      const formValue = {
        handlingOn,
        handlingName: handling.name || '',
        address1: handling.address1 || '',
        address2: handling.address2 || '',
        city: handling.city || '',
        stateProvRegCode: handling.state || '',
        countryCode: handling.country || '',
        postalCode: handling.postalCode || '',
        notes: handling.notes || '',
        reason: handling.reason || '',
        file: this.specialHandlingGroup.value.file
      };
      if (!isEqual(formValue, this.specialHandlingGroup.value)) {
        this.specialHandlingGroup.setValue(formValue);
      }
    }
  }

  prepareForTranslate () {
    // This logic is to overwite the existing label and instructions to our new language of Alternate Address
    if (this.label === 'Special Handling') {
      this.label = 'Alternate Address';
      this.comp.label = this.label;
    }
    if (
      !this.specialHandlingInstructions ||
      this.specialHandlingInstructions === 'Select this option if this application requires payment to a special address or name. If an organization is receiving funds and has ACH set up then special handling will send payments by check and not ACH.'
    ) {
      this.specialHandlingInstructions = SPECIAL_HANDLING_DESC;
      this.comp.specialHandlingInstructions = SPECIAL_HANDLING_DESC;
    }

    this.translations['Alternate Address'] = this.translations['Alternate Address'] ||
      'Alternate Address';
    this.translations[this.specialHandlingInstructions] = this.translations[this.specialHandlingInstructions] || this.specialHandlingInstructions;
    this.translations[NOTES_LABEL] = this.translations[NOTES_LABEL] || NOTES_LABEL;
    this.translations[FILE_LABEL] = this.translations[FILE_LABEL] || FILE_LABEL;
    this.translations[REASON_LABEL] = this.translations[REASON_LABEL] || REASON_LABEL;

    if (Object.keys(this.translations).length === 0) {
      this.setDefaultTranslations();
    }
  }

  setDefaultTranslations () {
    this.translations = {
      [this.label]: this.label,
      [this.specialHandlingInstructions]: this.specialHandlingInstructions,
      [SPECIAL_HANDLING_REQUIRED_DESC]: SPECIAL_HANDLING_REQUIRED_DESC,
      ...SPECIAL_HANDLING_FIELDS_MAP
    };
  }

  setDefaultHandling () {
    if (this.parentFields?.defaultSpecialHandling) {
      const defaultHandling = this.parentFields.defaultSpecialHandling ||
        {} as SpecialHandling;
      const formValue: SpecialHandlingFormGroup = {
        handlingOn: true,
        handlingName: defaultHandling.name || '',
        address1: defaultHandling.address1 || '',
        address2: defaultHandling.address2 || '',
        city: defaultHandling.city || '',
        stateProvRegCode: defaultHandling.state || '',
        countryCode: defaultHandling.country || '',
        postalCode: defaultHandling.postalCode || '',
        notes: defaultHandling.notes || '',
        reason: defaultHandling.reason || '',
        file: ''
      };
      // Set value will mark all controls as dirty,
      // but since in this scenario they just clicked the handling on checkbox
      // they should all be pristine
      setTimeout(() => {
        Object.keys(formValue).map((key) => {
          return key as keyof SpecialHandlingFormGroup;
        }).forEach((key) => {
          const control = this.specialHandlingGroup.get(key as any);
          control.setValue(formValue[key]);
          control.markAsPristine();
        });
      });

      return defaultHandling;
    }

    return BlankSpecialHandling;
  }

  async kickOffUpdate () {
    const value = await this.getHandlingFromFormGroup();
    if (this.comp && this.parentFields) {
      this.onValueChange.emit({
        value,
        updateFormGroup: false
      });
    }
  }

  getIsValid (isOn: boolean, handling: SpecialHandling): boolean {
    if (isOn) {
      return !!(handling &&
        handling.address1 &&
        handling.city &&
        handling.state &&
        handling.country &&
        handling.postalCode);
    }

    return true;
  }

  isEmpty (handling = {} as SpecialHandling) {
    return !handling.name &&
      !handling.address1 &&
      !handling.address2 &&
      !handling.city &&
      !handling.state &&
      !handling.postalCode &&
      !handling.notes &&
      !handling.fileUrl &&
      !handling.reason;
  }

  parentHasHandling () {
    return !!(
      this.handlingFromParent &&
      this.handlingFromParent.name &&
      this.handlingFromParent.address1 &&
      this.handlingFromParent.city &&
      this.handlingFromParent.state &&
      this.handlingFromParent.country &&
      this.handlingFromParent.postalCode
    );
  }

  async getFileUrl (value: any): Promise<string> {
    if (value) {
      const ycFile: YcFile<File> = value;
      const file = ycFile.file;
      const fileName = ycFile.fileName;
      let isNewFile = true;
      if (ycFile?.fileUrl) {
        const details = this.applicationFileService.breakDownloadUrlDownToObject(
          ycFile.fileUrl
        );
        if (details.fileId) {
          isNewFile = false;
        }
      } else if (ycFile?.pendingFileUploadId) {
        isNewFile = false;
      }
      if (isNewFile)  {
        // don't upload if not auto saving
        if (
          this.parentFields.applicationId &&
          this.parentFields.applicationFormId
        ) {
          try {
            const fileUrl = await this.applicationFileService.uploadFile(
              this.parentFields.applicationId,
              this.parentFields.applicationFormId,
              ycFile.file,
              ycFile.fileName,
              null
            );
            const fileDetail = this.applicationFileService.breakDownloadUrlDownToObject(
              fileUrl
            );
            this.specialHandlingGroup.get('file').setValue(new YcFile(
              fileName,
              file,
              fileUrl,
              +fileDetail.fileId
            ));

            return fileUrl;
          } catch (e) {
            this.logger.error(e);
            this.notifier.error(this.i18n.translate(
              'common:textErrorUploadingFile',
              {},
              'There was an error uploading the file'
            ));
            this.specialHandlingGroup.get('file').setValue(null);

            return null;
          }
        }
      } else {
        return ycFile.fileUrl || ('' + ycFile.pendingFileUploadId);
      }
    }

    return null;
  }

  async downloadFile () {
    this.spinnerService.startSpinner();
    await this.applicationAttachmentService.downloadReferenceFileUpload(
      this.specialHandlingGroup.value.file
    );
    this.spinnerService.stopSpinner();
  }

  formGroupCustomValidator () {
    return () => {
      if (this.hasCustomValidationError) {
        return {
          handlingOn: {
            customValidationError: {
              errorMessage: this.comp?.customValidation?.result ??
                this.comp?.validate.customMessage
            }
          }
        };
      }

      return null;
    };
  }

  requiredSpecialHandlingValidator () {
    return (group: AbstractControl) => {
      const value = group.value[this.compKey];
      if (value?.handlingOn) {
        const address1 = value.address1;
        const city = value.city;
        const stateProvRegCode = value.stateProvRegCode;
        const countryCode = value.countryCode;
        const postalCode = value.postalCode;
        const reason = value.reason;
        const file = value.file;
        const requiredError = {
          required: {
            i18nKey: 'common:textThisInputIsRequired',
            defaultValue: 'This input is required'
          }
        };
        const returnVal = {} as any;
        if (!address1) {
          returnVal.address1 = requiredError;
        }
        if (!city) {
          returnVal.city = requiredError;
        }
        if (!stateProvRegCode) {
          returnVal.stateProvRegCode = requiredError;
        }
        if (!countryCode) {
          returnVal.countryCode = requiredError;
        }
        if (!postalCode) {
          returnVal.postalCode = requiredError;
        }
        if (!reason && this.requireReason) {
          returnVal.reason = requiredError;
        }
        if (!file && this.requireFileUpload) {
          returnVal.file = requiredError;
        }
        const hasErrors = Object.keys(returnVal).length > 0;

        return hasErrors ? returnVal : null;
      } else {
        return null;
      }
    };
  }

  ngOnDestroy () {
    this.sub.unsubscribe();
  }
}
