import { Injectable } from '@angular/core';
import { FormMaskingService } from '@core/services/form-masking.service';
import { PortalDeterminationService } from '@core/services/portal-determination.service';
import { AdHocReportingUI } from '@core/typings/ui/ad-hoc-reporting.typing';
import { FormDefinitionForUi, ReportFieldDataOptions } from '@features/configure-forms/form.typing';
import { ReportFieldResources } from '@features/formio/report-field.resources';
import { AdHocReportingMappingService } from '@features/reporting/services/ad-hoc-reporting-mapping.service';
import { AdHocReportingService } from '@features/reporting/services/ad-hoc-reporting.service';
import { I18nService } from '@yourcause/common/i18n';
import { LogService } from '@yourcause/common/logging';
import { NotifierService } from '@yourcause/common/notifier';
import { ComponentHelperService } from '../component-helper/component-helper.service';

@Injectable({ providedIn: 'root' })
export class ReportFieldService {

   constructor (
    private logger: LogService,
    private adHocReportingService: AdHocReportingService,
    private adHocReportingMapper: AdHocReportingMappingService,
    private reportFieldResources: ReportFieldResources,
    private i18n: I18nService,
    private portal: PortalDeterminationService,
    private componentHelper: ComponentHelperService,
    private formMaskingService: FormMaskingService,
    private notifier: NotifierService
  ) { }

  getReportFieldValue (
    reportFieldDataOptions: ReportFieldDataOptions,
    row: AdHocReportingUI.ReportResponseRow
  ) {
    // get application buckets
    const buckets = this.adHocReportingService.getBuckets('application', [], []);
    // find bucket whose property matches the report field object, get columns from that bucket
    const cols: AdHocReportingUI.ColumnImplementation[] = buckets.find((bucket) => {
      return bucket.property === reportFieldDataOptions.reportFieldObject;
    }).columns;
    // find the column that matches our report field display
    const column = cols.find((a) => {
      return a.definition.column === reportFieldDataOptions.reportFieldDisplay;
    });
    // use column def and endpoint response to return value
    const value = this.adHocReportingMapper.getValueForColumnFromRow(
      column.definition,
      row,
      this.formMaskingService.getDefaultMaskSetting()
    );

    return value;
  }

  async getReportFieldResponseFromAPI (
    portal: 'portal'|'manager',
    applicationId: number
  ) {
    if (applicationId) {
      try {
        const response = await this.reportFieldResources.getReportFieldResponse(portal, applicationId);

        return response;
      } catch (e) {
        this.logger.error(e);
        this.notifier.error(this.i18n.translate(
          'FORMS:textErrorFetchingReportFieldData',
          {},
          'There was an error fetching report field data'
        ));

        return null;
      }
    } else {
      return null;
    }
  }

  formContainsReportField (formDefinition: FormDefinitionForUi[]): boolean {
    let formContainsReportField = false;
    formDefinition.forEach((tab) => {
      this.componentHelper.eachComponent(tab.components, (comp) => {
        if (comp.type === 'reportField') {
          formContainsReportField = true;
        }
      });
    });

    return formContainsReportField;
  }

  getReportFieldResponse (applicationId: number) {
    const portal = this.portal.isManager ? 'manager' : 'portal';

    return this.getReportFieldResponseFromAPI(portal, applicationId);
  }

  getReportFieldResponseBulk (applicationIds: number[]) {
    return this.reportFieldResources.getReportFieldResponsesBulk(
      applicationIds
    );
  }

  handleReportFieldData (
    formDefinition: FormDefinitionForUi[],
    applicationId: number
  ) {
    const formHasReportField = this.formContainsReportField(formDefinition);
    if (formHasReportField) {
      return this.getReportFieldResponse(applicationId);
    }

    return null;
  }
}
