import { Component, Input, OnInit } from '@angular/core';
import { Validators } from '@angular/forms';
import { ClientSettingsService } from '@features/client-settings/client-settings.service';
import { CustomDataTableModalPages } from '@features/custom-data-tables/custom-data-tables.typing';
import { TypeSafeFormBuilder, TypeSafeFormGroup } from '@yourcause/common';
import { AnalyticsService, EventType } from '@yourcause/common/analytics';
import { I18nService } from '@yourcause/common/i18n';
import { YCModalComponent } from '@yourcause/common/modals';
import * as parse from 'papaparse';
import { InKindService } from '../in-kind.service';
import { ImportCategoryItem, InKindCategories } from '../in-kind.typing';
interface AddInKindCategoryFormGroup {
  categoryName: string;
  type: InKindCategories;
  defaultLang: string;
}

@Component({
  selector: 'gc-add-in-kind-category-modal',
  templateUrl: './add-in-kind-category-modal.component.html',
  styleUrls: ['./add-in-kind-category-modal.component.scss']
})
export class AddInKindCategoryModalComponent extends YCModalComponent<
  number|string
> implements OnInit {
  @Input() importOnly = false;
  @Input() id: number;
  @Input() context: 'add'|'edit';
  @Input() name?: string;

  modalHeader: string;
  secondaryButtonText: string;
  primaryButtonText: string;
  formGroup: TypeSafeFormGroup<AddInKindCategoryFormGroup>;
  CustomDataTableModalPages = CustomDataTableModalPages;
  page = CustomDataTableModalPages.NAME;
  primaryDisabled: boolean;
  selectedFile: File;
  hasValidated = false;
  fileIsValid = false;
  secondaryDisabled: boolean;
  parsedFile: ImportCategoryItem[];
  langKeys: string[];
  hasInternational = this.clientSettingsService.clientSettings.hasInternational;
  categoryOptions = [{
    label: this.inKindService.categoryTypeMap[InKindCategories.Goods],
    value: InKindCategories.Goods
  }, {
    label: this.inKindService.categoryTypeMap[InKindCategories.Services],
    value: InKindCategories.Services
  }, {
    label: this.inKindService.categoryTypeMap[InKindCategories.Space],
    value: InKindCategories.Space
  }, {
    label: this.inKindService.categoryTypeMap[InKindCategories.Other],
    value: InKindCategories.Other
  }];

  constructor (
    private inKindService: InKindService,
    private formBuilder: TypeSafeFormBuilder,
    private clientSettingsService: ClientSettingsService,
    private i18n: I18nService,
    private analyticsService: AnalyticsService
  ) {
    super();
  }

  get isName () {
    return this.page === CustomDataTableModalPages.NAME;
  }

  ngOnInit () {
    if (this.hasInternational) {
      let langKeys = this.clientSettingsService.get('selectedLanguages');
      if (langKeys.length === 0) {
        langKeys = [this.clientSettingsService.defaultLanguage];
      }
      this.langKeys = langKeys;
    }
    this.page = this.importOnly ?
      CustomDataTableModalPages.IMPORT :
      CustomDataTableModalPages.NAME;
    this.formGroup = this.formBuilder.group<AddInKindCategoryFormGroup>({
      categoryName: [this.name ? this.name : '', Validators.required],
      type: InKindCategories.Goods,
      defaultLang: this.clientSettingsService.defaultLanguage
    });
    this.setModalScenario();
    this.setIsDisabled();
  }

  goToImport () {
    this.page = CustomDataTableModalPages.IMPORT;
    this.setIsDisabled();
    this.setModalScenario();
  }

  goToName () {
    this.page = CustomDataTableModalPages.NAME;
    this.setIsDisabled();
    this.setModalScenario();
  }

  handlePrimaryClick () {
    if (this.context === 'add') {
      if (this.isName) {
        this.createCategoryOnly();
        this.analyticsService.emitEvent({
          eventName: 'Create in kind category',
          eventType: EventType.Click,
          extras: null
        });
      } else {
        if (this.importOnly) {
          this.import();
          this.analyticsService.emitEvent({
            eventName: 'Import in kind items',
            eventType: EventType.Click,
            extras: null
          });
        } else {
          this.createAndImport();
          this.analyticsService.emitEvent({
            eventName: 'Create in kind category and import items',
            eventType: EventType.Click,
            extras: null
          });
        }
      }
    } else {
      this.updateName();
    }
  }

  updateName () {
    this.closeModal.emit(this.formGroup.value.categoryName);
  }

  setIsDisabled () {
    if (this.isName) {
      this.primaryDisabled = this.formGroup.invalid;
      this.secondaryDisabled = this.formGroup.invalid;
    } else {
      this.primaryDisabled = !this.selectedFile ||
        !this.hasValidated ||
        !this.fileIsValid;
    }
  }

  setModalScenario () {
    if (this.context === 'add') {
      if (this.isName) {
        this.modalHeader = this.i18n.translate(
          'CONFIG:hdrCreateNewCategory',
          {},
          'Create New Category'
        );
        this.secondaryButtonText = this.i18n.translate(
          'GLOBAL:textCreateAndImportData',
          {},
          'Create and import data'
        );
        this.primaryButtonText = this.i18n.translate(
          'common:linkCreate',
          {},
          'Create'
        );
      } else {
        this.modalHeader = this.i18n.translate(
          'common:hdrImportCategoryItems',
          {},
          'Import Category Items'
        );
        if (this.importOnly) {
          this.secondaryButtonText = '';
          this.primaryButtonText = this.i18n.translate(
            'GLOBAL:textImportData',
            {},
            'Import data'
          );
        } else {
          this.secondaryButtonText = this.i18n.translate(
            'common:btnBack',
            {},
            'Back'
          );
          this.primaryButtonText = this.i18n.translate(
            'GLOBAL:textCreateAndImportData',
            {},
            'Create and import data'
          );
        }
      }
    } else {
      this.modalHeader = this.i18n.translate(
        'GLOBAL:hdrEditCategory',
        {},
        'Edit Category'
      );
      this.secondaryButtonText = '';
      this.primaryButtonText = this.i18n.translate(
        'GLOBAL:textUpdate',
        {},
        'Update'
      );
    }
  }

  downloadTemplate () {
    this.inKindService.getTemplateForDownload();
  }

  selectFileForImport () {
    const input = document.createElement('input');
    input.type = 'file';
    input.hidden = true;
    input.addEventListener('change', (event: any) => {
      const [file] = event.target.files;
      const fileReader = new FileReader();
      fileReader.addEventListener('loadend', async (loadEvent: any) => {
        this.hasValidated = false;
        this.fileIsValid = false;
        this.selectedFile = file;
        const contents: string = loadEvent.target.result;
        const parsed = parse.parse<ImportCategoryItem>(contents, {
          header: true
        });
        this.parsedFile = parsed.data;
        this.setIsDisabled();
      });
      fileReader.readAsText(file);
    });
    document.body.appendChild(input);
    input.click();
  }

  removeFile () {
    this.selectedFile = null;
    this.hasValidated = false;
    this.fileIsValid = false;
    this.parsedFile = null;
    this.setIsDisabled();
  }

  validateFile () {
    this.fileIsValid = this.inKindService.isFileValid(this.parsedFile);
    this.hasValidated = true;
    this.setIsDisabled();
  }

  async createCategoryOnly () {
    const id = await this.inKindService.createCategory(
      this.formGroup.value.categoryName,
      this.formGroup.value.type,
      this.formGroup.value.defaultLang
    );
    this.closeModal.emit(id);
  }

  async import () {
    await this.inKindService.importData(
      this.id,
      this.selectedFile
    );
    this.closeModal.emit(this.id);
  }

  async createAndImport () {
    const id = await this.inKindService.createAndImportData(
      this.formGroup.value.categoryName,
      this.formGroup.value.type,
      this.selectedFile,
      this.formGroup.value.defaultLang
    );
    this.closeModal.emit(id);
  }

  onCancel () {
    this.closeModal.emit();
  }
}
