import { Component, Input, OnChanges } from '@angular/core';
import { CurrencyService } from '@core/services/currency.service';
import { ApplicantTop10 } from '@core/typings/applicant.typing';
import { ApplicantManagerService } from '@features/applicant/applicant-manager.service';
import { ClientSettingsService } from '@features/client-settings/client-settings.service';
import { AutoTableRepository, AutoTableRepositoryFactory, ChartService, DashboardTableData, DebounceFactory, FileService, ValueComparisonService } from '@yourcause/common';
import { I18nService } from '@yourcause/common/i18n';
import { NotifierService } from '@yourcause/common/notifier';
import { ChartData, ChartOptions, ChartType, TooltipItem } from 'chart.js';

@Component({
  selector: 'gc-applicant-top-10',
  templateUrl: './applicant-top-10.component.html',
  styleUrls: ['./applicant-top-10.component.scss']
})

export class ApplicantTop10Component implements OnChanges {
  @Input() cycleIds: number[] = [];
  @Input() showDownload = false;

  key = 'APPLICANT_STATS';
  baseChart: ChartOptions = {
    responsive: true,
    plugins: {
      legend: {
        display: false
      }
    }
  };
  branding = this.clientSettingsService.get('clientBranding');
  colors = [
    this.branding.brandPrimary,
    this.branding.brandSecondary,
    this.branding.brandUtility
  ];
  defaultCurrency = this.clientSettingsService.defaultCurrency;
  awardsTotal: number;
  applicants: ApplicantTop10[] = [];
  applicantRepo: AutoTableRepository<ApplicantTop10>;
  applicantLabels: string[] = [];
  applicantData: number[] = [];
  applicantColors: string[] = [];
  applicantTableData: DashboardTableData<ApplicantTop10>[];
  applicantChartOptions = {
    ...this.baseChart,
    tooltips: {
      callbacks: {
        label: (tooltipItem: TooltipItem<ChartType>) => {
          const numberOfAwards = this.applicants[tooltipItem.dataIndex].numberOfAwards;

          return this.i18n.translate(
            'common:hdrAwards'
          ) + ': ' + numberOfAwards;
        },
        beforeLabel: (tooltipItem: TooltipItem<ChartType>, data: ChartData) => {
          return this.chartService.getLabelForTooltip(
            tooltipItem,
            data
          );
        },
        afterLabel: (tooltipItem: TooltipItem<ChartType>, data: ChartData) => {
          const amount = this.chartService.getAmountForTooltip(
            tooltipItem,
            data
          );

          return `${this.currencyService.formatMoney(amount)} (${
            this.chartService.getPercent(amount, this.awardsTotal)
          })`;
        }
      }
    }
  } as ChartOptions;

  constructor (
    private i18n: I18nService,
    private chartService: ChartService,
    private applicantManagerService: ApplicantManagerService,
    private autoTableFactory: AutoTableRepositoryFactory,
    private notifierService: NotifierService,
    private valueComparisonService: ValueComparisonService,
    private clientSettingsService: ClientSettingsService,
    private currencyService: CurrencyService,
    private fileService: FileService
  ) { }

  async ngOnChanges () {
    this.applicants = await this.applicantManagerService.getApplicantsForDashboard(
      this.cycleIds
    );
    this.setupApplicantStats();
  }

  setupApplicantStats () {
    if (!this.applicantRepo) {
      this.applicantRepo = this.autoTableFactory.create({
        key: this.key,
        columns: [],
        notifier: this.notifierService,
        rowsPerPage: 1000,
        valueComparisonService: this.valueComparisonService,
        rows: this.applicants,
        tableCSVFactory: this.getTableCsvFactory()
      });
    } else {
      this.applicantRepo.rows = this.applicants;
      this.applicantRepo.reset();
    }
    this.applicantLabels = this.applicants.map((applicant) => {
      return applicant.fullName;
    });
    this.applicantData = this.applicants.map((applicant) => {
      return applicant.awardsTotal;
    });
    this.applicantColors = this.chartService.getFixedAmountOfColors(
      this.applicants.length,
      this.colors
    );
    const totalNumber = this.applicants.reduce((acc, applicant) => {
        return acc + applicant.numberOfAwards;
      }, 0);
    this.awardsTotal = this.applicants.reduce((acc, applicant) => {
      return acc + applicant.awardsTotal;
    }, 0);
    const totalString = `${totalNumber} (${
      this.currencyService.formatMoney(this.awardsTotal)
    })`;
    const requestTotal = this.applicants.reduce((acc, applicant) => {
      return acc + (applicant.requestAmount || 0);
    }, 0);
    this.applicantTableData = [{
      columnName: this.i18n.translate(
        'common:lblApplicant'
      ),
      isNumber: false,
      isMoney: false,
      getRouterLink: (row) => {
        return row.applicantId !== 0 ?
          `/management/insights/applicants/applicant-profile/${
            row.applicantId
          }/applications` :
          undefined;
      },
      isLegendRef: true,
      key: 'fullName'
    }, {
      columnName: this.i18n.translate(
        'GLOBAL:lblRequestAmount'
      ),
      isNumber: false,
      isMoney: false,
      isMixedVal: true,
      getRouterLink: null,
      isLegendRef: false,
      key: 'requestAmount',
      getRowValue: (row) => {
        return this.currencyService.formatMoney(
          row.requestAmount
        );
      },
      total: this.currencyService.formatMoney(
        requestTotal
      )
    }, {
      columnName: this.i18n.translate(
        'common:hdrAwards'
      ),
      isNumber: false,
      isMoney: false,
      isMixedVal: true,
      getRouterLink: null,
      isLegendRef: false,
      key: 'awardNumberAndTotal',
      total: totalString
    }];
  }

  getTableCsvFactory () {
    return DebounceFactory.createSimple(() => {
      const currentRepo = this.autoTableFactory.getRepository(this.key);
      const filteredRows = currentRepo.clientSideFilteredRows as ApplicantTop10[];
      let totalRequestAmount = 0;
      let totalNumberOfAwards = 0;
      let totalAwardsAmount = 0;
      const adapted = filteredRows.map((row) => {
        totalRequestAmount = totalRequestAmount + (row.requestAmount || 0);
        totalNumberOfAwards = totalNumberOfAwards + (row.numberOfAwards || 0);
        totalAwardsAmount = totalAwardsAmount + (row.awardsTotal || 0);

        return {
          'Applicant Name': row.fullName,
          'Request Amount': row.requestAmount,
          'Number of Awards': row.numberOfAwards,
          'Awards Total': row.awardsTotal
        };
      });
      const totalsRow = {
        'Applicant Name': 'Total',
        'Request Amount': totalRequestAmount,
        'Number of Awards': totalNumberOfAwards,
        'Awards Total': totalAwardsAmount
      };

      return this.fileService.convertObjectArrayToCSVString([
        ...adapted,
        totalsRow
      ]);
    });
  }
}
