import { Component, ElementRef, OnDestroy, OnInit, ViewChild } from '@angular/core';
import { CurrencyService } from '@core/services/currency.service';
import { PolicyService } from '@core/services/policy.service';
import { SpinnerService } from '@core/services/spinner.service';
import { CyclesAPI } from '@core/typings/api/cycles.typing';
import { ProgramDropdownOptionsForInsights } from '@core/typings/program.typing';
import { AwardService } from '@features/awards/award.service';
import { ClientSettingsService } from '@features/client-settings/client-settings.service';
import { ProgramService } from '@features/programs/program.service';
import { AutoTableRepository, AutoTableRepositoryFactory, ChartService, DashboardTableData, DebounceFactory, FileService, Html2CanvasService, SelectOption, TypeaheadSelectOption, TypeSafeFormBuilder, TypeSafeFormGroup, ValueComparisonService } from '@yourcause/common';
import { I18nService } from '@yourcause/common/i18n';
import { NotifierService } from '@yourcause/common/notifier';
import { ChartData, ChartOptions, ChartType, TooltipItem } from 'chart.js';
import { Subscription } from 'rxjs';
interface AwardGroup {
  programStatusFilter: ProgramDropdownOptionsForInsights;
  cycleStatusFilter: CyclesAPI.CycleStatuses;
  programCycleFilter: number[];
}

@Component({
  selector: 'gc-award-insights-page',
  templateUrl: './award-insights-page.component.html',
  styleUrls: ['./award-insights-page.component.scss']
})
export class AwardInsightsPageComponent implements OnInit, OnDestroy {
  key = 'CYCLE_STATS';
  cycleIds: number[] = [];
  formGroup: TypeSafeFormGroup<AwardGroup>;
  programCycleOptions: SelectOption[] = [];
  currentProgramStatus: ProgramDropdownOptionsForInsights = ProgramDropdownOptionsForInsights.ACTIVE;
  currentCycleStatus: CyclesAPI.CycleStatuses = CyclesAPI.CycleStatuses.Open;
  programStatusOptions: TypeaheadSelectOption<string>[] = [{
    value: 'all',
    label: this.i18n.translate(
      'common:textAllPrograms',
      {},
      'All programs'
    )
  }, {
    value: 'active',
    label: this.i18n.translate(
      'common:lblActivePrograms',
      {},
      'Active programs'
    )
  }, {
    value: 'archived',
    label: this.i18n.translate(
      'common:lblArchivePrograms',
      {},
      'Archived programs'
    )
  }];
  cycleStatusOptions: TypeaheadSelectOption<CyclesAPI.CycleStatuses>[] = [{
    value: CyclesAPI.CycleStatuses.All,
    label: this.i18n.translate(
      'PROGRAM:textAllCycles',
      {},
      'All cycles'
    )
  }, {
    value: CyclesAPI.CycleStatuses.Open,
    label: this.i18n.translate(
      'PROGRAM:textOpenCycles',
      {},
      'Open cycles'
    )
  }, {
    value: CyclesAPI.CycleStatuses.Past,
    label: this.i18n.translate(
      'PROGRAM:textCloseCycles',
      {},
      'Closed cycles'
    )
  }];
  sub = new Subscription();
  loading = false;
  branding = this.clientSettingsService.get('clientBranding');
  defaultCurrency = this.clientSettingsService.defaultCurrency;
  canViewProgram =
    (
      this.policyService.grantApplication.canAccessApplicationManager()
    ) &&
    this.policyService.insights.canViewPrograms();

  colors = [
    this.branding.brandPrimary,
    this.branding.brandSecondary,
    this.branding.brandUtility
  ];
  baseChart: ChartOptions = {
    responsive: true,
    plugins: {
      legend: {
        display: false
      }
    }
  };
  awardsTotal: number;
  topCycles: CyclesAPI.CycleTop10[] = [];
  cycleRepo: AutoTableRepository<CyclesAPI.CycleTop10>;
  cycleLabels: string[] = [];
  cycleData: number[] = [];
  cycleColors: string[] = [];
  cycleTableData: DashboardTableData<CyclesAPI.CycleTop10>[];
  cycleChartOptions = {
    ...this.baseChart,
    tooltips: {
      callbacks: {
        label: (tooltipItem: TooltipItem<ChartType>) => {
          const cycleName = this.topCycles[tooltipItem.dataIndex].cycleName;

          return cycleName;
        },
        afterLabel: (tooltipItem: TooltipItem<ChartType>, data: ChartData) => {
          const totalAwardAmount = this.topCycles[tooltipItem.dataIndex].awardAmount;

          return [
            this.currencyService.formatMoney(
              totalAwardAmount
            )
          ];
        }
      }
    }
  };
  clientDefaultTz = this.clientSettingsService.clientSettings.defaultTimezone;
  downloadingChart = false;

  @ViewChild('chartWrapper') chartWrapperEl: ElementRef<HTMLDivElement>;

  constructor (
    private programService: ProgramService,
    private policyService: PolicyService,
    private chartService: ChartService,
    private clientSettingsService: ClientSettingsService,
    private formBuilder: TypeSafeFormBuilder,
    private autoTableFactory: AutoTableRepositoryFactory,
    private notifierService: NotifierService,
    private valueComparisonService: ValueComparisonService,
    private i18n: I18nService,
    private currencyService: CurrencyService,
    private awardService: AwardService,
    private spinnerService: SpinnerService,
    private html2CanvasService: Html2CanvasService,
    private fileService: FileService
  ) { }

  get cycleStatus () {
    if (this.formGroup) {
      return this.formGroup.value.cycleStatusFilter;
    }

    return CyclesAPI.CycleStatuses.Open;
  }

  async ngOnInit () {
    this.setProgramCycleOptions();
    const cycleIds = this.getDefaultSelectedCycleIds();
    this.formGroup = this.formBuilder.group({
      programStatusFilter: this.currentProgramStatus,
      cycleStatusFilter: this.currentCycleStatus,
      programCycleFilter: [
        cycleIds
      ]
    });
    await this.setCycleIds(cycleIds);
  }

  setProgramCycleOptions () {
    this.programCycleOptions = this.programService.getProgramCycleOptionsForInsights(
      this.currentProgramStatus,
      false,
      this.currentCycleStatus
    );
    if (this.formGroup) {
      const cycleIds = this.getDefaultSelectedCycleIds();
      this.formGroup.get('programCycleFilter').setValue(cycleIds);
    }
  }

  getDefaultSelectedCycleIds (): number[] {
    const cycleIds: number[] = [];
    this.programCycleOptions.forEach((opt) => {
      opt.dependentOptions.forEach((dependent) => {
        cycleIds.push(dependent.value);
      });
    });

    return cycleIds;
  }

  programStatusChanged (status: ProgramDropdownOptionsForInsights) {
    if (status !== this.currentProgramStatus) {
      this.currentProgramStatus = status;
      this.setProgramCycleOptions();
    }
  }

  cycleStatusChanged (status: CyclesAPI.CycleStatuses) {
    if (status !== this.currentCycleStatus) {
      this.currentCycleStatus = status;
      this.setProgramCycleOptions();
    }
  }

  async setCycleIds (cycleIds: number[]) {
    this.cycleIds = cycleIds;
    await this.setTopCycles();
  }

  async setTopCycles () {
    this.spinnerService.startSpinner();
    this.topCycles = await this.awardService.getTopCyclesByAward(
      this.cycleIds
    );
    this.setupAwardStats();
    this.spinnerService.stopSpinner();
  }

  setupAwardStats () {
    if (!this.cycleRepo) {
      this.cycleRepo = this.autoTableFactory.create({
        key: this.key,
        columns: [],
        notifier: this.notifierService,
        rowsPerPage: 1000,
        valueComparisonService: this.valueComparisonService,
        rows: this.topCycles,
        tableCSVFactory: this.getTableCsvFactory()
      });
    } else {
      this.cycleRepo.rows = this.topCycles;
      this.cycleRepo.reset();
    }

    this.sub.add(
      this.cycleRepo.loading.subscribe((val) => {
        this.loading = val;
      })
    );
    this.cycleLabels = this.topCycles.map((item) => {
      return item.cycleName;
    });
    this.cycleData = this.topCycles.map((item) => {
      return item.paymentAmount;
    });
    this.cycleColors = this.chartService.getFixedAmountOfColors(
      this.topCycles.length,
      this.colors
    );
      // total at the bottom of the table
    this.awardsTotal = this.topCycles.reduce((acc, item) => {
      return acc + item.awardAmount;
    }, 0);
    const totalString = `${this.currencyService.formatMoney(
      this.awardsTotal
    )}`;

    this.cycleTableData = [{
      columnName: this.i18n.translate(
        'common:lblCycle'
      ),
      isNumber: false,
      isMoney: false,
      getRouterLink: null,
      isLegendRef: true,
      key: 'cycleName'
    }, {
      columnName: this.i18n.translate(
        'common:hdrAwards'
      ),
      isNumber: true,
      isMoney: false,
      getRouterLink: null,
      isLegendRef: false,
      key: 'numberOfAwards',
      total: this.topCycles.reduce((acc, item) => {
        return acc + (item.numberOfAwards || 0);
      }, 0)
    }, {
      columnName: this.i18n.translate(
        'AWARDS:textTotalAwardAmount'
      ),
      isNumber: false,
      isMoney: false,
      isMixedVal: true,
      getRouterLink: null,
      isLegendRef: false,
      key: 'awardAmount',
      total: totalString,
      getRowValue: (row) => {
        return this.currencyService.formatMoney(
          row.awardAmount
        );
      }
    }];
  }

  getTableCsvFactory () {
    return DebounceFactory.createSimple(() => {
      const currentRepo = this.autoTableFactory.getRepository(this.key);
      const filteredRows = currentRepo.clientSideFilteredRows as CyclesAPI.CycleTop10[];
      let numberOfAwards = 0;
      let totalAwardAmount = 0;
      const adapted = filteredRows.map((row) => {
        numberOfAwards = numberOfAwards + (row.numberOfAwards || 0);
        totalAwardAmount = totalAwardAmount + (row.awardAmount || 0);

        return {
          'Cycle Name': row.cycleName,
          'Number of Awards': row.numberOfAwards,
          'Total Award Amount': row.awardAmount
        };
      });
      const totalsRow = {
        'Cycle Name': 'Total',
        'Number of Awards': numberOfAwards,
        'Total Award Amount': totalAwardAmount
      };

      return this.fileService.convertObjectArrayToCSVString([
        ...adapted,
        totalsRow
      ]);
    });
  }


  async downloadCharts () {
    this.spinnerService.startSpinner();
    this.downloadingChart = true;
    setTimeout(async () => {
      // Set timeout so native element has time to react to changes to showDownload
      await this.html2CanvasService.downloadSnapshot(this.chartWrapperEl.nativeElement);
      this.downloadingChart = false;
      this.spinnerService.stopSpinner();
    });
  }

  ngOnDestroy () {
    this.sub.unsubscribe();
  }
}

