import { Component, ElementRef, OnInit, ViewChild } from '@angular/core';
import { CurrencyService } from '@core/services/currency.service';
import { SpinnerService } from '@core/services/spinner.service';
import { Budget, BudgetTopLevelStats, FundingSourceTypes, TopLevelStats } from '@core/typings/budget.typing';
import { BudgetResources } from '@features/budgets/budget.resources';
import { BudgetService } from '@features/budgets/budget.service';
import { ClientSettingsService } from '@features/client-settings/client-settings.service';
import { ArrayHelpersService, AutoTableRepositoryFactory, DebounceFactory, FileService, Html2CanvasService, TableCSVFactory, TopLevelFilter, TypeaheadSelectOption, TypeSafeFormBuilder, TypeSafeFormGroup } from '@yourcause/common';
import { I18nService } from '@yourcause/common/i18n';
interface BudgetGroup {
  budgetsFilter: number[];
  budgetStatusFilter: 'all'|'open'|'closed';
}

@Component({
  selector: 'gc-budgets-insights-page',
  templateUrl: './budgets-insights-page.component.html',
  styleUrls: ['./budgets-insights-page.component.scss']
})
export class BudgetsInsightsPageComponent implements OnInit {
  budgets: Budget[] = [];
  formGroup: TypeSafeFormGroup<BudgetGroup>;
  topLevelStats: TopLevelStats;
  FundingSourceTypes = FundingSourceTypes;
  budgetOptions: TypeaheadSelectOption[] = [];
  budgetStatusOptions: TypeaheadSelectOption<string>[] = [{
    value: 'all',
    label: this.i18n.translate(
      'BUDGET:textAllBudgets'
    )
  }, {
    value: 'open',
    label: this.i18n.translate(
      'BUDGET:textOpenBudgets'
    )
  }, {
    value: 'closed',
    label: this.i18n.translate(
      'BUDGET:textClosedBudgets'
    )
  }];
  topLevelFilters: TopLevelFilter[] = [
    new TopLevelFilter(
      'text',
      'name',
      '',
      this.i18n.translate(
        'BUDGET:textSearchByBudgetName',
        {},
        'Search by budget name'
      )
    )
  ];
  budgetTopLevelStats: BudgetTopLevelStats;
  totalSpentFooterText: string;
  totalAllocatedFooterText: string;
  amountUnavailableText: string;
  budgetCsvFactory: TableCSVFactory<string>;
  key = 'BUDGET_INSIGHTS';
  downloadingChart = false;

  @ViewChild('chartWrapper') chartWrapperEl: ElementRef<HTMLDivElement>;


  constructor (
    private budgetService: BudgetService,
    private i18n: I18nService,
    private formBuilder: TypeSafeFormBuilder,
    private budgetResources: BudgetResources,
    private spinnerService: SpinnerService,
    private arrayHelper: ArrayHelpersService,
    private currencyService: CurrencyService,
    private clientSettingsService: ClientSettingsService,
    private fileService: FileService,
    private autoTableRepoFactory: AutoTableRepositoryFactory,
    private html2CanvasService: Html2CanvasService
  ) { }

  get budgetsFromState () {
    return this.budgetService.budgetsForDashboard;
  }

  get canReserveFunds () {
    return this.clientSettingsService.get('clientSettings').reserveFunds;
  }

  async ngOnInit () {
    this.setBudgetCsvFactory();
    this.formGroup = this.formBuilder.group<BudgetGroup>({
      budgetStatusFilter: 'open',
      budgetsFilter: [[]]
    });
    await this.updateBudgetFilter('open');
  }

  setTopLevelStats () {
    this.topLevelStats = {
      totalSpent: this.budgetTopLevelStats.totalSpent,
      totalRemaining: this.budgetTopLevelStats.totalRemaining,
      totalAllocated: this.budgetTopLevelStats.totalAllocated,
      totalUnavailableAmount: this.budgetTopLevelStats.totalUnavailableAmount,
      totalReserved: this.budgetTopLevelStats.reservedAmount
    };
  }

  setBudgetCsvFactory () {
    this.budgetCsvFactory = DebounceFactory.createSimple(() => {
      const currentRepo = this.autoTableRepoFactory.getRepository(this.key);
      const filteredRows = currentRepo.clientSideFilteredRows as Budget[];
      const adapted = filteredRows.map((row) => {
        return {
          Name: row.name,
          Type: row.fundingSourceType === FundingSourceTypes.DOLLARS ?
            'Cash' :
            'In kind',
          Allocated: row.totalAmount,
          'Number of Payments': row.numberOfPayments,
          'Payments Total': row.totalSpent,
          Unavailable: row.totalUnavailableAmount,
          Remaining: row.totalAmountAvailable,
          Reserved: row.reservedAmount,
          'Number of Funding Sources': row.numberOfFundingSources,
          Status: row.isClosed ? 'Closed' : 'Open'
        };
      });

      return this.fileService.convertObjectArrayToCSVString(adapted);
    });
  }

  async budgetFilterChanged () {
    const filterValue = this.formGroup.value.budgetStatusFilter;
    this.budgets = this.filterByStatus(filterValue).filter((budget) => {
      const budgets = this.formGroup.value.budgetsFilter;

      return budgets.length === 0 ||
        budgets.includes(budget.id);
    });
    this.spinnerService.startSpinner();
    this.budgetTopLevelStats = await this.budgetResources.getBudgetStats(
      this.budgets.map((budget) => budget.id)
    );
    this.setTopLevelStats();
    this.totalSpentFooterText = this.i18n.translate(
      'common:textNumberOfPayments',
      {
        number: this.budgetTopLevelStats.numberOfPayments
      },
      '__number__ payments'
    );
    this.totalAllocatedFooterText = this.i18n.translate(
      'BUDGET:textNumberOfFundingSourcesDynamic',
      {
        number: this.budgetTopLevelStats.numberOfFundingSources
      },
      '__number__ funding sources'
    );
    const unavailable = this.budgetTopLevelStats.totalUnavailableAmount;
    this.amountUnavailableText = this.i18n.translate(
      'BUDGET:textTotalUnavailableAmountDynamic',
      {
        amount: this.currencyService.formatMoney(
          unavailable
        )
      },
      '__amount__ unavailable'
    );
    this.spinnerService.stopSpinner();
  }

  async updateBudgetFilter (filterValue: 'all'|'open'|'closed') {
    const filtered = this.filterByStatus(filterValue);
    this.budgetOptions = this.arrayHelper.sort(
      filtered.map((budget) => {
        return {
          label: budget.name,
          value: budget.id
        };
      }),
      'label'
    );
    this.formGroup.get('budgetsFilter').setValue([]);
    await this.budgetFilterChanged();
  }

  getReservedApplications (row: Budget) {
    if (row.reservedApplications === 1) {
      return this.i18n.translate(
        'common:textNumberOfApplication',
        {
          number: row.reservedApplications
        },
        '__number__ application'
      );
    } else if (row.reservedApplications > 1) {
      return this.i18n.translate(
        'common:textNumberOfApplications',
        {
          number: row.reservedApplications
        },
        '__number__ applications'
      );
    }

    return '';
  }

  filterByStatus (
    filterValue: 'all'|'open'|'closed'
  ) {
    return this.budgetsFromState.filter((budget) => {
      switch (filterValue) {
        case 'all':
        default:
          return budget;
        case 'open':
          return !budget.isClosed;
        case 'closed':
          return budget.isClosed;
      }
    });
  }

  async downloadCharts () {
    this.spinnerService.startSpinner();
    this.downloadingChart = true;
    setTimeout(async () => {
      await this.html2CanvasService.downloadSnapshot(this.chartWrapperEl.nativeElement);
      this.downloadingChart = false;
      this.spinnerService.stopSpinner();
    });
  }
}


