import { Component, OnInit } from '@angular/core';
import { ActivatedRoute } from '@angular/router';
import { CurrencyService } from '@core/services/currency.service';
import { PolicyService } from '@core/services/policy.service';
import { StatusService } from '@core/services/status.service';
import { FundingSourceBudgetInfo, FundingSourceDashboardBudget, FundingSourceDashboardInfo, FundingSourceTypes, PaymentTableType, TopLevelStats } from '@core/typings/budget.typing';
import { PaymentForProcess } from '@core/typings/payment.typing';
import { BudgetService } from '@features/budgets/budget.service';
import { ClientSettingsService } from '@features/client-settings/client-settings.service';
import { SystemTagsService } from '@features/system-tags/system-tags.service';
import { SystemTags } from '@features/system-tags/typings/system-tags.typing';
import { AutoTableRepository, AutoTableRepositoryFactory, ChartService, DashboardTableData, PanelTypes, TableDataFactory, TopLevelFilter, TypeaheadSelectOption, ValueComparisonService } from '@yourcause/common';
import { I18nService } from '@yourcause/common/i18n';
import { NotifierService } from '@yourcause/common/notifier';
import { ChartData, ChartOptions, ChartType, TooltipItem } from 'chart.js';
import { uniqBy } from 'lodash';

@Component({
  selector: 'gc-funding-source-dashboard',
  templateUrl: 'funding-source-dashboard.component.html',
  styleUrls: ['./funding-source-dashboard.component.scss']
})
export class FundingSourceDashboardComponent implements OnInit {
  fundingSource: FundingSourceDashboardInfo;
  topLevelStats: TopLevelStats;
  status: string;
  id: number;
  name: string;
  PaymentTableType = PaymentTableType;
  FundingSourceTypes = FundingSourceTypes;
  PanelTypes = PanelTypes;
  branding = this.clientSettingsService.get('clientBranding');
  colors = [
    this.branding.brandPrimary,
    this.branding.brandSecondary,
    this.branding.brandUtility
  ];
  defaultCurrency = this.clientSettingsService.defaultCurrency;
  baseChart: ChartOptions = {
    responsive: true,
    plugins: {
      legend: {
        display: false
      }
    }
  };
  totalRemaining = this.i18n.translate(
    'BUDGET:textTotalRemaining',
    {},
    'Total remaining'
  );
  totalUnavailable = this.i18n.translate(
    'GLOBAL:textTotalUnavailable',
    {},
    'Total unavailable'
  );
  permissions = this.policyService.insights;
  totalUnallocatedText: string;
  numberOfBudgetsText: string;
  amountReservedText: string;
  oneBudgetText = this.i18n.translate(
    'BUDGET:textOneBudget',
    {},
    '1 budget'
  );

  // Funding Sources //
  budgetRepo: AutoTableRepository<FundingSourceDashboardBudget>;
  budgetLabels: string[] = [];
  budgetData: number[] = [];
  budgetColors: string[] = [];
  budgetTableData: DashboardTableData<FundingSourceBudgetInfo>[];
  budgetChartOptions = {
    ...this.baseChart,
    tooltips: {
      callbacks: {
        label: (tooltipItem: TooltipItem<ChartType>, data: ChartData) => {
          const amount = this.chartService.getAmountForTooltip(
            tooltipItem,
            data
          );
          const label = this.chartService.getLabelForTooltip(
            tooltipItem,
            data
          );
          const value = this.currencyService.formatMoney(amount);
          if (label === this.totalRemaining) {
            const percent = this.chartService.getPercent(
              this.fundingSource.stats.totalRemaining,
              this.fundingSource.stats.total
            );

            return value + ` (${percent})`;
          }  else if (label === this.totalUnavailable) {
            const percent = this.chartService.getPercent(
              this.fundingSource.stats.totalUnavailableAmount,
              this.fundingSource.stats.total
            );

            return value + ` (${percent})`;
          } else {
            return this.i18n.translate(
              'BUDGET:textTotalSpent',
              {},
              'Total spent'
            ) + ': ' + value;
          }
        },
        beforeLabel: (tooltipItem: TooltipItem<ChartType>, data: ChartData) => {
          return this.chartService.getLabelForTooltip(
            tooltipItem,
            data
          );
        },
        afterLabel: (tooltipItem: TooltipItem<ChartType>, data: ChartData) => {
          const label = this.chartService.getLabelForTooltip(
            tooltipItem,
            data
          );
          if (
            label !== this.totalRemaining &&
            label !== this.totalUnavailable
          ) {
            const amount = this.chartService.getAmountForTooltip(
              tooltipItem,
              data
            );

            return ' (' +
              this.i18n.translate(
                'BUDGET:textPercentOfBudget',
                {
                  percent: this.chartService.getPercent(
                    amount,
                    this.fundingSource.stats.allocatedTotal
                  )
                },
                '__percent__ of budget'
              ) + ')';
          }

          return '';
        }
      }
    }
  } as ChartOptions;

  // Payments //
  paymentStatusMap = this.statusService.paymentStatusMap;
  tableDataFactory: TableDataFactory<PaymentForProcess>;
  paymentsKey: string;
  programOptions: TypeaheadSelectOption[] = [];
  tagOptions = uniqBy(
    this.systemTagsService.getTagsForBucket(SystemTags.Buckets.Payment, true)
      .concat(this.systemTagsService.getTagsForBucket(SystemTags.Buckets.Application, true))
      .concat(this.systemTagsService.getTagsForBucket(SystemTags.Buckets.NonprofitProfile, true)),
      'id'
    ).map(tag => {
      return {
      label: tag.name,
      value: tag.id
      };
    });
  topLevelFilters: TopLevelFilter[] = [];
  clientDefaultTz = this.clientSettingsService.clientSettings.defaultTimezone || 'UTC';

  constructor (
    private budgetService: BudgetService,
    private activatedRoute: ActivatedRoute,
    private clientSettingsService: ClientSettingsService,
    private autoTableFactory: AutoTableRepositoryFactory,
    private chartService: ChartService,
    private i18n: I18nService,
    private statusService: StatusService,
    private systemTagsService: SystemTagsService,
    private valueComparisonService: ValueComparisonService,
    private notifierService: NotifierService,
    private policyService: PolicyService,
    private currencyService: CurrencyService
  ) { }


  get fundingSourcesFromState () {
    return this.budgetService.get('fundingSourcesForDashboard');
  }

  get canReserveFunds () {
    return this.clientSettingsService.get('clientSettings').reserveFunds;
  }

  ngOnInit () {
    this.id = this.activatedRoute.snapshot.params.id;
    const found = this.fundingSourcesFromState.find((fs) => {
      return +fs.fundingSourceId === +this.id;
    });
    this.name = found.fundingSourceName;
    this.status = this.i18n.translate(
      found.isClosed ?  'GLOBAL:textClosed' : 'GLOBAL:textOpen'
    );
    this.paymentsKey = `FUNDING_SOURCE_${this.id}_PAYMENTS`;
    this.fundingSource = this.budgetService.get(
      'fundingSourceDashboardMap'
      )[this.id];
    this.setTopLevelStats();
    this.totalUnallocatedText = this.i18n.translate(
      'BUDGET:textAmountUnallocated',
      {
        number: this.currencyService.formatMoney(
          this.fundingSource.stats.unallocatedTotal
        )
      },
      ' __number__ unallocated'
    );
    this.numberOfBudgetsText = this.i18n.translate(
      'BUDGET:textNumberOfBudgets',
      {
        number: this.fundingSource.stats.numberOfBudgets
      },
      '__number__ budgets'
    );
    if (!!this.fundingSource.stats.reservedAmount && this.canReserveFunds) {
      this.amountReservedText = this.i18n.translate(
        'common:textAmountReserved',
        {
          amount: this.currencyService.formatMoney(
            this.fundingSource.stats.reservedAmount
          )
        },
        '__amount__ reserved'
      );
    }
    this.setupBudgets();
  }

  setTopLevelStats () {
    this.topLevelStats = {
      total: this.fundingSource.stats.total,
      totalSpent: this.fundingSource.stats.totalSpent,
      totalRemaining: this.fundingSource.stats.totalRemaining,
      totalAllocated: this.fundingSource.stats.allocatedTotal,
      totalUnavailableAmount: this.fundingSource.stats.totalUnavailableAmount,
      totalReserved: this.fundingSource.stats.reservedAmount
    };
  }

  setupBudgets () {
    this.budgetRepo = this.autoTableFactory.create({
      key: 'FUNDING_SOURCE_BUDGET_STATS',
      columns: [],
      notifier: this.notifierService,
      rowsPerPage: 1000,
      valueComparisonService: this.valueComparisonService,
      rows: this.fundingSource.budgets
    });
    this.budgetLabels = this.fundingSource.budgets.map((budget) => {
      return budget.budgetName;
    });
    this.budgetLabels = [
      ...this.budgetLabels,
      this.totalRemaining,
      this.totalUnavailable
    ];
    this.budgetData = this.fundingSource.budgets.map((budget) => {
      return budget.totalSpent;
    });
    this.budgetData = [
      ...this.budgetData,
      this.fundingSource.stats.allocatedTotalRemaining,
      this.fundingSource.stats.totalUnavailableAmount
    ];
    this.budgetColors = this.chartService.getFixedAmountOfColors(
      this.budgetData.length,
      this.colors
    );
    this.budgetTableData = [{
      columnName: this.i18n.translate(
        'common:lblBudget'
      ),
      isNumber: false,
      isMoney: false,
      getRouterLink: this.permissions.canViewBudgets() ?
        ((row: FundingSourceBudgetInfo) => {
          return '/management/insights/budgets/budget/' +
            row.budgetId;
        }) :
        null,
      isLegendRef: true,
      key: 'budgetName',
      getRowValue: (row: FundingSourceBudgetInfo) => {
        const closedText = ` (${this.i18n.translate('GLOBAL:textClosed')})`;

        return `${row.budgetName}${row.isClosed ? closedText : ''}`;
      }
    }, {
      columnName: this.i18n.translate(
        'GLOBAL:textTotalAllocated'
      ),
      isNumber: false,
      isMoney: false,
      isMixedVal: true,
      getRouterLink: null,
      isLegendRef: false,
      key: 'totalAllocated',
      getRowValue: (row) => {
        return this.currencyService.formatMoney(
          row.totalAllocated
        );
      },
      total: this.currencyService.formatMoney(
        this.fundingSource.budgets.reduce((acc, budget) => {
          return acc + (budget.totalAllocated || 0);
        }, 0)
      )
    }, {
      columnName: this.i18n.translate(
        'GLOBAL:textTotalSpent'
      ),
      isNumber: false,
      isMoney: false,
      isMixedVal: true,
      getRouterLink: null,
      isLegendRef: false,
      key: 'totalSpent',
      getRowValue: (row) => {
        return this.currencyService.formatMoney(
          row.totalSpent
        );
      },
      total: this.currencyService.formatMoney(
        this.fundingSource.budgets.reduce((acc, budget) => {
          return acc + (budget.totalSpent || 0);
        }, 0)
      )
    }, {
      columnName: this.i18n.translate(
        'GLOBAL:textTotalRemaining'
      ),
      isNumber: false,
      isMoney: false,
      isMixedVal: true,
      getRouterLink: null,
      isLegendRef: false,
      key: 'totalRemaining',
      subKey: this.canReserveFunds ? 'reservedAmount' : null,
      subKeyType: 'money',
      getRowValue: (row) => {
        return this.currencyService.formatMoney(
          row.totalRemaining
        );
      },
      getRowSubValue: (row) => {
        return this.i18n.translate(
          'common:textReserved',
          {},
          'reserved'
        );
     },
      total: this.currencyService.formatMoney(
        this.fundingSource.budgets.reduce((acc, budget) => {
          return acc + budget.totalRemaining;
        }, 0)
      ),
      subTotalText: this.canReserveFunds ? this.i18n.translate(
        'GLOBAL:textAmountReserved',
        {
          amount: this.currencyService.formatMoney(
            this.fundingSource.budgets.reduce((acc, budget) => {
              return acc + budget.reservedAmount;
            }, 0)
          )
        },
        '__amount__ reserved'
      ) : null
    }, {
      columnName: this.i18n.translate(
        'BUDGET:hdrTotalUnavailable'
      ),
      isNumber: false,
      isMoney: false,
      isMixedVal: true,
      getRouterLink: null,
      isLegendRef: false,
      key: 'totalUnavailableAmount',
      getRowValue: (row) => {
        return this.currencyService.formatMoney(
          row.totalUnavailableAmount
        );
      },
      total: this.currencyService.formatMoney(
        this.fundingSource.budgets.reduce((acc, budget) => {
          return acc + budget.totalUnavailableAmount;
        }, 0)
      )
    }];
  }
}
