import { Component, ElementRef, OnInit, ViewChild } from '@angular/core';
import { CurrencyService } from '@core/services/currency.service';
import { SpinnerService } from '@core/services/spinner.service';
import { FundingSource, FundingSourceInsights, FundingSourceTopLevelStats, FundingSourceTypes, TopLevelStats } from '@core/typings/budget.typing';
import { BudgetResources } from '@features/budgets/budget.resources';
import { BudgetService } from '@features/budgets/budget.service';
import { ClientSettingsService } from '@features/client-settings/client-settings.service';
import { ArrayHelpersService, AutoTableRepositoryFactory, DebounceFactory, FileService, Html2CanvasService, TableCSVFactory, TopLevelFilter, TypeaheadSelectOption, TypeSafeFormBuilder, TypeSafeFormGroup } from '@yourcause/common';
import { I18nService } from '@yourcause/common/i18n';
interface FundingGroup {
  fundingSourceStatusFilter: 'all'|'open'|'closed';
  allocationType: 'allFunds'|'allocation';
  sourcesFilter: number[];
}

@Component({
  selector: 'gc-funding-source-insights-page',
  templateUrl: './funding-source-insights-page.component.html',
  styleUrls: ['./funding-source-insights-page.component.scss']
})
export class FundingSourceInsightsPageComponent implements OnInit {
  sources: FundingSourceInsights[] = [];
  formGroup: TypeSafeFormGroup<FundingGroup>;
  sourcesCsvFactory: TableCSVFactory<string>;
  totalByAllFunds = this.i18n.translate(
    'BUDGET:labelTotalByAllFunds',
    {},
    'Total by all funds'
  );
  totalByAllocation = this.i18n.translate(
    'BUDGET:labelTotalByAllocation',
    {},
    'Total by allocation'
  );
  FundingSourceTypes = FundingSourceTypes;
  fundingSourcesOptions: TypeaheadSelectOption[] = [];
  allocationTypeOptions: TypeaheadSelectOption[] = [{
    label: this.totalByAllocation,
    value: 'allocation'
  }, {
    label: this.totalByAllFunds,
    value: 'allFunds'
  }];

  topLevelFilters: TopLevelFilter[] = [
    new TopLevelFilter(
      'text',
      'fundingSourceName',
      '',
      this.i18n.translate(
        'BUDGET:textSearchByFundingSourceName',
        {},
        'Search by funding source name'
      )
    )
  ];
  fundingSourceStatusOptions = [{
    value: 'all',
    label: this.i18n.translate(
      'BUDGET:textAllSources'
    )
  }, {
    value: 'open',
    label: this.i18n.translate(
      'BUDGET:textOpenSources'
    )
  }, {
    value: 'closed',
    label: this.i18n.translate(
      'BUDGET:textClosedSources'
    )
  }];
  fundingSourceTopLevelStats: FundingSourceTopLevelStats;
  topLevelStats: TopLevelStats;
  amountUnallocatedText: string;
  numberOfBudgetsText: string;
  amountUnavailableText: string;
  oneBudgetText = this.i18n.translate(
    'BUDGET:textOneBudget',
    {},
    '1 budget'
  );
  key = 'FUNDING_SOURCE_INSIGHTS';
  downloadingChart = false;

  @ViewChild('chartWrapper') chartWrapperEl: ElementRef<HTMLDivElement>;


  constructor (
    private budgetService: BudgetService,
    private i18n: I18nService,
    private formBuilder: TypeSafeFormBuilder,
    private arrayHelper: ArrayHelpersService,
    private spinnerService: SpinnerService,
    private budgetResources: BudgetResources,
    private currencyService: CurrencyService,
    private clientSettingsService: ClientSettingsService,
    private fileService: FileService,
    private autoTableRepoFactory: AutoTableRepositoryFactory,
    private html2CanvasService: Html2CanvasService
  ) { }

  get fundingSourcesFromState () {
    return this.budgetService.get('fundingSourcesForDashboard');
  }

  get canReserveFunds () {
    return this.clientSettingsService.get('clientSettings').reserveFunds;
  }

  async ngOnInit () {
    this.setSourcesCsvFactory();
    this.formGroup = this.formBuilder.group<FundingGroup>({
      sourcesFilter: [[]],
      fundingSourceStatusFilter: 'open',
      allocationType: 'allFunds'
    });
    this.updateFundingSourceFilter('open');
    await this.sourceFilterChanged();
  }

  setSourcesCsvFactory () {
    this.sourcesCsvFactory = DebounceFactory.createSimple(() => {
      const currentRepo = this.autoTableRepoFactory.getRepository(this.key);
      const filteredRows = currentRepo.clientSideFilteredRows as FundingSourceInsights[];
      const adapted = filteredRows.map((row) => {
        return {
          Name: row.fundingSourceName,
          Type: row.fundingSourceType === FundingSourceTypes.DOLLARS ?
            'Cash' :
            'In kind',
          Total: row.total,
          'Total Allocated': row.allocatedTotal,
          'Reserved Amount': row.reservedAmount,
          'Reserved Applications': row.reservedApplications,
          'Total Spent': row.totalSpent,
          Unavailable: row.totalUnavailableAmount,
          'Number of Payments': row.numberOfPayments,
          'Number of Budgets': row.numberOfBudgets,
          Status: row.isClosed ? 'Closed' : 'Open'
        };
      });

      return this.fileService.convertObjectArrayToCSVString(adapted);
    });
  }

  setTopLevelStats (data: FundingSourceTopLevelStats) {
    const totalBy = this.formGroup.value.allocationType;
    if (totalBy === 'allFunds') {
      this.topLevelStats = {
        total: data.total,
        totalSpent: data.totalSpent,
        totalRemaining: data.totalRemaining,
        totalUnavailableAmount: data.totalUnavailableAmount,
        totalReserved: data.reservedAmount
      };
     } else {
      this.topLevelStats = {
        total: data.allocatedTotal,
        totalSpent: (data.allocatedTotal - data.allocatedTotalRemaining),
        totalRemaining: data.allocatedTotalRemaining,
        totalUnavailableAmount: data.totalUnavailableAmount,
        totalReserved: data.reservedAmount
      };
    }
  }

  async sourceFilterChanged () {
    const filterValue = this.formGroup.value.fundingSourceStatusFilter;
    this.sources = this.filterByStatus(filterValue).filter((source) => {
      const sources = this.formGroup.value.sourcesFilter;

      return sources.length === 0 ||
        sources.includes(source.fundingSourceId);
    });
    this.spinnerService.startSpinner();
    this.fundingSourceTopLevelStats = await this.budgetResources.getFundingSourceStats(
      this.sources.map((fundingSource) => fundingSource.fundingSourceId)
    );
    this.setTopLevelStats(this.fundingSourceTopLevelStats);
    this.amountUnallocatedText = this.i18n.translate(
      'BUDGET:textAmountUnallocated',
      {
        number: this.currencyService.formatMoney(
          this.fundingSourceTopLevelStats.unallocatedTotal
        )
      },
      '__number__ unallocated'
    );

    const unavailable = this.fundingSourceTopLevelStats.totalUnavailableAmount;
    if (unavailable) {
     this.amountUnavailableText = this.i18n.translate(
        'BUDGET:textTotalUnavailableAmountDynamic',
        {
          amount: this.currencyService.formatMoney(
            unavailable
          )
        },
        '__amount__ unavailable'
      );
    } else {
      this.amountUnavailableText = '';
    }
    this.numberOfBudgetsText = this.i18n.translate(
      'BUDGET:textNumberOfBudgets',
      {
        number: this.fundingSourceTopLevelStats.numberOfBudgets
      },
      '__number__ budgets'
    );
    this.spinnerService.stopSpinner();
  }

  async updateFundingSourceFilter (filterValue: 'all'|'open'|'closed') {
    const filtered = this.filterByStatus(filterValue);
    this.fundingSourcesOptions = this.arrayHelper.sort(
      filtered.map((source) => {
        return {
          label: source.fundingSourceName,
          value: source.fundingSourceId
        };
      }),
      'label'
    );
    this.formGroup.get('sourcesFilter').setValue([]);
    await this.sourceFilterChanged();
  }

  filterByStatus (
    filterValue: 'all'|'open'|'closed'
  ) {
    return this.fundingSourcesFromState.filter((source) => {
      switch (filterValue) {
        case 'all':
        default:
          return source;
        case 'open':
          return !source.isClosed;
        case 'closed':
          return source.isClosed;
      }
    });
  }

  getReservedApplications (row: FundingSource) {
    if (row.reservedApplications === 1) {
      return this.i18n.translate(
        'common:textNumberOfApplication',
        {
          number: row.reservedApplications
        },
        '__number__ application'
      );
    } else if (row.reservedApplications > 1) {
      return this.i18n.translate(
        'common:textNumberOfApplications',
        {
          number: row.reservedApplications
        },
        '__number__ applications'
      );
    }

    return '';
  }

  async downloadCharts () {
    this.spinnerService.startSpinner();
    this.downloadingChart = true;
    setTimeout(async () => {
      await this.html2CanvasService.downloadSnapshot(this.chartWrapperEl.nativeElement);
      this.downloadingChart = false;
      this.spinnerService.stopSpinner();
    });
  }
}


