import { Component, ElementRef, OnInit, ViewChild } from '@angular/core';
import { SpinnerService } from '@core/services/spinner.service';
import { CyclesAPI } from '@core/typings/api/cycles.typing';
import { ProgramDropdownOptionsForInsights } from '@core/typings/program.typing';
import { ClientSettingsService } from '@features/client-settings/client-settings.service';
import { ProgramService } from '@features/programs/program.service';
import { Html2CanvasService, SelectOption, TypeaheadSelectOption, TypeSafeFormBuilder, TypeSafeFormGroup } from '@yourcause/common';
import { I18nService } from '@yourcause/common/i18n';
interface OrgGroup {
  programStatusFilter: ProgramDropdownOptionsForInsights;
  cycleStatusFilter: CyclesAPI.CycleStatuses;
  programCycleFilter: number[];
}

@Component({
  selector: 'gc-organization-insights-page',
  templateUrl: './organization-insights-page.component.html',
  styleUrls: ['./organization-insights-page.component.scss']
})
export class OrganizationInsightsPageComponent implements OnInit {
  cycleIds: number[] = [];
  formGroup: TypeSafeFormGroup<OrgGroup>;
  programCycleOptions: SelectOption[] = [];
  currentProgramStatus: ProgramDropdownOptionsForInsights = ProgramDropdownOptionsForInsights.ACTIVE;
  currentCycleStatus: CyclesAPI.CycleStatuses = CyclesAPI.CycleStatuses.Open;
  programStatusOptions: TypeaheadSelectOption<string>[] = [{
    value: 'all',
    label: this.i18n.translate(
      'common:textAllPrograms',
      {},
      'All programs'
    )
  }, {
    value: 'active',
    label: this.i18n.translate(
      'common:lblActivePrograms',
      {},
      'Active programs'
    )
  }, {
    value: 'archived',
    label: this.i18n.translate(
      'common:lblArchivePrograms',
      {},
      'Archived programs'
    )
  }];
  cycleStatusOptions: TypeaheadSelectOption<CyclesAPI.CycleStatuses>[] = [{
    value: CyclesAPI.CycleStatuses.All,
    label: this.i18n.translate(
      'PROGRAM:textAllCycles',
      {},
      'All cycles'
    )
  }, {
    value: CyclesAPI.CycleStatuses.Open,
    label: this.i18n.translate(
      'PROGRAM:textOpenCycles',
      {},
      'Open cycles'
    )
  }, {
    value: CyclesAPI.CycleStatuses.Past,
    label: this.i18n.translate(
      'PROGRAM:textCloseCycles',
      {},
      'Closed cycles'
    )
  }];
  clientDefaultTz = this.clientSettingsService.clientSettings.defaultTimezone;
  downloadingChart = false;

  @ViewChild('chartWrapper') chartWrapperEl: ElementRef<HTMLDivElement>;


  constructor (
    private formBuilder: TypeSafeFormBuilder,
    private programService: ProgramService,
    private i18n: I18nService,
    private clientSettingsService: ClientSettingsService,
    private spinnerService: SpinnerService,
    private html2CanvasService: Html2CanvasService
  ) { }


  get cycleStatus () {
    if (this.formGroup) {
      return this.formGroup.value.cycleStatusFilter;
    }

    return CyclesAPI.CycleStatuses.Open;
  }

  async ngOnInit () {
    this.setProgramCycleOptions();
    const cycleIds = this.getDefaultSelectedCycleIds();
    this.formGroup = this.formBuilder.group({
      programStatusFilter: this.currentProgramStatus,
      cycleStatusFilter: this.currentCycleStatus,
      programCycleFilter: [
        cycleIds
      ]
    });
    this.setCycleIds(cycleIds);
  }

  setProgramCycleOptions () {
    this.programCycleOptions = this.programService.getProgramCycleOptionsForInsights(
      this.currentProgramStatus,
      true,
      this.currentCycleStatus
    );
    if (this.formGroup) {
      const cycleIds = this.getDefaultSelectedCycleIds();
      this.formGroup.get('programCycleFilter').setValue(cycleIds);
    }
  }

  getDefaultSelectedCycleIds (): number[] {
    const cycleIds: number[] = [];
    this.programCycleOptions.forEach((opt) => {
      opt.dependentOptions.forEach((dependent) => {
        cycleIds.push(+dependent.value);
      });
    });

    return cycleIds;
  }

  programStatusChanged (status: ProgramDropdownOptionsForInsights) {
    if (status !== this.currentProgramStatus) {
      this.currentProgramStatus = status;
      this.setProgramCycleOptions();
    }
  }

  cycleStatusChanged (status: CyclesAPI.CycleStatuses) {
    if (status !== this.currentCycleStatus) {
      this.currentCycleStatus = status;
      this.setProgramCycleOptions();
    }
  }

  setCycleIds (cycleIds: number[]) {
    this.cycleIds = cycleIds;
  }

  async downloadCharts () {
    this.spinnerService.startSpinner();
    this.downloadingChart = true;
    setTimeout(async () => {
      await this.html2CanvasService.downloadSnapshot(this.chartWrapperEl.nativeElement);
      this.downloadingChart = false;
      this.spinnerService.stopSpinner();
    });
  }
}
