import { Component, Input, OnChanges } from '@angular/core';
import { CurrencyService } from '@core/services/currency.service';
import { NonprofitTop10 } from '@core/typings/organization.typing';
import { ClientSettingsService } from '@features/client-settings/client-settings.service';
import { NonprofitService } from '@features/nonprofit/nonprofit.service';
import { AutoTableRepository, AutoTableRepositoryFactory, ChartService, DashboardTableData, DebounceFactory, FileService, ValueComparisonService } from '@yourcause/common';
import { I18nService } from '@yourcause/common/i18n';
import { NotifierService } from '@yourcause/common/notifier';
import { ChartData, ChartOptions, ChartType, TooltipItem } from 'chart.js';

@Component({
  selector: 'gc-organization-top-10',
  templateUrl: './organization-top-10.component.html',
  styleUrls: ['./organization-top-10.component.scss']
})

export class OrganizationTop10Component implements OnChanges {
  @Input() cycleIds: number[] = [];
  @Input() showDownload = false;

  key = 'NONPROFIT_STATS';
  baseChart: ChartOptions = {
    responsive: true,
    plugins: {
      legend: {
        display: false
      }
    }
  };
  branding = this.clientSettingsService.get('clientBranding');
  defaultCurrency = this.clientSettingsService.defaultCurrency;
  colors = [
    this.branding.brandPrimary,
    this.branding.brandSecondary,
    this.branding.brandUtility
  ];
  awardsTotal: number;
  nonprofits: NonprofitTop10[] = [];
  nonprofitRepo: AutoTableRepository<NonprofitTop10>;
  nonprofitLabels: string[] = [];
  nonprofitData: number[] = [];
  nonprofitColors: string[] = [];
  nonprofitTableData: DashboardTableData<NonprofitTop10>[];
  nonprofitChartOptions = {
    ...this.baseChart,
    tooltips: {
      callbacks: {
        label: (tooltipItem: TooltipItem<ChartType>) => {
          const numberOfAwards = this.nonprofits[tooltipItem.dataIndex].numberOfAwards;

          return this.i18n.translate(
            'common:hdrAwards'
          ) + ': ' + numberOfAwards;
        },
        beforeLabel: (tooltipItem: TooltipItem<ChartType>, data: ChartData) => {
          return this.chartService.getLabelForTooltip(
            tooltipItem,
            data
          );
        },
        afterLabel: (tooltipItem: TooltipItem<ChartType>, data: ChartData) => {
          const amount = this.chartService.getAmountForTooltip(
            tooltipItem,
            data
          );

          return `${this.currencyService.formatMoney(
            amount
          )} (${
            this.chartService.getPercent(amount, this.awardsTotal)
          })`;
        }
      }
    }
  } as ChartOptions;

  constructor (
    private i18n: I18nService,
    private chartService: ChartService,
    private autoTableFactory: AutoTableRepositoryFactory,
    private notifierService: NotifierService,
    private valueComparisonService: ValueComparisonService,
    private clientSettingsService: ClientSettingsService,
    private nonprofitService: NonprofitService,
    private currencyService: CurrencyService,
    private fileService: FileService
  ) { }

  async ngOnChanges () {
    this.nonprofits = await this.nonprofitService.getOrgsForDashboard(
      this.cycleIds
    );
    if (!this.nonprofitRepo) {
      this.nonprofitRepo = this.autoTableFactory.create({
        key: this.key,
        columns: [],
        notifier: this.notifierService,
        rowsPerPage: 1000,
        valueComparisonService: this.valueComparisonService,
        rows: this.nonprofits,
        tableCSVFactory: this.getTableCsvFactory()
      });
    } else {
      this.nonprofitRepo.rows = this.nonprofits;
      this.nonprofitRepo.reset();
    }
    this.nonprofitLabels = this.nonprofits.map((nonprofit) => {
      return nonprofit.name;
    });
    this.nonprofitData = this.nonprofits.map((nonprofit) => {
      return nonprofit.awardsTotal;
    });
    this.nonprofitColors = this.chartService.getFixedAmountOfColors(
      this.nonprofits.length,
      this.colors
    );
    const totalNumber = this.nonprofits.reduce((acc, nonprofit) => {
        return acc + nonprofit.numberOfAwards;
      }, 0);
    this.awardsTotal = this.nonprofits.reduce((acc, nonprofit) => {
      return acc + nonprofit.awardsTotal;
    }, 0);
    const totalString = `${totalNumber} (${
      this.currencyService.formatMoney(
        this.awardsTotal
      )
    })`;
    this.nonprofitTableData = [{
      columnName: this.i18n.translate(
        'common:lblOrganization'
      ),
      isNumber: false,
      isMoney: false,
      getRouterLink: (row) => {
        return row.organizationId !== 0 ?
          `/management/insights/organizations/nonprofit/${
            row.guid || row.organizationId
          }/profile` :
          undefined;
      },
      isLegendRef: true,
      key: 'name'
    }, {
      columnName: this.i18n.translate(
        'GLOBAL:lblRequestAmount'
      ),
      isNumber: false,
      isMoney: false,
      isMixedVal: true,
      getRouterLink: null,
      isLegendRef: false,
      key: 'requestAmount',
      getRowValue: (row) => {
        return this.currencyService.formatMoney(
          row.requestAmount
        );
      },
      total: this.currencyService.formatMoney(
        (this.nonprofits.reduce((acc, nonprofit) => {
        return acc + (nonprofit.requestAmount || 0);
        }, 0))
      )
    }, {
      columnName: this.i18n.translate(
        'common:hdrAwards'
      ),
      isNumber: false,
      isMoney: false,
      isMixedVal: true,
      getRouterLink: null,
      isLegendRef: false,
      key: 'awardNumberAndTotal',
      total: totalString
    }];
  }

  getTableCsvFactory () {
    return DebounceFactory.createSimple(() => {
      const currentRepo = this.autoTableFactory.getRepository(this.key);
      const filteredRows = currentRepo.clientSideFilteredRows as NonprofitTop10[];
      let totalRequestAmount = 0;
      let totalNumberOfAwards = 0;
      let totalAwardsAmount = 0;
      const adapted = filteredRows.map((row) => {
        totalRequestAmount = totalRequestAmount + (row.requestAmount || 0);
        totalNumberOfAwards = totalNumberOfAwards + (row.numberOfAwards || 0);
        totalAwardsAmount = totalAwardsAmount + (row.awardsTotal || 0);

        return {
          'Organization Name': row.name,
          'Request Amount': row.requestAmount,
          'Number of Awards': row.numberOfAwards,
          'Awards Total': row.awardsTotal
        };
      });
      const totalsRow = {
        'Organization Name': 'Total',
        'Request Amount': totalRequestAmount,
        'Number of Awards': totalNumberOfAwards,
        'Awards Total': totalAwardsAmount
      };

      return this.fileService.convertObjectArrayToCSVString([
        ...adapted,
        totalsRow
      ]);
    });
  }
}
