import { Component, Input, OnInit } from '@angular/core';
import { StatusService } from '@core/services/status.service';
import { TranslationService } from '@core/services/translation.service';
import { FundingSourceTypes, PaymentTableType } from '@core/typings/budget.typing';
import { PaymentForProcess } from '@core/typings/payment.typing';
import { PaymentStatus } from '@core/typings/status.typing';
import { BudgetResources } from '@features/budgets/budget.resources';
import { ProgramResources } from '@features/programs/program.resources';
import { ProgramService } from '@features/programs/program.service';
import { SystemTagsService } from '@features/system-tags/system-tags.service';
import { SystemTags } from '@features/system-tags/typings/system-tags.typing';
import { ALL_SKIP_FILTER, ArrayHelpersService, DebounceFactory, PaginatedResponse, PaginationOptions, TableDataFactory, TopLevelCustomLink, TopLevelFilter, TopLevelFilterOptionsConfig, TypeaheadSelectOption } from '@yourcause/common';
import { CachedAttr, CACHE_TYPES } from '@yourcause/common/cache';
import { I18nService } from '@yourcause/common/i18n';
import { uniqBy } from 'lodash';

@Component({
  selector: 'gc-payments-insights-table',
  templateUrl: 'payments-insights-table.component.html',
  styleUrls: ['./payments-insights-table.component.scss']
})

export class PaymentsInsightsTableComponent implements OnInit {
  @Input() paymentTableType: PaymentTableType;
  @Input() id: number;
  @Input() standalone = true;
  insightsRoute = 'budgets';
  paymentsKey: string;
  PaymentTableType = PaymentTableType;
  FundingSourceTypes = FundingSourceTypes;
  tableDataFactory: TableDataFactory<PaymentForProcess>;
  topLevelFilters: TopLevelFilter[] = [];
  programOptions: TypeaheadSelectOption[] = [];
  paymentStatusMap = this.statusService.paymentStatusMap;
  paymentSubStatusMap = this.statusService.paymentSubStatusMap;
  tagOptions = uniqBy(
    this.systemTagsService.getTagsForBucket(SystemTags.Buckets.Payment, true)
      .concat(this.systemTagsService.getTagsForBucket(SystemTags.Buckets.Application, true))
      .concat(this.systemTagsService.getTagsForBucket(SystemTags.Buckets.NonprofitProfile, true)),
      'id'
    ).map(tag => {
      return {
      label: tag.name,
      value: tag.id
      };
    });
  paymentStatusOptions: TopLevelFilterOptionsConfig = {
    selectOptions: [{
      value: ALL_SKIP_FILTER,
      display: this.i18n.translate('common:lblAllCap')
    }, {
      value: PaymentStatus.Pending,
      display: this.i18n.translate('common:lblPending')
    }, {
      value: PaymentStatus.Scheduled,
      display: this.i18n.translate('GLOBAL:textScheduled')
    }, {
      value: PaymentStatus.Processing,
      display: this.i18n.translate('GLOBAL:textProcessing')
    }, {
      value: PaymentStatus.Outstanding,
      display: this.i18n.translate('GLOBAL:textOutstanding')
    }, {
      value: PaymentStatus.Cleared,
      display: this.i18n.translate('GLOBAL:lblCleared')
    }, {
      value: PaymentStatus.Voided,
      display: this.i18n.translate('GLOBAL:textVoided')
    }]
  };
  viewTranslations = this.translationService.viewTranslations;
  programTranslationMap = this.viewTranslations.Grant_Program;
  cycleTranslationMap = this.viewTranslations.Grant_Program_Cycle;
  hasMaskedAppsOnPage = false;
  showMaskedApplicantsIcon = 'user-shield';
  hideMaskedApplicantsIcon = 'user-times';
  hideMaskedApplicantsTooltip = this.i18n.translate(
    'MANAGE:textShowApplicantInformation',
    {},
    'Show applicant information'
  );
  showMaskedApplicantsTooltip = this.i18n.translate(
    'MANAGE:textHideApplicantInformation',
    {},
    'Hide applicant information'
  );
  topLevelCustomLinks: TopLevelCustomLink[] = [];
  @CachedAttr(
    CACHE_TYPES.LOCALSTORAGE,
    false
  ) showMaskedApplicants: boolean;


  constructor (
    private arrayHelper: ArrayHelpersService,
    private budgetResources: BudgetResources,
    private systemTagsService: SystemTagsService,
    private i18n: I18nService,
    private statusService: StatusService,
    private programResources: ProgramResources,
    private translationService: TranslationService,
    private programService: ProgramService
  ) { }

  get isBudget () {
    return this.paymentTableType === PaymentTableType.BUDGETS;
  }
  get cycleIds () {
    return this.programService.currentProgramDashCycleIds;
  }

  get isFundingSource () {
    return this.paymentTableType === PaymentTableType.FUNDING_SOURCES;
  }

  get isProgram () {
    return this.paymentTableType === PaymentTableType.PROGRAM;
  }

  ngOnInit () {
    if (this.isFundingSource) {
      this.insightsRoute = 'funding-sources';
    } else if (this.isProgram) {
      this.insightsRoute = 'program-manager';
    }
    this.paymentsKey = `DYNAMIC_${this.id}_PAYMENTS`;
    this.startSetup();
  }

  setupFactory () {
    this.tableDataFactory = DebounceFactory.createSimple(
      async (options: PaginationOptions<PaymentForProcess>) => {
        const needsProgramOptions = !this.isProgram &&
          !this.programOptions.length;
        options = this.systemTagsService.formatPaginationOptions(options);
        let result: PaginatedResponse<PaymentForProcess>;
        if (this.isBudget || this.isFundingSource) {
          const func = this.isBudget ?
            'getBudgetDashboardPayments' :
            'getFundingSourceDashboardPayments';
          result = await this.budgetResources[func](
            this.id,
            options,
            needsProgramOptions
          );
          if (needsProgramOptions && result.programFacets) {
            this.programOptions = this.arrayHelper.sort(Object.keys(result.programFacets)
              .map(id => {
                const map = this.programTranslationMap[id];

                return {
                  value: +id,
                  label: map && map.Name ? map.Name : result.programFacets[id]
                };
            }), 'label');
          }
        } else {
          result = await this.programResources.getPaymentsForProgram(
            options,
            +this.id,
            this.cycleIds
          );
        }
        this.handleMasking(result.records);

        return {
          success: true,
          data: {
            recordCount: result.recordCount,
            records: result.records
          }
        };
      }
    );
  }

  handleMasking (records: PaymentForProcess[]) {
    let hasMaskedApps = false;
    records.forEach((row) => {
      if (row.isMasked && row.canViewMaskedApplicantInfo) {
        hasMaskedApps = true;
      }
    });
    this.hasMaskedAppsOnPage = hasMaskedApps;
    if (this.hasMaskedAppsOnPage) {
      this.setMaskedIconAndTooltip();
    }
  }

    toggleShowMaskedApplicants () {
    this.showMaskedApplicants = !this.showMaskedApplicants;
    this.setMaskedIconAndTooltip();
  }

  setMaskedIconAndTooltip () {
    const tooltip = this.showMaskedApplicants ?
      this.showMaskedApplicantsTooltip :
      this.hideMaskedApplicantsTooltip;
    const item = {
      name: tooltip,
      tooltip,
      icon: this.showMaskedApplicants ?
        this.hideMaskedApplicantsIcon :
        this.showMaskedApplicantsIcon,
      isIcon: true,
      onClick: () => {
        this.toggleShowMaskedApplicants();
      }
    };
    if (this.topLevelCustomLinks.length === 0) {
      this.topLevelCustomLinks = [item];
    } else {
      this.topLevelCustomLinks[0] = item;
    }
  }

  startSetup () {
    this.setupFactory();
    this.topLevelFilters = [
      new TopLevelFilter(
        'text',
        'applicantInfo.fullName',
        '',
        this.isBudget ?
          this.i18n.translate(
            'MANAGE:textSearchByApplicantOrgProgramSource2',
            {},
            'Search by applicant, organization, program, or source'
          ) :
          this.i18n.translate(
          'MANAGE:textSearchByApplicantOrgBudget',
          {},
          'Search by applicant, organization or budget'
        ),
        undefined,
        undefined,
        this.isBudget ?
          [{
            column: 'applicantInfo.fullName',
            filterType: 'cn'
          }, {
            column: 'organizationInfo.name',
            filterType: 'cn'
          }, {
            column: 'fundingSourceName',
            filterType: 'cn'
          }, {
            column: 'programName',
            filterType: 'cn'
          }] :
          [{
          column: 'applicantInfo.fullName',
          filterType: 'cn'
        }, {
          column: 'organizationInfo.name',
          filterType: 'cn'
        }, {
          column: 'budgetName',
          filterType: 'cn'
        }]
      ),
      new TopLevelFilter(
        'typeaheadSingleEquals',
        'statusId',
        ALL_SKIP_FILTER,
        '',
        this.paymentStatusOptions,
        this.i18n.translate(
          'MANAGE:textPaymentStatus',
          {},
          'Payment status'
        )
      ),
      new TopLevelFilter(
        'dateRange',
        'createdDate',
        '',
        this.i18n.translate(
          'GLOBAL:textPaymentDateRange',
          {},
          'Payment date range'
        )
      )
    ];
  }
}
