
import { HttpErrorResponse } from '@angular/common/http';
import { Component, OnInit } from '@angular/core';
import { Validators } from '@angular/forms';
import { Router } from '@angular/router';
import { AddressFormGroup } from '@core/components/address-block/address-block.component';
import { ApplicantService } from '@core/services/auth-user/applicant.service';
import { FormBuilderFactoryService } from '@core/services/form-builder-factory.service';
import { SpinnerService } from '@core/services/spinner.service';
import { TokenService } from '@core/services/token/token.service';
import { UpdateProfile } from '@core/typings/applicant.typing';
import { ClientSettingsService } from '@features/client-settings/client-settings.service';
import { SignatureService } from '@features/signature/signature.service';
import { UserService } from '@features/users/user.service';
import { Address, PanelTypes, PasswordService, PasswordValidator, ReflectedDataValidator, TypeSafeFormBuilder, TypeSafeFormGroup } from '@yourcause/common';
import { I18nService } from '@yourcause/common/i18n';
import { LogService } from '@yourcause/common/logging';
import { ModalFactory } from '@yourcause/common/modals';
import { NotifierService } from '@yourcause/common/notifier';
import { SignatureModalComponent } from '@yourcause/common/signature';
import { DeleteAccountModalComponent } from '../delete-account-modal/delete-account-modal.component';
import { PersonalDataModalComponent } from '../personal-data-modal/personal-data-modal.component';
interface MyAccountGroup {
  firstName: string;
  lastName: string;
  phoneNumber: string;
}
interface PasswordGroup {
  currentPassword: string;
  newPassword: string;
}

@Component({
  selector: 'gc-my-account-applicant',
  templateUrl: './my-account-applicant.component.html',
  styleUrls: ['./my-account-applicant.component.scss']
})
export class MyAccountApplicantComponent implements OnInit {
  PanelTypes = PanelTypes;
  myAccountForm: TypeSafeFormGroup<MyAccountGroup>;
  passwordForm: TypeSafeFormGroup<PasswordGroup>;
  addressFormGroup: TypeSafeFormGroup<AddressFormGroup>;
  currentPasswordVisible = false;
  newPasswordVisible = false;
  langKeys = this.clientSettingsService.langKeys;
  errorOnSubmit = false;
  containsUserInfo = false;
  isPreviousPassword = false;

   constructor (
    private logger: LogService,
    private userService: UserService,
    private tokenService: TokenService,
    private router: Router,
    private applicantService: ApplicantService,
    private formBuilder: TypeSafeFormBuilder,
    private formBuilderFactory: FormBuilderFactoryService,
    private notifier: NotifierService,
    private i18n: I18nService,
    private spinnerService: SpinnerService,
    private modalFactory: ModalFactory,
    private passwordService: PasswordService,
    private clientSettingsService: ClientSettingsService,
    private signatureService: SignatureService
  ) { }

  get applicant () {
    return this.userService.applicant;
  }

  async ngOnInit () {
    if (!this.applicant) {
      this.router.navigate(['/apply/auth/signin']);
    } else {
      const address: Address = {
        address1: this.applicant.address,
        address2: this.applicant.address2,
        city: this.applicant.city,
        postalCode: this.applicant.postalCode,
        countryCode: this.applicant.country,
        stateProvRegCode: this.applicant.state
      };
      this.addressFormGroup = this.formBuilderFactory.createAddressGroup(address);
      this.myAccountForm = this.formBuilder.group({
        firstName: [this.applicant.firstName, Validators.required],
        lastName: [this.applicant.lastName, Validators.required],
        phoneNumber: this.applicant.phoneNumber
      }, {
        validators: [
          ReflectedDataValidator('firstName'),
          ReflectedDataValidator('lastName')
        ]
      });
      this.passwordForm = this.formBuilder.group({
        currentPassword: '',
        newPassword: [
          '',
          [
            PasswordValidator(this.passwordService)
          ]
        ]
      });
    }
  }

  toggleCurrentPasswordVisible = () => {
    this.currentPasswordVisible = !this.currentPasswordVisible;
  };

  toggleNewPasswordVisible = () => {
    this.newPasswordVisible = !this.newPasswordVisible;
  };

  organizationNotifierError (action: 'Adding'|'Removing') {
    const adding = action === 'Adding';
    this.notifier.error(this.i18n.translate(
      adding ?
        'nonprofits:textErrorAddingNonprofitToApplicant' :
        'nonprofits:textErrorRemovingNonprofitToApplicant',
      {},
      `There was an error ${action.toLowerCase()} the organization`
    ));
  }

  passwordChanged () {
    this.errorOnSubmit = false;
    this.containsUserInfo = false;
    this.isPreviousPassword = false;
  }

  async onSubmit (isPassword = false) {
    this.spinnerService.startSpinner();
    const values = this.myAccountForm.value;
    const data: UpdateProfile = {
      firstName: (values.firstName || '').trim(),
      lastName: (values.lastName || '').trim(),
      existingPassword: isPassword ?
        this.passwordForm.value.currentPassword :
        undefined,
      password: isPassword ?
        this.passwordForm.value.newPassword :
        undefined,
      confirmPassword: isPassword ?
        this.passwordForm.value.newPassword :
        undefined,
      address: this.addressFormGroup.value.address1,
      address2: this.addressFormGroup.value.address2,
      country: this.addressFormGroup.value.countryCode,
      state: this.addressFormGroup.value.stateProvRegCode,
      city: this.addressFormGroup.value.city,
      postalCode: this.addressFormGroup.value.postalCode,
      phoneNumber: values.phoneNumber,
      culture: this.i18n.language
    };

    await this.updateProfile(data, isPassword);
  }

  async updateProfile (data: UpdateProfile, isPassword = false) {
    try {
      const response = await this.applicantService.updateProfile(data);
      if (isPassword) {
        if (response.validPassword) {
          this.passwordForm.reset();
          Object.keys(this.passwordForm.controls).forEach(key => {
            this.passwordForm.controls[key].setErrors(null);
          });
          this.notifier.success(this.i18n.translate(
            'ACCOUNT:notificationSuccessfullyUpdatedPassword',
            {},
            'Successfully updated your password'
          ));
        } else {
          this.containsUserInfo = response.containsUserInfo;
          this.isPreviousPassword = response.passwordPreviouslyUsed;
          this.errorOnSubmit = this.containsUserInfo || this.isPreviousPassword;
        }
      } else {
        this.notifier.success(this.i18n.translate(
          'ACCOUNT:notificationSuccessfullyUpdatedAccount',
          {},
          'Successfully updated your account'
        ));
      }
    } catch (err) {
      const e = err as HttpErrorResponse;
      if (e.error.message === 'Existing password is not valid') {
        this.notifier.error(this.i18n.translate(
          'ACCOUNT:textPasswordIncorrect',
          {},
          'Incorrect password'
        ));
      } else {
        this.logger.error(e);
        this.notifier.error(this.i18n.translate(
          'ACCOUNT.notificationErrorUpdatingAccount',
          {},
          'There was an error updating your account'
        ));
      }
    }

    await this.applicantService.getApplicant(true);
    this.spinnerService.stopSpinner();
  }

  async viewData () {
    await this.modalFactory.open(
      PersonalDataModalComponent, {
        user: this.applicant
    });
  }

  async deleteAccount () {
    const proceed = await this.modalFactory.open(
      DeleteAccountModalComponent,
      { user: this.applicant }
    );
    if (proceed) {
      this.notifier.success(this.i18n.translate(
        'ACCOUNT:textSuccessfullyDeletedAccount',
        {},
        'Successfully deleted account'
      ));
      this.tokenService.logout();
    }
  }

  async updateSignature () {
    this.spinnerService.startSpinner();
    await this.signatureService.setSignature();
    const savedSignature = this.signatureService.savedSignatureFile;
    this.spinnerService.stopSpinner();

    const response = await this.modalFactory.open(
      SignatureModalComponent,
      {
        savedSignature,
        signatureDescription: this.i18n.translate(
          'common:textChooseOptionUpdateElectronicSignature',
          {},
          'Choose an option to update your electronic signature.'
        ),
        modalHeader: this.i18n.translate(
          'common:hdrUpdateSignature',
          {},
          'Update Signature'
        ),
        primaryButtonText: this.i18n.translate(
          'common:btnSave',
          {},
          'Save'
        ),
        supportsBypass: false
      }
    );
    if (response) {
      this.spinnerService.startSpinner();
      await this.signatureService.handleUpdateSignature(response);
      this.spinnerService.stopSpinner();
    }
  }
}
