
import { HttpErrorResponse } from '@angular/common/http';
import { Component, OnInit } from '@angular/core';
import { Validators } from '@angular/forms';
import { SpinnerService } from '@core/services/spinner.service';
import { TokenService } from '@core/services/token/token.service';
import { UpdateUserProfile, User } from '@core/typings/client-user.typing';
import { ClientSettingsService } from '@features/client-settings/client-settings.service';
import { ClientUserService } from '@features/client-user/client-user.service';
import { SignatureService } from '@features/signature/signature.service';
import { UserService } from '@features/users/user.service';
import { PanelTypes, PasswordService, PasswordValidator, TypeSafeFormBuilder, TypeSafeFormGroup } from '@yourcause/common';
import { I18nService } from '@yourcause/common/i18n';
import { LogService } from '@yourcause/common/logging';
import { ModalFactory } from '@yourcause/common/modals';
import { NotifierService } from '@yourcause/common/notifier';
import { SignatureModalComponent } from '@yourcause/common/signature';
import { DeleteAccountModalComponent } from '../delete-account-modal/delete-account-modal.component';
import { PersonalDataModalComponent } from '../personal-data-modal/personal-data-modal.component';
interface MyAccountGroup {
  firstName: string;
  lastName: string;
}
interface PasswordGroup {
  currentPassword: string;
  newPassword: string;
}

@Component({
  selector: 'gc-my-account-manager',
  templateUrl: './my-account-manager.component.html',
  styleUrls: ['./my-account-manager.component.scss']
})
export class MyAccountManagerComponent implements OnInit {
  user: User;
  formGroup: TypeSafeFormGroup<MyAccountGroup>;
  passwordForm: TypeSafeFormGroup<PasswordGroup>;
  PanelTypes = PanelTypes;
  currentPasswordVisible = false;
  newPasswordVisible = false;
  langKeys = this.clientSettingsService.langKeys;
  errorOnSubmit = false;
  containsUserInfo = false;
  isPreviousPassword = false;

   constructor (
    private logger: LogService,
    private userService: UserService,
    private tokenService: TokenService,
    private modalFactory: ModalFactory,
    private notifier: NotifierService,
    private i18n: I18nService,
    private formBuilder: TypeSafeFormBuilder,
    private clientUserService: ClientUserService,
    private clientSettingsService: ClientSettingsService,
    private spinnerService: SpinnerService,
    private passwordService: PasswordService,
    private signatureService: SignatureService
  ) { }

  get isImpersonating () {
    return this.tokenService.hasImpersonationToken();
  }

  ngOnInit () {
    this.user = this.userService.user;
    this.formGroup = this.formBuilder.group({
      firstName: [this.user.firstName, Validators.required],
      lastName: [this.user.lastName, Validators.required]
    });
    this.passwordForm = this.formBuilder.group({
      currentPassword: '',
      newPassword: [
        '',
        [
          PasswordValidator(this.passwordService)
        ]
      ]
    });
  }

  toggleCurrentPasswordVisible = () => {
    this.currentPasswordVisible = !this.currentPasswordVisible;
  };

  toggleNewPasswordVisible = () => {
    this.newPasswordVisible = !this.newPasswordVisible;
  };

  getLangOnSubmit () {
    if (this.isImpersonating) {
      return this.user.culture;
    } else {
      return this.i18n.language;
    }
  }

  async onSubmit (isPassword = false) {
    this.spinnerService.startSpinner();
    const langOnSubmit = this.getLangOnSubmit();
    const data: UpdateUserProfile = {
      firstName: (this.formGroup.value.firstName || '').trim(),
      lastName: (this.formGroup.value.lastName || '').trim(),
      existingPassword: isPassword ?
        this.passwordForm.value.currentPassword :
        undefined,
      password: isPassword ?
        this.passwordForm.value.newPassword :
        undefined,
      confirmPassword: isPassword ?
        this.passwordForm.value.newPassword :
        undefined,
      culture: langOnSubmit
    };

    await this.updateProfile(data, isPassword);
  }

  passwordChanged () {
    this.errorOnSubmit = false;
    this.containsUserInfo = false;
    this.isPreviousPassword = false;
  }

  async updateProfile (data: UpdateUserProfile, isPassword = false) {
    try {
      const response = await this.clientUserService.updateProfile(data);
      if (isPassword) {
        if (response.validPassword) {
          this.passwordForm.reset();
          Object.keys(this.passwordForm.controls).forEach(key => {
            this.passwordForm.controls[key].setErrors(null);
          });
          this.notifier.success(this.i18n.translate(
            'ACCOUNT.notificationSuccessfullyUpdatedPassword',
            {},
            'Successfully updated your password'
          ));
        } else {
          this.containsUserInfo = response.containsUserInfo;
          this.isPreviousPassword = response.passwordPreviouslyUsed;
          this.errorOnSubmit = this.containsUserInfo || this.isPreviousPassword;
        }
      } else {
        this.notifier.success(this.i18n.translate(
          'ACCOUNT:notificationSuccessfullyUpdatedAccount',
          {},
          'Successfully updated your account'
        ));
      }
    } catch (err) {
      const e = err as HttpErrorResponse;
      if (e.error.message === 'Existing password is not valid') {
        this.notifier.error(this.i18n.translate(
          'ACCOUNT:textPasswordIncorrect',
          {},
          'Incorrect password'
        ));
      } else {
        this.logger.error(e);
        this.notifier.error(this.i18n.translate(
          'ACCOUNT.notificationErrorUpdatingAccount',
          {},
          'There was an error updating your account'
        ));
      }
    }

    await this.clientUserService.getUser(true, true);
    this.spinnerService.stopSpinner();
  }

  async viewData () {
    await this.modalFactory.open(
      PersonalDataModalComponent,
      { user: this.user }
    );
  }

  async deleteAccount () {
    const proceed = await this.modalFactory.open(
      DeleteAccountModalComponent,
      { user: this.user }
    );
    if (proceed) {
      this.notifier.success(this.i18n.translate(
        'ACCOUNT:textSuccessfullyDeletedAccount',
        {},
        'Successfully deleted account'
      ));
      this.tokenService.logout();
    }
  }

  async updateSignature () {
    this.spinnerService.startSpinner();
    await this.signatureService.setSignature();
    const savedSignature = this.signatureService.savedSignatureFile;
    this.spinnerService.stopSpinner();

    const response = await this.modalFactory.open(
      SignatureModalComponent,
      {
        savedSignature,
        signatureDescription: this.i18n.translate(
          'common:textChooseOptionUpdateElectronicSignature',
          {},
          'Choose an option to update your electronic signature.'
        ),
        modalHeader: this.i18n.translate(
          'common:hdrUpdateSignature',
          {},
          'Update Signature'
        ),
        primaryButtonText: this.i18n.translate(
          'common:btnSave',
          {},
          'Save'
        ),
        supportsBypass: false
      }
    );
    if (response) {
      this.spinnerService.startSpinner();
      await this.signatureService.handleUpdateSignature(response);
      this.spinnerService.stopSpinner();
    }
  }

}
