import { Injectable } from '@angular/core';
import { GCMockModule } from '@core/mocks/gc-module.mock';
import { Column, FilterColumn, TypeaheadSelectOption } from '@yourcause/common';
import { BeforeEach, Spec, TestCase } from '@yourcause/test-decorators';
import { DescribeAngularService } from '@yourcause/test-decorators/angular';
import { expect } from 'chai';
import { MyFilterService } from './my-filters.service';
import { FilterPages, MyFilterDetailFromApi, MyFilterFromApi } from './my-filters.typing';


@Injectable({ providedIn: 'root' })
@DescribeAngularService(MyFilterService, {
  imports: [
    GCMockModule
  ]
})
export class MyFilterServiceSpec implements Spec<MyFilterServiceSpec, MyFilterService> {
  firstFilterId = 1;
  secondFilterId = 2;
  myFilters: MyFilterFromApi[] = [{
    id: this.firstFilterId,
    name: '1st Filter',
    pageId: FilterPages.APPLICATIONS
  }, {
    id: this.secondFilterId,
    name: '2nd Filter',
    pageId: FilterPages.APPLICATIONS
  }];
  columnProp = 'formResponse';
  childProp = '52347';
  grandchildProp = '25682';
  grandchildType = 'multiListFuzzyText';
  columnGrandchildOption: TypeaheadSelectOption<Column> = {
    label: 'How many times a year do you travel? (NEW RF)',
    value: {
      columnName: 'How many times a year do you travel? (NEW RF)',
      filterOnly: true,
      label: 'How many times a year do you travel? (NEW RF)',
      labelOnly: true,
      onSelectedAsFilter: null,
      options: [{
        label: 'Always travelling',
        value: 'always'
      }, {
        label: 'Occassionally',
        value: 'occassionally'
      }],
      prop: this.grandchildProp,
      type: this.grandchildType,
      visible: true
    }
  };
  columnChildOption: TypeaheadSelectOption<Column> = {
    label: 'Toys for Toys Donations',
    value: {
      columnName: 'Toys for Tots Donations',
      dependentColumnOptions: [
        this.columnGrandchildOption
      ],
      filterOnly: true,
      label: 'Toys for Tots Donations',
      labelOnly: true,
      options: [],
      prop: this.childProp,
      type: 'text',
      visible: true
    }
  };
  formResponseColumn: Column = {
    columnClass: undefined,
    columnName: 'formResponse',
    dependentColumnOptions: [
      this.columnChildOption
    ],
    display: undefined,
    filterModalKey: undefined,
    filterOnly: true,
    headerVisible: false,
    i18nKey: 'common:hdrFormResponse',
    i18nValue: undefined,
    isActive: true,
    label: 'common:hdrFormResponse',
    labelOnly: undefined,
    locked: false,
    noFiltering: undefined,
    onSelectedAsFilter: undefined,
    options: [],
    prop: this.columnProp,
    rowProp: undefined,
    sortAscending: false,
    templateRef: null,
    type: 'text',
    visible: false
  };
  filterDetail: MyFilterDetailFromApi = {
    id: 1,
    name: 'Test',
    pageId: FilterPages.APPLICATIONS,
    filters: [{
      columnName: 'grantProgramId',
      filters: [{
        filterType: 'eq',
        filterValue: '1234'
      }]
    }]
  };

  @BeforeEach()
  async mock (service: MyFilterService) {
    service['myFilterResources']['getMyFilters'] = async () => {
      return this.myFilters;
    };
    service['myFilterResources']['getFilterDetail'] = async (id) => {
      return {
        id,
        name: 'Eastern Division Applicants',
        pageId: FilterPages.APPLICATIONS,
        filters: [{
        columnName: `formResponse.52347.25682`,
          filters: [{
            filterType: 'eq',
            filterValue: 'always'
          }, {
            filterType: 'eq',
            filterValue: 'occassionally'
          }]
        }]
      };
    };
    await service.setMyFilters();
  }

  @TestCase('should be able to set my filters')
  async setMyFilters (service: MyFilterService) {
    expect(service.myFilters.length).to.be.equal(this.myFilters.length);
    const has2ndFilter = service.myFilters.some((filter) => {
      return filter.id === this.secondFilterId;
    });
    expect(has2ndFilter).to.be.true;
  }

  @TestCase('should be able to set page text on my filters')
  async setPageText (service: MyFilterService) {
    const filtersHavePageText = service.myFilters.every((filter) => {
      return !!filter.pageText;
    });
    expect(filtersHavePageText).to.be.true;
  }

  @TestCase('should be able to get pageText from pageId')
  async getPageText (service: MyFilterService) {
    const pageText = service.getPageTextFromPageId(
      FilterPages.APPLICATIONS
    );
    expect(pageText).to.be.equal('Applications / Nominations');
  }


  @TestCase('should be able to adapt filters to active filters')
  adaptFitlersToActiveFilters (service: MyFilterService) {
    const filterColumns: FilterColumn<any>[] = [{
      columnName: `${this.columnProp}.${this.childProp}.${this.grandchildProp}`,
      filters: [{
        filterType: 'eq',
        filterValue: 'always'
      }, {
        filterType: 'eq',
        filterValue: 'occassionally'
      }]
    }];
    const adapted = service.adaptFitlersToActiveFilters(
      filterColumns,
      [this.formResponseColumn]
    );
    const expectedFilterLength = 1;
    expect(adapted.length).to.equal(expectedFilterLength);
    const adaptedFilter = adapted[0];
    const hasChild = !!adaptedFilter.columnChild;
    const hasGrandchild = !!adaptedFilter.columnGrandchild;
    expect(hasChild).to.be.true;
    expect(hasGrandchild).to.be.true;
    const expectedFilterValue =  ['always', 'occassionally'];
    expect(adaptedFilter.value).to.eql(expectedFilterValue);
  }

  @TestCase('should be able to update name in filter map')
  async updateNameInDetailMap (service: MyFilterService) {
    await service.setFilterDetail(this.firstFilterId);
    service.updateNameInDetailMap(this.firstFilterId, 'New name');
    expect(service.myFilterDetailMap[this.firstFilterId].name).to.be.equal('New name');
  }

  @TestCase('should be able to adapt filter value')
  adaptFilterValue (service: MyFilterService) {
    const adapted = service.adaptFilterDetail(this.filterDetail);

    const adaptedFilterVal = adapted.filters[0].filters[0].filterValue;
    expect(adaptedFilterVal).to.be.equal(1234);
  }

}

