import { ComponentFactoryResolver, Injectable, Injector } from '@angular/core';
import { AppManagerTypes } from '@core/typings/application.typing';
import { ApplicationManagerTableComponent } from '@features/application-manager/application-manager-table/application-manager-table.component';
import { ApplicationManagerResolver } from '@features/application-manager/resolvers/application-manager.resolver';
import { ApplicationManagerService } from '@features/application-manager/services/application-manager/application-manager.service';
import { AutoTableRepositoryFactory, Column, ColumnFilterRow, Columns, FilterColumn, FilterHelpersService, TableFiltersModalResponse, ValueComparisonService } from '@yourcause/common';
import { I18nService } from '@yourcause/common/i18n';
import { LogService } from '@yourcause/common/logging';
import { NotifierService } from '@yourcause/common/notifier';
import { AttachYCState, BaseYCService } from '@yourcause/common/state';
import { MyFilterResources } from './my-filters.resources';
import { MyFilterState } from './my-filters.state';
import { FilterPages, MyFilterDetailFromApi, MyFilterForUi } from './my-filters.typing';


@AttachYCState(MyFilterState)
@Injectable({ providedIn: 'root' })
export class MyFilterService extends BaseYCService<MyFilterState> {

   constructor (
    private logger: LogService,
    private myFilterResources: MyFilterResources,
    private i18n: I18nService,
    private notifier: NotifierService,
    private componentResolver: ComponentFactoryResolver,
    private autoTableFactory: AutoTableRepositoryFactory,
    private filterHelperService: FilterHelpersService,
    private applicationManagerService: ApplicationManagerService,
    private valueComparisonService: ValueComparisonService,
    private applicationManagerResolver: ApplicationManagerResolver
  ) {
    super();
  }

  get myFilters () {
    return this.get('myFilters');
  }

  get myFilterDetailMap () {
    return this.get('myFilterDetailMap');
  }

  get applicationColumns () {
    return this.get('applicationColumns');
  }

  async resetMyFilters () {
    this.set('myFilters', undefined);
    await this.setMyFilters();
  }

  async resetFilterDetail (id: number) {
    this.set('myFilterDetailMap', {
      ...this.myFilterDetailMap,
      [id]: undefined
    });
    await this.setFilterDetail(id);
  }

  async setMyFilters () {
    if (!this.myFilters) {
      const myFilters = await this.myFilterResources.getMyFilters();

      const adapted = myFilters.map<MyFilterForUi>((filter) => {
        return {
          ...filter,
          pageText: this.getPageTextFromPageId(filter.pageId)
        };
      });
      this.set('myFilters', adapted);
    }
  }

  getPageTextFromPageId (pageId: FilterPages) {
    switch (pageId) {
      default:
      case FilterPages.APPLICATIONS:
        return this.i18n.translate(
          'common:hdrApplicationsNominations',
          {},
          'Applications / Nominations'
        );
    }
  }

  async setFilterDetail (id: number) {
    if (!this.myFilterDetailMap[id]) {
      const detail = await this.myFilterResources.getFilterDetail(id);
      const adapted = this.adaptFilterDetail(detail);
      this.set('myFilterDetailMap', {
        ...this.myFilterDetailMap,
        [id]: {
          ...adapted
        }
      });
    }

    return this.myFilterDetailMap[id];
  }

  adaptFilterDetail (filterDetail: MyFilterDetailFromApi) {
    // Backend converts filter values to string,
    // So if the value is a number, we need to convert it to one
    filterDetail.filters.forEach((filter) => {
      filter.filters.forEach((innerFilter) => {
        if (this.valueComparisonService.isNumber(innerFilter.filterValue)) {
          innerFilter.filterValue = +innerFilter.filterValue;
        }
      });
    });

    return filterDetail;
  }

  async resetAllAfterFilterChange (id: number) {
    await Promise.all([
      this.resetMyFilters(),
      this.resetFilterDetail(id)
    ]);
  }

  async ensureDependentOptionsAreSet (columns: Columns) {
    await Promise.all(columns.map(async (col) => {
      if (col.getDependentColumnOptions) {
        const depColOptions = await col.getDependentColumnOptions(col);
        col.dependentColumnOptions = depColOptions;
      }
    }));
  }

  adaptFitlersToActiveFilters (
    filterColumns: FilterColumn<any>[],
    allColumns: Columns
  ): ColumnFilterRow[] {
    return filterColumns.reduce<ColumnFilterRow[]>((acc, filterColumn) => {
      const filterColumnName = filterColumn.columnName;
      const filterColumnProps = filterColumnName.split('.');
      let column = allColumns.find((col) => {
        return col.columnName === filterColumnName;
      });
      let columnChild: Column;
      let columnGrandchild: Column;
      if  (!column) {
        // This means there are children and possible grandchildren to find
        allColumns.forEach((col) => {
          const columnName = col.columnName;
          const columnAttr = filterColumnProps[0];
          const childAttr = filterColumnProps[1];
          const grandchildAttr = filterColumnProps[2];
          if (columnAttr === columnName) {
            column = col;
            col.dependentColumnOptions.forEach((option) => {
              if (option.value.prop === childAttr) {
                columnChild = option.value;
                columnChild.dependentColumnOptions.forEach((opt) => {
                  if (opt.value.prop === grandchildAttr) {
                    columnGrandchild = opt.value;
                  }
                });
              }
            });
          }
        });
      }

      return [
        ...acc,
        ...this.filterHelperService.adaptColumnToColumnFilterRow(
          column,
          filterColumn.filters,
          columnChild,
          columnGrandchild
        )
      ];
    }, []);
  }

  getFilterColumnsByPageType (
    pageId: FilterPages,
    injector: Injector
  ) {
    switch (pageId) {
      default:
      case FilterPages.APPLICATIONS:
        return this.getApplicationColumns(injector);
    }
  }

  getApplicationColumns (injector: Injector) {
    if (!this.applicationColumns) {
      const factory = this.componentResolver.resolveComponentFactory(
        ApplicationManagerTableComponent
      );
      const comp = factory.create(injector);
      const table = comp.instance;
      table.type = AppManagerTypes.ColumnsOnly;
      comp.changeDetectorRef.detectChanges();
      const repo = this.autoTableFactory.getRepository('COLUMNS_ONLY');
      this.set('applicationColumns', repo.uiFilterColumns);
    }

    return this.applicationColumns;
  }

  getTopLevelFiltersByPageType (pageId: FilterPages) {
    switch (pageId) {
      default:
      case FilterPages.APPLICATIONS:
        return this.applicationManagerService.getTopLevelFiltersForApplicationManager();
    }
  }

  updateNameInDetailMap (
    id: number,
    name: string
  ) {
    this.set('myFilterDetailMap', {
      ...this.myFilterDetailMap,
      [id]: {
        ...this.myFilterDetailMap[id],
        name
      }
    });
  }

  async handleAddOrUpdateFilterConfig (
    existingId: number,
    name: string,
    pageId = FilterPages.APPLICATIONS,
    skipSuccessToastr = false
  ) {
    try {
      const filterId = await this.myFilterResources.addOrUpdateFilterConfig(
        existingId,
        name,
        pageId
      );
      await this.resetMyFilters();
      if (existingId) {
        this.updateNameInDetailMap(
          existingId,
          name
        );
      }
      if (!skipSuccessToastr) {
        this.notifier.success(this.i18n.translate(
          'common:textSuccessfullySavedFilter',
          {},
          'Successfully saved the filter'
        ));
      }

      return filterId;
    } catch (e) {
      this.logger.error(e);
      this.notifier.error(this.i18n.translate(
        'common:textErrorSavingFilter',
        {},
        'There was an error saving the filter'
      ));

      return null;
    }
  }

  async saveFilterDetail (
    id: number,
    filtersToSave: ColumnFilterRow[],
    skipToastr = false
  ) {
    try {
      const filters = this.filterHelperService.mapToFilterColumns(filtersToSave);
      await this.myFilterResources.saveFilterDetails(
        id,
        filters
      );
      await this.resetAllAfterFilterChange(id);
      if (!skipToastr) {
        this.notifier.success(this.i18n.translate(
          'common:textSuccessfullySavedFilter',
          {},
          'Successfully saved the filter'
        ));
      }
    } catch (e) {
      this.logger.error(e);
      this.notifier.error(this.i18n.translate(
        'common:textErrorSavingFilter',
        {},
        'There was an error saving the filter'
      ));
    }
  }

  async saveNewFilterWithDetail (
    modalResponse: TableFiltersModalResponse,
    skipToastr = false
  ): Promise<number> {
    try {
      const id = await this.handleAddOrUpdateFilterConfig(
        null,
        modalResponse.filterName,
        FilterPages.APPLICATIONS,
        true
      );
      await this.saveFilterDetail(id, modalResponse.filters, skipToastr);
      await this.resetAllAfterFilterChange(id);

      return id;
    } catch (e) {
      this.logger.error(e);
      this.notifier.error(this.i18n.translate(
        'common:textErrorSavingFilter',
        {},
        'There was an error saving the filter'
      ));

      return null;
    }
  }

  async handleCopyFilter (
    row: MyFilterForUi,
    injector: Injector
  ) {
    try {
      await this.applicationManagerResolver.resolve();
      const detail = await this.setFilterDetail(row.id);
      const columns = this.getFilterColumnsByPageType(
        detail.pageId,
        injector
      );
      const filters = this.adaptFitlersToActiveFilters(
        detail.filters,
        columns
      );
      const copyText = this.i18n.translate('common:textCopy', {}, 'Copy');
      const id = await this.saveNewFilterWithDetail(
        {
          filterName: `${row.name} - ${copyText}`,
          filters
        },
        true
      );
      this.notifier.success(this.i18n.translate(
        'common:textSuccessfullyCopiedFilter',
        {},
        'Successfully copied the filter'
      ));

      return id;
    } catch (e) {
      this.logger.error(e);
      this.notifier.error(this.i18n.translate(
        'common:textErrorCopyingFilter',
        {},
        'There was an error copying the filter'
      ));

      return null;
    }
  }

  async handleDeleteFilter (id: number) {
    try {
      await this.myFilterResources.deleteFilter(id);
      await this.resetMyFilters();
      this.notifier.success(this.i18n.translate(
        'common:textSuccessDeleteFilter',
        {},
        'Successfully deleted the filter'
      ));
    } catch (e) {
      this.logger.error(e);
      this.notifier.error(this.i18n.translate(
        'common:textErrorDeletingFilter',
        {},
        'There was an error deleting the filter'
      ));
    }
  }
}

