import { Component, ElementRef, OnDestroy, OnInit, ViewChild } from '@angular/core';
import { AdminUsersService } from '@core/services/admin-users.service';
import { SpinnerService } from '@core/services/spinner.service';
import { APIAdminClient } from '@core/typings/api/admin-client.typing';
import { AdminClientService } from '@features/platform-admin/clients/admin-client.service';
import { UserService } from '@features/users/user.service';
import { ChartService, DateService, PanelTypes, TableDataDownloadFormat, TypeaheadSelectOption, TypeSafeFormBuilder, TypeSafeFormGroup } from '@yourcause/common';
import { CachedAttr, CACHE_TYPES } from '@yourcause/common/cache';
import { I18nService } from '@yourcause/common/i18n';
import { ModalFactory } from '@yourcause/common/modals';
import { ChartOptions, ChartType } from 'chart.js';
import moment from 'moment';
import { AtRiskActivityTrackingEmailModalComponent } from '../at-risk-activity-tracking-email-modal/at-risk-activity-tracking-email-modal.component';
import { CustomerInsightsService } from '../customer-insights.service';
import { CustomerInsights } from '../customer-insights.typing';

@Component({
  selector: 'gc-customer-insights-page',
  templateUrl: './customer-insights-page.component.html',
  styleUrls: ['./customer-insights-page.component.scss']
})
export class CustomerInsightsPageComponent implements OnInit, OnDestroy {
  clientSelectOptions: TypeaheadSelectOption<number>[] = [];
  dateRangeSelectOptions: TypeaheadSelectOption<string>[] = [];
  viewSelectOptions: TypeaheadSelectOption<string>[] = [
    {
      value: 'chart',
      label: this.i18n.translate(
        'common:lblViewAsChart',
        {},
        'View as chart'
      )
    },
    {
      value: 'table',
      label: this.i18n.translate(
        'common:lblViewAsTable',
        {},
        'View as table'
      )
    }
  ];
  textForLoginComparison: string;
  textForApplicationsComparison: string;
  viewStyleForLogins: 'table'|'chart' = 'chart';
  viewStyleForApplications: 'table'|'chart' = 'chart';
  userId = this.userService.currentUser.id;
  formGroup: TypeSafeFormGroup<{
    client: number;
    year: string;
  }>;
  PanelTypes = PanelTypes;
  applicationData: CustomerInsights.DataSetForChart[];
  loginData: CustomerInsights.DataSetForChart[];
  TableDataDownloadFormat = TableDataDownloadFormat;
  @ViewChild('applicationsChart', {static: true}) applicationsChart: ElementRef<HTMLCanvasElement>;
  @ViewChild('loginChart', {static: true}) loginChart: ElementRef<HTMLCanvasElement>;
  yearSelected: string;
  previousYear: string;
  applicationPercentageChange: number;
  loginPercentageChange: number;
  currentYearOption: TypeaheadSelectOption<string>;
  months = this.dateService.getTranslatedMonths();
  currentMonthIndex = +moment().format('M') - 1;
  chartOptions: ChartOptions<ChartType> = {
    plugins: {
      title: {
        display: true
      },
      tooltip: {},
      legend: {
        position: 'bottom' as any
      }
    },
    scales: {
      yAxes: {
        beginAtZero: true
      }
    },
    elements: {
      line: {
        fill: false
      }
    }
  };

  @CachedAttr(CACHE_TYPES.LOCALSTORAGE)
  selectedClientId: number;

  selectedClient: APIAdminClient.Client;
  snoozeStatusText: string;

  constructor (
    private userService: UserService,
    private adminUserService: AdminUsersService,
    private adminClientService: AdminClientService,
    private formBuilder: TypeSafeFormBuilder,
    private dateService: DateService,
    private customerInsightsService: CustomerInsightsService,
    private spinnerService: SpinnerService,
    private i18n: I18nService,
    private chartService: ChartService,
    private modalFactory: ModalFactory
  ) { }

  async ngOnInit () {
    this.spinnerService.startSpinner();
    await this.setClientOptions();
    this.setDateRangeOptions();
    this.setSelectedClient(this.selectedClientId);
    this.setSnoozeStatusText();
    this.setFormGroup();
    await this.setDashboardCharts(true);
    this.spinnerService.stopSpinner();
  }

  async setClientOptions () {
    const adminDetail = await this.adminUserService.getAdminDetail(this.userId);
    this.clientSelectOptions = this.adminClientService.getClientOptionsForAdmin(adminDetail);
  }

  setSelectedClient (clientId: number) {
    if (!clientId) {
      clientId = this.clientSelectOptions[0].value;
    }
    this.selectedClientId = clientId;
    this.selectedClient = this.adminClientService.clients.find((client) => {
      return client.clientId === clientId;
    });
  }

  setSnoozeStatusText () {
    if (!this.selectedClient.activityAlertsSnoozed) {
      this.snoozeStatusText = this.i18n.translate(
        'common:textActive',
        {},
        'Active'
      );
    } else {
      this.snoozeStatusText = this.i18n.translate(
        'common:textSnoozedUntilDate',
        {
          date: moment(this.selectedClient.snoozedTillDate).format('ll')
        },
        'Snoozed until __date__'
      );
    }
  }

  setDateRangeOptions () {
    // TODO: should this be a hardcoded year?
    this.dateRangeSelectOptions = this.dateService.getYearTypeaheadOptions(2017);
  }

  setFormGroup () {
    this.currentYearOption = this.dateRangeSelectOptions[0];
    this.formGroup = this.formBuilder.group({
      client: this.selectedClientId,
      year: this.currentYearOption.value
    });
    this.setCurrentAndPreviousYear(this.currentYearOption.value);
  }

  setCurrentAndPreviousYear (yearSelected: string) {
    this.yearSelected = moment(yearSelected)
      .get('year')
      .toString();
    this.previousYear = moment(yearSelected)
      .subtract(1, 'year')
      .get('year')
      .toString();
  }

  onClientChange () {
    this.setSelectedClient(this.formGroup.value.client);
    this.setDashboardCharts(false);
  }

  async openAtRiskModal () {
    const clientActivityAlertsSnoozed = this.selectedClient.activityAlertsSnoozed;
    const response = await this.modalFactory.open(
      AtRiskActivityTrackingEmailModalComponent,
      {
        clientName: this.selectedClient.name,
        clientActivityAlertsSnoozed,
        clientSnoozedTilDate: this.selectedClient.snoozedTillDate
      }
    );
    if (response) {
      if (clientActivityAlertsSnoozed && !response.isSnoozed) {
        this.spinnerService.startSpinner();
        await this.adminClientService.handleUnsnooze(
          this.selectedClient.clientId
        );
        this.spinnerService.stopSpinner();
      } else if (!clientActivityAlertsSnoozed && response.isSnoozed) {
        this.spinnerService.startSpinner();
        await this.adminClientService.handleSnoozeSetting(
          this.selectedClient.clientId,
          response
        );
        this.spinnerService.stopSpinner();
      }

      this.setSelectedClient(this.selectedClient.clientId);
      this.setSnoozeStatusText();
    }
  }

  async setDashboardCharts  (onInit: boolean) {
    this.spinnerService.startSpinner();
    const comparingCurrentYear = this.formGroup.value.year === this.currentYearOption.value;
    const { appData, loginData } = await this.customerInsightsService.getAndSetCustomerInsightsData({
      clientId: this.formGroup.value.client,
      date: this.formGroup.value.year
    });
    this.applicationData = appData;
    const applicationUsageTrend = this.customerInsightsService.getUsageTrendData(
      this.applicationData,
      this.formGroup.value.year === this.currentYearOption.value,
      this.currentMonthIndex
    );
    this.applicationPercentageChange = applicationUsageTrend.percentageChange;
    if (this.viewStyleForApplications === 'chart') {
      const applicationChartContext = this.applicationsChart?.nativeElement?.getContext('2d');
      if (this.applicationsChart && onInit) {
        this.chartService.getChartInstance(
          applicationChartContext, {
          type: 'line' as 'line',
          data: {
            labels: this.months,
            datasets: this.applicationData
          }
        } as any);
      }
    }

    this.loginData = loginData;
    const loginUsageTrend = this.customerInsightsService.getUsageTrendData(
      this.loginData,
      this.formGroup.value.year === this.currentYearOption.value,
      this.currentMonthIndex
    );
    this.loginPercentageChange = loginUsageTrend.percentageChange;

    if (this.viewStyleForLogins === 'chart') {
      const loginChartContext = this.loginChart?.nativeElement?.getContext('2d');
      if (loginChartContext && onInit) {
        this.chartService.getChartInstance(
          loginChartContext, {
            type: 'line',
            data: {
              labels: this.months,
              datasets: this.loginData
            },
            options: this.chartOptions
          }
        );
      }
    }
    this.setCurrentAndPreviousYear(this.formGroup.value.year);
    this.setComparisonLanguage(
      comparingCurrentYear,
      loginUsageTrend.previousYearTotal,
      applicationUsageTrend.previousYearTotal
    );

    this.spinnerService.stopSpinner();
  }

  setComparisonLanguage (
    comparingCurrentYear: boolean,
    previousTotalLogins: number,
    previousTotalApplications: number
  ) {
    this.textForLoginComparison = this.i18n.translate(
      'common:ComparedToThisTimeLastYearUpdate',
      {
        date: comparingCurrentYear ? moment().subtract(1, 'year').format('M/YYYY') : this.previousYear,
        previousTotal: previousTotalLogins
      },
      'Compared to __date__ (__previousTotal__)'
    );
    this.textForApplicationsComparison = this.i18n.translate(
      'common:ComparedToThisTimeLastYearUpdate',
      {
        date: comparingCurrentYear ? moment().subtract(1, 'year').format('M/YYYY') : this.previousYear,
        previousTotal: previousTotalApplications
      },
      'Compared to __date__ (__previousTotal__)'
    );
  }

  setViewStyleForApplications (value: 'chart'|'table') {
    this.viewStyleForApplications = value;
  }

  setViewStyleForLogins (value: 'chart'|'table') {
    this.viewStyleForLogins = value;
  }

  async downloadChartData (
    chart: 'applications'|'logins',
    exportFileTypeId: TableDataDownloadFormat
  ) {
    this.spinnerService.startSpinner();
    const adminInsightsType = chart === 'applications' ?
      CustomerInsights.AdminInsightsType.APPS_SUBMITTED :
      CustomerInsights.AdminInsightsType.LOGINS;
    await this.customerInsightsService.getAdminInsightsFile({
      adminInsightsRequestModel: {
        clientId: this.formGroup.value.client,
        date: this.formGroup.value.year
        },
        adminInsightsType,
        exportFileTypeId
      },
      this.selectedClient.name
    );
    this.spinnerService.stopSpinner();
  }

  ngOnDestroy () {
    this.customerInsightsService.resetCustomerInsightData();
  }
}
