import { GCMockModule } from '@core/mocks/gc-module.mock';
import { TableDataDownloadFormat } from '@yourcause/common';
import { BeforeEach, Spec, TestCase } from '@yourcause/test-decorators';
import { DescribeAngularService } from '@yourcause/test-decorators/angular';
import { expect, spy } from 'chai';
import { CustomerInsightsService } from './customer-insights.service';
import { CustomerInsights } from './customer-insights.typing';

@DescribeAngularService(
  CustomerInsightsService,
  {
    imports: [
      GCMockModule
    ]
  }
)
export class CustomerInsightsServiceSpec implements Spec<CustomerInsightsServiceSpec, CustomerInsightsService> {
  clientId = 1234;
  clientDataKey = '12343/11/1990';
  mockDataForLogin: CustomerInsights.DataSetForChart[] = [{
    label: 'providedYear',
    data: [1],
    backgroundColor: 'abc',
    borderColor: 'abc',
    pointBackgroundColor: 'abc',
    pointHoverBackgroundColor: 'abc',
    pointHoverBorderColor: 'abc'
  }, {
    label: 'previousYear',
    data: [1],
    backgroundColor: 'cde',
    borderColor: 'cde',
    pointBackgroundColor: 'cde',
    pointHoverBackgroundColor: 'cde',
    pointHoverBorderColor: 'cde'
  }];
  mockDataForApplications: CustomerInsights.DataSetForChart[] = [{
    label: '1990',
    data: [1],
    backgroundColor: 'abc',
    borderColor: 'abc',
    pointBackgroundColor: 'abc',
    pointHoverBackgroundColor: 'abc',
    pointHoverBorderColor: 'abc'
  }, {
    label: '1989',
    data: [2],
    backgroundColor: 'cde',
    borderColor: 'cde',
    pointBackgroundColor: 'cde',
    pointHoverBackgroundColor: 'cde',
    pointHoverBorderColor: 'cde'
  }];
  unadaptedResponse = {
    clientId: this.clientId,
    providedYearData: {
      '1990/01': 1
    },
    previousYearData: {
      '1989/01': 2
    }
  };
  @BeforeEach()
  prep (service: CustomerInsightsService) {
    service['customerInsightsResources'].getAdminInsightsFile = async () => {
      return 'insightsfileurl';
    };

    service['customerInsightsResources'].getApplicationSubmittedInfo = async () => {
      return this.unadaptedResponse;
    };

    service['customerInsightsResources'].getUserLoginInfo = async () => {
      return {
        clientId: this.clientId,
        providedYearData: {
          January: 1
        },
        previousYearData: {
          January: 1
        }
      };
    };
    service['clientSettingsService'].defaultBranding = {
      brandUtility: 'abc',
      brandPrimary: 'bcd',
      brandSecondary: 'cde'
    };
  }
  @TestCase('should be able to return client data key')
  testReturnClientDataKey (service: CustomerInsightsService) {
    const payload = {
      clientId: this.clientId,
      date: '3/11/1990'
    };
    const key = service.getClientDataKey(payload);
    const expectedKey = '' + payload.clientId + payload.date;
    expect(key).to.equal(expectedKey);
  }

  @TestCase('should be able to return client insights data from state')
  async testReturnClientInsightsDataFromState (service: CustomerInsightsService) {
    service.setCustomerInsightsData(
      this.clientDataKey,
      this.mockDataForApplications,
      this.mockDataForLogin
    );
    const payload = {
      clientId: this.clientId,
      date: '3/11/1990'
    };
    const appEndpointSpy = spy.on(service, 'getApplicationSubmittedDataSet');
    const loginEndpointSpy = spy.on(service, 'getLoginDataSet');
    const dataFromState = await service.getAndSetCustomerInsightsData(payload);
    expect(dataFromState).to.deep.equal({
      appData: this.mockDataForApplications,
      loginData: this.mockDataForLogin
    });
    expect(appEndpointSpy).to.not.have.been.called();
    expect(loginEndpointSpy).to.not.have.been.called();
  }

  @TestCase('should be able to fetch client insights data')
  async testFetchClientInsightsData (service: CustomerInsightsService) {
    const payload = {
      clientId: this.clientId,
      date: '3/11/1990'
    };
    const dataFromState = await service.getAndSetCustomerInsightsData(payload);
    expect(dataFromState).to.deep.equal({
      appData: this.mockDataForApplications,
      loginData: this.mockDataForLogin
    });
  }

  @TestCase('should be able to adapt insights data for charts')
  testAdaptInsightsDataForCharts (service: CustomerInsightsService) {
    const adaptedData = service.adaptDataForChart(this.unadaptedResponse);
    expect(adaptedData).to.deep.equal(this.mockDataForApplications);
  }

  @TestCase('should be able to get insights file for download')
  async testGetInsightsFileForDownload (service: CustomerInsightsService) {
    const appEndpointSpy = spy.on(service['customerInsightsResources'], 'getAdminInsightsFile');
    await service.getAdminInsightsFile(
      {
        adminInsightsType: CustomerInsights.AdminInsightsType.APPS_SUBMITTED,
        exportFileTypeId: TableDataDownloadFormat.XLSX,
        adminInsightsRequestModel: {
          clientId: this.clientId,
          date: '03/11/1990'
        }
      },
      ''
    );
    expect(appEndpointSpy).to.have.been.called();
  }

  @TestCase('should be able to get usage trend data')
  testGetUsageTrendData (service: CustomerInsightsService) {
    const usageData = service.getUsageTrendData(this.mockDataForApplications, false, 1);
    expect(usageData.previousYearTotal).to.equal(2);
    expect(usageData.percentageChange).to.equal(50);
  }

  @TestCase('should be able to reset client insights data')
  async testResttingClientInsightsData (service: CustomerInsightsService) {
    service.setCustomerInsightsData(
      this.clientDataKey,
      this.mockDataForApplications,
      this.mockDataForLogin
    );
    const payload = {
      clientId: this.clientId,
      date: '3/11/1990'
    };
    const dataFromState = await service.getAndSetCustomerInsightsData(payload);
    expect(dataFromState).to.deep.equal({
      appData: this.mockDataForApplications,
      loginData: this.mockDataForLogin
    });
    service.resetCustomerInsightData();
    expect(service.customerInsightData).to.be.null;
  }

}
