import { Injectable } from '@angular/core';
import { GCMockModule } from '@core/mocks/gc-module.mock';
import { APIAdminClient } from '@core/typings/api/admin-client.typing';
import { APIDataImport } from '@core/typings/api/data-import.typing';
import { ReferenceFieldAPI } from '@core/typings/api/reference-fields.typing';
import { ApplicationFormsImportModel, ApplicationsInKindImportModel, AwardsImportModel, EmployeeApplicantInfoModel } from '@core/typings/data-import.typing';
import { ReferenceFieldsUI } from '@core/typings/ui/reference-fields.typing';
import { FormAudience, FormTypes } from '@features/configure-forms/form.typing';
import { DataHubReportStatus } from '@features/reporting/data-hub/data-hub.typing';
import { YcFile } from '@yourcause/common';
import { BeforeEach, Spec, TestCase } from '@yourcause/test-decorators';
import { DescribeAngularService } from '@yourcause/test-decorators/angular';
import { expect, spy } from 'chai';
import { HistoricalImportService } from './historical-import.service';

@Injectable({ providedIn: 'root' })
@DescribeAngularService(HistoricalImportService, {
  imports: [GCMockModule]
})
export class HistoricalImportServiceSpec implements Spec<HistoricalImportServiceSpec, HistoricalImportService> {
  mockBatches: APIDataImport.DataImportBatchModel[] = [{
    id: 123,
    name: 'mockBatch',
    clientName: 'clientName',
    isProcessed: false,
    dateProcessed: '5/5/2005',
    clientId: 456,
    applicantsEmployeesOfClient: true,
    scheduledJobId: 101,
    scheduledJobStatus: DataHubReportStatus.PENDING,
    scheduledJobStartDate: '1/1/2001',
    scheduledJobCompletionDate: '2/2/2002',
    scheduledJobCreatedDate: '3/3/2003',
    scheduledJobUpdatedDate: '4/4/2004'
  }];

  applicationFormsImportData: ApplicationFormsImportModel[] = [{
    ImportId: '1',
    FormId: 2,
    FormRevisionId: 3,
    IsDraft: false,
    SubmittedDate: '10/10/2020',
    ApplicationFormStatusId: 1,
    StatusUpdatedDate: '10/10/2020',
    StatusNotes: 'nice',
    WorkFlowLevelId: 123,
    ApplicationImportId: '123',
    FormData: '{ "key": "abc" }'
  }];
  numberField: ReferenceFieldAPI.ReferenceFieldDisplayModel = {
    isEncrypted: false,
    isMasked: false,
    aggregateType: null,
    referenceFieldId: 1,
    parentReferenceFieldId: null,
    formCount: 0,
    usedOnReports: false,
    createdBy: '',
    createDate: '',
    updatedBy: '',
    updateDate: '',
    customDataTableGuid: '',
    name: 'Number',
    description: '',
    type: ReferenceFieldsUI.ReferenceFieldTypes.Number,
    key: 'number',
    supportsMultiple: false,
    categoryId: 3,
    formAudience: FormAudience.APPLICANT,
    isSingleResponse: true,
    tableAllowsImport: false,
    isTableField: false,
    aggregateTableReferenceFieldId: null,
    referenceFieldTableId: null,
    standardComponentIsPublished: true,
    isStandardProductField: false,
    subsetCollectionType: null
  };
  tableField: ReferenceFieldAPI.ReferenceFieldDisplayModel = {
    isEncrypted: false,
    isMasked: false,
    aggregateType: null,
    referenceFieldId: 4,
    parentReferenceFieldId: null,
    formCount: 0,
    usedOnReports: false,
    createdBy: '',
    createDate: '',
    updatedBy: '',
    updateDate: '',
    customDataTableGuid: '',
    name: 'Table field',
    description: '',
    type: ReferenceFieldsUI.ReferenceFieldTypes.Table,
    key: 'tableKey',
    supportsMultiple: false,
    categoryId: null,
    formAudience: FormAudience.APPLICANT,
    isSingleResponse: true,
    tableAllowsImport: false,
    isTableField: false,
    aggregateTableReferenceFieldId: null,
    referenceFieldTableId: null,
    standardComponentIsPublished: true,
    isStandardProductField: false,
    subsetCollectionType: null
  };
  referenceFields = [
    this.numberField,
    this.tableField
  ];
  applicationFormsImportSelectedForm: APIAdminClient.GrantProgramForm = {
    formId: 2,
    formName: 'My Form',
    formDescription: 'It is mine',
    formRevisionId: 2,
    formType: {
      id: FormTypes.REQUEST,
      name: 'formtype',
      audience: FormAudience.APPLICANT
    },
    isDefault: true,
    formDefinition: [{
      tabName: 'tab1',
      logic: null,
      uniqueId: '123',
      index: 0,
      components: [{
        type: `referenceFields-${this.numberField.key}`,
        key: this.numberField.key
      }, {
        type: 'amountRequested',
        key: 'key2'
      }]
    }]
  };

  mockCSVReplacementFile = 'ImportId,IsDraft,SubmittedDate,ApplicationFormStatusId,StatusUpdatedDate,StatusNotes,WorkFlowLevelId,ApplicationImportId,FormData,FormId,FormRevisionId\n1,false,10/10/2020,1,10/10/2020,nice,123,123,"{""FormId"":2,""FormRevisionId"":3,""FormData"":""{ \\""key\\"": \\""abc\\"" }""}",2,2';

  applicationFormImportAttrs = ['ImportId', 'IsDraft', 'SubmittedDate', 'ApplicationFormStatusId', 'StatusUpdatedDate', 'StatusNotes', 'WorkFlowLevelId', 'ApplicationImportId'];

  private firstHalf = Array(Math.floor(HistoricalImportService['maxBatchFileSize'] / 6 /* 6 chars per row */) - 1 /* compensate for header */).fill('a,b,c').join('\n');
  private secondHalf = Array(Math.floor(HistoricalImportService['maxBatchFileSize'] / 6 /* 6 chars per row */) - 1 /* compensate for header */).fill('d,e,f').join('\n');
  private headers = '1,2,3\n';

  private generateLargeCsv () {
    const firstHalf = this.firstHalf;
    const secondHalf = this.secondHalf;
    const headers = this.headers;
    // simulate a 10k csv with headers 1, 2, 3 and values a, b, c (for first half) and d, e, f (for second half)
    const rows = headers + firstHalf + '\n' + secondHalf;

    return { rows, headers };
  }

  private mockBatchFunctions (service: HistoricalImportService) {
    const mockFileBatchId = 0;
    service['getFileBatchIdentifier'] = async () => {
      return mockFileBatchId;
    };
    service['commitFileBatch'] = async () => { };

    const eyeSpy = spy(async () => { });
    service['historicalImportResources']['addFileToBatch'] = eyeSpy;

    return { eyeSpy, mockFileBatchId };
  }

  @BeforeEach()
  async mock (service: HistoricalImportService) {
    service['referenceFieldService']['referenceFieldsResources'].searchReferenceFields = async () => {
      return {
        records: this.referenceFields,
        recordCount: this.referenceFields.length
      };
    };
    service['referenceFieldService']['referenceFieldsResources'].getCategories = async () => {
      return [];
    };
    await service['referenceFieldService'].resolve();
    service['historicalImportResources']['getBatches'] = async () => {
      return {
        records: this.mockBatches,
        recordCount: this.mockBatches.length
      };
    };
  }

  @TestCase('should be able to properly batch files')
  async testShouldBatchFiles (service: HistoricalImportService) {
    const { eyeSpy } = this.mockBatchFunctions(service);

    const { rows } = this.generateLargeCsv();

    const batchId = 1;
    const fileType = APIDataImport.DataImportFileType.ApplicationReferenceFields;
    const file = rows;
    const timezone = '';

    await service.addFileToBatch(batchId, fileType, file, timezone, '');

    // make sure it was called twice
    expect(eyeSpy).to.have.been.called.twice;
  }


  @TestCase('should not batch when a file is used')
  async testShouldNotBatchFileFiles (service: HistoricalImportService) {
    const { eyeSpy } = this.mockBatchFunctions(service);

    const batchId = 1;
    const fileType = APIDataImport.DataImportFileType.ApplicationReferenceFields;
    const file = new YcFile('', new File([], ''));
    const timezone = '';

    await service.addFileToBatch(batchId, fileType, file, timezone, file.fileName);

    // make sure it was called once
    expect(eyeSpy).to.have.been.called.once;
    // and with only 4 arguments (since we aren't batching)
    expect(eyeSpy).to.have.been.called.with(batchId, fileType, file, timezone);
  }

  @TestCase('should be able to set current batch')
  async testSetCurrentBatch (service: HistoricalImportService) {
    const getBatchesSpy = spy.on(service['historicalImportResources'], 'getBatches');
    await service.setCurrentBatch(404);

    expect(getBatchesSpy).to.have.been.called();
    expect(service.currentBatch).to.equal(this.mockBatches[0]);
  }

  @TestCase('should be able to get batches')
  async testGetBatches (service: HistoricalImportService) {
    const getBatchesSpy = spy.on(service['historicalImportResources'], 'getBatches');
    await service.getBatches(null);
    expect(getBatchesSpy).to.have.been.called();
  }

  @TestCase('should be able to get the number of pending batches')
  async getNumberOfPendingBatches (service: HistoricalImportService) {
    const getBatchesSpy = spy.on(service['historicalImportResources'], 'getBatches');
    const numberOfPendingBatches = await service.getNumberOfPendingBatches(1);
    expect(getBatchesSpy).to.have.been.called();
    expect(numberOfPendingBatches).to.equal(this.mockBatches.length);
  }

  @TestCase('should be able to delete batch')
  async testDeleteBatch (service: HistoricalImportService) {
    const deleteBatchSpy = spy.on(service['historicalImportResources'], 'deleteBatch');
    await service.handleDeleteBatch({id: null} as APIDataImport.DataImportBatchModel);

    expect(deleteBatchSpy).to.have.been.called();
  }

  @TestCase ('should be able to get file batch identifier')
  async testGetFileBatchIdentifier (service: HistoricalImportService) {
    const dataImportBatchFileBatchId = 1;
    service['historicalImportResources']['createFileBatch'] = async () => {
      return {
        dataImportBatchFileBatchId
      };
    };
    const createFileBatchSpy = spy.on(service['historicalImportResources'], 'createFileBatch');
    const result = await service.getFileBatchIdentifier(1, APIDataImport.DataImportFileType.Applicants, 'fileName');

    expect(createFileBatchSpy).to.have.been.called();
    expect(result).to.equal(dataImportBatchFileBatchId);
  }

  @TestCase('should be able to commit file batch')
  async testCommitFileBatch (service: HistoricalImportService) {
    const commitFileBatchSpy = spy.on(service['historicalImportResources'], 'commitFileBatch');
    await service.commitFileBatch(null, null);

    expect(commitFileBatchSpy).to.have.been.called();
  }

  @TestCase('should be able to create batch')
  async testCreateBatch (service: HistoricalImportService) {
    const createBatchSpy = spy.on(service['historicalImportResources'], 'createBatch');
    await service.handleCreateBatch(null);

    expect(createBatchSpy).to.have.been.called();
  }

  @TestCase('should be able to get model by model type')
  testGetModelByModelType (service: HistoricalImportService) {
    const applicationsInKindType = service.getModelByModelType(
      'ApplicationInKindItems'
    );
    const awardsType = service.getModelByModelType(
      'Awards'
    );
    const employeeApplicationInfoType = service.getModelByModelType(
      'EmployeeApplicantInfo'
    );

    expect(applicationsInKindType).to.equal(ApplicationsInKindImportModel);
    expect(awardsType).to.equal(
      AwardsImportModel
    );
    expect(employeeApplicationInfoType).to.equal(
      EmployeeApplicantInfoModel
    );
  }

  @TestCase('should be able to map application form data')
  testMappingApplicationFormData (service: HistoricalImportService) {
    const { replacementFile } = service.mapApplicationForm(this.applicationFormsImportData, this.applicationFormsImportSelectedForm);

    expect(replacementFile).to.equal(this.mockCSVReplacementFile);
  }

  @TestCase('should be able to get sample CSV')
  testGettingSampleCSV (service: HistoricalImportService) {
    const getAttrsSpy = spy.on(service, 'getAttrs');
    const downloadStringSpy = spy.on(service['fileService'], 'downloadString');
    service.getSampleCSV(AwardsImportModel);

    expect(getAttrsSpy).to.have.been.called();
    expect(downloadStringSpy).to.have.been.called();
  }

  @TestCase('should be able to get attributes')
  testGetAttrs (service: HistoricalImportService) {
    const attrs = service.getAttrs(ApplicationFormsImportModel);

    expect(attrs).to.deep.equal(this.applicationFormImportAttrs);
  }

  @TestCase('should be able to extract columns from form')
  testExtractingColumnsFromForm (service: HistoricalImportService) {
    const columns = service.extractColumnsFromForm(this.applicationFormsImportSelectedForm);

    expect(columns).to.deep.equal([this.numberField.key]);
  }

  @TestCase('should be able to call validate endpoint')
  testCallingValidateEndpoint (service: HistoricalImportService) {
    const validateBatchSpy = spy.on(service['historicalImportResources'], 'validateBatch');
    service.getValidationErrors(1);

    expect(validateBatchSpy).to.have.been.called();
  }
}
