import { Injectable, Type } from '@angular/core';
import { APIAdminClient } from '@core/typings/api/admin-client.typing';
import { APIDataImport } from '@core/typings/api/data-import.typing';
import { ApplicationFormsImportModel, FormFieldResponse, HistoricalDataImportModel, HistoricalDataImportModels } from '@core/typings/data-import.typing';
import { ReferenceFieldsUI } from '@core/typings/ui/reference-fields.typing';
import { FormDefinitionComponent } from '@features/configure-forms/form.typing';
import { CustomDataTablesService } from '@features/custom-data-tables/custom-data-table.service';
import { ComponentHelperService } from '@features/formio/services/component-helper/component-helper.service';
import { ReferenceFieldsService } from '@features/reference-fields/services/reference-fields.service';
import { DataHubReportStatus } from '@features/reporting/data-hub/data-hub.typing';
import { DynamicCSVImportService, FileService, getKeys, PaginationOptions, SimpleStringMap, YcFile } from '@yourcause/common';
import { I18nService } from '@yourcause/common/i18n';
import { LogService } from '@yourcause/common/logging';
import { NotifierService } from '@yourcause/common/notifier';
import { AttachYCState, BaseYCService } from '@yourcause/common/state';
import { GuidService } from '@yourcause/common/utils';
import { plainToClass } from 'class-transformer';
import { unparse } from 'papaparse';
import { HistoricalImportResources } from './historical-import.resources';
import { HistoricalImportState } from './historical-import.state';


@AttachYCState(HistoricalImportState)
@Injectable({ providedIn: 'root' })
export class HistoricalImportService extends BaseYCService<HistoricalImportState> {
  static maxBatchFileSize? = 1048576;

  constructor (
    private logger: LogService,
    private historicalImportResources: HistoricalImportResources,
    private notifier: NotifierService,
    private i18n: I18nService,
    private csvService: DynamicCSVImportService,
    private fileService: FileService,
    private referenceFieldService: ReferenceFieldsService,
    private customDataTableService: CustomDataTablesService,
    private guidService: GuidService,
    private componentHelper: ComponentHelperService
  ) {
    super();
  }

  get currentBatch () {
    return this.get('currentBatch');
  }

  async setCurrentBatch (
    batchId: number
  ): Promise<APIDataImport.DataImportBatchModel|null> {
    if (this.currentBatch?.id !== batchId) {
      const paginationOptions: PaginationOptions<APIDataImport.DataImportBatchModel> = {
        returnAll: false,
        rowsPerPage: 1,
        pageNumber: 1,
        sortColumns: [{
          columnName: 'id',
          sortAscending: false
        }],
        filterColumns: [{
        columnName: 'id',
        filters: [{
            filterType: 'eq',
            filterValue: batchId
          }]
        }],
        retrieveTotalRecordCount: false
      };
      const results = await this.historicalImportResources.getBatches(
        paginationOptions
      );

      if (results.records.length > 0) {
        const batch = results.records[0];
        this.set('currentBatch', batch);

        return batch;
      }

      return null;
    }

    return this.currentBatch;
  }

  getBatches (
    paginationOptions: PaginationOptions<APIDataImport.DataImportBatchModel>,
    clientId?: number
  ) {
    return this.historicalImportResources.getBatches(
      paginationOptions,
      clientId
    );
  }

  async getNumberOfPendingBatches (clientId: number): Promise<number> {
    const paginationOptions: PaginationOptions<APIDataImport.DataImportBatchModel> = {
      returnAll: false,
      rowsPerPage: 1,
      pageNumber: 1,
      sortColumns: [{
        columnName: 'dateProcessed',
        sortAscending: false
      }],
      filterColumns: [{
        columnName: 'scheduledJobStatus',
        filters: [{
          filterType: 'eq',
          filterValue: DataHubReportStatus.PENDING
        }, {
          filterType: 'eq',
          filterValue: DataHubReportStatus.PROCESSING
        }]
      }],
      retrieveTotalRecordCount: true
    };
    const results = await this.historicalImportResources.getBatches(
      paginationOptions,
      clientId
    );

    return results.recordCount;
  }

  async handleDeleteBatch (batch: APIDataImport.DataImportBatchModel) {
    try {
      await this.historicalImportResources.deleteBatch(batch.id);
      this.notifier.success(this.i18n.translate(
        'MANAGE:notificatonBatchDeleteSuccess',
        {},
        'Batch deleted successfully'
      ));
    } catch (e) {
      this.logger.error(e);
      this.notifier.error(this.i18n.translate(
        'MANAGE:notificationBatchDeleteError',
        {},
        'There was an error deleting the batch'
      ));
    }
  }

  getValidationErrors (batchId: number): Promise<APIDataImport.ValidationError[]> {
    return this.historicalImportResources.validateBatch(batchId);
  }

  async getFileBatchIdentifier (
    batchId: number,
    fileType: APIDataImport.DataImportFileType,
    fileName: string
  ): Promise<number> {
    const result = await this.historicalImportResources.createFileBatch(
      batchId,
      fileType,
      fileName
    );

    return result.dataImportBatchFileBatchId;
  }

  async commitFileBatch (fileBatchId: number, batchId: number) {
    await this.historicalImportResources.commitFileBatch(
      fileBatchId,
      batchId
    );
  }

  async addFileToBatch (
    fileImportBatchId: number,
    fileType: APIDataImport.DataImportFileType,
    file: YcFile|string,
    timezone: string,
    fileName: string,
    paymentBatchName?: string
  ) {
    const fileBatchId = await this.getFileBatchIdentifier(
      fileImportBatchId,
      fileType,
      fileName
    );
    if (typeof file === 'string') {
      const batches = await this.csvService.breakUpFile(file, HistoricalImportService.maxBatchFileSize);

      for (const batch of batches) {
        await this.historicalImportResources.addFileToBatch(
          fileImportBatchId,
          fileType,
          batch,
          timezone,
          fileBatchId,
          paymentBatchName
        );
      }
    } else {
      await this.historicalImportResources.addFileToBatch(
        fileImportBatchId,
        fileType,
        file,
        timezone,
        fileBatchId,
        paymentBatchName
      );
    }

    return fileBatchId;
  }

  async handleCreateBatch (batch: APIDataImport.CreateImportBatch) {
    try {
      await this.historicalImportResources.createBatch(batch);
      this.notifier.success(this.i18n.translate(
        'MANAGE:textSuccessfullyCreatedTheBatch',
        {},
        'Successfully created the batch'
      ));
    } catch (e) {
      this.logger.error(e);
      this.notifier.error(this.i18n.translate(
        'MANAGE:notificationBatchCreateError',
        {},
        'There was an error creating the batch'
      ));
    }
  }

  getModelByModelType (
    type: keyof typeof APIDataImport.DataImportFileType
  ): HistoricalDataImportModel {
    const actualType = APIDataImport.DataImportFileType[type];

    return HistoricalDataImportModels.find(model => model.FileType === actualType);
  }

  mapApplicationForm (
    data: ApplicationFormsImportModel[],
    selectedForm: APIAdminClient.GrantProgramForm
  ): {
    replacementFile: string;
    ancillaryFile: string;
  } {
    const mandatoryKeys = this.getAttrs(ApplicationFormsImportModel);

    const responses: FormFieldResponse[] = [];

    const components: SimpleStringMap<FormDefinitionComponent> = {};
    const definition = selectedForm.formDefinition instanceof Array ?
      selectedForm.formDefinition :
      [selectedForm.formDefinition];
    definition.forEach((tab) => {
      this.componentHelper.eachComponent(tab.components, (comp) => {
        components[comp.key] = comp;
      });
    });

    const result = unparse(data.map(row => {

      row.ImportId = row.ImportId || this.guidService.nonce();

      row.FormData = JSON.stringify(Object.keys(row)
        .filter(key => !mandatoryKeys.includes(key))
        .map(key => key as keyof ApplicationFormsImportModel)
        .reduce((acc, key: keyof ApplicationFormsImportModel) => {
          let val: any = row[key];
          const comp = components[key];
          if (comp) {
            if (comp.type === 'content') {
              // all comps come through (including content components)
              // skip comps that don't have a ref field
              return acc;
            }
            const refFieldKey = this.componentHelper.getRefFieldKeyFromCompType(comp.type);
            const refField = this.referenceFieldService.getReferenceFieldByKey(refFieldKey);

            if (!refField) {
              return acc;
            }

            if (
              refField.type === ReferenceFieldsUI.ReferenceFieldTypes.SelectBoxes ||
              refField.type === ReferenceFieldsUI.ReferenceFieldTypes.CustomDataTable ||
              refField.type === ReferenceFieldsUI.ReferenceFieldTypes.Radio
            ) {
              const dataTable = this.customDataTableService.customDataTables
                .find(cdt => cdt.guid === refField.customDataTableGuid);

              const values = this.customDataTableService.customDataTableMap[dataTable.id];
              val = ('' + (<string>row[key] || ''))
                .split(',') // client admins will upload the CSVs as comma separated
                .map((_val) => {
                  const foundValue = values.find(value => [value.key, ...value.values.map(({ text }) => text)].includes(_val));

                  return foundValue ? foundValue.key : _val;
                })
                .join(',');
            } else if (
              refField.supportsMultiple && (
                refField.type === ReferenceFieldsUI.ReferenceFieldTypes.TextArea ||
                refField.type === ReferenceFieldsUI.ReferenceFieldTypes.TextField
              )
            ) {
              val = ('' + (<string>row[key] || ''))
                .split(
                  ','
                )
                .join(this.referenceFieldService.getSupportsMultipleSeparator(refField)) as any;
            }

            responses.push({
              ApplicationFormsImportId: row.ImportId,
              ApplicationImportId: row.ApplicationImportId,
              DateValue: refField.type === ReferenceFieldsUI.ReferenceFieldTypes.Date ?
                '' + val :
                null,
              CurrencyValue: '',
              ImportId: null,
              ReferenceFieldId: refField.referenceFieldId,
              Value: '' + val
            });
          }
          const returnVal = {
            ...acc,
            [key]: val
          };

          delete row[key];

          return returnVal;
        }, {}));

      row.FormId = selectedForm.formId;
      row.FormRevisionId = selectedForm.formRevisionId;

      return row;
    }), {
      header: true,
      newline: '\n'
    });

    const filteredResponses = responses.filter(value => {
      return !!value.Value && value.Value !== 'null';
    });

    const ancillaryFile = unparse(filteredResponses, { header: true });

    return {
      ancillaryFile,
      replacementFile: result
    };
  }

  getSampleCSV<T> (
    model: Type<T>,
    selectedForm?: APIAdminClient.GrantProgramForm
  ) {
    const attrs = this.getAttrs(model, selectedForm);
    const csv = attrs.join(',');

    this.fileService.downloadString(csv, 'text/csv', `sample_${model.name}.csv`);
  }

  getAttrs (
    model: Type<any>,
    selectedForm?: APIAdminClient.GrantProgramForm
  ) {
    const validationAttrs = getKeys(model);

    const nonValidationAttrs = Object.keys(plainToClass(model, {}));

    const allWithDupes = validationAttrs.concat(nonValidationAttrs);

    return allWithDupes
      .filter((attr, index) => allWithDupes.indexOf(attr) === index)
      .concat(selectedForm ? this.extractColumnsFromForm(selectedForm) : [])
      .filter((attr) => attr.toLowerCase() !== 'formdata');
  }

  extractColumnsFromForm (
    form: APIAdminClient.GrantProgramForm
  ): string[] {
    const attrs: string[] = [];
    let formDefinition = form.formDefinition;

    if (!(formDefinition instanceof Array)) {
      formDefinition = [formDefinition];
    }

    formDefinition.forEach((tab) => {
      this.componentHelper.eachComponent(tab.components, (comp) => {
        const refField = this.referenceFieldService.getReferenceFieldFromCompType(comp.type);
        const passesType = ![
          ReferenceFieldsUI.ReferenceFieldTypes.Table,
          ReferenceFieldsUI.ReferenceFieldTypes.Subset,
          ReferenceFieldsUI.ReferenceFieldTypes.ExternalAPI,
          ReferenceFieldsUI.ReferenceFieldTypes.Aggregate,
          ReferenceFieldsUI.ReferenceFieldTypes.FileUpload
        ].includes(refField?.type);
        if (refField && passesType) {
          attrs.push(comp.key);
        }
      });
    });

    return attrs;
  }
}
