import { Injectable } from '@angular/core';
import { PermissionSet } from '@core/typings/permission.typing';
import { BaseRole, RoleToSave } from '@core/typings/roles.typing';
import { RolesService } from '@features/roles/roles.service';
import { UserService } from '@features/users/user.service';
import { ArrayHelpersService, TypeaheadSelectOption } from '@yourcause/common';
import { I18nService } from '@yourcause/common/i18n';
import { LogService } from '@yourcause/common/logging';
import { NotifierService } from '@yourcause/common/notifier';
import { AttachYCState, BaseYCService } from '@yourcause/common/state';
import { startCase } from 'lodash';
import { RolesAndPermissionResources } from './roles-and-permissions.resources';
import { RolesAndPermissionsPlatformState } from './roles-and-permissions.state';

export const PLATFORM_PERMISSION_SET_TYPE = 20; // Platform only has 1 set. Could not start at 1 because GM permission translations use this number in the key.

@AttachYCState(RolesAndPermissionsPlatformState)
@Injectable({ providedIn: 'root' })
export class RolesAndPermissionsPlatformService extends BaseYCService<RolesAndPermissionsPlatformState> {

   constructor (
    private logger: LogService,
    private roleAndPermissionResources: RolesAndPermissionResources,
    private i18n: I18nService,
    private notifier: NotifierService,
    private userService: UserService,
    private roleService: RolesService,
    private arrayHelper: ArrayHelpersService
  ) {
    super();
  }

  get roles () {
    return this.get('roles');
  }

  get permissions () {
    return this.get('permissions');
  }

  async setAllRolesAndPermissions () {
    await Promise.all([
      this.setAllRoles(),
      this.setAllPermissions()
    ]);
  }

  async setAllRoles () {
    if (!this.roles) {
      const roles = await this.roleAndPermissionResources.getRoles();
      this.set('roles', roles);
    }
  }

  async setAllPermissions () {
    if (!this.permissions) {
      const permissions = await this.roleAndPermissionResources.getPermissions();
      permissions.forEach((permission) => {
        permission.permissionName = startCase(permission.permissionName);
      });
      this.set('permissions', this.arrayHelper.sort(permissions, 'permissionName'));
    }
  }

  async resetRoles () {
    this.set('roles', undefined);
    await this.setAllRoles();
  }

  getPermissionSets () {
    return [{
      type: PLATFORM_PERMISSION_SET_TYPE,
      name: this.i18n.translate(
        'GLOBAL:textSystem',
        {},
        'System'
      ),
      description: '',
      permissions: this.permissions.map((permission) => {
        return {
          type: permission.permissionId,
          name: permission.permissionName,
          description: ''
        };
      })
    }];
  }

  getRoleById (id: number): BaseRole {
    const found = this.get('roles').find((role) => role.systemRole.id === id);

    return {
      roleName: found.systemRole.name,
      roleDescription: found.systemRole.description,
      allowedPermissions: found.systemRole.permissions.map((permission) => {
        return {
          permissionSetType: PLATFORM_PERMISSION_SET_TYPE,
          permissionType: permission.permissionId,
          allowed: true
        };
      }),
      deniedPermissions: []
    };
  }

  getPermissionSetOptions () {
    return this.getPermissionSets().map<TypeaheadSelectOption<PermissionSet>>((set) => ({
      label: this.i18n.translate(
        this.roleService.getPermissionSetKey(set.type),
        {},
        set.name
      ),
      value: set
    }));
  }

  async saveRole (role: RoleToSave) {
    try {
      const roleToSave = {
        roleId: role.roleId,
        roleName: role.roleName,
        roleDescription: role.roleDescription,
        active: true,
        permissionIds: role.allowedPermissions.map((permission) => {
          return permission.permissionType;
        })
      };
      await this.roleAndPermissionResources.saveRole(roleToSave);
      await this.userService.setAdminPermissions();
      await this.resetRoles();
      this.notifier.success(this.i18n.translate(
        'common:textSuccessfullySavedRole',
        {},
        'Successfully saved the role'
      ));
    } catch (e) {
      this.logger.error(e);
      this.notifier.error(this.i18n.translate(
        'common:textErrorSavingRole',
        {},
        'There was an error saving the role'
      ));
    }
  }
}
