
import { Component, OnInit } from '@angular/core';
import { Router } from '@angular/router';
import { PolicyService } from '@core/services/policy.service';
import { SpinnerService } from '@core/services/spinner.service';
import { AdHocReportingAPI } from '@core/typings/api/ad-hoc-reporting.typing';
import { ClientSettingsService } from '@features/client-settings/client-settings.service';
import { StandardProductConfigurationService } from '@features/platform-admin/standard-product-configuration/standard-product-configuration.service';
import { StandardProductTypes } from '@features/platform-admin/standard-product-configuration/standard-product-configuration.typing';
import { ALL_SKIP_FILTER, TableDataDownloadFormat, TopLevelFilter } from '@yourcause/common';
import { CachedAttr, CACHE_TYPES } from '@yourcause/common/cache';
import { I18nService } from '@yourcause/common/i18n';
import { ConfirmationModalComponent, ModalFactory } from '@yourcause/common/modals';
import { NotifierService } from '@yourcause/common/notifier';
import { AdHocReportingService } from '../../services/ad-hoc-reporting.service';
import { ManageAdHocReportModalComponent } from '../manage-ad-hoc-report-modal/manage-ad-hoc-report-modal.component';
import { SendReportModalComponent } from '../send-report-modal/send-report-modal.component';
import { ShareReportModalComponent } from '../share-report-modal/share-report-modal.component';

@Component({
  selector: 'gc-report-list-view-page',
  templateUrl: './ad-hoc-report-list-view-page.component.html',
  styleUrls: ['./ad-hoc-report-list-view-page.component.scss']
})
export class AdHocReportListViewPageComponent implements OnInit {
  rows: AdHocReportingAPI.UserSavedReportForUi[];
  canManageAdHoc = this.policyService.dataExport.canManageAdHoc();
  AccessType = AdHocReportingAPI.AdHocReportAccessType;
  TableDataDownloadFormat = TableDataDownloadFormat;
  topLevelFilters = [
    new TopLevelFilter(
      'text',
      'name',
      '',
      this.i18n.translate(
        'reporting:lblSearchByReportName',
        {},
        'Search by report name'
      )
    )
  ];
  @CachedAttr(
    CACHE_TYPES.LOCALSTORAGE,
    ALL_SKIP_FILTER
  )
  reportTypeFilter: boolean;

  constructor (
    private i18n: I18nService,
    private notifier: NotifierService,
    private spinnerService: SpinnerService,
    private modalFactory: ModalFactory,
    private adHocReportingService: AdHocReportingService,
    private router: Router,
    private policyService: PolicyService,
    private clientSettingsService: ClientSettingsService,
    private standardProductService: StandardProductConfigurationService

  ) { }

  get isRootZone () {
    return this.clientSettingsService.clientSettings.isRootClient;
  }

  ngOnInit () {
    this.setTableRows();
    this.setTopLevelFilters();
  }

  setTopLevelFilters () {
    const showStandardFilter = !this.isRootZone &&
      this.standardProductService.standardReportTemplates.length > 0;
    if (this.canManageAdHoc) {
      this.topLevelFilters = [
        ...this.topLevelFilters,
        new TopLevelFilter(
          'typeaheadSingleEquals',
          'isUserOwnedReport',
          this.reportTypeFilter,
          '',
          {
            selectOptions: [{
              display: this.i18n.translate(
                'reporting:lblAllReports',
                {},
                'All reports'
              ),
              value: ALL_SKIP_FILTER
            }, {
              display: this.i18n.translate(
                'reporting:lblMyReports',
                {},
                'My reports'
              ),
              value: true,
              customColumn: 'isMyReport'
            }, showStandardFilter ? {
              display: this.i18n.translate(
                'common:textStandardReportTemplates',
                {},
                'Standard report templates'
              ),
              value: 'isStandardReportTemplate',
              customColumn: 'isStandardReportTemplate',
              customValue: true
            } : undefined].filter((item) => !!item)
          },
          this.i18n.translate(
            'GLOBAL:textReportType',
            {},
            'Report type'
          )
        )
      ];
    }
  }

  onTopLevelFilterChange (filter: TopLevelFilter) {
    if (filter.column === 'isUserOwnedReport') {
      this.reportTypeFilter = filter.value;
    }
  }

  setTableRows () {
    this.rows = this.standardProductService.getReportsAndStandardTemplates();
  }

  async showCreateModal () {
    this.spinnerService.startSpinner();
    await this.adHocReportingService.resolveModalDeps();
    this.spinnerService.stopSpinner();
    const report = await this.modalFactory.open(
      ManageAdHocReportModalComponent,
      {}
    );
    if (report) {
      this.spinnerService.startSpinner();
      const id = await this.adHocReportingService.saveReportDetails(report);
      this.spinnerService.stopSpinner();
      this.router.navigate([
        `/management/reporting/ad-hoc-data-feeds/ad-hoc/edit/${id}`
      ]);
    }
  }

  async copyReport (report: AdHocReportingAPI.UserSavedReportForUi) {
    const deps = {
      confirmButtonText: this.i18n.translate(
        'common:textCopy',
        {},
        'Copy'
      ),
      confirmText: this.i18n.translate(
        'reporting:textAreYouSureCopyReport2',
        {},
        'This action will create a new copy of the report including any shared or scheduled settings. Are you sure you want to copy the report?'
      ),
      modalHeader: this.i18n.translate(
        'reporting:hdrCopyReport',
        {},
        'Copy Report'
      ),
      modalSubHeader: report.name
    };
    const response = await this.modalFactory.open(
      ConfirmationModalComponent,
      deps
    );
    if (response) {
      this.spinnerService.startSpinner();
      await this.adHocReportingService.handleCopyReport(report.id);
      this.spinnerService.stopSpinner();
      this.setTableRows();
    }
  }

  async removeReport (report: AdHocReportingAPI.UserSavedReportForUi) {
    const deps = {
      confirmButtonText: this.i18n.translate(
        'common:btnDelete',
        {},
        'Delete'
      ),
      confirmText: this.i18n.translate(
        'reporting:textAreYouSureDeleteReport',
        {},
        'Are you sure you want to delete this report? This action cannot be undone.'
      ),
      modalHeader: this.i18n.translate(
        'reporting:hdrDeleteReport',
        {},
        'Delete Report'
      )
    };

    const proceed = await this.modalFactory.open(
      ConfirmationModalComponent,
      deps
    );
    if (proceed) {
      this.spinnerService.startSpinner();
      await this.adHocReportingService.deleteReport(report.id);
      this.spinnerService.stopSpinner();
      this.notifier.success(this.i18n.translate(
        'reporting:notificationSuccessfullyDeletedReport',
        {},
        'Successfully deleted report'
      ));
      this.setTableRows();
    }
  }

  async shareReportModal (report: AdHocReportingAPI.UserSavedReportForUi) {
    this.spinnerService.startSpinner();
    const passed = await this.adHocReportingService.fetchReportById(report.id);
    this.spinnerService.stopSpinner();
    if (passed) {
      const response = await this.modalFactory.open(
        ShareReportModalComponent,
        { report }
      );
      if (response) {
        this.spinnerService.startSpinner();
        await this.adHocReportingService.saveSharedUsersForReport(
          report.id,
          response
        );
        this.setTableRows();
        this.spinnerService.stopSpinner();
      }
    }
  }

  async manualSendReportModal (report: AdHocReportingAPI.UserSavedReportForUi) {
    this.spinnerService.startSpinner();
    const passed = await this.adHocReportingService.fetchReportById(report.id);
    this.spinnerService.stopSpinner();
    if (passed) {
      const response = await this.modalFactory.open(
        SendReportModalComponent,
        { report, isSchedule: false }
      );
      if (response) {
        this.spinnerService.startSpinner();
        await this.adHocReportingService.handleSendReport(
          response as AdHocReportingAPI.SendReportPayload
        );
        this.spinnerService.stopSpinner();
      }
    }
  }

  async confirmDeleteScheduleModal (report: AdHocReportingAPI.UserSavedReportForUi) {
    const deps = {
      confirmButtonText: this.i18n.translate(
        'common:textDeleteSchedule',
        {},
        'Delete schedule'
      ),
      confirmText: this.i18n.translate(
        'reporting:textDeleteScheduleDesc',
        {},
        'Deleting the schedule will prevent it from being sent on the specific timetable. Are you sure you want to delete the schedule?'
      ),
      modalHeader: this.i18n.translate(
        'common:textDeleteScheduleHeader',
        {},
        'Delete Schedule'
      ),
      modalSubheader: report.name
    };
    const proceed = await this.modalFactory.open(
      ConfirmationModalComponent,
      deps
    );
    if (proceed) {
      this.spinnerService.startSpinner();
      await this.adHocReportingService.handleDeleteScheduleReport(report.id);
      this.setTableRows();
      this.spinnerService.stopSpinner();
    }
  }

  async scheduleReportModal (report: AdHocReportingAPI.UserSavedReportForUi) {
    this.spinnerService.startSpinner();
    const passed = await this.adHocReportingService.fetchReportById(report.id);
    this.spinnerService.stopSpinner();
    if (passed) {
      const response = await this.modalFactory.open(
        SendReportModalComponent,
        { report, isSchedule: true }
      );
      if (response) {
        this.spinnerService.startSpinner();
        await this.adHocReportingService.handleScheduleReport(
          response as AdHocReportingAPI.ScheduleReportPayload,
          report.hasSchedule
        );
        this.setTableRows();
        this.spinnerService.stopSpinner();
      }
    }
  }

  editReport (report: AdHocReportingAPI.UserSavedReportForUi) {
    this.router.navigateByUrl(
      `/management/reporting/ad-hoc-data-feeds/ad-hoc-edit/${report.id}`
    );
  }

  viewReport (report: AdHocReportingAPI.UserSavedReportForUi) {
    this.router.navigateByUrl(
      `/management/reporting/ad-hoc-data-feeds/ad-hoc-view/${report.id}`
    );
  }

  async downloadReport (
    report: AdHocReportingAPI.UserSavedReportForUi,
    format: TableDataDownloadFormat
  ) {
    this.spinnerService.startSpinner();
    await this.adHocReportingService.downloadAdHocReport(
      null,
      this.adHocReportingService.getObjectByReportModelType(report.reportModelType),
      report.id,
      null,
      [],
      null,
      report.name,
      format
    );
    this.spinnerService.stopSpinner();
  }

  async publishOrUnpublish (row: AdHocReportingAPI.UserSavedReportForUi) {
    const toUnpublish = row.standardComponentIsPublished;
    const actionText = this.i18n.translate(
      toUnpublish ?
        'GLOBAL:textUnpublish' :
        'GLOBAL:textPublish',
      {},
      toUnpublish ? 'Unpublish' : 'Publish'
    );
    const confirmText = this.i18n.translate(
      toUnpublish ?
        'common:textConfirmUnpublishReport' :
        'common:textConfirmPublishReport',
      {},
      toUnpublish ?
        'Are you sure you want to unpublish this report?' :
        'Are you sure you want to publish this report?'
    );
    const response = await this.modalFactory.open(
      ConfirmationModalComponent,
      {
        modalHeader: actionText,
        modalSubHeader: row.name,
        confirmText,
        confirmButtonText: actionText
      }
    );
    if (response) {
      this.spinnerService.startSpinner();
      if (toUnpublish) {
        await this.standardProductService.handleUnpublishStandardComponents(
          row.id,
          StandardProductTypes.REPORTS
        );
      } else {
        await this.standardProductService.handlePublishStandardComponents(
          [row.id],
          StandardProductTypes.REPORTS
        );
      }
      this.setTableRows();
      this.spinnerService.stopSpinner();
    }
  }

  async createFromTemplate (row: AdHocReportingAPI.UserSavedReportForUi) {
    const response = await this.modalFactory.open(
      ManageAdHocReportModalComponent,
      {
        startFromTemplate: row
      }
    );

    if (response) {
      this.spinnerService.startSpinner();
      const id = await this.adHocReportingService.handleCreateReportFromTemplate(
        response
      );
      this.spinnerService.stopSpinner();
      if (id) {
        this.router.navigate([
          `/management/reporting/ad-hoc-data-feeds/ad-hoc-edit/${id}`
        ]);
      }
    }
  }
}
