import { Component, OnDestroy, OnInit } from '@angular/core';
import { ActivatedRoute, Router } from '@angular/router';
import { SpinnerService } from '@core/services/spinner.service';
import { FooterState } from '@core/states/footer.state';
import { AdHocReportingUI } from '@core/typings/ui/ad-hoc-reporting.typing';
import { ClientSettingsService } from '@features/client-settings/client-settings.service';
import { StandardProductConfigurationService } from '@features/platform-admin/standard-product-configuration/standard-product-configuration.service';
import { AdHocReportingDefinitions } from '@features/reporting/services/ad-hoc-reporting-definitions.service';
import { AdvancedFilterGroup, FooterStates, SwitchState } from '@yourcause/common';
import { I18nService } from '@yourcause/common/i18n';
import { ModalFactory } from '@yourcause/common/modals';
import { AdHocReportingService } from '../../services/ad-hoc-reporting.service';
import { ManageAdHocReportModalComponent } from '../manage-ad-hoc-report-modal/manage-ad-hoc-report-modal.component';

@Component({
  selector: 'gc-manage-ad-hoc-report-page',
  templateUrl: './manage-ad-hoc-report-page.component.html',
  styleUrls: ['./manage-ad-hoc-report-page.component.scss']
})
export class ManageAdHocReportPageComponent implements OnInit, OnDestroy {
  maxColumns = this.adHocReportingService.MAX_COLUMNS;
  BUILDER_MODES = AdHocReportingUI.AdHocBuilderComponentMode;
  id = this.activatedRoute.snapshot.params.id;
  detail = this.adHocReportingService.getReportById(this.id);
  columns: AdHocReportingUI.ColumnImplementation[];
  filtersUsingAnd: SwitchState;
  filterGroups: AdvancedFilterGroup[];
  type = this.detail.object;
  clientDefaultTz = this.clientSettingsService.clientSettings.defaultTimezone;
  recordIds: number[];
  exceedsMaxColumns = false;
  isStandardTemplate: boolean;

  constructor (
    private modalFactory: ModalFactory,
    private activatedRoute: ActivatedRoute,
    private adHocReportingService: AdHocReportingService,
    private i18n: I18nService,
    private footerState: FooterState,
    private spinnerService: SpinnerService,
    private router: Router,
    private clientSettingsService: ClientSettingsService,
    private adHocReportingDefinitions: AdHocReportingDefinitions,
    private standardProductService: StandardProductConfigurationService
  ) { }

  get isViewMode () {
    return this.activatedRoute.snapshot.data.isViewMode;
  }

  get reportDetails () {
    return this.adHocReportingService.get('reportDetails')[this.id];
  }

  async ngOnInit () {
    this.spinnerService.startSpinner();
    this.setRecordIds();
    this.isStandardTemplate = !!this.standardProductService.getReportTemplateById(
      +this.id
    );
    if (!this.isViewMode) {
      this.footerState.set('footerState', FooterStates.ACTION);
      this.footerState.setPrimaryAction(() => this.save());
      this.footerState.setActionLabel(this.i18n.translate('common:btnSave', {}, 'Save'));
      this.footerState.setCancelAction(() => this.cancel());
      this.footerState.setCancelLabel(this.i18n.translate('common:btnCancel', {}, 'Cancel'));
    } else {
      this.footerState.clearAll();
    }
    this.columns = this.adHocReportingService.getReportColumnsById(+this.id);
    this.handleColumnsChanged(this.columns);

    const defs = this.columns.map(column => column.definition);
    this.filterGroups = this.adHocReportingService.adaptAdvancedAPIFiltersToUIGroups(
      this.reportDetails.advancedFilterColumns,
      defs
    );
    this.filtersUsingAnd = this.reportDetails.useLogicalOperatorAnd ?
      SwitchState.Toggled :
      SwitchState.Untoggled;

    this.spinnerService.stopSpinner();
  }

  setRecordIds () {
    this.recordIds = this.adHocReportingService.getRecordIdsFromReportId(+this.id);
  }

  async manage () {
    const invalidColumnHeaders = this.getInvalidColumnHeaders();
    await this.adHocReportingService.resolveModalDeps();
    const newDetail = await this.modalFactory.open(
      ManageAdHocReportModalComponent,
      {
        existingDetail: this.detail,
        invalidColumnHeadersAlert: this.adHocReportingService.getInvalidColumnHeadersText(
          invalidColumnHeaders
        )
      }
    );
    if (newDetail) {
      const formIds = [
        ...newDetail.forms,
        newDetail.primaryFormId
      ].filter((form) => !!form);
      await this.adHocReportingService.resolveFormInfo(
        this.adHocReportingService.getRecordIdsFromModalResponse(newDetail),
        this.adHocReportingDefinitions[this.type]
      );
      this.columns = (this.columns || []).filter(({ definition }) => {
        // If form field column, only keep it if form still exists
        // Allow all standard fields
        if (
          !definition.isStandardField &&
          definition.parentBucket.includes('category.')
        ) {
          const formStillExists = formIds.some(formId => {
            return definition.formIds.includes(formId);
          });

          return formStillExists;
        }

        return true;
      });

      this.detail = newDetail;
      await this.save();
      await this.ngOnInit();
    }
  }

  getInvalidColumnHeaders () {
    return this.adHocReportingService.getInvalidColumnHeaders(
      this.columns.map((column) => {
        return column.columnNameOverride ||
          `${column.definition.display} (${column.definition.parentBucketName })`;
      })
    );
  }

  handleColumnsChanged (columns: AdHocReportingUI.ColumnImplementation[]) {
    if (columns.length > this.adHocReportingService.MAX_COLUMNS) {
      this.exceedsMaxColumns = true;
    } else {
      this.exceedsMaxColumns = false;
    }

    this.footerState.setPrimaryDisabled(this.exceedsMaxColumns);
  }

  async save (): Promise<boolean> {
    if (this.exceedsMaxColumns) {
      return false;
    }
    const invalidColumnHeaders = this.getInvalidColumnHeaders();
    if (invalidColumnHeaders.length > 0) {
      await this.adHocReportingService.openInvalidColumnHeadersModal(
        invalidColumnHeaders
      );

      return false;
    }
    this.spinnerService.startSpinner();
    this.footerState.setPrimaryDisabled(true);
    await this.adHocReportingService.saveReportDetails(
      this.detail,
      this.columns,
      this.filterGroups,
      this.filtersUsingAnd,
      +this.id
    );
    this.footerState.setPrimaryDisabled(false);
    this.spinnerService.stopSpinner();

    return true;
  }

  async onToggleViewMode () {
    if (!this.isViewMode) {
      const isValid = await this.save();
      if (!isValid) {
        return;
      }
    }
    const route = `/management/reporting/ad-hoc-data-feeds/${
      this.isViewMode ? 'ad-hoc-edit' : 'ad-hoc-view'
    }/${this.reportDetails.id}`;
    this.router.navigate([route]);
  }

  cancel () {
    const route = `/management/reporting/ad-hoc-data-feeds/ad-hoc`;
    this.router.navigate([route]);
  }

  ngOnDestroy () {
    this.footerState.clearAll();
  }
}
