import { Component, EventEmitter, Input, OnDestroy, OnInit, Output } from '@angular/core';
import { Validators } from '@angular/forms';
import { FooterState } from '@core/states/footer.state';
import { PermissionSet } from '@core/typings/permission.typing';
import { BasePermissionImplementation, BaseRole, DetachedPermissionSet, PermissionMap, RoleToSave } from '@core/typings/roles.typing';
import { TypeaheadSelectOption, TypeSafeFormBuilder, TypeSafeFormGroup } from '@yourcause/common';
import { I18nService } from '@yourcause/common/i18n';
import { DragulaService } from 'ng2-dragula';
import { Subscription } from 'rxjs';
import { RolesService } from '../roles.service';

interface RoleInfoGroup {
  roleName: string;
  roleDescription: string;
}
interface PermissionTypeGroup {
  type: any;
}
@Component({
  selector: 'yc-new-edit-roles-page',
  templateUrl: './new-edit-roles-page.component.html',
  styleUrls: ['./new-edit-roles-page.component.scss']
})
export class NewEditRolesPageComponent implements OnInit, OnDestroy {
  @Input() roleId: number;
  @Input() role: BaseRole;
  @Input() permissionSets: DetachedPermissionSet[];
  @Input() supportsDeniedPermissions = true;
  @Input() descriptionRequired = true;
  @Input() permissionTypeOptions: TypeaheadSelectOption<PermissionSet>[];
  @Output() onSave = new EventEmitter<RoleToSave>();

  isNew: boolean;
  permissionMap: PermissionMap = {};
  unsetPermissions: BasePermissionImplementation[];
  allowedPermissions: BasePermissionImplementation[];
  deniedPermissions: BasePermissionImplementation[];
  roleInfoGroup: TypeSafeFormGroup<RoleInfoGroup>;
  permissionTypeGroup: TypeSafeFormGroup<PermissionTypeGroup>;
  sub = new Subscription();

  constructor (
    private footerState: FooterState,
    private formBuilder: TypeSafeFormBuilder,
    private dragulaService: DragulaService,
    private i18n: I18nService,
    private roleService: RolesService
  ) { }

  ngOnInit () {
    this.isNew = !this.roleId;
    this.permissionMap = this.roleService.getPermissionMap(this.permissionSets);
    this.roleInfoGroup = this.formBuilder.group<RoleInfoGroup>({
      roleName: [this.role.roleName, Validators.required],
      roleDescription: this.descriptionRequired ?
        [this.role.roleDescription, Validators.required] :
        [this.role.roleDescription]
    });
    this.permissionTypeGroup = this.formBuilder.group({
      type: [this.permissionSets[0]]
    });
    this.sub.add(this.roleInfoGroup.statusChanges.subscribe(() => {
      this.setDisableSave();
    }));
    this.sub.add(this.dragulaService.drop().subscribe(() => {
      this.setDisableSave();
    }));
    this.setFooterState();
    this.allowedPermissions = [
      ...this.role.allowedPermissions.filter((permission) => {
        return this.permissionMap[
          permission.permissionSetType
        ][permission.permissionType];
      })
    ];
    this.deniedPermissions = [
      ...this.role.deniedPermissions.filter((permission) => {
        return this.permissionMap[
          permission.permissionSetType
        ][permission.permissionType];
      })
    ];

    this.permissionTypeChanged();
  }

  setFooterState () {
    this.footerState.set('footerState', this.footerState.FOOTER_STATES.ACTION);
    this.footerState.set('footerAction', () => this.save());
    this.setDisableSave();
    this.footerState.set(
      'footerActionLabel',
      this.i18n.translate('common:btnSave', {}, 'Save')
    );
  }

  setDisableSave () {
    const disabled = this.roleInfoGroup.invalid || (
      (this.allowedPermissions && this.allowedPermissions.length === 0) &&
      (this.deniedPermissions && this.deniedPermissions.length === 0)
    );

    this.footerState.set(
      'footerActionDisabled',
      disabled
    );
  }

  permissionTypeChanged () {
    const permissionSet: PermissionSet = this.permissionTypeGroup.value.type;
    this.unsetPermissions = permissionSet.permissions
      .filter((permission) => !this.deniedPermissions.find(deniedPermission => (
        deniedPermission.permissionSetType === permissionSet.type &&
        deniedPermission.permissionType === permission.type
      )) && !this.allowedPermissions
        .find(allowedPermission => (
          allowedPermission.permissionSetType === permissionSet.type &&
          allowedPermission.permissionType === permission.type
        ))
      ).map(unsetPermission => ({
        permissionSetType: permissionSet.type,
        permissionType: unsetPermission.type
      }));
  }

  allowOrDenyPermission (
    permission: BasePermissionImplementation,
    wasUnset: boolean,
    allow: boolean
  ) {
    this.unsetPermissions = wasUnset ?
      this.unsetPermissions.filter(unsetPerm => unsetPerm !== permission) :
      [
        ...this.unsetPermissions,
        permission
      ];

    const otherPermissions = allow ? 'allowedPermissions' : 'deniedPermissions';

    this[otherPermissions] = wasUnset ?
      [
        ...this[otherPermissions],
        permission
      ] : this[otherPermissions].filter(otherPerm => otherPerm !== permission);

    this.setDisableSave();
  }

  async save () {
    const roleToSave: RoleToSave = {
      ...this.roleInfoGroup.value,
      roleId: this.isNew ? undefined : this.roleId,
      allowedPermissions: this.allowedPermissions,
      deniedPermissions: this.deniedPermissions
    };
    this.onSave.emit(roleToSave);
  }

  ngOnDestroy () {
    this.footerState.clearAll();
    this.sub.unsubscribe();
  }
}
