import { Injectable } from '@angular/core';
import { GCMockModule } from '@core/mocks/gc-module.mock';
import { UserSignature } from '@core/typings/user.typing';
import { FormAudience } from '@features/configure-forms/form.typing';
import { YcFile } from '@yourcause/common';
import { SignatureTabType } from '@yourcause/common/signature';
import { BeforeEach, Spec, TestCase } from '@yourcause/test-decorators';
import { DescribeAngularService } from '@yourcause/test-decorators/angular';
import { expect } from 'chai';
import { SignatureService } from './signature.service';

@Injectable({ providedIn: 'root' })
@DescribeAngularService(SignatureService, {
  imports: [
    GCMockModule
  ]
})
export class SignatureServiceSpec implements Spec<SignatureServiceSpec, SignatureService> {
  userSignature: UserSignature =  {
    userSignatureId: 1,
    userId: 2,
    userSignatureUrl: 'url'
  };
  uploadedSignatureId = 14;

  @BeforeEach()
  mock (service: SignatureService) {
    service['signatureResources'].getUserSignature = async () => {
      return this.userSignature;
    };
    service['signatureResources'].uploadSignature = async () => {
      return this.uploadedSignatureId;
    };
    service['fileService'].getBlob = async () =>{
      return new Blob([]);
    };
  }

  @TestCase('should be able to get default form description applicant')
  getDefaultFormDescriptionApplicant (service: SignatureService) {
    const descriptionNom = service.getDefaultFormDescriptionApplicant(true);
    const descriptionApp = service.getDefaultFormDescriptionApplicant(false);
    const expectedNomDesc = 'In order to proceed, please use one of the options below to sign and submit your nomination. By clicking Submit I understand that this is a legal representation of my signature.';
    const expectedAppDesc = 'In order to proceed, please use one of the options below to sign and submit your application. By clicking Submit I understand that this is a legal representation of my signature.';
    expect(descriptionNom).to.be.equal(expectedNomDesc);
    expect(descriptionApp).to.be.equal(expectedAppDesc);
  }

  @TestCase('should be able to get default form description manager')
  getDefaultFormDescriptionManager (service: SignatureService) {
    const description = service.getDefaultFormDescriptionManager();
    const expectedDesc = 'In order to proceed, please use one of the options below to sign and submit your form. By clicking Submit I understand that this is a legal representation of my signature.';
    expect(description).to.be.equal(expectedDesc);
  }

  @TestCase('should be able to get supports bypass')
  getSignatureFormDescriptionManagerPortal (service: SignatureService) {
    const formName = 'form name';
    const {
      supportsBypass
    } = service.getSignatureFormDescriptionManagerPortal(
      formName,
      FormAudience.APPLICANT,
      'signature description',
      false
    );
    const expectedBypassSetting = true;
    expect(supportsBypass).to.be.equal(expectedBypassSetting);
  }

  @TestCase('should be able to set signature')
  async setSignature (service: SignatureService) {
    await service.setSignature();
    expect(service.userSignature).to.deep.equal(this.userSignature);
  }

  @TestCase('should be able to reset signature')
  async resetSignature (service: SignatureService) {
    const newUserSignatureId = 12;
    service['signatureResources'].getUserSignature = async () => {
      return {
        ...this.userSignature,
        userSignatureId: newUserSignatureId
      };
    };
    await service.resetUserSignature();
    expect(service.userSignature.userSignatureId).to.be.equal(newUserSignatureId);
  }

  @TestCase('should be able to handle signature modal response for saved signature')
  async handleSignatureModalResponseSaved (service: SignatureService) {
    const id = await service.handleSignatureModalResponse({
      file: service.savedSignatureFile,
      type: SignatureTabType.Saved
    });
    // In this case, we do not upload. We use the existing users signature ID
    expect(id).to.be.equal(service.userSignature.userSignatureId);
  }

  @TestCase('should be able to handle signature modal response for uploaded signature')
  async handleSignatureModalResponseUpload (service: SignatureService) {
    const id = await service.handleSignatureModalResponse({
      file: new YcFile(
        'signature.png',
        null
      ),
      type: SignatureTabType.Draw
    });

    expect(id).to.be.equal(this.uploadedSignatureId);
  }

  @TestCase('upload signature should return file ID')
  async uploadFile (service: SignatureService) {
    const id = await service.uploadSignature({
      file: new YcFile(
        'signature.png',
        null
      ),
      type: SignatureTabType.Upload
    });
    expect(id).to.be.equal(this.uploadedSignatureId);
  }

  @TestCase('should be able to set form signature')
  setFormSignature (service: SignatureService) {
    // TEST
  }

  @TestCase('should be able to set signature on map')
  setSignatureOnMap (service: SignatureService) {
    // TEST
  }

  @TestCase('should be able to get signature for application form response')
  getSignatureForApplicationFormResponse (service: SignatureService) {
    // TEST
  }
}
