import { Component, EventEmitter, Input, OnChanges, Output, SimpleChanges } from '@angular/core';
import { ControlContainer, FormGroup } from '@angular/forms';
import { SpinnerService } from '@core/services/spinner.service';
import { MergeDocumentModalPreviewComponent } from '@features/application-view/application-attachments/merge-document-preview-modal/merge-document-preview-modal.component';
import { DocumentTemplateService } from '@features/document-templates/document-template.service';
import { DocumentTemplateForUi } from '@features/document-templates/document-template.typing';
import { DocumentTemplatesResolver } from '@features/document-templates/resolvers/document-templates.resolver';
import { EmailService } from '@features/system-emails/email.service';
import { ClientEmailTemplateFromAPI, Email, EmailNotificationType, ProgramEmailTemplateForUI, SimpleEmail } from '@features/system-emails/email.typing';
import { EmailResolver } from '@features/system-emails/resolvers/email.resolver';
import { ViewEmailModalComponent } from '@features/system-emails/view-email-modal/view-email-modal.component';
import { UserService } from '@features/users/user.service';
import { ArrayHelpersService, GenericFile, maxFileSizeForEmailAttachments, SelectOption } from '@yourcause/common';
import { I18nService } from '@yourcause/common/i18n';
import { LogService } from '@yourcause/common/logging';
import { ModalFactory } from '@yourcause/common/modals';
import { NotifierService } from '@yourcause/common/notifier';

@Component({
  selector: 'gc-select-email-template',
  templateUrl: './select-email-template.component.html',
  styleUrls: ['./select-email-template.component.scss']
})
export class SelectEmailTemplateComponent implements OnChanges {
  @Input() forceFormGroupValueToDefault = true;
  // Unless you set this boolean to false in the template,
  // the formControl value will be set to the default template
  @Input() programId: number;
  @Input() controlName: 'clientEmailTemplateId'|'cashClientEmailTemplateId'|'inKindClientEmailTemplateId' = 'clientEmailTemplateId';
  @Input() ccControlName: 'cc'|'cashCC'|'inKindCC'|'ccEmails' = 'cc';
  @Input() bccControlName: 'bcc'|'cashBCC'|'inKindBCC'|'bccEmails' = 'bcc';
  @Input() customMessageControlName = 'customMessage';
  @Input() attachmentsControlName = 'attachments';
  @Input() allowMailMerge = true;
  @Input() applicationIds: number[] = [];
  @Input() emailType: EmailNotificationType;
  @Input() notifyApplicant = true;
  @Input() hasCustomMessage = false;
  @Input() customMessageLabel: string;
  @Input() formGroup: FormGroup;
  @Input() singlePreviewMargin = '0';
  @Input() allowAttachments = true;
  @Input() showSingleOption = false;
  @Input() activeTempsFromChangesMap: number[] = [];
  @Input() allowCCandBCC = true;
  @Input() forceUpdateEmailTemplates = true;
  @Output() onPreviewModalOpenOrClose = new EventEmitter<boolean>();
  @Output() onDefaultAttachmentsChange = new EventEmitter<GenericFile[]>();
  attachments: GenericFile[] = [];
  maxAttachmentsFileSize = maxFileSizeForEmailAttachments;
  private maxDescription = 170;
  emailOptions: SelectOption[] = [];
  templates: SimpleEmail[];
  userLang = this.userService.getCurrentUserCulture();
  showCcBcc = false;
  allowCcUpdates = false;
  mailMergeTemplate: DocumentTemplateForUi;

  constructor (
    private logger: LogService,
    private modalFactory: ModalFactory,
    private i18n: I18nService,
    private notifier: NotifierService,
    private spinnerService: SpinnerService,
    private arrayHelper: ArrayHelpersService,
    private controlContainer: ControlContainer,
    private emailService: EmailService,
    private userService: UserService,
    private documentTemplateService: DocumentTemplateService,
    private emailResolver: EmailResolver,
    private documentTemplateResolver: DocumentTemplatesResolver
  ) { }

  get templateMap () {
    return this.emailService.templateMap[this.emailType];
  }

  get isClientTemplate () {
    return !this.programId;
  }

  get email (): Email {
    if (this.emailService.templateMap) {
      return this.templateMap.email;
    }

    return null;
  }

  get clientEmailTemplateId () {
    return this.formGroup.value[this.controlName];
  }

  async ngOnChanges (changes: SimpleChanges) {
    if (!this.emailService.emails || this.documentTemplateService.documentTemplates) {
      await Promise.all([
        this.emailResolver.resolve(),
        this.documentTemplateResolver.resolve()
      ]);
    }
    if (!this.formGroup) {
      this.formGroup = <FormGroup>this.controlContainer.control;
    }
    this.spinnerService.startSpinner();
    if (changes.emailType || changes.notifyApplicant) {
      // if simple changes includes emailType
      let templates: SimpleEmail[] = [];
      try {
        // force an update to the template map
        await this.emailService.setTemplateMap(
          this.emailType,
          this.forceUpdateEmailTemplates,
          this.userLang
        );
        if (this.programId) {
          // if this is a program-specific email, update that map as well
          await this.emailService.setProgramTemplatesByType(
            this.programId,
            this.emailType
          );

          templates = this.templateMap.programTemplates[this.programId]
            .map(template => this.getSimpleEmail(template, true));
        } else {
          const standard: SimpleEmail = {
            id: 0,
            emailNumber: '' + this.email.emailNotificationType,
            subject: this.email.subject,
            emailNotificationType: this.email.emailNotificationType,
            audienceType: this.email.audienceType,
            title: this.email.title,
            body: this.templateMap.template,
            description: this.email.description,
            active: true,
            ccEmails: [],
            bccEmails: [],
            attachments: [],
            allowCarbonCopyUpdates: false,
            supportsCarbonCopy: false,
            allowDocumentTemplates: false
          };
          templates = [standard].concat(this.templateMap.clientTemplates
            .map((template) => this.getSimpleEmail(template, false)));
        }
        this.setEmailOptions(templates, this.activeTempsFromChangesMap);
        templates = templates.filter((temp) => !!temp.active || this.activeTempsFromChangesMap.includes(temp.id));
        this.templates = this.arrayHelper.sort(templates, 'id');
        if (this.programId) {
          await this.setDefaultTemplate();
        }
      } catch (e) {
        this.logger.error(e);
        this.notifier.error(this.i18n.translate(
          'GLOBAL:textErrorLoadingEmail',
          {},
          'There was an error loading the copies'
        ));
      }
    }
    if (this.programId && changes.programId) {
      await this.setDefaultTemplate();
    }
    this.spinnerService.stopSpinner();
  }

  async setDefaultTemplate () {
    const formControlVal = this.formGroup.get(this.controlName).value;
    if (this.forceFormGroupValueToDefault || !formControlVal) {
      const defaultTemplate = await this.emailService.getProgramDefault(
        this.programId,
        this.emailType
      );
      setTimeout(() => {
        this.formGroup.get(this.controlName).setValue(defaultTemplate);
        // Fire off email select in the scenario where the setValue won't
        if (this.templates.length === 1) {
          this.onEmailSelect();
        }
      });
    } else {
      this.onEmailSelect();
    }
  }

  setEmailOptions (templates: SimpleEmail[], activeTempsFromChangesMap: number[]) {
    const inactiveIds = templates.filter((temp) => {
      return !temp.active;
    }).map((temp) => temp.id);
    this.emailOptions = templates.map((temp) => {
      return {
        value: temp.id,
        display: `
          <div class="d-flex" style="height: 60px">
            <div class="mr-3" style="width: 40px;">
              ${temp.emailNumber}
            </div>
            <div class="mr-3" style="width: 350px;">
              ${temp.subject}
              <small class="d-block text-muted">
                ${
                  temp.description.length > this.maxDescription ?
                    temp.description.slice(0, this.maxDescription) + '...' :
                    temp.description
                }
              </small>
            </div>
          </div>
        `
      };
    }).filter((option) => {
      return !inactiveIds.includes(option.value) || activeTempsFromChangesMap.includes(option.value);
    });
  }

  viewProgramEmail (row: SelectOption) {
    const template = this.templates.find((temp) => {
      return temp.id === row.value;
    });

    this.viewEmail(template);
  }

  async viewEmail (email: SimpleEmail) {
    this.onPreviewModalOpenOrClose.emit(true);
    await this.modalFactory.open(
      ViewEmailModalComponent,
      { email },
      { backdrop: false, class: 'modal-xl' }
    );
    this.onPreviewModalOpenOrClose.emit(false);
  }

  getSimpleEmail (
    email: ClientEmailTemplateFromAPI|ProgramEmailTemplateForUI,
    forProgram: boolean
  ): SimpleEmail {
    const type = this.emailType;
    const found = this.emailService.emails.find((item) => {
      return item.emailNotificationType === type;
    });
    const template = 'clientEmailTemplate' in email ? email.clientEmailTemplate : email;

    return {
      subject: template.subject,
      body: template.body,
      emailNumber: template.emailNumber,
      description: template.description,
      active: forProgram ? email.active : template.active,
      id: template.id,
      title: template.title,
      ccEmails: template.ccEmails,
      bccEmails: template.bccEmails,
      attachments: 'clientEmailTemplate' in email ?
        email.clientEmailTemplate.attachments :
        email.emailAttachments,
      allowCarbonCopyUpdates: template.allowCarbonCopyUpdates,
      emailNotificationType: type,
      audienceType: found.audienceType,
      supportsCarbonCopy: found.supportsCarbonCopy,
      allowDocumentTemplates: found.allowDocumentTemplates,
      documentTemplates: template.documentTemplates
    };
  }

  onEmailSelect () {
    const control = this.formGroup.get(this.controlName);
    let found: SimpleEmail;
    if (control.value !== 0 && this.notifyApplicant && this.allowCCandBCC) {
      found = this.templates.find((temp) => {
        return temp.id === control.value;
      });
      this.showCcBcc = found.supportsCarbonCopy &&
        !!(
          (found.ccEmails.length || found.bccEmails.length) ||
          found.allowCarbonCopyUpdates
        );
      this.updateAttachmentsList(found.attachments);
    } else {
      // Cc Bcc not available on standard email (0)
      this.showCcBcc = false;
      this.updateAttachmentsList([]);
      // emit attachments []
    }
    this.mailMergeTemplate = this.getMailMergeTemplate();
    this.allowCcUpdates = this.showCcBcc ?
      found.allowCarbonCopyUpdates :
      false;
    const ccControl = this.formGroup.get(this.ccControlName);
    const bccControl = this.formGroup.get(this.bccControlName);
    if (ccControl && bccControl) {
      ccControl.setValue(
        this.showCcBcc ? found.ccEmails : []
      );
      bccControl.setValue(
        this.showCcBcc ? found.bccEmails : []
      );
    }
  }

  updateAttachmentsFormControl (attachments: GenericFile[]) {
    const attachmentsControl = this.formGroup.get(this.attachmentsControlName);
    attachmentsControl?.setValue(attachments);
  }

  updateAttachmentsList (attachments: GenericFile[]) {
    this.attachments = [
      ...attachments
    ];
    this.updateAttachmentsFormControl(this.attachments);
  }

  addAttachment (payload: {attachment: GenericFile}) {
    this.attachments = [
      ...this.attachments,
      payload.attachment
    ];
    this.updateAttachmentsFormControl(this.attachments);
  }

  removeAttachment (index: number) {
    this.attachments = [
      ...this.attachments.slice(0, index),
      ...this.attachments.slice(index + 1)
    ];
    this.updateAttachmentsFormControl(this.attachments);
  }

  getMailMergeTemplate (): DocumentTemplateForUi {
    const selectedEmail = this.formGroup.get(this.controlName).value;
    if (selectedEmail !== 0 && this.notifyApplicant) {
      const template = this.templates.find((temp) => {
        return temp.id === selectedEmail;
      });
      if (template?.documentTemplates?.length > 0) {
        const selectedDocId = template.documentTemplates[0]?.documentTemplateId;
        const foundTemplate = this.documentTemplateService.documentTemplates
          .find((temp) => {
            return temp.id === selectedDocId;
          });
        if (foundTemplate) {
          return foundTemplate;
        }
      }
    }

    return null;
  }

  async openMailMergePreview () {
    this.onPreviewModalOpenOrClose.emit(true);
    await this.modalFactory.open(
      MergeDocumentModalPreviewComponent,
      {
        templateId: this.mailMergeTemplate.id,
        applicationIds: this.applicationIds
      }
    );
    this.onPreviewModalOpenOrClose.emit(false);
  }
}
