import { Component, Input, OnInit } from '@angular/core';
import { FormGroup, Validators } from '@angular/forms';
import { TypeaheadSelectOption, TypeSafeFormBuilder, TypeSafeFormGroup } from '@yourcause/common';
import { AnalyticsService, EventType } from '@yourcause/common/analytics';
import { I18nService } from '@yourcause/common/i18n';
import { YCModalComponent } from '@yourcause/common/modals';
import { SystemTagsService } from '../system-tags.service';
import { SystemTags } from '../typings/system-tags.typing';
interface CreateTagFormGroup {
  bucket: SystemTags.Buckets;
  tagName: string;
  convertToGlobal: boolean;
}

@Component({
  selector: 'gc-create-system-tag-modal',
  templateUrl: './create-system-tag-modal.component.html',
  styleUrls: ['./create-system-tag-modal.component.scss']
})
export class CreateSystemTagModalComponent extends YCModalComponent<SystemTags.AddTagModalResult> implements OnInit {
  @Input() previousBucket = SystemTags.Buckets.Global;

  formGroup: TypeSafeFormGroup<CreateTagFormGroup>;
  color: SystemTags.TagColor = SystemTags.TagColor.Yellow;
  existingTag: SystemTags.TagDefinition;
  showConvertCheckbox = false;
  showConvertHelpText = false;
  convertHelpText = this.i18n.translate(
    'tags:textTagInUseAndWillBeConverted',
    {},
    'This tag is currently in use in another bucket and will be converted to a Global tag'
  );
  bucketOptions = this.systemTagsService.buckets
    .map(bucket => <TypeaheadSelectOption<SystemTags.Buckets>>({
      value: bucket.id,
      label: bucket.name
    }));

  constructor (
    private systemTagsService: SystemTagsService,
    private formBuilder: TypeSafeFormBuilder,
    private i18n: I18nService,
    private analyticsService: AnalyticsService
  ) {
    super();
  }

  ngOnInit () {
    if (this.existingTag) {
      this.color = this.existingTag.color;
    }
    this.formGroup = this.formBuilder.group<CreateTagFormGroup>({
      bucket: [this.previousBucket],
      tagName: ['', Validators.required],
      convertToGlobal: false
    }, {
      validator: [this.tagExistsValidator()]
    });
  }

  saveAndAddAnother () {
    this.submit(true);
    this.analyticsService.emitEvent({
      eventName: 'Save tag and add another',
      eventType: EventType.Click,
      extras: null
    });
  }

  save () {
    this.submit(false);
    this.analyticsService.emitEvent({
      eventName: 'Save tag',
      eventType: EventType.Click,
      extras: null
    });
  }

  cancel () {
    this.closeModal.emit();
  }

  tagExistsValidator () {
    return (group: FormGroup) => {
      const tagNameControl = group.get('tagName');
      const promoteToGlobalControl = group.get('convertToGlobal');
      const bucketControl = group.get('bucket');
      this.showConvertHelpText = false;

      if (
        tagNameControl &&
        promoteToGlobalControl &&
        bucketControl
      ) {
        const currentBucket = this.systemTagsService.buckets
          .find(bucket => bucket.id === bucketControl.value);

        const enteredTagName = tagNameControl.value ?
          tagNameControl.value :
          '';

        const existingTag = this.existingTag = this.systemTagsService.summaries
          .find(tag => tag.name === enteredTagName);

        const existingBucket = bucketControl.value && existingTag ?
          this.systemTagsService.buckets.find(bucket => bucket.id === existingTag.bucketId) :
          null;


        this.showConvertCheckbox = false;
        const wantsToConvert = promoteToGlobalControl.value;
        const error = (() => {
          if (!existingTag || !existingBucket) {
            return null;
          } else if (existingTag.bucketId === currentBucket.id) {
            return {
              tagName: {
                tagAlreadyInBucket: {
                  i18nKey: 'tags:textAlreadyInBucketError',
                  defaultValue: 'This tag is already in use in this bucket and cannot be added again.'
                }
              }
            };
          } else if (existingBucket.id === SystemTags.Buckets.Global) {
            return {
              tagName: {
                tagInGlobal: {
                  i18nKey: 'tags:textTagIsInGlobalError',
                  defaultValue: 'This tag is currently in use in the Global bucket and cannot be added to this bucket'
                }
              }
            };
          } else {
            this.showConvertCheckbox = true;
            if (!wantsToConvert) {
              this.showConvertHelpText = false;

              return {
                tagName: {
                  tagAlreadyExists: {
                    i18nKey: 'tags:textTagIsInUseError',
                    context: {
                      existingBucketName: existingBucket.name
                    },
                    defaultValue: 'This tag is currently in use in the __existingBucketName__ bucket and cannot be added to this bucket. To convert to a Global tag, click the checkbox below'
                  }
                }
              };
            } else {
              this.showConvertHelpText = true;

              return null;
            }
          }
        })();

        if (!this.showConvertCheckbox && wantsToConvert) {
          promoteToGlobalControl.setValue(false);
        }

        return error;
      }

      return null;
    };
  }

  private submit (reopenModal: boolean) {
    const {
      bucket,
      tagName,
      convertToGlobal
    } = this.formGroup.value;

    this.closeModal.emit({
      tag: convertToGlobal ? this.existingTag : {
        active: false,
        enabled: true,
        bucketId: bucket,
        color: this.color,
        name: tagName,
        description: ''
      },
      reopenModal,
      wantsToConvert: convertToGlobal
    });
  }
}
