import { Component } from '@angular/core';
import { SpinnerService } from '@core/services/spinner.service';
import { I18nService } from '@yourcause/common/i18n';
import { LogService } from '@yourcause/common/logging';
import { ConfirmationModalComponent, ModalFactory } from '@yourcause/common/modals';
import { NotifierService } from '@yourcause/common/notifier';
import { CreateSystemTagModalComponent } from '../create-system-tag-modal/create-system-tag-modal.component';
import { InactiveTagsModalComponent } from '../inactive-tags-modal/inactive-tags-modal.component';
import { SystemTagsService } from '../system-tags.service';
import { SystemTags } from '../typings/system-tags.typing';



@Component({
  selector: 'gc-manage-system-tags-page',
  templateUrl: './manage-system-tags-page.component.html',
  styleUrls: ['./manage-system-tags-page.component.scss']
})
export class ManageSystemTagsPageComponent {
  tagsMarkedForDelete: SystemTags.TagDefinition[];

   constructor (
    private logger: LogService,
    private notifier: NotifierService,
    private i18n: I18nService,
    private systemTagsService: SystemTagsService,
    private modalFactory: ModalFactory,
    private spinnerService: SpinnerService
  ) { }

  buckets$ = this.systemTagsService.bucketsWithTags$;

  async addTag (previousBucket?: SystemTags.Buckets) {
    const tag = await this.modalFactory.open(
      CreateSystemTagModalComponent,
      {
        previousBucket
      }
    );
    if (tag) {
      this.spinnerService.startSpinner();
      if (tag.wantsToConvert) {
        await this.systemTagsService.updateRecord(tag.tag, 'Promote', null);
        this.notifier.success(
          this.i18n.translate(
            'tags:notificationSuccessfullyPromotedTag',
            {},
            'Successfully promoted tag'
          )
        );
      } else {
        await this.systemTagsService.addRecord(tag.tag);
        this.notifier.success(
          this.i18n.translate(
            'tags:notificationSuccessfullyCreatedTag',
            {},
            'Successfully created tag'
          )
        );
      }

      if (tag.reopenModal) {
        this.addTag(tag.tag.bucketId);
      }
      this.spinnerService.stopSpinner();
    }
  }

  async changeTagColor (tag: SystemTags.TagDefinition, color: SystemTags.TagColor) {
    this.spinnerService.startSpinner();
    try {
      await this.systemTagsService.updateRecord(tag, 'ChangeColor', color);

      this.notifier.success(this.i18n.translate('tags:notificationSuccessfullyUpdatedTagColor', {}, 'Successfully updated this tag\'s color'));
    } catch (e) {
      this.logger.error(e);
      this.notifier.error(this.i18n.translate('tags:notificationErrorUpdatingTagColor', {}, 'There was an error updating this tag\'s color'));
    }
    this.spinnerService.stopSpinner();

    tag.color = color;
  }

  async showInactiveTags (bucket: SystemTags.TagBucketWithTags) {
    const results = await this.modalFactory.open(
      InactiveTagsModalComponent,
      {
        bucket
      }
    );
    if (results && results.length) {
      this.spinnerService.startSpinner();
      await Promise.all(results.map(async result => {
        if (result.change === 'delete') {
          await this.systemTagsService.removeRecord(result.tag);
        } else {
          await this.systemTagsService.updateRecord(result.tag, result.change, null);
        }
      }));

      this.notifier.success(
        this.i18n.translate(
          'tags:notificationSuccessfullyUpdatedTags',
          {},
          'Successfully updated tags'
        )
      );
      this.spinnerService.stopSpinner();
    }
  }

  async markTag (tag: SystemTags.TagDefinition, action: 'Enable'|'Disable'|'delete') {
    const isDisable = action === 'Disable';
    const proceed = await this.modalFactory.open(
      ConfirmationModalComponent, {
      confirmText: this.i18n.translate(
        isDisable ?
          'tags:textAreYouSureInactivate' :
          'tags:textAreYouSureRemove',
        {},
        isDisable ?
          'Are you sure you want to mark this tag as inactive?' :
          'Are you sure you want to remove this tag?'
      ),
      confirmButtonText: this.i18n.translate(
        isDisable ?
          'GLOBAL:btnMarkAsInactive' :
          'common:textRemove',
        {},
        isDisable ?
          'Mark as inactive' :
          'Remove'
      )
    });
    if (proceed) {
      this.spinnerService.startSpinner();
      if (action === 'Disable') {
        await this.systemTagsService.updateRecord(tag, action, null);
        this.notifier.success(this.i18n.translate(
          'tags:notificationSuccessfullyMarkedInactive',
          {},
          'Successfully marked this tag as inactive'
        ));
      } else {
        await this.systemTagsService.removeRecord(tag);
        this.notifier.success(this.i18n.translate(
          'tags:notificationSuccessfullyRemoved',
          {},
          'Successfully removed this tag'
        ));
      }
      this.spinnerService.stopSpinner();
    }
  }
}
