import { Injectable } from '@angular/core';
import { GCMockModule } from '@core/mocks/gc-module.mock';
import { BeforeEach, Spec, TestCase } from '@yourcause/test-decorators';
import { DescribeAngularService } from '@yourcause/test-decorators/angular';
import { expect } from 'chai';
import { SystemTagsService } from './system-tags.service';
import { SystemTags } from './typings/system-tags.typing';


@Injectable({ providedIn: 'root' })
@DescribeAngularService(SystemTagsService, {
  imports: [
    GCMockModule
  ]
})
export class SystemTagsServiceSpec implements Spec<SystemTagsServiceSpec, SystemTagsService> {
  globalTagId = 1;
  appTagId = 2;
  nomTagId = 3;
  nonprofitTagId = 4;
  paymentTagId = 5;
  globalTag = {
    id: this.globalTagId,
    name: 'Global tag',
    description: '',
    bucketId: SystemTags.Buckets.Global,
    enabled: true,
    active: true,
    color: SystemTags.TagColor.Yellow
  };
  appTag = {
    id: this.appTagId,
    name: 'Application tag',
    description: '',
    bucketId: SystemTags.Buckets.Application,
    enabled: true,
    active: true,
    color: SystemTags.TagColor.Yellow
  };
  nomTag = {
    id: this.nomTagId,
    name: 'Nomination tag',
    description: '',
    bucketId: SystemTags.Buckets.Nomination,
    enabled: true,
    active: true,
    color: SystemTags.TagColor.Yellow
  };
  nonprofitTag = {
    id: this.nonprofitTagId,
    name: 'Nonprofit tag',
    description: '',
    bucketId: SystemTags.Buckets.NonprofitProfile,
    enabled: true,
    active: true,
    color: SystemTags.TagColor.Yellow
  };
  paymentTag = {
    id: this.paymentTagId,
    name: 'Payment tag',
    description: '',
    bucketId: SystemTags.Buckets.Payment,
    enabled: true,
    active: true,
    color: SystemTags.TagColor.Yellow
  };
  allTags: SystemTags.TagDefinition[] = [
    this.globalTag,
    this.appTag,
    this.nomTag,
    this.nonprofitTag,
    this.paymentTag
  ];
  nonprofitId = 12;
  tagsForNonprofitRecord = [{
    active: true,
    tagId: this.nonprofitTagId,
    id: this.nonprofitTagId,
    tag: this.nonprofitTag
  }];


  @BeforeEach()
  async mock (service: SystemTagsService) {
    service['systemTagsResources'].getAll = async () => {
      return this.allTags;
    };
    service['systemTagsResources'].getTags = async () => {
      return this.tagsForNonprofitRecord;
    };
    await service.setSummaries();
  }

  @TestCase('should be able to find tag')
  findTag (service: SystemTagsService) {
    const found = service.findTag(this.nomTagId);
    expect(found).to.be.equal(this.nomTag);
  }

  @TestCase('should be able to set buckets')
  setBuckets (service: SystemTagsService) {
    service.getBuckets();
    expect(service.buckets.length).to.be.equal(6);
    expect(service.buckets[0].id).to.be.equal(1);
  }

  @TestCase('should be able to get tag options')
  getTagOptions (service: SystemTagsService) {
    const options = service.getTagOptions(SystemTags.Buckets.Payment);
    const hasPaymentTags = options.some((opt) => {
      return opt.value === this.paymentTagId;
    });
    expect(hasPaymentTags).to.be.true;
    const hasGlobalTags = options.some((opt) => {
      return opt.value === this.globalTagId;
    });
    expect(hasGlobalTags).to.be.true;
    const hasNonprofitTags = options.some((opt) => {
      return opt.value === this.nonprofitTagId;
    });
    expect(hasNonprofitTags).to.be.false;
  }

  @TestCase('should be able to get tags for bucket')
  getTagsForBucket (service: SystemTagsService) {
    const tags = service.getTagsForBucket(SystemTags.Buckets.Application, false);
    const hasAppTags = tags.some((tag) => {
      return tag.id === this.appTagId;
    });
    expect(hasAppTags).to.be.true;
    const hasGlobalTags = tags.some((tag) => {
      return tag.id === this.globalTagId;
    });
    expect(hasGlobalTags).to.be.true;
    const hasPaymentTags = tags.some((tag) => {
      return tag.id === this.paymentTagId;
    });
    expect(hasPaymentTags).to.be.false;
  }

  @TestCase('should be able to get and set tags for record')
  async getAndSetTagsForRecord (service: SystemTagsService) {
    const tagIds = await service.getAndSetTagsForRecord(
      SystemTags.Buckets.NonprofitProfile,
      this.nonprofitId
    );
    expect(tagIds[0]).to.be.equal(this.nonprofitTagId);
  }

  @TestCase('should be able to get new tags')
  getNewTags (service: SystemTagsService) {
    const newTags = service.getNewTags(
      [
        this.globalTag
      ],
      [this.appTagId]
    );
    expect(newTags.length).to.be.equal(1);
    expect(newTags[0]).to.be.equal(this.appTagId);
  }

  @TestCase('should be able to get removed tags')
  getRemovedTags (service: SystemTagsService) {
    const removedTags = service.getRemovedTags(
      [
        this.globalTag
      ],
      [this.appTagId]
    );
    expect(removedTags.length).to.be.equal(1);
    expect(removedTags[0]).to.be.equal(this.globalTagId);
  }

  @TestCase('should be able to get current tags for PDF')
  async getCurrentTagsForPDF (service: SystemTagsService) {
    const nonprofitTags = await service.getCurrentTagsForPDF(this.nonprofitId);
    expect(nonprofitTags[0]).to.be.equal(this.nonprofitTag.name);
  }
}

