import { Injectable } from '@angular/core';
import { ArrayHelpersService, PaginationOptions, TypeaheadSelectOption } from '@yourcause/common';
import { I18nService } from '@yourcause/common/i18n';
import { LogService } from '@yourcause/common/logging';
import { NotifierService } from '@yourcause/common/notifier';
import { AttachYCState, BaseYCService } from '@yourcause/common/state';
import { uniqBy } from 'lodash';
import { combineLatest, filter, map, Observable } from 'rxjs';
import { SystemTagsResources } from './system-tags.resources';
import { SystemTagsState } from './system-tags.state';
import { SystemTags } from './typings/system-tags.typing';


@AttachYCState(SystemTagsState)
@Injectable({ providedIn: 'root' })
export class SystemTagsService extends BaseYCService<SystemTagsState> {
  readonly tagColorClassMap: Record<SystemTags.TagColor, string> = {
    [SystemTags.TagColor.Red]: 'alert alert-danger',
    [SystemTags.TagColor.Green]: 'alert alert-success',
    [SystemTags.TagColor.Blue]: 'alert alert-info',
    [SystemTags.TagColor.Yellow]: 'alert alert-warning'
  };
  bucketsWithTags$: Observable<SystemTags.TagBucketWithTags[]> = combineLatest(
    this.changesTo$('buckets'),
    this.changesTo$('summaries')
  )
  .pipe(filter(([buckets, tags]) => !!buckets && !!tags))
  .pipe(map(([buckets, tags]) => {
    return buckets.map(bucket => ({
      ...bucket,
      activeTags: tags
        .filter(tag => tag.bucketId === bucket.id)
        .filter(tag => tag.enabled),
      inactiveTags: tags
        .filter(tag => tag.bucketId === bucket.id)
        .filter(tag => !tag.enabled)
    }));
  }));

   constructor (
    private logger: LogService,
    private i18n: I18nService,
    private arrayHelper: ArrayHelpersService,
    private systemTagsResources: SystemTagsResources,
    private notifier: NotifierService
  ) {
    super();
  }

  get buckets () {
    return this.get('buckets');
  }

  get summaries () {
    return this.get('summaries');
  }

  get tagMap () {
    return this.get('tagMap');
  }

  getTagOptionsForAppManager (isNomination: boolean) {
    const applicationTags = this.getTagsForBucket(
      isNomination ?
        SystemTags.Buckets.Nomination :
        SystemTags.Buckets.Application,
      true
    );
    const nonprofitTags = this.getTagsForBucket(
      SystemTags.Buckets.NonprofitProfile,
      true
    );
    const applicantTags = this.getTagsForBucket(
      SystemTags.Buckets.Applicant,
      true
    );
    const availableTags = uniqBy([
      ...applicationTags,
      ...nonprofitTags,
      ...applicantTags
    ], 'id').map((tag) => {
      return {
        label: tag.name,
        value: tag.id
      };
    });

    return this.arrayHelper.sort(availableTags, 'label');
  }

  findTag (tagId: number) {
    return this.summaries.find((tag) => {
      return tag.id === tagId;
    });
  }

  private tagToAddTag (tag: SystemTags.TagDefinition): SystemTags.AddTagRequest {
    return {
      name: tag.name,
      color: tag.color,
      description: tag.description,
      tagBucketId: tag.bucketId
    };
  }

  async setSummaries () {
    if (!this.summaries) {
      const summaries = await this.systemTagsResources.getAll();

      summaries.forEach(tag => {
        tag.color = tag.color || SystemTags.TagColor.Yellow;
      });

      this.set('summaries', summaries);
    }
  }

  async resetSummaries () {
    this.set('summaries', undefined);
    await this.setSummaries();
  }

  getBuckets () {
    const buckets = [{
      name: this.i18n.translate('tags:lblGlobal'),
      id: SystemTags.Buckets.Global
    }, {
      name: this.i18n.translate('common:hdrApplications'),
      id: SystemTags.Buckets.Application
    }, {
      name: this.i18n.translate('common:lblApplicantsSlashNominators', {}, 'Applicants/Nominators'),
      id: SystemTags.Buckets.Applicant
    }, {
      name: this.i18n.translate('common:hdrNominations'),
      id: SystemTags.Buckets.Nomination
    }, {
      name: this.i18n.translate('tags:lblNonprofitProfile', {}, 'Nonprofit profile'),
      id: SystemTags.Buckets.NonprofitProfile
    }, {
      name: this.i18n.translate('common:hdrPayments'),
      id: SystemTags.Buckets.Payment
    }];
    this.set('buckets', buckets);
  }

  formatPaginationOptions<T> (
    options: PaginationOptions<T>&{ tags?: number[] },
    columnName = 'tags'
  ): PaginationOptions<T> & { tags?: number[] } {
    let tags: number[] = options?.tags || [];
    tags = this.filterOutFilterColumns<T>(options, columnName, tags, 'filterColumns');
    tags = this.filterOutFilterColumns<T>(options, columnName, tags, 'orFilterColumns');

    tags = tags.filter((tag, index) => tags.indexOf(tag) === index);

    return {
      ...options,
      tags
    };
  }

  getTagOptions (bucket?: SystemTags.Buckets): TypeaheadSelectOption[] {
    return this.arrayHelper.sort((this.summaries || [])
      .filter(tag => bucket ? [bucket, SystemTags.Buckets.Global].includes(tag.bucketId) : true)
      .map(tag => ({
        label: tag.name,
        value: tag.id
      })), 'label');
  }

  private filterOutFilterColumns<T> (
    options: PaginationOptions<T>,
    columnName: string,
    tags: number[],
    prop: 'filterColumns'|'orFilterColumns'
  ) {
    options[prop] = options[prop]
      .filter(column => {
        if (column.columnName === columnName) {
          tags = [
            ...tags,
            ...column.filters
              .reduce((acc, { filterValue }) => {
                return [
                  ...acc,
                  filterValue
                ];
              }, [])
          ];

          return false;
        }

        return true;
      });

    return tags;
  }

  getTagsForBucket (
    bucketType: SystemTags.Buckets,
    includeDisabled = false
  ) {
    const tags = this.summaries.filter((tag) =>  {
      return (!includeDisabled ? tag.enabled : true) &&
        [bucketType, SystemTags.Buckets.Global].includes(tag.bucketId);
    });

    return this.arrayHelper.sort(tags, 'name');
  }

  async fetchTagsForRecord (
    tagType: SystemTags.Buckets,
    entityId: string|number
  ) {
    const tags = await this.systemTagsResources.getTags(tagType, entityId);

    tags.forEach(({tag}) => {
      tag.color = SystemTags.TagColor.Yellow;
    });

    const subTagMap = this.tagMap[tagType] || {};

    this.set('tagMap', {
      ...this.tagMap,
      [tagType]: {
        ...subTagMap,
        [entityId]: tags.map(tag => tag.tag.id)
      }
    });
  }

  getTagsForRecord (
    tagType: SystemTags.Buckets,
    entityId: string|number
  ) {
    return this.tagMap[tagType] ?
      this.tagMap[tagType][entityId] || null :
      null;
  }

  async getAndSetTagsForRecord (
    tagType: SystemTags.Buckets,
    entityId: string|number
  ) {
    let currentTagIds = this.getTagsForRecord(
      tagType,
      entityId
    );
    if (!currentTagIds) {
      await this.fetchTagsForRecord(
        tagType,
        entityId
      );
      currentTagIds = this.getTagsForRecord(
        tagType,
        entityId
      );
    }

    return currentTagIds;
  }

  getNewAndRemovedTags (
    currentTags: SystemTags.TagDefinition[],
    tagIds: number[]
  ) {
    return {
      removedTags: this.getRemovedTags(currentTags, tagIds),
      newTags: this.getNewTags(currentTags, tagIds)
    };
  }

  getNewTags (
    currentTags: SystemTags.TagDefinition[],
    tagIds: number[]
  ) {
    return tagIds.filter((potentialNewTag) => {
      return !currentTags.some(tag => tag.id === potentialNewTag);
    });
  }

  getRemovedTags (
    currentTags: SystemTags.TagDefinition[],
    tagIds: number[]
  ) {
    return currentTags.filter((existingTag) => {
      return !tagIds.includes(existingTag.id);
    }).map(tag => tag.id);
  }

  async setTagsOnRecord (
    tagType: SystemTags.Buckets,
    entityId: string|number,
    newTags: number[],
    removedTags: number[],
    subTagType?: SystemTags.Buckets,
    subEntityId?: string|number
  ) {
    try {
      if (subTagType && subEntityId) {
        // find the records that have been added to the sub entity
        const newTagsForSubEntity = newTags.filter(tagId => {
          return this.summaries.find(summary => {
            return summary.id === tagId && summary.bucketId === subTagType;
          });
        });

        const removedTagsForSubEntity = removedTags.filter(tagId => {
          return this.summaries.find(summary => {
            return summary.id === tagId && summary.bucketId === subTagType;
          });
        });

        if (
          (newTagsForSubEntity.length > 0) ||
          (removedTagsForSubEntity.length > 0)
        ) {
          await this.setTags(
            subTagType,
            subEntityId,
            newTagsForSubEntity,
            removedTagsForSubEntity
          );
          await this.fetchTagsForRecord(subTagType, subEntityId);

          newTags = newTags.filter(newTag => {
            return !newTagsForSubEntity.includes(newTag);
          });

          removedTags = removedTags.filter(removedTag => {
            return !removedTagsForSubEntity.includes(removedTag);
          });
        }
      }

      await this.setTags(
        tagType,
        entityId,
        newTags,
        removedTags
      );
      await this.fetchTagsForRecord(tagType, entityId);
      this.notifier.success(this.i18n.translate(
        'GLOBAL:textSuccessSetTagsOnRecord',
        {},
        'Successfully set the selected tags'
      ));
    } catch (e) {
      this.notifier.error(this.i18n.translate(
        'GLOBAL:textErrorSettingTags',
        {},
        'There was an error setting the tags'
      ));
    }
  }

  async addRecord (tag: SystemTags.TagDefinition): Promise<number> {
    try {
      const body = this.tagToAddTag(tag);
      const tagId = await this.systemTagsResources.addOrUpdateRecord(body);
      await this.resetSummaries();

      return tagId;
    } catch (e) {
      this.logger.error(e);
      this.notifier.error(this.i18n.translate(
        'GLOBAL:textErrorAddingTheTag',
        {},
        'There was an error adding the tag'
      ));

      return null;
    }
  }
  async updateRecord<T extends 'Promote'|'Enable'|'Disable'|'ChangeColor'> (
    tag: SystemTags.TagDefinition,
    action: T,
    color: T extends 'ChangeColor' ? SystemTags.TagColor : null
  ): Promise<void> {
    if (action === 'ChangeColor') {
      await this.systemTagsResources.addOrUpdateRecord({
        id: tag.id,
        name: tag.name,
        description: tag.description,
        tagBucketId: tag.bucketId,
        color
      });
    } else {
      await this.systemTagsResources.updateRecord(tag, action, color);
    }


    // changing the color does not need to reset the table
    if (action !== 'ChangeColor') {
      await this.resetSummaries();
    }
  }

  async removeRecord (tag: SystemTags.TagDefinition) {
    await this.systemTagsResources.removeRecord(tag);
    await this.resetSummaries();
  }

  private tagAttributeByTagType (
    tagType: SystemTags.Buckets
  ) {
    switch (tagType) {
      case SystemTags.Buckets.Application:
      case SystemTags.Buckets.Nomination:
        return 'applicationId';
      case SystemTags.Buckets.NonprofitProfile:
        return 'organizationId';
      case SystemTags.Buckets.Payment:
        return 'paymentId';
      case SystemTags.Buckets.Applicant:
        return 'applicantId';
      case SystemTags.Buckets.Global:
        return '';
    }
  }

  async setTags (
    tagType: SystemTags.Buckets,
    entityId: string|number,
    newTags: number[],
    removedTags: number[]
  ) {
    const body = {
      [this.tagAttributeByTagType(tagType)]: entityId,
      tags: newTags.map(tag => ({
        tagId: tag,
        active: true
      })).concat(removedTags.map(tag => ({
        tagId: tag,
        active: false
      })))
    };

    return this.systemTagsResources.setTags(
      tagType,
      body
    );
  }

  async getCurrentTagsForPDF (
    entityId: string|number
  ) {
    const tagType = SystemTags.Buckets.Application;
    let currentTagIds = this.getTagsForRecord(
      tagType,
      entityId
    );
    if (!currentTagIds) {
      await this.fetchTagsForRecord(tagType, entityId);
      currentTagIds = this.getTagsForRecord(
        tagType,
        entityId
      );
    }

    return currentTagIds.map((id) => {
      return this.summaries.find(tag => tag.id === id)?.name;
    }).filter((name) => !!name);
  }


  async getTagsBulk (
    applicationIds: number[]
  ) {
    const tags = await this.systemTagsResources.getTagsBulk(applicationIds);
    const applicationTagMap: Record<number, string[]> = {};
    applicationIds.forEach((appId) => {
      const appTagsIds = tags.filter((tag) => {
        return tag.applicationId === appId;
      }).map((tag) => {
        return tag.tag.id;
      });
      applicationTagMap[appId] = appTagsIds.map((id) => {
        return this.summaries.find(tag => tag.id === id)?.name;
      }).filter((name) => !!name);
    });

    return applicationTagMap;
  }


  async handleBulkAddTags (
    tagIds: number[],
    appIds: number[]
  ) {
    try {
      await this.systemTagsResources.bulkAddTags(
        tagIds,
        appIds
      );
      this.notifier.success(this.i18n.translate(
        'GLOBAL:textSuccessSettingTags',
        {},
        'Successfully added the selected tags'
      ));
    } catch (e) {
      this.logger.error(e);
      this.notifier.error(this.i18n.translate(
        'GLOBAL:textErrorSettingTags',
        {},
        'There was an error setting the tags'
      ));
    }
  }
}
