
import { Component, OnDestroy } from '@angular/core';
import { ActivatedRoute, Router } from '@angular/router';
import { SpinnerService } from '@core/services/spinner.service';
import { WorkflowLevelAdvancementAPI } from '@core/typings/api/workflow-level-advancement.typing';
import { WorkflowLevelAutomationAPI } from '@core/typings/api/workflow-level-automation.typing';
import { ApplicationStatuses } from '@core/typings/status.typing';
import { FormsService } from '@features/configure-forms/services/forms/forms.service';
import { ProgramService } from '@features/programs/program.service';
import { WorkflowService } from '@features/workflow/workflow.service';
import { ALL_SKIP_FILTER, DEFAULT_DOUBLE_VALUE, TableDataFactory, TopLevelFilter } from '@yourcause/common';
import { I18nService } from '@yourcause/common/i18n';
import { LogService } from '@yourcause/common/logging';
import { ModalFactory } from '@yourcause/common/modals';
import { NotifierService } from '@yourcause/common/notifier';
import { Subscription } from 'rxjs';
import { RerouteApplicationModalComponent } from '../reroute-application-modal/reroute-application-modal.component';
import { WorkflowLevelAutomationService } from '../workflow-level-automation.service';


@Component({
  selector: 'gc-view-history-page',
  templateUrl: './view-history-page.component.html',
  styleUrls: ['./view-history-page.component.scss']
})
export class ViewHistoryPageComponent implements OnDestroy {
  isAdvancement: boolean;
  sub = new Subscription();
  tableKey: string;
  tableDataFactory: TableDataFactory<WorkflowLevelAutomationAPI.WorkflowLevelAutomationApplication|WorkflowLevelAdvancementAPI.WorkflowLevelAutomationApplication>;
  AppStatuses = ApplicationStatuses;
  options = [{
    display: this.i18n.translate('CONFIG:textAllRules', {}, 'All rules'),
    value: DEFAULT_DOUBLE_VALUE,
    dependentOptions: [{
      display: '',
      value: ALL_SKIP_FILTER
    }]
  }, {
    display: this.i18n.translate('GLOBAL:textForm', {}, 'Form'),
    value: 'formId',
    dependentOptions: [{
      display: this.i18n.translate('CONFIG:textAllForms', {}, 'All forms'),
      value: ALL_SKIP_FILTER
    }].concat(this.formService.myFormOptions.map(form => ({
      display: form.label,
      value: '' + form.value
    })))
  }, {
    display: this.i18n.translate('common:lblProgram', {}, 'Program'),
    value: 'programId',
    dependentOptions: [{
      display: this.i18n.translate('common:textAllPrograms', {}, 'All programs'),
      value: ALL_SKIP_FILTER
    }].concat(this.programService.allManagerPrograms.map(program => ({
      display: program.grantProgramName,
      value: program.grantProgramId
    })))
  }, {
    display: this.i18n.translate('common:lblWorkflow', {}, 'Workflow'),
    value: 'workflowId',
    dependentOptions: [{
      display: this.i18n.translate('CONFIG:textAllWorkflows', {}, 'All workflows'),
      value: ALL_SKIP_FILTER
    }].concat(this.workflowService.myWorkflowOptions.map(workflow => ({
      display: workflow.label,
      value: '' + workflow.value
    })))
  }];

  topLevelFilters: TopLevelFilter[] = [];

   constructor (
    private logger: LogService,
    private i18n: I18nService,
    private notifier: NotifierService,
    private spinnerService: SpinnerService,
    private modalFactory: ModalFactory,
    private workflowLevelAutomationService: WorkflowLevelAutomationService,
    private workflowService: WorkflowService,
    private formService: FormsService,
    private programService: ProgramService,
    private router: Router,
    private activatedRoute: ActivatedRoute
  ) {
    this.sub.add(this.router.events.subscribe(() => {
      this.setIsAdvancement();
      this.setTopLevelFilters();
      this.setTableDataFactoryAndKey();
    }));
  }

  setIsAdvancement () {
    const isAdvancement = this.activatedRoute.snapshot.data.isRoutingHistory;
    this.isAdvancement = !!isAdvancement;
  }

  setTopLevelFilters () {
    this.topLevelFilters = [
      new TopLevelFilter(
        'doubleDropdownEquals',
        'ruleProgramFormWorkflowFilter',
        null,
        null,
        {
          selectOptions: this.options
        },
        this.i18n.translate(
          'GLOBAL:textSearchRulesByFormProgramOrWorkflow',
          {},
          'Search rules by form, program, or workflow'
        ),
        null
      ),
      new TopLevelFilter(
        'dateRange',
        this.isAdvancement ?
          'workflowLevelDate' :
          'initialWorkflowLevelDate',
        '',
        this.i18n.translate(
          'CONFIG:lblRoutedOn',
          {},
          'Routed on'
        )
      )
    ];
  }

  setTableDataFactoryAndKey () {
    this.tableKey = this.isAdvancement ?
      'ROUTING_HISTORY' :
      'ASSIGNMENT_HISTORY';
    setTimeout(() => {
      this.tableDataFactory = this.isAdvancement ?
        this.workflowLevelAutomationService.routingHistoryTableDataFactory as any :
        this.workflowLevelAutomationService.assignmentHistoryTableDataFactory as any;
    }, 100);
  }

  handleNpoSelect (
    record: WorkflowLevelAutomationAPI.WorkflowLevelAutomationApplication|WorkflowLevelAdvancementAPI.WorkflowLevelAutomationApplication
  ) {
    this.router.navigate([
      `/management/nonprofit/${record.nonprofitGuid}/profile`
    ]);
  }

  async rerouteApplication (
    record: WorkflowLevelAutomationAPI.WorkflowLevelAutomationApplication|WorkflowLevelAdvancementAPI.WorkflowLevelAutomationApplication
  ) {
    const data = await this.modalFactory.open(
      RerouteApplicationModalComponent,
      {
        currentWorkflowLevelId: record.currentWorkflowLevelId,
        workflowId: record.workflowId
      }
    );
    if (data) {
      this.spinnerService.startSpinner();
      try {
        await this.workflowLevelAutomationService.rerouteApplication(
          data.workflowLevelId,
          record.applicationId,
          data.comments
        );
        this.notifier.success(this.i18n.translate(
          'CONFIG:notificationSuccessfullyReroutedApplication',
          {},
          'Successfully rerouted application'
        ));
      } catch (e) {
        this.logger.error(e);

        this.notifier.success(this.i18n.translate(
          'CONFIG:notificationErrorOnRerouteApplication',
          {},
          'There was an error rerouting this application'
        ));
      }
      this.spinnerService.stopSpinner();
    }
  }

  ngOnDestroy () {
    this.sub.unsubscribe();
  }
}
