import { Component, EventEmitter, Input, OnChanges, Output } from '@angular/core';
import { WorkflowDetail, WorkflowLevel } from '@core/typings/workflow.typing';
import { FormAudience, FormTypes, WorkflowLevelFormAdd, WorkflowLevelFormApi } from '@features/configure-forms/form.typing';
import { FormsService } from '@features/configure-forms/services/forms/forms.service';
import { ProgramService } from '@features/programs/program.service';
import { EmailNotificationType } from '@features/system-emails/email.typing';
import { I18nService } from '@yourcause/common/i18n';
import { ModalFactory } from '@yourcause/common/modals';
import { ProgramFormModalComponent } from '../../programs/program-form-modal/program-form-modal.component';

@Component({
  selector: 'gc-workflow-accordion',
  templateUrl: 'workflow-accordion.component.html',
  styleUrls: ['./workflow-accordion.component.scss']
})

export class WorkflowAccordionComponent implements OnChanges {
  @Input() workflowDetail: WorkflowDetail;
  @Input() workflowFormMap: {
    [x: string]: WorkflowLevelFormApi[];
  };
  @Input() defaultForm: WorkflowLevelFormApi;
  @Input() hideMasking = false;
  @Input() maskingOn = false;
  @Input() showDisabled = false;
  @Input() programId: number;
  @Input() isNomination: boolean;
  @Output() onFormChangeAdd = new EventEmitter();
  @Output() onFormChangeRemove = new EventEmitter();
  @Output() onFormOrderChange = new EventEmitter<{id: number; updatedList: WorkflowLevelFormApi[]}>();

  workflowActionsMap: {
    [x: string]: string[];
  } = {};
  workflowMaskMap: {
    [x: string]: boolean;
  } = {};
  levels: WorkflowLevel[];
  FormAudience = FormAudience;
  FormTypes = FormTypes;
  typeAudienceMap = this.formService.typeAudienceMap;
  EmailNotificationType = EmailNotificationType;

  constructor (
    private i18n: I18nService,
    private modalFactory: ModalFactory,
    private formService: FormsService,
    private programService: ProgramService
  ) { }

  get formTypeMap () {
    return this.formService.formTypeMap;
  }

  ngOnChanges () {
    this.levels = [];
    this.workflowActionsMap = {};
    this.workflowDetail.levels.forEach((level) => {

      if (this.showDisabled || (!this.showDisabled && !level.disabled)) {
        this.levels.push(level);
      }
      Object.keys(level).forEach((key) => {
        this.setWorkflowMap(level, key);
        if (key === 'showMaskedApplicantInfo') {
          this.workflowMaskMap[level.id] = level[key];
        }
        if (key === 'subLevels') {
          level[key].forEach((sub) => {
            sub.parentName = level.name;
            if (this.showDisabled || (!this.showDisabled && !sub.disabled)) {
              this.levels.push(sub);
            }
            Object.keys(sub).forEach((subKey) => {
              if (subKey === 'showMaskedApplicantInfo') {
                this.workflowMaskMap[sub.id] = sub[subKey];
              }
              this.setWorkflowMap(sub, subKey);
            });
          });
        }
      });
    });
  }

  setWorkflowMap (level: WorkflowLevel, key: string) {
    if (level[key as keyof WorkflowLevel] && key.startsWith('allow')) {
      let name = '';
      switch (key) {
        case 'allowDeclination':
          name = this.i18n.translate(
            'WORKFLOW:textAllowDeclination',
            {},
            'Allow declination'
          );
          break;
        case 'allowApproval':
          name = this.i18n.translate(
            'WORKFLOW:textAllowApproval',
            {},
            'Allow approval'
          );
          break;
        case 'allowAward':
          name = this.i18n.translate(
            'WORKFLOW:textAllowAward',
            {},
            'Allow award'
          );
          break;
        case 'allowUserToArchiveAndUnarchive':
          name = this.i18n.translate(
            'WORKFLOW:textAllowUserToArchiveAndUnarchive',
            {},
            'Allow user to archive and unarchive'
          );
          break;
        case 'allowUserToRequestRevision':
          name = this.i18n.translate(
            'WORKFLOW:textAllowUserToRequestRevision',
            {},
            'Allow user to request revision'
          );
          break;
        case 'allowRecommendedFunding':
          name = this.i18n.translate(
            'WORKFLOW:textAllowUserToRecommendFunding',
            {},
            'Allow user to recommend funding'
          );
          break;
        default:
          name = '';
      }
      if (this.workflowActionsMap[level.id]) {
        this.workflowActionsMap[level.id].push(name);
      } else {
        this.workflowActionsMap[level.id] = [name];
      }
    }
  }

  handleMoveItem (
    row: WorkflowLevel,
    itemIndex: number,
    isUp: boolean
  ) {
    const adapted = this.programService.updateFormsListSortOrder(
      this.workflowFormMap[row.id],
      itemIndex,
      isUp
    );
    this.onFormOrderChange.emit({
      id: row.id,
      updatedList: adapted
    });
  }

  getFormName (id: number) {
    const found = this.formService.published.find((form) => {
      return +form.formId === +id;
    });

    return found ? found.name : '';
  }

  async addEditForm (
    row: WorkflowLevel,
    form?: WorkflowLevelFormApi,
    index?: number
  ) {
    const existingForms = (this.workflowFormMap[row.id] || []).filter((item) => {
      if (form) {
        return form.formId !== item.formId;
      }

      return item;
    });
    const deps = {
      form,
      workflowLevel: row,
      defaultForm: this.defaultForm,
      existingForms,
      programId: this.programId,
      isNomination: this.isNomination
    };
    const response = await this.modalFactory.open(
      ProgramFormModalComponent,
      deps
    );
    if (response) {
      const formChange: WorkflowLevelFormAdd = {
        response,
        row,
        index,
        form
      };
      this.onFormChangeAdd.emit(formChange);
    }
  }

  removeForm (row: WorkflowLevel, index: number) {
    this.onFormChangeRemove.emit({
      row,
      index
    });
  }
}
